{
  ALMailSlot v1.02

  (C)1999 Andrew Leigh
  http://www.alphalink.com.au/~leigh/components

  Description:
    ALMailSlot is a component which allows applications to communicate across a
    network using mailslots.

  History:
    v1.0  26-Jun-1999 Inital release.
    v1.01 08-Aug-1999 Made the thread execution procedure more time efficient.
    v1.02 03-Oct-1999 Fixed problem when sending multiple lines of text that
                      contained carriage returns.
}

unit ALMailSlot;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TALMailSlot = class;
  TNewMessage = procedure(Sender: TObject; Origin, Time, Text: String) of Object;

  TCheckThread = class(TThread)
  private
    MailSlot: TALMailSlot;
  protected
    procedure Execute; override;
  end;

  TALMailSlot = class(TComponent)
  private
    fMailBox: String;
    fActive: Boolean;
    CheckThread: TCheckThread;
    LocalHandle, RemoteHandle: THandle;
    LocalPath, RemotePath: String;
    fCheckInterval: Integer;
    MessageSize, MessageCount: DWord;
    InMessage: TStringList;
    OutMessage: TStringList;
    fComputer: String;
    fNewMessage: TNewMessage;
    CheckThreadRunning: Boolean;
    fThreadPriority: TThreadPriority;
    procedure StartThread;
    procedure SetActive(const Value: Boolean);
    procedure ReadMessage;
    procedure MailStrings(Recipient: String);
    procedure SetMailBox(const Value: String);
    procedure SetThreadPriority(const Value: TThreadPriority);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SendMessage(Recipient, Text: String);
    procedure Broadcast(Text: String);
    property Computer: String                 read fComputer;
  published
    property MailBox: String                  read fMailBox         write SetMailBox;
    property Active: Boolean                  read fActive          write SetActive         default False;
    property CheckInterval: Integer           read fCheckInterval   write fCheckInterval    default 1000;
    property ThreadPriority: TThreadPriority  read fThreadPriority  write SetThreadPriority default tpNormal;
    property OnNewMessage: TNewMessage        read fNewMessage      write fNewMessage;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('ALComps', [TALMailSlot]);
end;

{ TALMailSlot }

constructor TALMailSlot.Create(AOwner: TComponent);
var
  Temp: Array[0..255] of Char;
  Size: DWord;
begin
  inherited;

  fMailBox := 'MailBox';
  fActive := False;
  fCheckInterval := 1000;
  fThreadPriority := tpNormal;

  OutMessage := TStringList.Create;
  InMessage := TStringList.Create;
  Size := 255;
  GetComputerName(Temp, Size);
  fComputer := StrPas(Temp);

  if not(csDesigning in ComponentState) then
  begin
    CheckThread := TCheckThread.Create(True);
    CheckThread.MailSlot := Self;
    CheckThreadRunning := False;
  end;
end;

destructor TALMailSlot.Destroy;
begin
  if fActive then
    Active := False;

  InMessage.Free;
  OutMessage.Free;
  if not(csDesigning in ComponentState) then
    CheckThread.Free;

  inherited;
end;

procedure TALMailSlot.SendMessage(Recipient, Text: String);
begin
  InMessage.Text := Text;
  with InMessage do
  begin
    Insert(0, 'Message');
    Insert(1, TimeToStr(Now));
    Insert(2, fComputer);
  end;
  MailStrings(Recipient);
end;

procedure TALMailSlot.Broadcast(Text: String);
begin
  InMessage.Text := Text;
  with InMessage do
  begin
    Insert(0, 'Message');
    Insert(1, TimeToStr(Now));
    Insert(2, fComputer);
  end;
  MailStrings('*');
end;

procedure TALMailSlot.MailStrings(Recipient: String);
var
  Bytes: DWord;
begin
  RemotePath := '\\' + Recipient + '\mailslot\' + fMailBox;
  RemoteHandle := CreateFile(PChar(RemotePath), GENERIC_WRITE, FILE_SHARE_READ, nil, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
  if RemoteHandle = INVALID_HANDLE_VALUE then
    Exit
  else
    WriteFile(RemoteHandle, Pointer(InMessage.Text)^, Length(InMessage.Text), Bytes, nil);
end;

procedure TALMailSlot.SetActive(const Value: Boolean);
begin
  if fActive <> Value then
  begin
    fActive := Value;
    if fActive and CheckThreadRunning then
      Exit
    else if fActive and not(csDesigning in ComponentState) then
      StartThread
    else if not fActive and not(csDesigning in ComponentState) then
    begin
      CheckThreadRunning := False;
      CheckThread.Terminate;
      CheckThread.WaitFor;
      CloseHandle(LocalHandle);
    end;
  end;
end;

procedure TALMailSlot.StartThread;
begin
  LocalPath := '\\.\mailslot\' + fMailBox;
  LocalHandle := CreateMailSlot(PChar(LocalPath), 0, 0, nil);
  if LocalHandle = INVALID_HANDLE_VALUE then
    fActive := False
  else
  begin
    CheckThread.Priority := fThreadPriority;
    CheckThread.Resume;
  end;
end;

procedure TALMailSlot.ReadMessage;
var
  NewMessage, TempComputer, TempTime: String;
  Bytes: DWord;
begin
  SetLength(NewMessage, MessageSize);
  ReadFile(LocalHandle, PChar(NewMessage)^, MessageSize, Bytes, nil);
  OutMessage.Clear;
  OutMessage.Text := NewMessage;
  if (OutMessage[0] = 'Message') and Assigned(fNewMessage) then
  begin
    TempComputer := OutMessage[2];
    TempTime := OutMessage[1];
    OutMessage.Delete(0); OutMessage.Delete(0); OutMessage.Delete(0);
    fNewMessage(Self, TempComputer, TempTime, OutMessage.Text);
  end;
end;

procedure TALMailSlot.SetMailBox(const Value: String);
begin
  if fMailBox <> Value then
  begin
    fMailBox := Value;
    if fActive then
    begin
      Active := False;
      Active := True;
    end;
  end;
end;

procedure TALMailSlot.SetThreadPriority(const Value: TThreadPriority);
begin
  if fThreadPriority <> Value then
  begin
    fThreadPriority := Value;
    if not(csDesigning in ComponentState) then
      CheckThread.Priority := fThreadPriority;
  end;
end;

{ TCheckThread }

procedure TCheckThread.Execute;
var
  ThreadWaitInterval, NextTime: Integer;
begin
  if MailSlot.fCheckInterval > 1000 then
    ThreadWaitInterval := 1000
  else
    ThreadWaitInterval := MailSlot.fCheckInterval;

  MailSlot.CheckThreadRunning := True;

  if ThreadWaitInterval = 1000 then
  begin
    NextTime := MaxInt;
    while not Terminated do
    begin
      if NextTime >= MailSlot.fCheckInterval then
      begin
        GetMailSlotInfo(MailSlot.LocalHandle, nil, MailSlot.MessageSize, @MailSlot.MessageCount, nil);
        if MailSlot.MessageCount > 0 then
          Synchronize(MailSlot.ReadMessage);
        NextTime := 0;
      end;
      Sleep(1000);
      Inc(NextTime, 1000);
    end;
  end
  else
  begin
    while not Terminated do
    begin
      GetMailSlotInfo(MailSlot.LocalHandle, nil, MailSlot.MessageSize, @MailSlot.MessageCount, nil);
      if MailSlot.MessageCount > 0 then
        Synchronize(MailSlot.ReadMessage);
      Sleep(ThreadWaitInterval);
    end;
  end;
end;

end.
