////////////////////////////////////////////////////////////////////////////////
// Main unit of the HITDemo project
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit HITDemoMain;

interface
{$A+,H+}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Buttons, Menus;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Memo1: TMemo;
    AbortBtn: TSpeedButton;
    ClearBtn: TSpeedButton;
    CloseBtn: TSpeedButton;
    HelpBtn: TSpeedButton;
    MenuBtn: TSpeedButton;
    PopupMenu1: TPopupMenu;
    PopItemA: TMenuItem;
    PopItemB: TMenuItem;
    PopItemC: TMenuItem;
    PopItemD: TMenuItem;
    PopItemE: TMenuItem;
    PopItemF: TMenuItem;
    PopItemG: TMenuItem;
    PopItemH: TMenuItem;
    PopItemI: TMenuItem;
    PopItemJ: TMenuItem;
    PopItemK: TMenuItem;
    PopItemL: TMenuItem;
    PopItemM: TMenuItem;
    PopItemN: TMenuItem;
    PopItemO: TMenuItem;
    PopItemP: TMenuItem;
    PopItemQ: TMenuItem;
    procedure AbortBtnClick(Sender: TObject);
    procedure ClearBtnClick(Sender: TObject);
    procedure CloseBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure HelpBtnClick(Sender: TObject);
    procedure MenuBtnClick(Sender: TObject);
    procedure PopItemAClick(Sender: TObject);
    procedure PopItemBClick(Sender: TObject);
    procedure PopItemCClick(Sender: TObject);
    procedure PopItemDClick(Sender: TObject);
    procedure PopItemEClick(Sender: TObject);
    procedure PopItemFClick(Sender: TObject);
    procedure PopItemGClick(Sender: TObject);
    procedure PopItemHClick(Sender: TObject);
    procedure PopItemIClick(Sender: TObject);
    procedure PopItemJClick(Sender: TObject);
    procedure PopItemKClick(Sender: TObject);
    procedure PopItemLClick(Sender: TObject);
    procedure PopItemMClick(Sender: TObject);
    procedure PopItemNClick(Sender: TObject);
    procedure PopItemOClick(Sender: TObject);
    procedure PopItemPClick(Sender: TObject);
    procedure PopItemQClick(Sender: TObject);
    procedure Run(Sender: TObject; Num: Char);
    procedure RunA(Sender: TObject);
    procedure RunB(Sender: TObject);
    procedure RunC(Sender: TObject);
    procedure RunD(Sender: TObject);
    procedure RunE(Sender: TObject);
    procedure RunF(Sender: TObject);
    procedure RunG(Sender: TObject);
    procedure RunH(Sender: TObject);
    procedure RunI(Sender: TObject);
    procedure RunJ(Sender: TObject);
    procedure RunK(Sender: TObject);
    procedure RunL(Sender: TObject);
    procedure RunM(Sender: TObject);
    procedure RunN(Sender: TObject);
    procedure RunO(Sender: TObject);
    procedure RunP(Sender: TObject);
    procedure RunQ(Sender: TObject);
  end;

var Form1: TForm1;


implementation
//==============================================================================
{$C-,I-,O+,Q-,R-,W-}
{$R *.DFM}
//------------------------------------------------------------------------------
uses
  HIT_Lists,
  Integers,
  Integers_FFT,
  Integers_Primes,
  EllipticCurvesGFP,
  GF2,
  EllipticCurvesGF2N;

const
  CHECKMSG : array[Boolean] of string[5] = ('ERROR','ok');
  LOGICAL  : array[Boolean] of string[5] = ('FALSE','TRUE');


////////////////////////////////////////////////////////////////////////////////
// TForm1 class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Abort the running of some HIT routines
//==============================================================================
procedure TForm1.AbortBtnClick(Sender: TObject);
begin
  HITAbort;
end;

//==============================================================================
// Clear the editor
//==============================================================================
procedure TForm1.ClearBtnClick(Sender: TObject);
begin
  Memo1.Lines.Clear;
end;

//==============================================================================
// Terminate the application
//==============================================================================
procedure TForm1.CloseBtnClick(Sender: TObject);
begin
  Close;
end;

//==============================================================================
// Create the form
//==============================================================================
procedure TForm1.FormCreate(Sender: TObject);
begin
  Application.HintColor := clYELLOW;

  //----- String format (Used by the procedure Integers.IStr)
  //IStrFormat.SpaceAsPlus := false; //'+' instead of a space
  //IStrFormat.LowerCased := false; //For hexadecimal strings

  //----- String format (Used by the procedure GF2.GF2HexStr)
  //GF2StrFormat.LowerCased := false; //For hexadecimal strings
end;

//==============================================================================
// Open the HIT.HLP file
//==============================================================================
procedure TForm1.HelpBtnClick(Sender: TObject);
begin
  Application.HelpCommand(HELP_FINDER,0);
end;


////////////////////////////////////////////////////////////////////////////////
// Example A
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Generate and display a random integer
//==============================================================================
procedure TForm1.RunA(Sender: TObject);
  var A : TInteger;
begin
  //----- Create A equal to 0
  A := TInteger.Create;
  with Memo1.Lines do
  try
    //----- Random signed integer of at most 166 bits (~ 50 decimal digits)
    IRnd(A,166,false,true);
    //----- Display A
    Add('A (base 10) =' + IStr(A,10,5)); //5-digit blocks
    Add('');
    Add('A (base 16) =' + IStr(A,16,4)); //4-digit blocks
    Add('');
    Add('A (base 2)  =' + IStr(A,2,16)); //16-digit blocks
    Add('');
  finally
    //----- Free the integer
    A.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example B
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Generate and sort a list of random numbers
//==============================================================================
procedure TForm1.RunB(Sender: TObject);
  var
    A : TInteger;
    L : TIntList; //Implemented in "HIT_Lists.pas" file
    i : Integer;
    S : string;
begin
  A := nil;
  L := nil;
  try
    //----- Create A equal to 0
    A := TInteger.Create;
    //----- Create an empty list
    L := TIntList.Create;

    //===== Compute random integers
    for i := 1 to 10 do
    begin
      //----- Random signed integer (random size of at most 200 bits)
      IRnd(A,IRnd32 mod 201,false,true);
      //----- Append a copy of A to the list L
      L.Append(A);
    end;

    //===== Display the integers to the base 10
    Memo1.Lines.Add('List of random integers');
    S := '';
    for i := 0 to L.Count-1 do
      S := S + '> ' + IStr(L[i],10,0) + #13#10; //"NewLine"
    Memo1.Lines.Add(S);

    //===== Sort the list
    L.Sort(@IListCmp); //And not "L.Sort(@ICmp)" !

    //===== Display the sorted list
    Memo1.Lines.Add('Sorted list');
    S := '';
    for i := 0 to L.Count-1 do
      S := S + '> ' + IStr(L[i],10,0) + #13#10;
    Memo1.Lines.Add(S);
  finally
    //----- Free the number
    A.Free;
    //----- Free the list
    L.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example C
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Compute 2^(A-1) modulo A (A is a 664-bit random integer)
//==============================================================================
procedure TForm1.RunC(Sender: TObject);
  var
    S         : string;
    A, E, X   : TInteger;
    BSize, Hw : Longint;
    TickCount : Int64;
    Chrono    : Extended;
begin
  //----- Set to nil so that the "finally" part works okay in any case
  A := nil;
  E := nil;
  X := nil;
  with Memo1.Lines do
  try
    //----- Create integers equal to 0
    A := TInteger.Create;
    E := TInteger.Create;
    X := TInteger.Create;

    //----- Compute a 664-bit random unsigned integer
    BSize := 664; //664 bits ~ 200 decimal digits
    IRnd(A,BSize);
    A.Odd := true; //So that the number is odd

    //----- Display A
    Add('A =' + IStr(A,10,5)); //Base 10, 5-digit blocks

    //----- Display the different sizes of the number
    Add('Size (base 10)   = ' + IntToStr(A.DecimalSize));
    Add('Size (base 2)    = ' + IntToStr(A.BitSize));
    Add('Size (base 2^8)  = ' + IntToStr(A.ByteSize));
    Add('Size (base 2^32) = ' + IntToStr(A.LongSize));
    Add('');

    //----- E = A-1
    ISet(E,A); //Copy A to E
    IDec(E);   //E := E - 1
    //----- X = 2
    ISet(X,2);

    //----- Calculate X := 2^(A-1) mod A
    TickCount := HITTimer;
    IPowMod(X,E,A);
    Chrono := HITTimer(TickCount);

    //----- Display
    Add('2^(A-1) mod A =' + IStr(X,10,5));
    Add('Size (base 10)   = ' + IntToStr(X.DecimalSize));
    Add('Size (base 2)    = ' + IntToStr(X.BitSize));
    Add('Size (base 2^8)  = ' + IntToStr(X.ByteSize));
    Add('Size (base 2^32) = ' + IntToStr(X.LongSize));
    Add('');

    //----- Rare event: Proba(X = 1) ~ 2/N.BitSize
    if X.Is1 then Add('>>> A is pseudoprime for the base 2 <<<'#13#10);

    //----- The work done depends on the size and on the Hamming weight
    Dec(BSize);
    Hw := E.Weight-1;
    Add('The calculations required ' +
         IntToStr(BSize) + ' squares, ' +
         IntToStr(Hw) + ' multiplications and ' +
         IntToStr(BSize+Hw) + ' divisions.');

    //----- Display the running time
    S := FloatToStrF(Chrono,ffFixed,18,3) + ' ms';
    Add('Running time = ' + S);
    Add('');
  finally
    A.Free;
    E.Free;
    X.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example D
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Saving and loading integers
//==============================================================================
procedure TForm1.RunD(Sender: TObject);
  const FILENAME = 'random.hit';
  var A, B : TInteger;
begin
  A := nil;
  B := nil;
  with Memo1.Lines do
  try
    //----- Create A equal to 0
    A := TInteger.Create;
    B := TInteger.Create;

    //----- Random signed integer of at most 500 decimal digits
    IRnd(A,1660,false,true);

    //----- Display A
    Add('A =  ' + IStr(A,10,5));
    Add('');

    //----- Save A to the FILENAME file
    A.SaveToFile(FILENAME);

    //----- Load B from the FILENAME file
    B.LoadFromFile(FILENAME);

    //----- Display B
    Add('B =  ' + IStr(B,10,5));
    Add('');

    //----- Check
    Add('Check -> ' + CHECKMSG[IEqu(A,B)]);
    Add('');
  finally
    //----- Free the integers
    A.Free;
    B.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example E
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// GCD (Greatest Common Divisor)
//==============================================================================
procedure TForm1.RunE(Sender: TObject);
  var
    A, B, C, D : TInteger;
    S          : string;
    TickCount  : Int64;
    Chrono     : Extended;
begin
  //----- Set to nil so that the "finally" part works okay in any case
  A := nil;
  B := nil;
  C := nil;
  D := nil;
  //----- Allow to abort the running by a click on the Abort button.
  //      To "see" how it slows down the running of routines, you can
  //      disable this instruction.
  HITBreakWith(Application.ProcessMessages);

  with Memo1.Lines do
  try
    AbortBtn.Enabled := HITBreakOn;

    //----- Create integers equal to 0
    A := TInteger.Create;
    B := TInteger.Create;

    //----- Random unsigned integers
    IRnd(A,20000,false);
    IRnd(B,20000,false);

    //----- Create C and D equal to A and B
    C := TInteger.Create(A);
    D := TInteger.Create(B);

    //----- Display the decimal sizes of A and B
    Add('Decimal size of A = ' + IntToStr(A.DecimalSize));
    Add('Decimal size of B = ' + IntToStr(B.DecimalSize));
    Add('');

    Add('Computing GCD ...');

    //----- Compute the GCD of A and B (Euclidean algorithm)
    TickCount := HITTimer;
    while not A.Is0 do
    begin
      //----- (A', B') <- (B mod A, A)
      IMod(B,A);
      ISwp(B,A);
    end;
    Chrono := HITTimer(TickCount);

    //----- Display B (= GCD(A,B))
    Delete(Count-1);
    Add('GCD(A,B) =' + IStr(B,10,0));
    //----- Display the running time
    S := FloatToStrF(Chrono,ffFixed,18,3) + ' ms';
    Add('Running time = ' + S);
    Add('');

    Add('Computing GCD ...');

    //----- Compute the GCD of A and B (with the IGCD routine)
    TickCount := HITTimer;
    IGCD(C,D);
    Chrono := HITTimer(TickCount);

    //----- Display C (= GCD(A,B))
    Delete(Count-1);
    Add('GCD(A,B) =' + IStr(C,10,0));
    //----- Display the running time
    S := FloatToStrF(Chrono,ffFixed,18,3) + ' ms';
    Add('Running time = ' + S);
    Add('');

    // The 2 previous running times show that IGCD is much more efficient
    // than the simple Euclidean algorithm. That's okay. IGCD uses the
    // Lehmer-Collins variant.

  finally
    //----- Default status
    AbortBtn.Enabled := false;
    HITNoBreak;
    //----- Free the integers
    A.Free;
    B.Free;
    C.Free;
    D.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example F
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Egyptian fractions N/D = 1/d1 + 1/d2 + ... + 1/dk
//------------------------------------------------------------------------------
// This routine doesn't compute fractions exactly like scribes did. The
// ancient Egyptians used not only fractions of type 1/D, they also knew
// the fraction 2/3.
// The implemented algorithm was used by Fibonacci.
//==============================================================================
procedure TForm1.RunF(Sender: TObject);
  const NSIZE = 16; //Binary size of N. Must be small enough. With the given
                    //value, the result can already be pharaonic.
                    //For instance, try with that:
                    //  ISet(N,31340);
                    //  ISet(D,2591571893);
  var
    N, D, X : TInteger;
    L       : TIntList; //To check result
    i       : Integer;
begin
  N := nil;
  D := nil;
  X := nil;
  L := nil;
  with Memo1.Lines do
  try
    //----- Create an empty list
    L := TIntList.Create;

    //----- Create N and D equal to 0
    N := TInteger.Create;
    D := TInteger.Create;

    //----- Generate N and D such that N > 1 and 0 < (N/D) < 1
    //      For perfectionists who wish obtain the real Egyptian
    //      fractions: either generate N and D such that
    //      0 < (N/D) < (2/3), either subtract 2/3 from N/D
    //      when (N/D) >= (2/3).
    repeat
      //----- Random unsigned N of at most NSIZE bits
      IRnd(N,NSIZE,false);
    until ICmp(N,1) > 0; //N must be greater than 1
    repeat
      //----- Random unsigned D of at most 2*NSIZE bits
      IRnd(D,NSIZE+NSIZE,false);
    until ICmp(D,N) > 0; //D must be greater than N

    //----- Display the initial fraction (base 10, 3-digit blocks)
    Add('N =' + IStr(N,10,3));
    Add('D =' + IStr(D,10,3));
    Add('');

    //----- Add N, D and GCD(N,D) to the list
    L.Append(N);
    L.Append(D);
    L.Append(N);
    IGCD(L[2],D); //L[2] := GCD(N,D).

    //===== Compute and display the "Egyptian" denominators
    Add('N / D =');
    X := TInteger.Create;
    repeat
      //----- Reduce N/D
      ISet(X,N);
      if not IGCD(X,D) then //X := GCD(X,D) and if X <> 1 then
      begin
        IDiv(N,X); //N := N div X
        IDiv(D,X); //D := N div X
      end;

      //---- Exit loop when N = 1
      if N.Is1 then Break;

      //----- Compute the greatest 1/X that is less than N/D
      ISet(X,D);
      IDiv(X,N);
      IInc(X); //X := (N div D) + 1 -> (N/D > 1/X > (N-1)/D)

      //----- Display the new denominator
      Add('1 /' + IStr(X,10,3));
      Add('  +');

      //----- Add the new denominator to the list
      L.Append(X);

      //----- Compute the residual N'/D' = N/D - 1/X
      IMul(N,X);
      ISub(N,D); //N := N*X-D
      IMul(D,X); //D := D*X
    until false;

    //----- Display the last denominator
    Add('1 /' + IStr(D,10,3));
    Add('');

    //----- Add the last denominator to the list
    L.Append(D);

    //----- Display the number of denominators
    Add('Denominators: ' + IntToStr(L.Count-3));
    Add('');

    //===== Check
    //----- We start with N/D = 0/1
    N.Set0;
    D.Set1;
    //----- i starts from 3: L[0],L[1] and L[2] contain initial values
    for i := 3 to L.Count-1 do
    begin
      IMul(N,L[i]);
      IAdd(N,D);    //N := N * L[i] + D
      IMul(D,L[i]); //D := D * L[i]
      //----- Reduce the fraction (if possible)
      ISet(X,N);
      if not IGCD(X,D) then //if GCD <> 1 (Here, it cannot be equal to 0)
      begin
        IDiv(N,X); //N := N div X
        IDiv(D,X); //D := N div X
      end;
    end;
    //----- Here, we have the reduced fraction N/D, so if the gcd
    //      of the initial values was different from 1, we have to
    //      multiply N and D by this gcd
    if not L[2].Is1 then
    begin
      IMul(N,L[2]);
      IMul(D,L[2]);
    end;
    Add('Check -> ' + CHECKMSG[IEqu(N,L[0]) and IEqu(D,L[1])]);
    Add('');
  finally
    //----- Free the list
    L.Free;

    //----- Free the integers
    N.Free;
    D.Free;
    X.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example G
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Compare the running times of ISqr and TFFT.Sqr
// ISqr makes use of the Karatsuba's algorithm
// TFFT.Sqr makes use of the Fast Fourier Transform
//==============================================================================
procedure TForm1.RunG(Sender: TObject);
  var
    FFT        : TFFT;
    X, Y       : TInteger;
    Len        : Longint;
    S          : string;
    TickCount  : Int64;
    Chrono     : Extended;
begin
  FFT := nil;
  X := nil;
  Y := nil;
  with Memo1.Lines do
  try
    //----- Create a TFFT multiplier
    FFT := TFFT.Create;

    //----- Create X equal to 0
    X := TInteger.Create;

    //----- Get a random integer
    // On CYRIX MII 300 processor, with the fast multiplication algorithm
    // used by INTEGERS (Karatsuba's algorithm), squaring a 10,000,000-bit
    // integer requires about 160s, with FFT it requires less than 48s.
    //-----
    // First, run this routine with the value below, then, increase this
    // value by 1. You will note that the running time of FFT is more
    // than twice as long.
    // That's okay: there are thresholds because of FFT vectors.
    //   4096 * 24 < 196608 <= 8192 * 24, the size of the FFT vectors
    //   is 8192 (to the base 2^24)
    //   8192 * 24 < 196609 <= 16384 * 24, the size of the FFT vectors
    //   is 16384 (to the base 2^24)
    Len := 196608;
    IRnd(X,Len);

    //----- Create Y equal to X
    Y := TInteger.Create(X);

    //----- Allow to abort calculations (slow down computation!)
    HITBreakWith(Application.ProcessMessages);
    try
      //----- Activate the Abort button (if need be)
      AbortBtn.Enabled := HITBreakOn;

      //----- Calculate X
      TickCount := HITTimer;
      ISqr(X);
      Chrono := HITTimer(TickCount);
      //----- Display the running times
      Add('Running times');
      S := FloatToStrF(Chrono,ffFixed,18,3) + ' ms';
      Add('ISqr     (Karatsuba)              = ' + S);

      //----- Calculate Y
      TickCount := HITTimer;
      FFT.Sqr(Y);
      Chrono := HITTimer(TickCount);
      //----- Display the running times
      S := FloatToStrF(Chrono,ffFixed,18,3) + ' ms';
      Add('TFFT.Sqr (Fast Fourier Transform) = ' + S);
      Add('');
    finally
      //----- Restore the default status
      HITNoBreak;
      AbortBtn.Enabled := false;
    end;

    //----- Check (Compare the 2 squares)
    Add('Check -> ' + CHECKMSG[IEqu(X,Y)]);
    Add('');

    //----- Display the memory currently used by the FFT multiplier
    //      ! Do not take in account FFT.InstanceSize
    Add('FFT used memory = ' + IntToStr(FFT.UsedMem) + ' bytes');
    Add('');

  finally
    //----- Free the FFT multiplier
    FFT.Free;
    //----- Free the integers
    X.Free;
    Y.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example H
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Modular square roots modulo a prime
//==============================================================================
procedure TForm1.RunH(Sender: TObject);
  const
    PSIZE = 332; //Prime binary size (332 bits ~ 100 decimal digits)
  var
    P, A, R1, R2, S1, S2 : TInteger;
    TickCount            : Int64;
    c1, c2, c3, c4       : Extended;
begin
  P := nil;
  A := nil;
  R1 := nil;
  S1 := nil;
  R2 := nil;
  S2 := nil;

  //----- Allow to abort calculations (slow down computation!)
  HITBreakWith(Application.ProcessMessages);

  with Memo1.Lines do
  try
    //----- Activate the Abort button (if need be)
    AbortBtn.Enabled := HITBreakOn;
    //----- No signed symbol in displayed numbers
    //      The space (as '+' symbol) is suppressed from strings
    IStrFormat.Signed := false;

    //----- Create integers equal to 0
    P := TInteger.Create;
    A := TInteger.Create;
    R1 := TInteger.Create;
    R2 := TInteger.Create;
    S1 := TInteger.Create;
    S2 := TInteger.Create;

    Add('Creating prime ...');

    //===== PSIZE-bit prime
    TickCount := HITTimer;
    IRnd(P,PSIZE);       //PSIZE-bit random integer
    IMakePrime(P,[2,3]); //Strong pseudoprime for the bases 2 and 3
    c1 := HITTimer(TickCount);

    Delete(Count-1); //Suppress last line from Memo1.Lines
    Add('Computing quadratic residue ...');

    //===== A such that (A < P) and (Jacobi symbol (A,P) = 1)
    TickCount := HITTimer;
    IRnd(A,PSIZE+PSIZE); //(PSIZE*2)-bit random integer
    IMod(A,P);           //So that A < P
    while IJacobi(A,P) <> 1 do
    begin
      IInc(A);
      if ICmpAbs(A,P) >= 0 then ISub(A,P); //So that A < P (probably never used)
    end;
    c2 := HITTimer(TickCount);

    Delete(Count-1);
    Add('Computing square roots ...');

    //===== R1 and R2 = P - R1
    TickCount := HITTimer;
    ISet(R1,A);
    ISqrRootMod(R1,P);
    ISet(R2,P);
    ISub(R2,R1);
    c3 := HITTimer(TickCount);

    Delete(Count-1);
    Add('Checking ...');

    //===== Squares
    TickCount := HITTimer;
    ISet(S1,R1);
    ISet(S2,R2);
    ISqr(S1);
    IMod(S1,P);
    ISqr(S2);
    IMod(S2,P);
    c4 := HITTimer(TickCount);

    //----- Display
    Delete(Count-1);
    Add('> Prime'#13#10 + IStr(P,10,5));
    Add('');
    Add('> Quadratic residue'#13#10 + IStr(A,10,5));
    Add('');
    Add('> Root 1'#13#10 + IStr(R1,10,5));
    Add('');
    Add('> Root 2'#13#10 + IStr(R2,10,5));
    Add('');
    //----- Check (compare the 2 squares)
    Add('Check 1 -> ' + CHECKMSG[IEqu(S1,A)]);
    Add('Check 2 -> ' + CHECKMSG[IEqu(S2,A)]);
    Add('');

    //----- Running times
    Add('> Running times (milliseconds)');
    Add('Prime ........... ' + FloatToStrF(c1,ffFixed,18,3));
    Add('Quadratic residue ' + FloatToStrF(c2,ffFixed,18,3));
    Add('Square roots .... ' + FloatToStrF(c3,ffFixed,18,3));
    Add('Checks .......... ' + FloatToStrF(c4,ffFixed,18,3));
    Add('');
  finally
    //----- Restore default status
    AbortBtn.Enabled := false;
    HITNoBreak;
    IStrFormat.Signed := true;
    //----- Free the integers
    P.Free;
    A.Free;
    R1.Free;
    R2.Free;
    S1.Free;
    S2.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example I
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// RSA cryptosystem.
// Do not use this (very) simplistic demo in an application:
// 1) it is not secure;
// 2) it is not optimized;
//
// 3) RSA is covered by patents in the USA until September 20, 2000
//    -------------------------------------------------------------
//
//==============================================================================
procedure TForm1.RunI(Sender: TObject);
  //----- Constants to display integers to the base 16, 4-digit blocks
  //      You can use (10,5) or even (2,8) instead of (16,4)
  const
    BASE   = 16;
    BLOCKS = 5;
  var
    P, Q, N, E, D, X, Y, Z : TInteger;
    S                      : string;
    TickCount              : Int64;
    c1, c2, c3, c4         : Extended;
begin
  P := nil;
  Q := nil;
  N := nil;
  E := nil;
  D := nil;
  X := nil;
  Y := nil;
  Z := nil;

  //----- No signed symbol in displayed numbers
  //      The space (as '+' symbol) is suppressed from strings
  IStrFormat.Signed := false;

  with Memo1.Lines do
  try
    //----- Create integers equal to 0
    P := TInteger.Create;
    Q := TInteger.Create;
    N := TInteger.Create;
    E := TInteger.Create;
    D := TInteger.Create;
    X := TInteger.Create;
    Y := TInteger.Create;
    Z := TInteger.Create;

    Add('Creating modulus ...');

    //===== 512-bit modulus (Product of 2 distinct primes)
    TickCount := HITTimer;
    IRnd(P,254);         //254-bit random integer
    P.Bit[252] := true;  //Set the second leading bit (See below)
    IMakePrime(P,[2,3]); //Strong pseudoprime for the bases 2 and 3
    IRnd(Q,258);
    Q.Bit[256] := true;
    IMakePrime(Q,[2,3]);
    //----- Modulus N = PQ
    ISet(N,P);
    IMul(N,Q);
    c1 := HITTimer(TickCount);

    // Setting the second leading bit of both factors implies that
    // N.BitSize = P.BitSize + Q.BitSize.
    // If the second leading bits were random, the binary size of N could be
    // equal to P.BitSize + Q.BitSize - 1 each other time.

    Delete(Count-1); //Suppress last line from Memo1.Lines
    Add('Creating exponents ...');

    //===== Exponents
    TickCount := HITTimer;
    //----- E := 2^k + 1
    //      We want a low Hamming weight to speed up encryption
    E.Set1;
    E.Bit[N.BitSize div 4] := true;
    //----- P := LCM(P-1, Q-1) (Carmichael function)
    //      All popular works concerning RSA suggest to take the decryption
    //      exponent D equal to the inverse of E modulo PHI(N) (Euler
    //      PHI-function). But it is better to take D equal to the inverse
    //      of E modulo L(N) (Carmichael LAMBDA-function). It generally
    //      minimizes the size of D.
    IDec(P);
    IDec(Q);
    ILCM(P,Q);
    repeat
      //----- Encryption exponent E
      IMakePrime(E,[2,3]);
      //----- Decryption exponent D := E^(-1) mod LCM(P-1, Q-1)
      ISet(D,E);
      IInvMod(D,P);
    until IGCD1(D,N);
    c2 := HITTimer(TickCount);

    Delete(Count-1);
    Add('Generating plain text ...');

    //===== Random message X
    //      The "message" is constituted with one repeated digit in order to
    //      make visual checking possible.
    S := StringOfChar(PAnsiChar('123456789abcdef')[IRnd32 mod (BASE-1)],150);
    ISet(X,S,BASE);
    while ICmpAbs(N,X) <= 0 do IDiv(X,BASE); //X must be less than N

    Delete(Count-1);
    Add('Encrypting ...');

    //===== Encrypt
    TickCount := HITTimer;
    ISet(Y,X);
    IPowMod(Y,E,N);
    c3 := HITTimer(TickCount);

    Delete(Count-1);
    Add('Decrypting ...');

    //===== Decrypt
    TickCount := HITTimer;
    ISet(Z,Y);
    IPowMod(Z,D,N);
    c4 := HITTimer(TickCount);

    //----- Display
    Delete(Count-1);
    Add('> Modulus'#13#10 + IStr(N,BASE,BLOCKS));
    Add('');
    Add('> Encryption exponent'#13#10 + IStr(E,BASE,BLOCKS));
    Add('');
    Add('> Decryption exponent'#13#10 + IStr(D,BASE,BLOCKS));
    Add('');
    Add('> Plain text'#13#10 + IStr(X,BASE,BLOCKS));
    Add('');
    Add('> Encrypted'#13#10 + IStr(Y,BASE,BLOCKS));
    Add('');
    Add('> Decrypted'#13#10 + IStr(Z,BASE,BLOCKS));
    Add('');

    //----- Running times
    Add('> Running times (milliseconds)');
    Add('Modulus ..... ' + FloatToStrF(c1,ffFixed,18,3));
    Add('Exponents ... ' + FloatToStrF(c2,ffFixed,18,3));
    Add('Encryption .. ' + FloatToStrF(c3,ffFixed,18,3));
    Add('Decryption .. ' + FloatToStrF(c4,ffFixed,18,3));
    Add('');
  finally
    //----- Restore default value
    IStrFormat.Signed := true;
    //----- Free the integers
    P.Free;
    Q.Free;
    N.Free;
    E.Free;
    D.Free;
    X.Free;
    Y.Free;
    Z.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example J
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Generate and display a random polynomial over GF(2)
// Polynomials over GF(2) are also called "Binary polynomials". The coefficients
// are always either 0 or 1, and the computations are done modulo 2. That's why
// it is very convenient to represent them as bit sequences.
//==============================================================================
procedure TForm1.RunJ(Sender: TObject);
  const INFO = '  (Parity is equivalent to "Weight and 1")';
  var
    P : TGF2Poly;
    D : TDynLongintArray;
    A : TInteger;
    S : string;
    i : Integer;
begin
  P := nil;
  A := nil;
  with Memo1.Lines do
  try
    //----- Create P equal to 0
    P := TGF2Poly.Create;
    //----- Random polynomial over GF(2) of (at most) degree 128
    GF2Rnd(P,128,false);

    //===== Display the polynomial

    //----- Binary representation: One bit = One coefficient
    Add('P (Binary representation)');
    Add('-------------------------');
    Add(GF2BinStr(P,8)); //8-bit blocks
    Add('Degree  = ' + IntToStr(P.Degree));
    Add('Density = ' + FloatToStrF(P.Density,ffFixed,18,3));
    Add('Index   = ' + IntToStr(P.Index));
    Add('Parity  = ' + IntToStr(Ord(P.Parity)) + INFO);
    Add('Weight  = ' + IntToStr(P.Weight));
    Add('');

    //----- Hexa representation: Same as binary representation but the
    //      values are packed in hexadecimal values
    Add('P (Hexadecimal representation)');
    Add('------------------------------');
    Add(GF2HexStr(P,2)); //2-digit blocks
    Add('');

    //----- Display the degrees of coefficients equal to 1
    Add('P (Degrees of coefs equal to 1)');
    Add('-------------------------------');
    D := P.Degs;
    if Length(D) = 0 then S := '0'
    else
    begin
      S := IntToStr(D[0]);
      for i := 1 to Length(D)-1 do S := S + ', ' + IntToStr(D[i]);
    end;
    Add(S);
    Add('');

    //----- Conventional representation of a polynomial
    //      "X^1" and "X^0" should be written "X" and "1". But this is
    //      only a demo :)
    Add('P (Standard representation)');
    Add('---------------------------');
    if Length(D) = 0 then S := '0'
    else
    begin
      S := 'X^' + IntToStr(D[0]);
      for i := 1 to Length(D)-1 do S := S + ' + X^' + IntToStr(D[i]);
    end;
    Add(S);
    Add('');

    //----- "Decimal" representation
    //      To display a polynomial to the base 10, you first need to
    //      cast it to TInteger type
    A := TInteger.Create;
    GF2PolyToInt(P,A); //Cast P to TInteger
    Add('P (Decimal representation)');
    Add('--------------------------');
    IStrFormat.Signed := false;
    Add(IStr(A,10,5)); //5-digit blocks
    IStrFormat.Signed := true;
    Add('');

  finally
    //----- Free the polynomial
    P.Free;
    //----- Free the integer
    A.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example K
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Factorization of polynomials over GF(2)
//==============================================================================
procedure TForm1.RunK(Sender: TObject);
  //----- Constants to display polynomials (4-digit blocks)
  const BLOCKS = 4;
  var
    P, Q : TGF2Poly;
    F    : TDynTGF2PolyArray;
    Test : Boolean;
    i    : Integer;
begin
  P := nil;
  Q := nil;
  with Memo1.Lines do
  try
    //----- Create 2 polynomials equal to 0
    P := TGF2Poly.Create;
    Q := TGF2Poly.Create;

    //----- Generate a composite polynomial of degree = 768
    GF2Rnd(P,256);
    for i := 1 to 2 do
    begin
      GF2Rnd(Q,256); //Random, degree equal to 256
      GF2Mul(P,Q);   //P := P * Q
    end;
    Add('P = ' + GF2HexStr(P,BLOCKS));
    Add('P.Degree = ' + IntToStr(P.Degree));
    Add('');

    //===== Factorization
    Add('List of the irreducible factors of P');
    Add('------------------------------------');
    GF2Split(P,F,true); //F <- Irreducible factors of P
    Q.Set1;
    for i := 0 to Length(F)-1 do
    begin
      //----- Build Q for checking
      GF2Mul(Q,F[i]);
      //----- Display facor #i
      Add('F = ' + GF2HexStr(F[i],BLOCKS));
    end;
    Add('');

    //===== Here, Q should be equal to P without squared factors
    GF2SqrFree(P);
    GF2Add(P,Q);
    Add('Check product -> ' + CHECKMSG[P.Is0]);

    //===== Check the irreducibility of the factors
    Test := true;
    for i := 0 to Length(F)-1 do
      if not F[i].Irreducible then
      begin
        Add('');
        Add('F = ' + GF2HexStr(F[i],BLOCKS) + ' is not irreducible');
        Test := false;
        Break;
      end;
    Add('Check irreducibility -> ' + CHECKMSG[Test]);
    Add('');

  finally
    //----- Free the factors
    for i := 0 to Length(F)-1 do F[i].Free;
    //----- Free the polynomials
    P.Free;
    Q.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example L
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Operations over GF(2^N)
//==============================================================================
procedure TForm1.RunL(Sender: TObject);
  //----- Constants to display polynomials (4-digit blocks)
  const BLOCKS = 4;
  var
    P, Q, R, S : TGF2Poly;
    gf         : TGF2N;
    E          : TInteger;
begin
  E := nil;
  P := nil;
  Q := nil;
  R := nil;
  S := nil;
  gf := nil;
  with Memo1.Lines do
  try
    //----- Create an integer equal to 0
    E := TInteger.Create;

    //----- Create 4 polynomials equal to 0
    P := TGF2Poly.Create;
    Q := TGF2Poly.Create;
    R := TGF2Poly.Create;
    S := TGF2Poly.Create;

    //----- Init the Galois field GF(2^197)
    // Here, we use the pentanomial X^197 + X^9 + X^4 + X^2 + 1 but
    // computations are a little faster with trinomials
    // See GF2.DAT for other irreducible polynomials

    //----- Create and initialize a Galois field
    gf := TGF2N.Create(197);
    {
    //----- Do the same than previous instruction but this way allows
    //      to init a Galois field with any irreducible polynomials
    gf := TGF2N.Create;
    gf.SetModulus([197,9,4,2,0],false);
    }

    //===== Distributivity: P(Q + R) = PQ + PR
    //----- Random P, Q, R
    gf.Rnd(P);
    gf.Rnd(Q);
    gf.Rnd(R);
    //----- Compute S := P(Q + R)
    GF2Set(S,Q);
    gf.Add(S,R);
    gf.Mul(S,P);
    //----- Compute P := PQ + PR
    gf.Mul(R,P);
    gf.Mul(P,Q);
    gf.Add(P,R);
    //------ Display
    Add('Distributivity test: P(Q + R) = PQ + PR ?');
    Add('-----------------------------------------');
    Add('Left  = ' + GF2HexStr(S,BLOCKS));
    Add('Right = ' + GF2HexStr(P,BLOCKS));
    gf.Add(P,S);
    Add('Check -> ' + CHECKMSG[P.Is0]);
    Add('');

    //===== Square root: (P^(1/2))^2 = P
    //----- Random P
    gf.Rnd(P);
    //----- Compute Q := (P^(1/2))^2
    GF2Set(Q,P);
    gf.SqrRoot(Q);
    gf.Sqr(Q);
    //------ Display
    Add('Square root test: P = (P^(1/2))^2 ?');
    Add('-----------------------------------');
    Add('Left  = ' + GF2HexStr(P,BLOCKS));
    Add('Right = ' + GF2HexStr(Q,BLOCKS));
    gf.Add(P,Q);
    Add('Check -> ' + CHECKMSG[P.Is0]);
    Add('');

    //===== Inversion: Q(P/Q) = P
    //----- Random P, Q
    gf.Rnd(P);
    repeat
      gf.Rnd(Q);
    until not Q.Is0;
    //----- Compute Q := Q(P/Q)
    GF2Set(R,Q);
    gf.Inv(Q);
    gf.Mul(Q,P);
    gf.Mul(Q,R);
    //------ Display
    Add('Inversion test: P = Q(P/Q) ?');
    Add('----------------------------');
    Add('Left  = ' + GF2HexStr(P,BLOCKS));
    Add('Right = ' + GF2HexStr(Q,BLOCKS));
    gf.Add(P,Q);
    Add('Check -> ' + CHECKMSG[P.Is0]);
    Add('');

    //===== Power: 1/P = P^(2^N - 2)
    //----- Random P
    repeat
      gf.Rnd(P);
    until not P.Is0;
    //----- Compute Q := P^(2^N - 2)
    GF2Set(Q,P);
    gf.Cardinality(E); //Number of elements of GF(2^N) = 2^N
    ISub(E,2);
    gf.Pow(Q,E);
    //----- Compute P := 1/P
    gf.Inv(P);
    //------ Display
    Add('Power test: 1/P = P^(2^N - 2) ?');
    Add('-------------------------------');
    Add('Left  = ' + GF2HexStr(P,BLOCKS));
    Add('Right = ' + GF2HexStr(Q,BLOCKS));
    gf.Add(P,Q);
    Add('Check -> ' + CHECKMSG[P.Is0]);
    Add('');

  finally
    //----- Free the field
    gf.Free;
    //----- Free the polynomials
    P.Free;
    Q.Free;
    R.Free;
    S.Free;
    //----- Free the integer
    E.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example M
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Elliptic Curves over GF(2^N)
// How to compute a point Q from a point P such that 2P = 2Q
//==============================================================================
procedure TForm1.RunM(Sender: TObject);
  //----- Constants to display polynomials (4-digit blocks)
  const BLOCKS = 4;
  var
    ec                   : TGF2NECurve;
    A, B                 : TGF2Poly;
    P, Q, D1, D2, H1, H2 : TGF2NEPoint;
    TickCount            : Int64;
    cd                   : array[1..6] of Extended;
begin
  ec := nil;
  A := nil;
  B := nil;
  P := nil;
  Q := nil;
  D1 := nil;
  D2 := nil;
  H1 := nil;
  H2 := nil;
  with Memo1.Lines do
  try
    //----- Create an elliptic curve over GF(2^N)
    ec := TGF2NECurve.Create;

    //----- Initialize the curve (field + coefs)
    // SetField may raise an exception if the polynomial is not an
    // irreducible trinomial or pentanomial over GF(2)
    // !!! Set the Check parameter to TRUE if you are not sure that the
    // polynomial is irreducible
    ec.SetField([199,34,0],false);
  //ec.SetField([197,9,4,2,0],false);

    ec.SetRndCoefs;

    //----- Create 2 polynomials equal to 0
    A := TGF2Poly.Create;
    B := TGF2Poly.Create;

    //------ Display the curve coefficients
    Add('Curve coefficients and modulus:');
    Add('-------------------------------');
    ec.GetCoefs(A,B);
    Add('A = ' + GF2HexStr(A,BLOCKS));
    Add('B = ' + GF2HexStr(B,BLOCKS));
    ec.GetModulus(A);
    Add('M = ' + GF2HexStr(A,BLOCKS));
    Add('');

    //----- Create 6 points equal to O (identity element)
    P := TGF2NEPoint.Create;
    Q := TGF2NEPoint.Create;
    D1 := TGF2NEPoint.Create;
    D2 := TGF2NEPoint.Create;
    H1 := TGF2NEPoint.Create;
    H2 := TGF2NEPoint.Create;

    // What follows show how to compute a point Q from a point P
    // such that (Q <> P) and (2Q = 2P)

    //----- Random P
    TickCount := HITTimer;
    ec.GetRndPoint(P);
    cd[1] := HITTimer(TickCount);

    //----- Get the only point of order 2 of the curve
    TickCount := HITTimer;
    ec.GetOrder2Point(Q);
    cd[2] := HITTimer(TickCount);

    //----- Compute Q := Q + P (Since 2Q = O, we'll have 2(Q + P) = 2P
    TickCount := HITTimer;
    ec.Add(Q,P);
    cd[3] := HITTimer(TickCount);

    //------ Display P and Q
    Add('Random point P:');
    Add('---------------');
    Add('X = ' + GF2HexStr(P.X,BLOCKS));
    Add('Y = ' + GF2HexStr(P.Y,BLOCKS));
    Add('');
    Add('Q = P + (0,SqrRoot(B)):');
    Add('-----------------------');
    Add('X = ' + GF2HexStr(Q.X,BLOCKS));
    Add('Y = ' + GF2HexStr(Q.Y,BLOCKS));
    Add('');

    //----- Double the 2 points
    D1.Copy(P);
    TickCount := HITTimer;
    ec.Double(D1);
    cd[4] := HITTimer(TickCount);
    D2.Copy(Q);
    TickCount := HITTimer;
    ec.Double(D2);
    cd[5] := HITTimer(TickCount);

    //------ Display (we should have P = Q)
    Add('2P:');
    Add('---');
    Add('X = ' + GF2HexStr(D1.X,BLOCKS));
    Add('Y = ' + GF2HexStr(D1.Y,BLOCKS));
    Add('');
    Add('2Q:');
    Add('---');
    Add('X = ' + GF2HexStr(D2.X,BLOCKS));
    Add('Y = ' + GF2HexStr(D2.Y,BLOCKS));
    Add('');

    //----- Compute the 2 halves of 2P (the result should be P and Q)
    TickCount := HITTimer;
    ec.Halves(H1,H2,D1);
    cd[6] := HITTimer(TickCount);

    //------ Display
    Add('H1:');
    Add('---');
    Add('X = ' + GF2HexStr(H1.X,BLOCKS));
    Add('Y = ' + GF2HexStr(H1.Y,BLOCKS));
    Add('');
    Add('H2:');
    Add('---');
    Add('X = ' + GF2HexStr(H2.X,BLOCKS));
    Add('Y = ' + GF2HexStr(H2.Y,BLOCKS));
    Add('');

    //===== Check
    Add('Check (2P = 2Q) -> ' + CHECKMSG[D1.Equ(D2)]);
    Add('Check halves -> ' + CHECKMSG[(H1.Equ(P) and H2.Equ(Q)) or
                                      (H2.Equ(P) and H1.Equ(Q))]);
    Add('');

    //===== Running times
    Add('Running times (milliseconds):');
    Add('-----------------------------');
    Add('Random point  ' + FloatToStrF(cd[1],ffFixed,18,3));
    Add('Order-2 point ' + FloatToStrF(cd[2],ffFixed,18,3));
    Add('Addition .... ' + FloatToStrF(cd[3],ffFixed,18,3));
    Add('Double ...... ' + FloatToStrF(cd[4],ffFixed,18,3));
    Add('Double ...... ' + FloatToStrF(cd[5],ffFixed,18,3));
    Add('Halves ...... ' + FloatToStrF(cd[6],ffFixed,18,3));
    Add('');

  finally
    //----- Free the curve
    ec.Free;
    //----- Free the polynomials
    A.Free;
    B.Free;
    //----- Free the points
    P.Free;
    Q.Free;
    D1.Free;
    D2.Free;
    H1.Free;
    H2.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example N
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Multiplication in Elliptic Curves over GF(P)
//==============================================================================
procedure TForm1.RunN(Sender: TObject);
  //----------------------------------------------------
  // Display the point P
  //----------------------------------------------------
  procedure DisplayPoint(P: TGFPEPoint; ec: TGFPECurve);
  begin
    with Memo1.Lines do
    begin
      if P.IsO then Add('Point at infinity')
      else
      begin
        Add('X =' + IStr(P.X,10,5));
        Add('Y =' + IStr(P.Y,10,5));
        if not ec.OnCurve(P) then
          Add('An error occured: The point is not on the curve!?');
      end;
      Add('');
    end;
  end;
  //----------------------------------------------------
  const
    JACOBIAN_COORDINATES = TRUE;
    BOOL : array[Boolean] of string[5] = ('FALSE','TRUE');
  var
    ec            : TGFPECurve;
    p1, p2        : TGFPEPoint;
    P, A, B, N, K : TInteger;
    TickCount     : Int64;
    cd            : array[1..5] of Extended;
begin
  ec := nil;
  p1 := nil;
  p2 := nil;
  P := nil;
  A := nil;
  B := nil;
  N := nil;
  K := nil;
  with Memo1.Lines do
  try
    //----- Create integers
    P := TInteger.Create;
    A := TInteger.Create;
    B := TInteger.Create;
    N := TInteger.Create;
    K := TInteger.Create;

    //===== Data
    // Elliptic curve: Y^2 = (X^3 + AX + B) mod P
    // This curve, that has order N, is a curve of "nearly" prime order.
    // N = 2^3 * 5 * Q, with Q prime
    //----- Assign values
    ISet(A,'041352927055184874031446617136206415722357986600930228476278',10);
    ISet(B,'014528684559820132426384121471094834872846186397178552086639',10);
    ISet(P,'120140331550274557544369149720257908456906000440432717419817',10);
    ISet(N,'120140331550274557544369149719741744758534518192873862189080',10);

    //----- Display
    Add('Curve coefficients, modulus and order:');
    Add('--------------------------------------');
    Add('A =' + IStr(A,10,5));
    Add('B =' + IStr(B,10,5));
    Add('P =' + IStr(P,10,5));
    Add('N =' + IStr(N,10,5));
    Add('');

    //----- Create an elliptic curve over GF(P)
    ec := TGFPECurve.Create;

    //----- Initialize the curve (prime + coefs)
    ec.SetModulus(P);
    ec.SetCoefs(A,B);

    //----- Create 2 points equal to "O" (identity element)
    p1 := TGFPEPoint.Create;
    p2 := TGFPEPoint.Create;

    //----- Get a random point (different from O) on the curve
    TickCount := HITTimer;
    ec.GetRndPoint(p1);
    cd[1] := HITTimer(TickCount);

    //----- Display
    Add('Random point p1:');
    Add('----------------');
    DisplayPoint(p1,ec);

    // What follows show how to compute a point p2 equal to p1*K and then
    // to re-compute p1 from p2

    //----- Random K such that 1 < K < N and GCD(K,N) = 1
    TickCount := HITTimer;
    repeat
      IRnd(K,P.BitSize);
      IMod(K,N);
    until (ICmp(K,1) > 0) and IGCD1(K,N);
    cd[2] := HITTimer(TickCount);

    //----- Compute p2 := p1*K
    p2.Copy(p1);
    TickCount := HITTimer;
    ec.Mul(p2,K,JACOBIAN_COORDINATES);
    cd[3] := HITTimer(TickCount);

    //----- Display
    Add('Point p2 = p1 * K:');
    Add('------------------');
    DisplayPoint(p2,ec);

    //----- Compute p2 := p2*(K^(-1) mod N)
    TickCount := HITTimer;
    IInvMod(K,N);
    cd[4] := HITTimer(TickCount);

    TickCount := HITTimer;
    ec.Mul(p2,K,JACOBIAN_COORDINATES);
    cd[5] := HITTimer(TickCount);

    //----- Display
    Add('Point p1 = p2 * (K^(-1) mod N):');
    Add('-------------------------------');
    DisplayPoint(p2,ec);

    //===== Check
    Add('Check -> ' + CHECKMSG[p1.Equ(p2)]);
    Add('');

    //===== Running times
    Add('Running times (milliseconds):');
    Add('-----------------------------');
    Add('Random point ........ ' + FloatToStrF(cd[1],ffFixed,18,3));
    Add('Random multiplicator  ' + FloatToStrF(cd[2],ffFixed,18,3));
    Add('Mul ................. ' + FloatToStrF(cd[3],ffFixed,18,3));
    Add('Inverse multiplicator ' + FloatToStrF(cd[4],ffFixed,18,3));
    Add('Mul ................. ' + FloatToStrF(cd[5],ffFixed,18,3));
    Add('');

  finally
    //----- Free the curve
    ec.Free;
    //----- Free the points
    p1.Free;
    p2.Free;
    //----- Free the integers
    P.Free;
    A.Free;
    B.Free;
    N.Free;
    K.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example O
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// This example is the same as the previous one, but it is over GF(2^N) and the
// used curve could still be a record.
//==============================================================================
procedure TForm1.RunO(Sender: TObject);
  const
    //----- Multiplications are faster (x3) when TRUE
    JACOBIAN_COORDINATES = TRUE;
    //----- For checks
    BOOL : array[Boolean] of string[5] = ('FALSE','TRUE');
    //
    // The curve data become from the PhD thesis (in French) of Reynald LERCIER
    // See page 210.
    // Title: "Algorithmique des Courbes Elliptiques dans les Corps Finis"
    // At:    ftp://lix.polytechnique.fr/pub/lercier/papers/these.ps.Z
    //
    // Elliptic curve: Y^2 + XY = X^3 + AX^2 + B over GF(2^N)
    // Field modulus = t^1301 + t^11 + t^10 + t + 1
    // Coef A = 0
    // Coef B = t^16 + t^14 + t^13 + t^9 + t^8 + t^7 + t^6 + t^5 + t^4 + t^3
    // The order of the curve is Z = 2^1301 + 1 + d
    //
    d = '1252020466045855099434238134741436458098159760255154193747525' +
        '1045091682467090516587665003925037001223752821603011654792837' +
        '2408257614374530929254301621371725666282688807678760468051211' +
        '77132601520639';

  //------------------------------------------------------
  // Display the point P
  //------------------------------------------------------
  procedure DisplayPoint(P: TGF2NEPoint; ec: TGF2NECurve);
  begin
    with Memo1.Lines do
    if P.IsO then Add('Point at infinity')
    else
    begin
      Add('X =  ' + GF2HexStr(P.X,4));
      Add('X.Degree = ' + IntToStr(P.X.Degree));
      Add('Y =  ' + GF2HexStr(P.Y,4));
      Add('Y.Degree = ' + IntToStr(P.Y.Degree));
      if not ec.OnCurve(P) then
        Add('An error occurred: The point is not on the curve!?');
    end;
  end;
  //------------------------------------------------------

  var
    ec        : TGF2NECurve;
    p1, p2    : TGF2NEPoint;
    Z, K      : TInteger;
    TickCount : Int64;
    Chrono    : Extended;
begin
  ec := nil;
  p1 := nil;
  p2 := nil;
  Z := nil;
  K := nil;
  with Memo1.Lines do
  try
    //----- Create an elliptic curve over GF(2^N)
    ec := TGF2NECurve.Create;

    //----- Init the curve
    //      Field modulus = t^1301 + t^11 + t^10 + t + 1
    //      The Check parameter is FALSE because we are sure that the field
    //      modulus is irreducible over GF(2)
    ec.SetField([1301,11,10,1,0],false);
    //      A = 0
    //      B = t^16 + t^14 + t^13 + t^9 + t^8 + t^7 + t^6 + t^5 + t^4 + t^3
    ec.SetCoefs([],[16,14,13,9,8,7,6,5,4,3]);

    //----- Create integers
    Z := TInteger.Create;
    K := TInteger.Create;

    //----- Z = 2^1301 + 1 + d
    Z.Bit[1301] := true; //okay since Z = 0
    IInc(Z);
    ISet(K,d,10);
    IAdd(Z,K);

    //----- Display
    Add('Order of the curve:');
    Add('-------------------');
    Add('Z =' + IStr(Z,10,3));
    Add('Decimal size = ' + IntToStr(Z.DecimalSize));
    Add('');

    // What follows show how to compute a point p2 equal to p1*K and then
    // to re-compute p1 from p2

    //----- Create 2 points equal to "O" (identity element)
    p1 := TGF2NEPoint.Create;
    p2 := TGF2NEPoint.Create;

    //----- Allow to abort calculations
    //  Computations are not instant, WE ARE WORKING WITH A RECORD!
    //  On CYRIX 300, one elliptic curve multiplication takes about:
    //  - 15s with the parameter Jacobian = FALSE
    //  -  6s with the parameter Jacobian = TRUE
    HITBreakWith(Application.ProcessMessages);
    try
      //----- Activate the Abort button (if need be)
      AbortBtn.Enabled := HITBreakOn;

      Add('Computing a random point point p1 on the curve ...');

      //----- Get a random point (different from O) on the curve
      TickCount := HITTimer;
      ec.GetRndPoint(p1);
      Chrono := HITTimer(TickCount);

      //----- Display
      Delete(Count-1);
      Add('Random point p1:');
      Add('----------------');
      DisplayPoint(p1,ec);
      Add('Running time ' + FloatToStrF(Chrono,ffFixed,18,3) + ' ms');
      Add('');

      //----- Random K such that 1 < K < N and GCD(K,N) = 1
      Add('Computing a random multiplicator K ...');
      TickCount := HITTimer;
      repeat
        IRnd(K,2000);
        IMod(K,Z);
      until (ICmp(K,1) > 0) and IGCD1(K,Z);
      Chrono := HITTimer(TickCount);
      //----- Display
      Delete(Count-1);
      Add('Random multiplicator:');
      Add('---------------------');
      Add('K =' + IStr(K,10,3));
      Add('Decimal size = ' + IntToStr(K.DecimalSize));
      Add('Running time ' + FloatToStrF(Chrono,ffFixed,18,3) + ' ms');
      Add('');

      //----- Compute p2 := p1*K
      Add('Computing p2 := p1 * K ...');
      p2.Copy(p1);
      TickCount := HITTimer;
      ec.Mul(p2,K,JACOBIAN_COORDINATES);
      Chrono := HITTimer(TickCount);

      //----- Display
      Delete(Count-1);
      Add('Point p2 = p1 * K:');
      Add('------------------');
      DisplayPoint(p2,ec);
      Add('Running time ' + FloatToStrF(Chrono,ffFixed,18,3) + ' ms');
      Add('');

      //----- Compute K := K^(-1) mod Z
      Add('Compute K^(-1) mod Z ...');
      TickCount := HITTimer;
      IInvMod(K,Z);
      Chrono := HITTimer(TickCount);

      //----- Display
      Delete(Count-1);
      Add('H := K^(-1) mod Z:');
      Add('------------------');
      Add('H =' + IStr(K,10,3));
      Add('Decimal size = ' + IntToStr(K.DecimalSize));
      Add('Running time ' + FloatToStrF(Chrono,ffFixed,18,3) + ' ms');
      Add('');

      //----- Compute p2 := p2*(K^(-1) mod Z)
      Add('Computing p2 := p2 * H ...');
      TickCount := HITTimer;
      ec.Mul(p2,K,JACOBIAN_COORDINATES);
      Chrono := HITTimer(TickCount);

      //----- Display
      Delete(Count-1);
      Add('Point p2 = p2 * (K^(-1) mod Z):');
      Add('-------------------------------');
      DisplayPoint(p2,ec);
      Add('Running time ' + FloatToStrF(Chrono,ffFixed,18,3) + ' ms');
      Add('');

      //===== Check
      Add('Check (p1 = p2) -> ' + CHECKMSG[p1.Equ(p2)]);
      Add('');

    finally
      //----- Restore the default status
      HITNoBreak;
      AbortBtn.Enabled := false;
    end;

  finally
    //----- Free the curve
    ec.Free;
    //----- Free the integers
    Z.Free;
    K.Free;
    //----- Free the points
    p1.Free;
    p2.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example P
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Create a list containing all the primes less than 2^E
//==============================================================================
procedure TForm1.RunP(Sender: TObject);
  const E = 24; //Should be in 1..30
  var
    L         : TPrimeList;
    i, j      : Longint;
    TickCount : Int64;
    Chrono    : Extended;
begin
  L := nil;
  with Memo1.Lines do
  try
    //----- Create an empty list
    L := TPrimeList.Create;

    //----- Allow to abort calculations
    HITBreakWith(Application.ProcessMessages);
    try
      //----- Activate the Abort button (if need be)
      AbortBtn.Enabled := HITBreakOn;

      Add('Generating the prime list ...');

      //----- Generate the list
      //      Pi(x) = number of primes less than or equal to x
      TickCount := HITTimer;
      L.Count := Pi2Pow[E];
      Chrono := HITTimer(TickCount); //Chrono in milliseconds

    finally
      //----- Restore the default status
      HITNoBreak;
      AbortBtn.Enabled := false;
    end;

    //----- Display infos
    Delete(Count-1);
    Add('The list contains ' + IntToStr(L.Count) + ' primes (Mem size = ' +
        IntToStr(L.Count * 4) +  ' bytes)');
    Add('The greatest prime is ' + IntToStr(L.LastPrime));
    Add('');

    //----- Display the running time
    Add('Creation time = ' + FloatToStrF(Chrono,ffFixed,18,3) + ' milliseconds');
    Add('On average, HIT found ' +
         FloatToStrF(L.Count / Chrono,ffFixed,18,3) + ' primes by millisecond');
    Add('');

    //----- Display some primes
    j := (Longint(IRnd32 and $7fffffff) mod L.Count) + 1;
    for i := 0 to 9 do
    begin
      Add('Prime[' + IntToStr(j) + '] = ' + IntToStr(L[j]));
      Inc(j);
      if j > L.Count then j := 1;
    end;
    Add('');
  finally
    //----- Free the list
    L.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example Q
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
//==============================================================================
procedure TForm1.RunQ(Sender: TObject);
  var
    P         : TGF2Poly;
    F         : TDynTGF2PolyArray;
    S         : string;
    i         : Longint;
    TickCount : Int64;
    Chrono    : Extended;
begin
  P := nil;
  with Memo1.Lines do
  try
    //----- Create a null polynomial
    P := TGF2Poly.Create;

    //----- Set the polynomial equal to X^32768 + X (A big polynomial!)
    P[32768] := true;
    P[1] := true;

    //----- Allow to abort calculations
    HITBreakWith(Application.ProcessMessages);
    try
      //----- Activate the Abort button (if need be)
      AbortBtn.Enabled := HITBreakOn;

      Add('Factorization in progress ...');

      //----- Factorization of P
      TickCount := HITTimer;
      GF2Split(P,F,true);
      Chrono := HITTimer(TickCount);

    finally
      //----- Restore the default status
      HITNoBreak;
      AbortBtn.Enabled := false;
    end;

    //----- Display the 2192 factors (and free them)
    Delete(Count-1);
    S := '';
    for i := 0 to Length(F)-1 do S := S + GF2HexStr(F[i],4) + ', ';
    Add(S);
    Add('');

    //----- Display infos
    Add(IntToStr(Length(F)) + ' factors');
    Add('Running time = ' + FloatToStrF(Chrono,ffFixed,18,3) + ' milliseconds');
    Add('');
  finally
    //----- Free the factors
    GF2FreeTGF2PolyArray(F);
    //----- Free the polynomial
    P.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Open the Popup menu
//==============================================================================
procedure TForm1.MenuBtnClick(Sender: TObject);
  var P : TPoint;
begin
  MenuBtn.Enabled := false;
  try
    GetCursorPos(P);
    PopupMenu1.Popup(P.X,P.Y);
  finally
    MenuBtn.Enabled := true;
  end;
end;

//==============================================================================
// Call one of the methods Run
//==============================================================================
procedure TForm1.Run(Sender: TObject; Num: Char);
begin
  MenuBtn.Enabled := false;
  ClearBtn.Enabled := false;
  CloseBtn.Enabled := false;
  HelpBtn.Enabled := false;
  PopupMenu1.AutoPopup := false;
  try
    Memo1.Lines.Add(#13#10'Example ' + Num);
    Memo1.Lines.Add('---------'#13#10);
    case Num of
      'A': RunA(Sender);
      'B': RunB(Sender);
      'C': RunC(Sender);
      'D': RunD(Sender);
      'E': RunE(Sender);
      'F': RunF(Sender);
      'G': RunG(Sender);
      'H': RunH(Sender);
      'I': RunI(Sender);
      'J': RunJ(Sender);
      'K': RunK(Sender);
      'L': RunL(Sender);
      'M': RunM(Sender);
      'N': RunN(Sender);
      'O': RunO(Sender);
      'P': RunP(Sender);
      'Q': RunQ(Sender);
    end;
  finally
    MenuBtn.Enabled := true;
    ClearBtn.Enabled := true;
    CloseBtn.Enabled := true;
    HelpBtn.Enabled := true;
    PopupMenu1.AutoPopup := true;
  end;
end;

//==============================================================================
// Call RunA method
//==============================================================================
procedure TForm1.PopItemAClick(Sender: TObject);
begin
  Run(Sender,'A');
end;

//==============================================================================
// Call RunB method
//==============================================================================
procedure TForm1.PopItemBClick(Sender: TObject);
begin
  Run(Sender,'B');
end;

//==============================================================================
// Call RunC method
//==============================================================================
procedure TForm1.PopItemCClick(Sender: TObject);
begin
  Run(Sender,'C');
end;

//==============================================================================
// Call RunD method
//==============================================================================
procedure TForm1.PopItemDClick(Sender: TObject);
begin
  Run(Sender,'D');
end;

//==============================================================================
// Call RunE method
//==============================================================================
procedure TForm1.PopItemEClick(Sender: TObject);
begin
  Run(Sender,'E');
end;

//==============================================================================
// Call RunF method
//==============================================================================
procedure TForm1.PopItemFClick(Sender: TObject);
begin
  Run(Sender,'F');
end;

//==============================================================================
// Call RunG method
//==============================================================================
procedure TForm1.PopItemGClick(Sender: TObject);
begin
  Run(Sender,'G');
end;

//==============================================================================
// Call RunH method
//==============================================================================
procedure TForm1.PopItemHClick(Sender: TObject);
begin
  Run(Sender,'H');
end;

//==============================================================================
// Call RunI method
//==============================================================================
procedure TForm1.PopItemIClick(Sender: TObject);
begin
  Run(Sender,'I');
end;

//==============================================================================
// Call RunJ method
//==============================================================================
procedure TForm1.PopItemJClick(Sender: TObject);
begin
  Run(Sender,'J');
end;

//==============================================================================
// Call RunK method
//==============================================================================
procedure TForm1.PopItemKClick(Sender: TObject);
begin
  Run(Sender,'K');
end;

//==============================================================================
// Call RunL method
//==============================================================================
procedure TForm1.PopItemLClick(Sender: TObject);
begin
  Run(Sender,'L');
end;

//==============================================================================
// Call RunM method
//==============================================================================
procedure TForm1.PopItemMClick(Sender: TObject);
begin
  Run(Sender,'M');
end;

//==============================================================================
// Call RunN method
//==============================================================================
procedure TForm1.PopItemNClick(Sender: TObject);
begin
  Run(Sender,'N');
end;

//==============================================================================
// Call RunO method
//==============================================================================
procedure TForm1.PopItemOClick(Sender: TObject);
begin
  Run(Sender,'O');
end;

//==============================================================================
// Call RunP method
//==============================================================================
procedure TForm1.PopItemPClick(Sender: TObject);
begin
  Run(Sender,'P');
end;

//==============================================================================
// Call RunQ method
//==============================================================================
procedure TForm1.PopItemQClick(Sender: TObject);
begin
  Run(Sender,'Q');
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

