{This is just a simple TGraphic control.  It draws a gradient that can be
 used as the background on a Form by setting the Align property to alClient.
 It only looks good in 16 or 24 bit color.  It has an optional Ctl3D
 property to draw a 'frame' around itself

 Version 1.2 - 9/98 - Clean up code, re-release.

 Version 1.1 - 2/98 - Added Radial Gradients

 Version 1.0 - 1/98 - Initial release to the public
 Harm:  harmans@.uswest.net}
unit HarmGradient;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
    Math, ExtCtrls;

type
  TGradientStyle = (gsHorizontal, gsVertical, gsElliptic, gsRectangle,
                    gsVertCenter, gsHorizCenter, gsRadial1, gsRadial2);

  THarmGradient = class(TGraphicControl)
  private
    { Private declarations }
    FBeginClr : TColor;    //Start color for gradient
    FEndClr : TColor;     //End color for gradient
    FGradientStyle : TGradientStyle;  //One of six choices
    FSteps : integer;  //Number of color iterations
    F3D : Boolean;   //3d ctl t/f
    FAngle : integer;  //starting angle for radial gradients
    FRedraw : Boolean;  //RePaint or just BitBlt?
    function Muldv(a,b,c : integer) : longint;
    procedure Paint; override;  //Override Paint - entire control is painted.
    procedure WMEraseBkgnd(Var Msg : TMessage); message WM_ERASEBKGND;
    procedure SetBeginClr(Value : TColor);  //Set start color of gradient
    procedure SetEndClr(Value : TColor);    //Set end color of gradient
    procedure SetGradient(Value : TGradientStyle);  //Choose gradient fill style
    procedure SetSteps(Value : integer); //Choose number of color iterations
    procedure Set3D(Value : Boolean);   //Toggle 3d
    procedure SetAngle(Value : integer); //Set start angle for radial
    procedure DoHorizontal(fr, fg, fb, dr, dg, db : Integer); //These all draw
    procedure DoVertical(fr, fg, fb, dr, dg, db : Integer); // the gradients.
    procedure DoElliptic(fr, fg, fb, dr, dg, db : Integer); // The fr fg fb etc
    procedure DoRectangle(fr, fg, fb, dr, dg, db : Integer); // are color values.
    procedure DoVertCenter(fr, fg, fb, dr, dg, db : Integer);
    procedure DoHorizCenter(fr, fg, fb, dr, dg, db : Integer);
    procedure DoRadial1(fr, fg, fb, dr, dg, db : Integer);
    procedure DoRadial2(fr, fg, fb, dr, dg, db : Integer);
  protected
    { Protected declarations }
  public
    { Public declarations }
    bm : TBitmap;          //WorkHorse internal bmp
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    property Canvas;
  published
    { Published declarations }
    property BeginColor : TColor read FBeginClr
       write SetBeginClr default clNavy;
    property Ctl3D : Boolean read F3D
       write Set3D default FALSE;
    property EndColor : TColor read FEndClr
       write SetEndClr default clAqua;
    property GradientStyle : TGradientStyle read FGradientStyle
       write SetGradient default gsVertical;
    property Steps : Integer read Fsteps
       write SetSteps default 255;
    property StartAngle : Integer read FAngle
       write SetAngle default 0;
    property Align;
    property Visible;
  end;

procedure Register;

implementation

constructor THarmGradient.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FBeginClr := clNavy;
  FEndClr := clAqua;
  Height := 75;
  Width := 75;
  bm := TBitmap.Create;
  bm.Canvas.Pen.Style:= psClear;
  bm.Canvas.Pen.Mode:= pmCopy;
  ControlStyle := ControlStyle + [csOpaque,csNoStdEvents];
  FGradientStyle := gsVertical;
  FSteps := 255;
  F3D := FALSE;
  FAngle := 0;
  FRedraw := TRUE;
end;

//This is an attempt to speed up some of the processing.  I don't think
// it helps.
function THarmGradient.Muldv(a,b,c : integer) : longint;
ASM

  MOV EAX, a
  IMUL b
  IDIV c

end;

//Paint is a huge procedure.  The bigger the gradient size, the slower
// the paint.
procedure THarmGradient.Paint;
var
  FromR, FromG, FromB : Integer;
  DiffR, DiffG, DiffB : Integer;
  rct : TRect;
{To speed things up and reduce flicker, I use a Bitmap to draw the gradient in
 its entirety, then BitBlt it to the canvas of the control.}
begin
  if (Width = bm.Width) and (Height = bm.Height) then
    if FRedraw = FALSE then begin //If size hasn't changed, don't redraw.
      BitBlt(Canvas.Handle,0,0,Width,Height,bm.Canvas.Handle,0,0,SRCCOPY);
      Exit;
    end;
  bm.Width := Width;          //Set BMP dimensions to match control's
  bm.Height := Height;
  rct := Rect(0,0,Width,Height);  //Set rectangle size for later use
  FromR := FBeginClr and $000000ff;  //Strip out separate RGB values
  FromG := (FBeginClr shr 8) and $000000ff;
  FromB := (FBeginClr shr 16) and $000000ff;
  DiffR := (FEndClr and $000000ff) - FromR;   //Find the difference
  DiffG := ((FEndClr shr 8) and $000000ff) - FromG;
  DiffB := ((FEndClr shr 16) and $000000ff) - FromB;
  //Depending on gradient style selected, go draw it on the Bitmap canvas.
  case FGradientStyle of
    gsHorizontal  : DoHorizontal(FromR, FromG, FromB, DiffR, DiffG, DiffB);
    gsVertical    : DoVertical(FromR, FromG, FromB, DiffR, DiffG, DiffB);
    gsElliptic    : DoElliptic(FromR, FromG, FromB, DiffR, DiffG, DiffB);
    gsRectangle   : DoRectangle(FromR, FromG, FromB, DiffR, DiffG, DiffB);
    gsVertCenter  : DoVertCenter(FromR, FromG, FromB, DiffR, DiffG, DiffB);
    gsHorizCenter : DoHorizCenter(FromR, FromG, FromB, DiffR, DiffG, DiffB);
    gsRadial1     : DoRadial1(FromR, FromG, FromB, DiffR, DiffG, DiffB);
    gsRadial2     : DoRadial2(FromR, FromG, FromB, DiffR, DiffG, DiffB);
  end;
  if F3D = TRUE then begin   //Do a 3D frame.  Only inset version for now.
    bm.Canvas.Pen.Style := psSolid;
    Frame3D(bm.Canvas,rct,clGray,clWhite,2);
    InflateRect(rct,+1,+1);
    Frame3D(bm.Canvas,rct,clBlack,clSilver,1);
    bm.Canvas.Pen.Style := psClear;
  end;
  BitBlt(Canvas.Handle,0,0,Width,Height,bm.Canvas.Handle,0,0,SRCCOPY);
  FRedraw := FALSE;    //This helps sped things up
end;

procedure THarmGradient.WMEraseBkgnd(var Msg:TMessage);
begin
  Msg.Result := 1;  //We intercept this message and tell
end;                // Windows it processed OK.

procedure THarmGradient.SetBeginClr(Value : TColor);
begin
  if FBeginClr <> Value then begin
    FBeginClr := Value;
    FRedraw := TRUE;
    Invalidate;
  end;
end;

procedure THarmGradient.SetEndClr(Value : TColor);
begin
  if FEndClr <> Value then begin
    FEndClr := Value;
    FRedraw := TRUE;
    Invalidate;
  end;
end;

procedure THarmGradient.SetGradient(Value : TGradientStyle);
begin
  if FGradientStyle <> Value then begin
    FGradientStyle := Value;
    FRedraw := TRUE;
    Invalidate;
  end;
end;

procedure THarmGradient.Set3D(Value : Boolean);
begin
  if F3D <> Value then begin
    F3D := Value;
    FRedraw := TRUE;
    Invalidate;
  end;
end;

procedure THarmGradient.SetSteps(Value : Integer);
begin
  if FSteps <> Value then begin
    FSteps := Value;
    FRedraw := TRUE;
    Invalidate;
  end;
end;

procedure THarmGradient.SetAngle(Value : Integer);
begin
  if FAngle <> Value then begin
    FAngle := Value;
    FRedraw := TRUE;
    Invalidate;
  end;
end;

procedure THarmGradient.DoHorizontal(fr, fg, fb, dr, dg, db : Integer);
var
  ColorRect: TRect;
  I: Integer;
  R, G, B : Byte;
begin
  ColorRect.Top:= 0;                //Set rectangle top
  ColorRect.Bottom := Height;
  for I := 0 to FSteps do begin         //Make lines (rectangles) of color
    ColorRect.Left:= Muldv (I, Width, FSteps+1);    //Find left for this color
    ColorRect.Right:= Muldv (I + 1, Width, FSteps+1);   //Find Right
    R := fr + Muldv(I, dr, FSteps);    //Find the RGB values
    G := fg + Muldv(I, dg, FSteps);
    B := fb + Muldv(I, db, FSteps);
    bm.Canvas.Brush.Color := RGB(R, G, B);   //Plug colors into brush
    bm.Canvas.FillRect(ColorRect);           //Draw on Bitmap
  end;
end;

procedure THarmGradient.DoVertical(fr, fg, fb, dr, dg, db : Integer);
var
  ColorRect: TRect;
  I: Integer;
  R, G, B : Byte;
begin
  ColorRect.Left:= 0;                //Set rectangle left&right
  ColorRect.Right:= Width;
  for I := 0 to FSteps do begin         //Make lines (rectangles) of color
    ColorRect.Top:= Muldv (I, Height, FSteps+1);    //Find top for this color
    ColorRect.Bottom:= Muldv (I + 1, Height, FSteps+1);   //Find Bottom
    R := fr + Muldv(I, dr, FSteps);    //Find the RGB values
    G := fg + Muldv(I, dg, FSteps);
    B := fb + Muldv(I, db, FSteps);
    bm.Canvas.Brush.Color := RGB(R, G, B);   //Plug colors into brush
    bm.Canvas.FillRect(ColorRect);           //Draw on Bitmap
  end;
end;

procedure THarmGradient.DoElliptic(fr, fg, fb, dr, dg, db : Integer);
var
  I: Integer;
  R, G, B : Byte;
  Pw, Ph : Real;
  x1,y1,x2,y2 : Real;
begin
 // bm.Canvas.Pen.Style := psClear;
 // bm.Canvas.Pen.Mode := pmCopy;
  x1 := 0 - (bm.Width / 4);
  x2 := bm.Width + (bm.Width / 4)+4;
  y1 := 0 - (bm.Height / 4);
  y2 := bm.Height + (bm.Height / 4)+4;
  Pw := ((bm.Width / 4) + (bm.Width / 2)) / 155;
  Ph := ((bm.Height / 4) + (bm.Height / 2)) / 155;
  for I := 0 to 155 do begin         //Make ellipses of color
    x1 := x1 + Pw;
    x2 := X2 - Pw;
    y1 := y1 + Ph;
    y2 := y2 - Ph;
    R := fr + Muldv(I, dr, 155);    //Find the RGB values
    G := fg + Muldv(I, dg, 155);
    B := fb + Muldv(I, db, 155);
    bm.Canvas.Brush.Color := R or (G shl 8) or (b shl 16);   //Plug colors into brush
    bm.Canvas.Ellipse(Trunc(x1),Trunc(y1),Trunc(x2),Trunc(y2));
  end;
 // bm.Canvas.Pen.Style := psSolid;
end;

procedure THarmGradient.DoRectangle(fr, fg, fb, dr, dg, db : Integer);
var
  I: Integer;
  R, G, B : Byte;
  Pw, Ph : Real;
  x1,y1,x2,y2 : Real;
begin
 // bm.Canvas.Pen.Style := psClear;
 // bm.Canvas.Pen.Mode := pmCopy;
  x1 := 0;
  x2 := bm.Width+2;
  y1 := 0;
  y2 := bm.Height+2;
  Pw := (bm.Width / 2) / FSteps;
  Ph := (bm.Height / 2) / FSteps;
  for I := 0 to FSteps do begin         //Make rectangles of color
    x1 := x1 + Pw;
    x2 := X2 - Pw;
    y1 := y1 + Ph;
    y2 := y2 - Ph;
    R := fr + Muldv(I, dr, FSteps);    //Find the RGB values
    G := fg + Muldv(I, dg, FSteps);
    B := fb + Muldv(I, db, FSteps);
    bm.Canvas.Brush.Color := RGB(R, G, B);   //Plug colors into brush
    bm.Canvas.FillRect(Rect(Trunc(x1),Trunc(y1),Trunc(x2),Trunc(y2)));
  end;
 // bm.Canvas.Pen.Style := psSolid;
end;

procedure THarmGradient.DoVertCenter(fr, fg, fb, dr, dg, db : Integer);
var
  ColorRect: TRect;
  I: Integer;
  R, G, B : Byte;
  Haf : Integer;
begin
  Haf := bm.Height Div 2;
  ColorRect.Left := 0;
  ColorRect.Right := bm.Width;
  for I := 0 to Haf do begin
    ColorRect.Top := Muldv (I, Haf, Haf);
    ColorRect.Bottom := Muldv (I + 1, Haf, Haf);
    R := fr + Muldv(I, dr, Haf);
    G := fg + Muldv(I, dg, Haf);
    B := fb + Muldv(I, db, Haf);
    bm.Canvas.Brush.Color := RGB(R, G, B);
    bm.Canvas.FillRect(ColorRect);
    ColorRect.Top := bm.Height - (Muldv (I, Haf, Haf));
    ColorRect.Bottom := bm.Height - (Muldv (I + 1, Haf, Haf));
    bm.Canvas.FillRect(ColorRect);
  end;
end;

procedure THarmGradient.DoHorizCenter(fr, fg, fb, dr, dg, db : Integer);
var
  ColorRect: TRect;
  I: Integer;
  R, G, B : Byte;
  Haf : Integer;
begin
  Haf := bm.Width Div 2;
  ColorRect.Top := 0;
  ColorRect.Bottom := bm.Height;
  for I := 0 to Haf do begin
    ColorRect.Left := Muldv (I, Haf, Haf);
    ColorRect.Right := Muldv (I + 1, Haf, Haf);
    R := fr + Muldv(I, dr, Haf);
    G := fg + Muldv(I, dg, Haf);
    B := fb + Muldv(I, db, Haf);
    bm.Canvas.Brush.Color := RGB(R, G, B);
    bm.Canvas.FillRect(ColorRect);
    ColorRect.Left := bm.Width - (Muldv (I, Haf, Haf));
    ColorRect.Right := bm.Width - (Muldv (I + 1, Haf, Haf));
    bm.Canvas.FillRect(ColorRect);
  end;
end;

procedure THarmGradient.DoRadial1(fr, fg, fb, dr, dg, db : Integer);
var
  R, G, B : Byte;
  I : integer;
  x1,y1,x2,y2 : integer;
  x3,y3,x4,y4 : integer;
  Rd : integer;
  t1, t2 : integer;
begin
  x1 := 0 - (bm.Width Div 4);
  y1 := 0 - (bm.Height Div 4);
  x2 := bm.Width + (bm.Width Div 4);
  y2 := bm.Height + (bm.Height Div 4);
  t1 := bm.Width Div 2;
  t2 := bm.Height Div 2;
  if bm.Width > bm.Height then
   Rd := bm.Width Div 2
  else
   Rd := bm.Height Div 2;

  for I := FAngle to FAngle+179 do begin  //First
    R := fr + Muldv(I-FAngle, dr, 180);
    G := fg + Muldv(I-FAngle, dg, 180);
    B := fb + Muldv(I-FAngle, db, 180);

    x3 := t1 + Round(Sin(DegToRad(I)) * Rd);
    y3 := t2 - Round(Cos(DegToRad(I)) * Rd);
    x4 := t1 + Round(Sin(DegToRad(I+2)) * Rd);
    y4 := t2 - Round(Cos(DegToRad(I+2)) * Rd);

    bm.Canvas.Brush.Color := R or (G shl 8) or (B shl 16);
    bm.Canvas.Pie(x1,y1,x2,y2,x4,y4,x3,y3);
  end;
  for I := FAngle+180 to FAngle+359 do begin  //Second
    R := fr + Muldv(FAngle+359-I, dr, 180);
    G := fg + Muldv(FAngle+359-I, dg, 180);
    B := fb + Muldv(FAngle+359-I, db, 180);

    x3 := t1 + Round(Sin(DegToRad(I)) * Rd);
    y3 := t2 - Round(Cos(DegToRad(I)) * Rd);
    x4 := t1 + Round(Sin(DegToRad(I+2)) * Rd);
    y4 := t2 - Round(Cos(DegToRad(I+2)) * Rd);

    bm.Canvas.Brush.Color := R or (G shl 8) or (B shl 16);
    bm.Canvas.Pie(x1,y1,x2,y2,x4,y4,x3,y3);
  end;

end;

procedure THarmGradient.DoRadial2(fr, fg, fb, dr, dg, db : Integer);
var
  R, G, B : Byte;
  I : integer;
  x1,y1,x2,y2 : integer;
  x3,y3,x4,y4 : integer;
  Rd : integer;
  t1, t2 : integer;
begin
  x1 := 0 - (bm.Width Div 4);
  y1 := 0 - (bm.Height Div 4);
  x2 := bm.Width + (bm.Width Div 4);
  y2 := bm.Height + (bm.Height Div 4);
  t1 := bm.Width Div 2;
  t2 := bm.Height Div 2;
  if bm.Width > bm.Height then
   Rd := bm.Width Div 2
  else
   Rd := bm.Height Div 2;

  for I := FAngle to FAngle+89 do begin  //First
    R := fr + Muldv(I-FAngle, dr, 90);
    G := fg + Muldv(I-FAngle, dg, 90);
    B := fb + Muldv(I-FAngle, db, 90);

    x3 := t1 + Round(Sin(DegToRad(I)) * Rd);
    y3 := t2 - Round(Cos(DegToRad(I)) * Rd);
    x4 := t1 + Round(Sin(DegToRad(I+2)) * Rd);
    y4 := t2 - Round(Cos(DegToRad(I+2)) * Rd);

    bm.Canvas.Brush.Color := R or (G shl 8) or (B shl 16);
    bm.Canvas.Pie(x1,y1,x2,y2,x4,y4,x3,y3);
  end;
  for I := FAngle+90 to FAngle+179 do begin
    R := fr + Muldv(FAngle+179-I, dr, 90);
    G := fg + Muldv(FAngle+179-I, dg, 90);
    B := fb + Muldv(FAngle+179-I, db, 90);

    x3 := t1 + Round(Sin(DegToRad(I)) * Rd);
    y3 := t2 - Round(Cos(DegToRad(I)) * Rd);
    x4 := t1 + Round(Sin(DegToRad(I+2)) * Rd);
    y4 := t2 - Round(Cos(DegToRad(I+2)) * Rd);

    bm.Canvas.Brush.Color := R or (G shl 8) or (B shl 16);
    bm.Canvas.Pie(x1,y1,x2,y2,x4,y4,x3,y3);
  end;
  for I := FAngle+180 to FAngle+269 do begin  //3rd
    R := fr + Muldv(I-FAngle-180, dr, 90);
    G := fg + Muldv(I-FAngle-180, dg, 90);
    B := fb + Muldv(I-FAngle-180, db, 90);

    x3 := t1 + Round(Sin(DegToRad(I)) * Rd);
    y3 := t2 - Round(Cos(DegToRad(I)) * Rd);
    x4 := t1 + Round(Sin(DegToRad(I+2)) * Rd);
    y4 := t2 - Round(Cos(DegToRad(I+2)) * Rd);

    bm.Canvas.Brush.Color := R or (G shl 8) or (B shl 16);
    bm.Canvas.Pie(x1,y1,x2,y2,x4,y4,x3,y3);
  end;
  for I := FAngle+270 to FAngle+359 do begin    //4th
    R := fr + Muldv(FAngle+359-I, dr, 90);
    G := fg + Muldv(FAngle+359-I, dg, 90);
    B := fb + Muldv(FAngle+359-I, db, 90);

    x3 := t1 + Round(Sin(DegToRad(I)) * Rd);
    y3 := t2 - Round(Cos(DegToRad(I)) * Rd);
    x4 := t1 + Round(Sin(DegToRad(I+2)) * Rd);
    y4 := t2 - Round(Cos(DegToRad(I+2)) * Rd);

    bm.Canvas.Brush.Color := R or (G shl 8) or (B shl 16);
    bm.Canvas.Pie(x1,y1,x2,y2,x4,y4,x3,y3);
  end;

end;

destructor THarmGradient.Destroy;
begin
  bm.Free;       //Cleanup the bitmap - no memory leaks
  inherited Destroy;
end;

procedure Register;
begin
  RegisterComponents('Samples', [THarmGradient]);
end;

end.
