{$define LITE}

unit ElHeader;

interface

{$I elpack.inc}
{$R ElHeader.res}
{$B-}
{$RANGECHECKS OFF}
{$ALIGN ON}
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
{$IFNDEF LITE}
  ElIni,
  ElImgLst,
{$ENDIF}
{$IFDEF D_4_UP}
  ImgList,
{$ENDIF}
  ElTools, ElList, Menus, StdCtrls, ElVCLUtils;

type TElSectionStyle = (ElhsText, ElhsOwnerDraw, ElhsPictureOnly);
type TElSSortMode = (hsmNone, hsmAscend, hsmDescend);
type TElSAlignment = (hsaLeft, hsaCenter, hsaRight);
type TElHResizingStates  = (trsBegin, trsMove, trsEnd);
type TSectionChangeMode = (scmCaption, scmFieldName, scmFieldType,
                          scmAlign, scmStyle, scmEditable, scmPassword);

type TFieldTypes  = (sftCustom, // custom field
                     sftText,   // text field
                     sftNumber, // Integer numbers
                     sftFloating, // Floating numbers
                     sftDateTime, // DateTime
                     sftDate,     // Date
                     sftTime,     // Time
                     sftPicture,  // Picture
                     sftEnum,     // enumeration
                     sftBLOB,     // BLOB data
                     sftUndef     // undefined field
                    );

type EElError = class(Exception)
     end;

type TCustomElHeader = class;
     TElHeaderSection = class;
     TElHeaderSectionEvent = procedure (Sender:TCustomElHeader; Section : TElHeaderSection) of object;

     TElHeaderLookupEvent = procedure (Sender : TObject; Section : TElHeaderSection; var Text : string) of object;
     TElHeaderLookupDoneEvent = procedure (Sender : TObject;
                                Section : TElHeaderSection;
                                Text : string;
                                Accepted : boolean) of object;

     TElHeaderSection = class (TPersistent)
      private
       FUseMainStyle: Boolean;
       FFontColor: TColor;
       FParentColor: Boolean;
       FColor: TColor;
       FLookupEnabled: boolean;
       FExpandable   : Boolean;
       FExpanded     : Boolean;
       FParentSection: TElHeaderSection;
       FParentIdx    : integer;
       FPopupMenu : TPopupMenu;
       FPopupName : string;
       FResizable : Boolean;
       FClickSelect : Boolean;
       FProtected : Boolean;
       FFieldName : String;
       FFieldType : TFieldTypes;
       FEditable : Boolean;
       FImageIndex : Integer;
       FPicAlign : TElSAlignment;
       FTag     : integer;
       FVisible : boolean;
       FStyle   : TElSectionStyle;
       FWidth   : integer;
       FMinWidth,
       FMaxWidth: integer;
       FSortMode: TElSSortMode;
       FAllowClick: boolean;
       FAlignment: TElSAlignment;
       FText    :  string;
       FData    : pointer;
       FOwner   : TCustomElHeader;
       FLookupHist : TStringList;

       FStickySize : single;
       ASaveSize   : integer;
       procedure SetWidth(value:integer);
       function GetLeft:integer;
       function GetRight:integer;
       procedure SetMaxWidth(value:integer);
       procedure SetMinWidth(value:integer);
       procedure SetText(value:string);
       procedure SetStyle(value:TElSectionStyle);
       procedure SetSortMode(value:TElSSortMode);
       procedure SetAlignment(value:TElSAlignment);
       procedure SetVisible(value:boolean);
       function GetIndex:integer;
       function GetPosition : Integer;
       procedure SetImageIndex(newValue : Integer);
       procedure SetProtected(newValue : Boolean);
       procedure SetExpandable(newValue: Boolean);
       procedure SetExpanded(newValue: Boolean);
       procedure SetParentSection(newValue: TElHeaderSection);
       procedure SetPopupMenu(newValue: TPopupMenu);
       function GetVisible : boolean;
       procedure SetLookupEnabled(newValue : boolean);
       procedure SetParentColor(newValue: Boolean);
       procedure SetColor(newValue: TColor);
       procedure SetFontColor(newValue: TColor);
       procedure SetUseMainStyle(newValue: Boolean);
      protected
       procedure SetFieldName(newValue : String); virtual;
       procedure SetFieldType(newValue : TFieldTypes); virtual;
       procedure SetEditable(newValue : boolean); virtual;
       procedure SetResizable(newValue : boolean); virtual;
       procedure SetSaveSize(newValue : integer);
       property FSaveSize : integer read ASaveSize write SetSaveSize;
      public
       constructor Create(AOwner:TCustomElHeader);
       destructor Destroy; override;
       procedure UpdateSection;
       procedure Assign(source:TPersistent); override;
       property Index : integer read GetIndex;
       property Text : string read FText write SetText;
       property Data : pointer read FData write FData;
       property Style: TElSectionStyle read FStyle write SetStyle;
       property Left : integer read GetLeft;
       property Width: integer read FWidth write SetWidth;
       property Right: integer read GetRight;
       property MaxWidth: integer read FMaxWidth write SetMaxWidth;
       property MinWidth: integer read FMinWidth write SetMinWidth;
       property SortMode: TElSSortMode read FSortMode write SetSortMode;
       property AllowClick : boolean read FAllowClick write FAllowClick;
       property Alignment: TElSAlignment read FAlignment write SetAlignment;
       property PictureAlign : TElSAlignment read FPicAlign write FPicAlign;
       property Visible : boolean read GetVisible write SetVisible;
       property Position : Integer read GetPosition;  { Public }
       property ImageIndex : Integer read FImageIndex write SetImageIndex;  { Published }
       property FieldName : String read FFieldName write SetFieldName;  { Public }
       property FieldType : TFieldTypes read FFieldType write SetFieldType;  { Public }
       property Editable : Boolean read FEditable write SetEditable;  { Public }
       property Password : Boolean read FProtected write SetProtected default False;  { Public }
       property Resizable : Boolean read FResizable write SetResizable;  { Public }
       property ClickSelect : Boolean read FClickSelect write FClickSelect;  { Public }
       property Expandable: Boolean read FExpandable write SetExpandable;
       property Expanded: Boolean read FExpanded write SetExpanded;
       property ParentSection: TElHeaderSection read FParentSection write SetParentSection;
       property PopupMenu: TPopupMenu read FPopupMenu write SetPopupMenu;
       property LookupEnabled : boolean read FLookupEnabled write SetLookupEnabled;
       property LookupHistory : TStringList read FLookupHist;
       property ParentColor: Boolean read FParentColor write SetParentColor;
       property Color: TColor read FColor write SetColor;
       property FontColor: TColor read FFontColor write SetFontColor;
       property UseMainStyle: Boolean read FUseMainStyle write SetUseMainStyle;
     end;

    TElHeaderSections = class (TPersistent)
      private
       FList : TElList;
       function GetCount:integer;
       function GetSectionsOrder : String;
       procedure SetSectionsOrder(newValue : String);
      protected
       FOwner: TCustomElHeader;
       procedure Clear;
       function GetSection(index:integer):TElHeaderSection; virtual;
       procedure SetSection(index:integer; Section :TElHeaderSection); virtual;
       function GetSectionByPos(index:integer):TElHeaderSection; virtual;
       function CreateSection:TElHeaderSection; virtual;
       procedure WriteData(Stream:TStream); virtual;
       procedure ReadData(Stream:TStream); virtual;
       procedure DefineProperties(Filer: TFiler); override;
       function LastVisibleSection:TElHeaderSection;
       function GetPrevVisibleSection(Section:TElHeaderSection):TElHeaderSection;
       function FindSection (tag : integer) : TElHeaderSection; virtual;
      public
       constructor Create(AOwner:TCustomElHeader);
       destructor Destroy; override;
       procedure Assign(source:TPersistent); override;
       function AddSection:TElHeaderSection;
       function InsertSection(index:integer):TElHeaderSection;
       procedure DeleteSection(Section:TElHeaderSection);
       procedure MoveSection(Section:TElHeaderSection; NewPos:integer);
       procedure LoadFromStream(Stream:TStream);
       procedure SaveToStream(Stream:TStream);
       procedure SaveToFile(FileName:String);
       procedure LoadFromFile(FileName:String);

       property Owner : TCustomElHeader read FOwner;
       property SectionsOrder : String read GetSectionsOrder write SetSectionsOrder;  { Protected }
       property Count : integer read GetCount;
       property ItemByPos[Index:integer] : TElHeaderSection read GetSectionByPos;
       property Item[Index:integer] : TElHeaderSection read GetSection write SetSection; default;
     end;

     TElSectionRedrawEvent = procedure (Sender:TCustomElHeader;
                                       Section : TElHeaderSection; R:TRect; Pressed:boolean) of object;
     TElSectionResizingEvent = procedure (Sender:TCustomElHeader;
                                       Section : TElHeaderSection; State:TElHResizingStates; Width:integer) of object;
     TElSectionMoveEvent = procedure (Sender:TCustomElHeader;
                                      Section : TElHeaderSection; OldPos, NewPos : integer) of object;
     TPictureNeededEvent = procedure(Sender:TCustomElHeader; Section : TElHeaderSection; var ImageIndex:integer) of object;
     TSectionChangeEvent = procedure(Sender:TCustomElHeader; Section : TElHeaderSection; Change : TSectionChangeMode) of object;

    TCustomElHeader = class(TCustomControl)
    private
      FRightAlignedText: Boolean;
      FRightAlignedOrder: Boolean;
      FStickySections : Boolean;
      FMoveOnDrag : Boolean;
{$IFNDEF LITE}
      FImages : TElImageList;
      FStorage : TElIniFile;
      FStoragePath : String;
{$ELSE}
      FImages : TImageList;
{$ENDIF}
      FImageChangeLink: TChangeLink;
      FSections : TElHeaderSections;
      FIsNT : boolean;
      FTracking : boolean;
      FAllowDrag   : boolean;
      FPainting    : boolean;
      FAscBmp,
      FDescBmp,
      FLeftBmp,
      FRightBmp,
      FPointBmp,
      DragBmp,
      DragBmpMask  : TBitmap;
      DragRect     : TRect;
      FDragCoord   : TPoint;
      FLookup      : TComboBox;
      FPressCoord  : TPoint;
      FPressed     : boolean;
      FPressedItem : TElHeaderSection;
      FLookupSection,
      FTrackSection: TElHeaderSection;
      FResizing    : boolean;
      FResizeSection: TElHeaderSection;
      FDropSrc,
      FDropTrg     : TElHeaderSection;
      FLineDC      : HDC;
      FHeaderLine  : integer;
      FLineTab     : integer;
      FResizeOnDrag: boolean;
      FHeaderLineVis: boolean;
      FIgnoreLookupChange : boolean;
      FDoingLookup : boolean;
      FFocusedCtl  : THandle;
      LoadingCount : integer; 

      FInStick : boolean;
      FOldWidth: integer;

      FOnSectionClick  : TElHeaderSectionEvent;
      FOnSectionResize : TElHeaderSectionEvent;
      FOnSectionDraw   : TElSectionRedrawEvent;
      FOnResize        : TNotifyEvent;
      FOnSectionResizing: TElSectionResizingEvent;
      FOnSectionDelete : TElHeaderSectionEvent;
      FOnSectionMove   : TElSectionMoveEvent;
      FOnVisibleChange : TElHeaderSectionEvent;
      FOnPictureNeeded : TPictureNeededEvent;
      FOnSectionChange : TSectionChangeEvent;
      FOnSectionCreate : TElHeaderSectionEvent;
      FOnHeaderLookup : TElHeaderLookupEvent;
      FOnHeaderLookupDone : TElHeaderLookupDoneEvent;
      FOnSectionExpand: TElHeaderSectionEvent;
      FOnSectionCollapse: TElHeaderSectionEvent;
      procedure DrawLine;
      procedure AllocateLineDC;
      procedure ReleaseLineDC;
      function GetColumnsWidth: integer;
      procedure InvalidateRight(value:integer);
      procedure SetTracking (newValue : boolean);
      procedure CMDrag(var Message: TCMDrag); message CM_DRAG;
      procedure RedrawSection(Canvas : TCanvas; Section:TElHeaderSection; R:TRect);
      procedure RedrawSections;
      procedure SetSections(value:TElHeaderSections);
{$IFNDEF LITE}
      procedure SetImages(newValue : TElImageList);
{$ELSE}
      procedure SetImages(newValue : TImageList);
{$ENDIF}
      procedure OnImageListChange(Sender : TObject);
      procedure GetDragImage(XPos : Integer);
      procedure WMEraseBkgnd(var Msg : TWMEraseBkgnd); message WM_ERASEBKGND;
{$IFNDEF LITE}
      procedure SetStorage(newValue : TElIniFile);
{$ENDIF}
      procedure EditExit(Sender : TObject);
      procedure EditKeyDown(Sender : TObject; var Key: Word; Shift: TShiftState);
      procedure EditKeyUp(Sender : TObject; var Key: Word; Shift: TShiftState);
      procedure EditChange(Sender : TObject);
      procedure WMRButtonDown(var Msg: TWMRButtonDown); message WM_RBUTTONDOWN;
      procedure WMSize(var Msg: TWMSize); message WM_SIZE;
      procedure SetStickySections(newValue : Boolean);
      procedure AdjustStickySize(Caller : TElHeaderSection);
      procedure WMRButtonUp(var Msg : TWMRButtonUp); message WM_RBUTTONUP;
      function IsLoading : Boolean;
      procedure SetRightAlignedText(newValue: Boolean);
      procedure SetRightAlignedOrder(newValue: Boolean);
    protected
      { Protected declarations }
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
      function  CreateSections:TElHeaderSections; virtual;
      function  InResizeArea (X:integer; var HitSection:TElHeaderSection):boolean; virtual;
      procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
      procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
      procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
      procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
      procedure Paint; override;
{$IFDEF D_4_UP}
      procedure Resize; override;
{$ELSE}
      procedure Resize; dynamic;
{$ENDIF}
      function  DoGetPicture(Section:TElHeaderSection):integer; virtual;
      procedure DoVisChanged(Section : TElHeaderSection); virtual;
      procedure DoSectionDelete(Section : TElHeaderSection); virtual;
      procedure DoSectionMove(Section:TElHeaderSection; OldPos, NewPos : integer); virtual;
      procedure DoSectionResizing(Section:TElHeaderSection; State: TElHResizingStates; NewWidth : integer); virtual;
      procedure DoSectionResize(Section : TElHeaderSection); virtual;
      procedure DoSectionClick(Section : TElHeaderSection); virtual;
      procedure DoSectionDraw(Section : TElHeaderSection; R:TRect; Pressed:boolean); virtual;
      procedure DoNotifySectionChange(Section : TElHeaderSection; Change : TSectionChangeMode); virtual;
      procedure DoSectionExpandEvent(Section : TElHeaderSection); virtual;
      procedure DoSectionCollapseEvent(Section : TElHeaderSection); virtual;
      procedure DoSectionCreate(Section : TElHeaderSection); virtual;
      procedure DoSectionLookupEvent(Section : TElHeaderSection; var Text : string); virtual;
      procedure DoSectionLookupDoneEvent(Section : TElHeaderSection; Text : string; Accepted : boolean); virtual;
      procedure OnFontChange(Sender : TObject);
      function GetResizableWidth : integer;
      procedure CreateParams(var Params: TCreateParams); override;

      property SectionsWidth : integer read GetColumnsWidth;
      property Sections : TElHeaderSections read FSections write SetSections;
      property ResizeOnDrag : boolean read FResizeOnDrag write FResizeOnDrag;
      property Tracking : boolean read FTracking write SetTracking;
      property AllowDrag: boolean read FAllowDrag write FAllowDrag;

{$IFNDEF LITE}
      property Images : TElImageList read FImages write SetImages;  { Published }
{$ELSE}
      property Images : TImageList read FImages write SetImages;  { Published }
{$ENDIF}

      property MoveOnDrag : Boolean read FMoveOnDrag write FMoveOnDrag;  { Published }
{$IFNDEF LITE}
      property StoragePath : String read FStoragePath write FStoragePath;
      property Storage : TElIniFile read FStorage write SetStorage;
{$ENDIF}
      property StickySections : Boolean read FStickySections write SetStickySections;

      property OnSectionChange : TSectionChangeEvent     read FOnSectionChange
                                                         write FOnSectionChange;
      property OnResize : TNotifyEvent                   read FOnResize
                                                         write FOnResize;
      property OnSectionShowHide : TElHeaderSectionEvent read FOnVisibleChange
                                                         write FOnVisibleChange;
      property OnSectionClick : TElHeaderSectionEvent read FOnSectionClick
                                                      write FOnSectionClick;
      property OnSectionResize : TElHeaderSectionEvent read FOnSectionResize
                                                      write FOnSectionResize;
      property OnSectionDraw   : TElSectionRedrawEvent read FOnSectionDraw
                                                      write FOnSectionDraw;
      property OnSectionDelete : TElHeaderSectionEvent read FOnSectionDelete
                                                       write FOnSectionDelete;
      property OnSectionResizing: TElSectionResizingEvent read FOnSectionResizing
                                                          write FOnSectionResizing;
      property OnSectionMove : TElSectionMoveEvent read FOnSectionMove
                                                   write FOnSectionMove;
      property OnPictureNeeded : TPictureNeededEvent read FOnPictureNeeded
                                                      write FOnPictureNeeded;
      property OnSectionCreate : TElHeaderSectionEvent read FOnSectionCreate
                                                       write FOnSectionCreate;
      property OnSectionExpand: TElHeaderSectionEvent   read FOnSectionExpand
                                                        write FOnSectionExpand;
      property OnSectionCollapse: TElHeaderSectionEvent read FOnSectionCollapse
                                                        write FOnSectionCollapse;
      property OnHeaderLookup : TElHeaderLookupEvent    read FOnHeaderLookup
                                                        write FOnHeaderLookup;
      property OnHeaderLookupDone : TElHeaderLookupDoneEvent read FOnHeaderLookupDone
                                                             write FOnHeaderLookupDone;

      property RightAlignedText: Boolean read FRightAlignedText write SetRightAlignedText;
      property RightAlignedOrder: Boolean read FRightAlignedOrder write SetRightAlignedOrder;
    public
      { Public declarations }
      constructor Create(AOwner:TComponent); override;
      destructor Destroy; override;
      procedure Loaded; override;
      function GetSectionAt(X,Y:integer):TElHeaderSection;
      function GetSectionRect(SectionNum:integer):TRect;
      property Canvas;
{$IFNDEF LITE}
      procedure Save;
      procedure Restore;
{$ENDIF}
      procedure MarkStickySections;
    end;

    TElHeader = class (TCustomElHeader)
      published
        property Color;
        property SectionsWidth;
        property Sections;
        property Tracking;
        property AllowDrag;
        property Align;
        property Enabled;
{$IFDEF D_4_UP}
        property Anchors;
        property Action;
        property Constraints;
        property DockOrientation;
        property Floating;
        property BevelKind;
        property DoubleBuffered;
        property DragKind;
{$ENDIF} 
//        property FixClick; // Left for future
        property MoveOnDrag;
        property Font;
        property Images;
        property ParentFont;
        property ParentShowHint;
        property PopupMenu;
        property ResizeOnDrag;
        property RightAlignedText;
        property RightAlignedOrder;
        property ShowHint;
        property StickySections;
{$IFNDEF LITE}
        property Storage;
        property StoragePath;
{$ENDIF}
        property Visible;

        property OnMouseDown;
        property OnMouseMove;
        property OnMouseUp;
        property OnResize;
        property OnSectionClick;
        property OnSectionResize;
        property OnSectionDraw;
        property OnSectionResizing;
        property OnSectionDelete;
        property OnSectionMove;
        property OnSectionShowHide;
        property OnPictureNeeded;
        property OnSectionChange;
        property OnSectionCreate;
        property OnSectionExpand;
        property OnSectionCollapse;
        property OnHeaderLookup;
        property OnHeaderLookupDone;
    end;

implementation

Const ResizeWidth = 5;

// =============================================================================

constructor TElHeaderSection.Create;
begin
  inherited Create;
  FStyle:=ElhsText;
  FVisible:=true;
  FOwner:=FOwner;
  FWidth:=120;
  FMinWidth:=0;
  FMaxWidth:=10000;
  FAlignment:=hsaLeft;
  FAllowClick:=true;
  FResizable:=true;
  FClickSelect := true;
  FLookupHist := TStringList.Create;
  FParentColor := True;
  FSaveSize := -1;
end;

procedure TElHeaderSection.UpdateSection;
var R:TRect;
begin
  if not Visible then exit;
  R:=Rect(Left,0,Left+FWidth,FOwner.Height-1);
  InvalidateRect(FOwner.Handle,@R,false);
  FOwner.Refresh;
end;

procedure TElHeaderSection.SetSortMode(value:TElSSortMode);
var i : integer;
begin
  if FSortMode=value then exit;
  for i:=0 to FOwner.FSections.Count-1 do FOwner.FSections[i].FSortMode:=hsmNone;
  FSortMode:=value;
  FOwner.Repaint;
end;

procedure TElHeaderSection.SetMaxWidth(value:integer);
begin
  if value=FMaxWidth then exit;
  FMaxWidth:=value;
  if FWidth>FMaxWidth then Width:=FMaxWidth;
end;

procedure TElHeaderSection.SetMinWidth(value:integer);
begin
  if value=FMinWidth then exit;
  FMinWidth:=value;
  if FWidth>FMinWidth then Width:=FMinWidth;
end;

procedure TElHeaderSection.SetWidth(value:integer);
var i, wo, wn : integer;
    HS : TElHeaderSection;
begin
  if (value>FMaxWidth) or (value<FMinWidth) then exit;
  if (not FOwner.FStickySections) or (not Visible) or (FOwner.FInStick) or (FOwner.IsLoading) then
  begin
    FWidth:=value;
    if Visible and (not FOwner.FInStick) then
    begin
      FOwner.InvalidateRight(Self.Index);
      FOwner.Refresh;
    end;
  end else
  begin
    if Position = FOwner.FSections.Count - 1 then
    begin
      // the last section, do nothing (we can't resize it)
    end else
    begin
      wo := FOwner.Width - (Left + FWidth);
      wn := FOwner.Width - (Left + Value);
      for i := Position + 1 to FOwner.Sections.Count - 1 do    // Iterate
      begin
        HS := FOwner.FSections.ItemByPos[i];
        if HS.Resizable and HS.Visible then HS.FWidth := Round((HS.Width / wo) * wn);
      end;    // for
      FWidth := value;
      FOwner.AdjustStickySize(Self);
    end;
    FOwner.MarkStickySections;
    if Visible and (not FOwner.FInStick) then
    begin
      FOwner.InvalidateRight(Index);
      FOwner.Refresh;
    end;
  end;
end;

procedure TElHeaderSection.SetLookupEnabled(newValue : boolean);
begin
  if FLookupEnabled <> newValue then
  begin
    FLookupEnabled := newValue;
    UpdateSection;
  end;
end;

function TElHeaderSection.GetVisible : boolean;
begin
  if FParentSection = nil then result := FVisible else result := FVisible and FParentSection.Visible and FParentSection.Expanded;
end;

procedure TElHeaderSection.SetVisible;
var ns, os, j, k, wo, wn, i : integer;
    HS : TElHeaderSection;
begin
  if Visible = Value then exit;
  if (not FOwner.FStickySections) or ((Self.FParentSection <> nil) and (not FParentSection.Visible)) or (FOwner.FInStick) or (FOwner.IsLoading) then
  begin
    FVisible:=value;
    if (not FOwner.FInStick) and ((ParentSection = nil) or (ParentSection.Visible)) then
    begin
      FOwner.InvalidateRight(Index);
      FOwner.Refresh;
    end;
  end else
  begin
    if Value then
    begin
      ns := Width;
      os := 0;
    end else
    begin
      ns := 0;
      os := Width;
    end;
    FVisible := value;
    wo := FOwner.Width - (Left + os);
    wn := FOwner.Width - (Left + ns);
    if Position = FOwner.FSections.Count - 1 then
    begin
      j := 0;
      k := FOwner.Sections.Count - 2;
    end else
    begin
      j := Position + 1;
      k := FOwner.Sections.Count - 1;
    end;
    for i := j to k do    // Iterate
    begin
      HS := FOwner.FSections.ItemByPos[i];
      if HS.Resizable then
      begin
        HS.FWidth := Round((HS.Width / wo) * wn);
      end;
    end;    // for
    FOwner.AdjustStickySize(Self);
    FOwner.MarkStickySections;
  end;
  FOwner.Repaint;
  FOwner.DoVisChanged(self);
end;

function TElHeaderSection.GetLeft:integer;
var ind:integer;
begin
  ind:=FOwner.Sections.FList.IndexOf(self);
  if ind=0 then result:=0 else result:=TElHeaderSection(FOwner.FSections.FList[ind-1]).Right;
end;

function TElHeaderSection.GetRight:integer;
var i:integer;
    ind:integer;
begin
  ind:=FOwner.FSections.FList.IndexOf(self);
  if ind=0 then i:=0 else i:=TElHeaderSection(FOwner.FSections.FList[ind-1]).Right;
  if Visible then result:=i+FWidth else result:=i;
end;

procedure TElHeaderSection.SetText(value:string);
begin
  if FText=value then exit;
  FText:=value;
  if FStyle=ElhsText then UpdateSection;
  FOwner.DoNotifySectionChange(self, scmCaption);
end;

procedure TElHeaderSection.SetStyle(value:TElSectionStyle);
begin
  if FStyle=value then exit;
  FStyle:=value;
  FOwner.DoNotifySectionChange(self, scmStyle);
  UpdateSection;
end;

procedure TElHeaderSection.SetAlignment(value:TElSAlignment);
begin
  if FAlignment=value then exit;
  FAlignment:=value;
  FOwner.DoNotifySectionChange(self, scmAlign);
  UpdateSection;
end;

function TElHeaderSection.GetIndex:integer;
begin
  result:=FTag;
end;

procedure TElHeaderSection.Assign(source:TPersistent);
var THS : TElHeaderSection;
begin
  if Source is TElHeaderSection then
  begin
    THS:=TElHeaderSection(source);
    FVisible:=THS.FVisible;
    FStyle:=THS.FStyle;
    FText:=THS.FText;
    FWidth:=THS.FWidth;
    FMinWidth:=THS.FMinWidth;
    FMaxWidth:=THS.FMaxWidth;
    FSortMode:=THS.FSortMode;
    FAllowClick:=THS.FAllowClick;
    FAlignment:=THS.FAlignment;
    FExpandable := THS.FExpandable;
    FExpanded := THS.FExpanded;
    if THS.FParentSection = nil
       then FParentSection := nil
       else FParentSection := FOwner.Sections.Item[THS.FParentSection.Index];
    FPopupMenu := THS.PopupMenu;
    FLookupEnabled := THS.FLookupEnabled;
    FUseMainStyle := THS.FUseMainStyle;
  end else inherited;
end;

function TElHeaderSection.GetPosition : Integer;
{ Returns the value of data member FPosition. }
begin
  result:=-1;
  if FOwner<>nil then result:=FOwner.FSections.FList.IndexOf(self);
end;  { GetPosition }

procedure TElHeaderSection.SetImageIndex(newValue : Integer);
{ Sets data member FImageIndex to newValue. }
begin
  if (FImageIndex <> newValue) then
  begin
    FImageIndex := newValue;
    UpdateSection;
  end;  { if }
end;  { SetImageIndex }

procedure TElHeaderSection.SetResizable(newValue : boolean);
begin
  FResizable := newValue;
  if Assigned(FOwner) then
     if FOwner.FStickySections then FOwner.SetStickySections(true);
end;

procedure TElHeaderSection.SetSaveSize(newValue : integer);
begin
  ASaveSize := newValue; 
end;

procedure TElHeaderSection.SetEditable(newValue : boolean);
begin
  if FEditable<>newValue then
  begin
    FEditable:=newValue;
    FOwner.DoNotifySectionChange(self, scmEditable);
  end;
end;

procedure TElHeaderSection.SetFieldName(newValue : String);
{ Sets data member FFieldName to newValue. }
begin
  if (FFieldName <> newValue) then
  begin
    FFieldName := newValue;
    FOwner.DoNotifySectionChange(self, scmFieldName);
  end;  { if }
end;  { SetFieldName }

procedure TElHeaderSection.SetFieldType(newValue : TFieldTypes);
{ Sets data member FFieldType to newValue. }
begin
  if (FFieldType <> newValue) then
  begin
    FFieldType := newValue;
    FOwner.DoNotifySectionChange(self, scmFieldType);
  end;  { if }
end;  { SetFieldType }

procedure TElHeaderSection.SetProtected(newValue : Boolean);
{ Sets data member FProtected to newValue. }
begin
  if (FProtected <> newValue) then
  begin
    FProtected := newValue;
    FOwner.DoNotifySectionChange(self, scmPassword);
  end;  { if }
end;  { SetProtected }

procedure TElHeaderSection.SetExpandable(newValue: Boolean);
var i : integer;
    b : boolean;
begin
  if (FExpandable <> newValue) then
  begin
    b := false;
    FExpandable := newValue;
    if not FExpandable then
    for i := 0 to FOwner.FSections.Count -1 do
    begin
      if TElHeaderSection(FOwner.FSections.FList[i]).FParentSection = Self then
      begin
        TElHeaderSection(FOwner.FSections.FList[i]).FParentSection := nil;
        b := true;
      end;
    end;
    if b and (not (csDestroying in FOwner.ComponentState)) then FOwner.Repaint;
  end;  {if}
end;

procedure TElHeaderSection.SetExpanded(newValue: Boolean);
begin
  if (FExpanded <> newValue) and Expandable then
  begin
    if (FExpandable and newValue) or (not newValue) then
    begin
      FExpanded := newValue;
      FOwner.Repaint;
    end;
  end;  {if}
end;

procedure TElHeaderSection.SetParentSection(newValue: TElHeaderSection);
var S : TElHeaderSection;
begin
  if (FParentSection <> newValue) then
  begin
    S := newValue;
    while S <> nil do
    begin
      if S = Self then exit;
      S := S.FParentSection;
    end;
    FParentSection := newValue;
    FOwner.DoVisChanged(Self);
    FOwner.Repaint;
  end;  {if}
end;

procedure TElHeaderSection.SetPopupMenu(newValue: TPopupMenu);
begin
  if (FPopupMenu <> newValue) then
  begin
    FPopupMenu := newValue;
    FPopupName := FPopupMenu.Name;
  end;  {if}
end;

procedure TElHeaderSection.SetParentColor(newValue: Boolean);
begin
  if (FParentColor <> newValue) then
  begin
    FParentColor := newValue;
    UpdateSection;
  end;  {if}
end;

procedure TElHeaderSection.SetColor(newValue: TColor);
begin
  if (FColor <> newValue) then
  begin
    FColor := newValue;
    ParentColor := false;
    UpdateSection;
  end;  {if}
end;

procedure TElHeaderSection.SetFontColor(newValue: TColor);
begin
  if (FFontColor <> newValue) then
  begin
    FFontColor := newValue;
    ParentColor := false;
    UpdateSection;
  end;  {if}
end;

procedure TElHeaderSection.SetUseMainStyle(newValue: Boolean);
begin
  if (FUseMainStyle <> newValue) then
  begin
    FUseMainStyle := newValue;
    if Assigned (FOwner.Parent) then FOwner.Parent.Repaint;
  end;  {if}
end;

destructor TElHeaderSection.Destroy;
begin
  Expandable := false;
  FLookupHist.Free;
  inherited;
end;

// =============================================================================
constructor TElHeaderSections.Create;
begin
  inherited Create;
  FOwner := AOwner;
  FList:= TElList.Create;
end;

destructor TElHeaderSections.Destroy;
begin
  Clear;
  FList.Free;
  inherited destroy;
end;

procedure TElHeaderSections.Clear;
begin
  while FList.Count>0 do
  begin
    TElHeaderSection(FList[0]).Free;
    FList.Delete(0);
  end;
end;

procedure TElHeaderSections.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('Data', ReadData, WriteData, true);
end;

function TElHeaderSections.CreateSection:TElHeaderSection;
begin
  result:=TElHeaderSection.Create(FOwner);
end;

function TElHeaderSections.AddSection:TElHeaderSection;
begin
  result:=InsertSection(FList.Count);
end;

function TElHeaderSections.InsertSection(index:integer):TElHeaderSection;
var i : integer;
begin
  result:=CreateSection;
  i:=0;
  while FindSection(i)<>nil do inc(i);
  Result.FTag:=i;
  Result.FOwner:=FOwner;
  FList.Insert(index, result);
  FOwner.InvalidateRight(i);
  FOwner.DoSectionCreate(result);
  FOwner.Update;
end;

procedure TElHeaderSections.DeleteSection(Section:TElHeaderSection);
var i : integer;
begin
 if Section=nil then Raise EElError.Create('Invalid section');
 FList.Remove(Section);
 FOwner.DoSectionDelete(Section);
 Section.Free;
 for i:=0 to FList.Count-1 do TElHeaderSection(FList[i]).FTag:=i;
 FOwner.Invalidate;
 FOwner.Update;
end;

procedure TElHeaderSections.MoveSection(Section:TElHeaderSection; NewPos:integer);
var i:integer;
begin
  i:=FList.IndexOf(Section);
  FList.Move(i,NewPos);
  FOwner.DoSectionMove(TElHeaderSection(FList[i]), i, NewPos);
  FOwner.Invalidate;
  FOwner.Update;
end;

function TElHeaderSections.GetCount:integer;
begin
  result:=FList.Count;
end;

procedure TElHeaderSections.SetSection(index:integer; Section :TElHeaderSection);
begin

end;

function TElHeaderSections.GetSection;
begin
  result:=FindSection(index);
end;

function TElHeaderSections.GetSectionByPos;
begin
  result:=TElHeaderSection(FList[index]);
end;

type PSectionData = ^TSectionData;
     TSectionData = record
       FFakeProperty1: Integer;
       FLookupEnabled: boolean; 
       FParentIdx: integer;
       FExpandable: Boolean;
       FExpanded: Boolean;
       FVisible : boolean;
       FClickSel: boolean;
       FResiable: boolean;
       FStyle   : TElSectionStyle;
       FWidth   : integer;
       FMinWidth,
       FMaxWidth: integer;
       FSortMode: TElSSortMode;
       FAllowClick: boolean;
       FAlignment: TElSAlignment;
       FTag      : integer;
       FFieldType : TFieldTypes;
       FEditable : Boolean;
       FImageIndex : Integer;
       FPicAlign : TElSAlignment;
       FPassword : boolean;
     end;

type P5SectionData = ^T5SectionData;
     T5SectionData = record
       FVisible : boolean;
       FStyle   : TElSectionStyle;
       FWidth   : integer;
       FMinWidth,
       FMaxWidth: integer;
       FSortMode: TElSSortMode;
       FAllowClick: boolean;
       FAlignment: TElSAlignment;
       FTag      : integer;
       FFieldType : TFieldTypes;
       FEditable : Boolean;
       FImageIndex : Integer;
       FPicAlign : TElSAlignment;
       FPassword : boolean;
     end;

type P6SectionData = ^T6SectionData;
     T6SectionData = record
       FVisible : boolean;
       FClickSel: boolean;
       FResiable: boolean;
       FStyle   : TElSectionStyle;
       FWidth   : integer;
       FMinWidth,
       FMaxWidth: integer;
       FSortMode: TElSSortMode;
       FAllowClick: boolean;
       FAlignment: TElSAlignment;
       FTag      : integer;
       FFieldType : TFieldTypes;
       FEditable : Boolean;
       FImageIndex : Integer;
       FPicAlign : TElSAlignment;
       FPassword : boolean;
     end;

procedure TElHeaderSections.WriteData(Stream:TStream);
var i,j:integer;
    P:PSectionData;
    TS : TElHeaderSection;
    P1:PChar;
    S:String;
begin
  inherited;
  GetMem(P,SizeOf(TSectionData));
  i:= -6;
  Stream.WriteBuffer(i,SizeOf(integer));
  i:=Count;
  Stream.WriteBuffer(i,SizeOf(integer));
  for i:=0 to count-1 do
  begin
    TS:=TElHeaderSection(FList[i]);
    P.FLookupEnabled := TS.FLookupEnabled;
    if TS.FParentSection <> nil
       then P.FParentIdx := TS.FParentSection.Index
       else P.FParentIdx := -1;
    P.FExpandable := TS.FExpandable;
    P.FExpanded   := TS.FExpanded;
    P.FVisible:=TS.FVisible;
    P.FStyle  :=TS.FStyle;
    P.FWidth  :=TS.FWidth;
    P.FResiable:= TS.FResizable;
    P.FClickSel:= TS.FClickSelect;
    P.FMinWidth:=TS.FMinWidth;
    P.FMaxWidth:=TS.FMaxWidth;
    P.FSortMode:=TS.FSortMode;
    P.FAllowClick:=TS.FAllowClick;
    P.FAlignment:=TS.FAlignment;
    P.FTag      := TS.FTag;
    P.FFieldType:= TS.FFieldType;
    P.FEditable := TS.FEditable;
    P.FImageIndex:=TS.FImageIndex;
    P.FPicAlign  :=TS.FPicAlign;
    P.FPassword  :=TS.FProtected;

    Stream.WriteBuffer(p^,SizeOf(TSectionData));
    S:=TElHeaderSection(FList[i]).Text;
    j:=Length(S)+1;
    GetMem(P1,j);
    StrPCopy(P1,S);
    Stream.WriteBuffer(j,SizeOf(integer));
    Stream.WriteBuffer(P1^,j);
    FreeMem(P1,j);

    S:=TElHeaderSection(FList[i]).FFieldName;
    j:=Length(S)+1;
    GetMem(P1,j);
    StrPCopy(P1,S);
    Stream.WriteBuffer(j,SizeOf(integer));
    Stream.WriteBuffer(P1^,j);
    FreeMem(P1,j);
    WriteStringToStream(Stream, TS.FPopupName);
  end;
  FreeMem(P,SizeOf(TSectionData));
end;

type P4SectionData = ^T4SectionData;
     T4SectionData = record
       FVisible : boolean;
       FStyle   : TElSectionStyle;
       FWidth   : integer;
       FMinWidth,
       FMaxWidth: integer;
       FSortMode: TElSSortMode;
       FAllowClick: boolean;
       FAlignment: TElSAlignment;
       FTag      : integer;
       FFieldType : TFieldTypes;
       FEditable : Boolean;
       FImageIndex : Integer;
       FPicAlign : TElSAlignment;
     end;

type P3SectionData = ^T3SectionData;
     T3SectionData = record
       FVisible : boolean;
       FStyle   : TElSectionStyle;
       FWidth   : integer;
       FMinWidth,
       FMaxWidth: integer;
       FSortMode: TElSSortMode;
       FAllowClick: boolean;
       FAlignment: TElSAlignment;
     end;


procedure TElHeaderSections.ReadData(Stream:TStream);
var i,j,k:integer;
    P:P4SectionData;
    Z:P3SectionData;
    Y:P5SectionData;
    Q:P6SectionData;
    T:PSectionData;
    P1:PChar;
    S:String;
    THS:TElHeaderSection;
    Ver : integer;
begin
  inherited;
  Clear;
  Stream.ReadBuffer(i,SizeOf(integer));
  ver := i;
  if (ver = -5) or (ver = -6) then
  begin
    Stream.ReadBuffer(i,SizeOf(integer));
    GetMem(T, SizeOf(TSectionData));
    while i>0 do
    begin
      Stream.ReadBuffer(T^,SizeOf(TSectionData));
      THS:=AddSection;
      THS.FLookupEnabled := T.FLookupEnabled;
      THS.FExpandable := T.FExpandable;
      THS.FExpanded := T.FExpanded;
      THS.FParentIdx := T.FParentIdx;
      THS.FResizable:=T.FResiable;
      THS.FClickSelect:=T.FClickSel;
      THS.FVisible:=T.FVisible;
      THS.FStyle  :=T.FStyle;
      THS.FWidth  :=T.FWidth;
      THS.FMinWidth:=T.FMinWidth;
      THS.FMaxWidth:=T.FMaxWidth;
      THS.FSortMode:=T.FSortMode;
      THS.FAllowClick:=T.FAllowClick;
      THS.FAlignment:=T.FAlignment;
      THS.FTag      := T.FTag;
      THS.FFieldType:= T.FFieldType;
      THS.FEditable := T.FEditable;
      THS.FImageIndex:=T.FImageIndex;
      THS.FPicAlign  :=T.FPicAlign;
      THS.FProtected :=T.FPassword;
      // read caption
      Stream.ReadBuffer(J,SizeOf(integer));
      GetMem(P1,j);
      Stream.ReadBuffer(P1^,j);
      S:=StrPas(P1);
      THS.FText:=S;
      FreeMem(P1,j);
      // read field name
      Stream.ReadBuffer(J,SizeOf(integer));
      GetMem(P1,j);
      Stream.ReadBuffer(P1^,j);
      S:=StrPas(P1);
      THS.FFieldName:=S;
      FreeMem(P1,j);
      dec(i);
      if Ver = -6 then ReadStringFromStream(Stream, THS.FPopupName);
    end;
    FreeMem(T);
    for i := 0 to Count - 1 do
    begin
      if Item[i].FParentIdx <> -1 then
      begin
        Item[i].FParentSection := Item[Item[i].FParentIdx];
      end;
    end;
  end else
  if i= -4 then
  begin
    Stream.ReadBuffer(i,SizeOf(integer));
    GetMem(q,SizeOf(T6SectionData));
    while i>0 do
    begin
      Stream.ReadBuffer(q^,SizeOf(T6SectionData));
      THS:=AddSection;
      THS.FResizable:=Q^.FResiable;
      THS.FClickSelect:=Q^.FClickSel;
      THS.FVisible:=q^.FVisible;
      THS.FStyle  :=q^.FStyle;
      THS.FWidth  :=q^.FWidth;
      THS.FMinWidth:=q^.FMinWidth;
      THS.FMaxWidth:=q^.FMaxWidth;
      THS.FSortMode:=q^.FSortMode;
      THS.FAllowClick:=q^.FAllowClick;
      THS.FAlignment:=q^.FAlignment;
      THS.FTag      := q^.FTag;
      THS.FFieldType:= q^.FFieldType;
      THS.FEditable := q^.FEditable;
      THS.FImageIndex:=q^.FImageIndex;
      THS.FPicAlign  :=q^.FPicAlign;
      THS.FProtected :=q^.FPassword;
      // read caption
      Stream.ReadBuffer(J,SizeOf(integer));
      GetMem(P1,j);
      Stream.ReadBuffer(P1^,j);
      S:=StrPas(P1);
      THS.FText:=S;
      FreeMem(P1,j);
      // read field name
      Stream.ReadBuffer(J,SizeOf(integer));
      GetMem(P1,j);
      Stream.ReadBuffer(P1^,j);
      S:=StrPas(P1);
      THS.FFieldName:=S;
      FreeMem(P1,j);
      dec(i);
    end;
    FreeMem(q);
  end else
  if i=-3 then
  begin
    Stream.ReadBuffer(i,SizeOf(integer));
    GetMem(Y,SizeOf(T5SectionData));
    while i>0 do
    begin
      Stream.ReadBuffer(y^,SizeOf(T5SectionData));
      THS:=AddSection;
      THS.FVisible:=y^.FVisible;
      THS.FStyle  :=y^.FStyle;
      THS.FWidth  :=y^.FWidth;
      THS.FMinWidth:=y^.FMinWidth;
      THS.FMaxWidth:=y^.FMaxWidth;
      THS.FSortMode:=y^.FSortMode;
      THS.FAllowClick:=y^.FAllowClick;
      THS.FAlignment:=y^.FAlignment;
      THS.FTag      := y^.FTag;
      THS.FFieldType:= y^.FFieldType;
      THS.FEditable := y^.FEditable;
      THS.FImageIndex:=y^.FImageIndex;
      THS.FPicAlign  :=y^.FPicAlign;
      THS.FProtected :=y^.FPassword;
      // read caption
      Stream.ReadBuffer(J,SizeOf(integer));
      GetMem(P1,j);
      Stream.ReadBuffer(P1^,j);
      S:=StrPas(P1);
      THS.FText:=S;
      FreeMem(P1,j);
      // read field name
      Stream.ReadBuffer(J,SizeOf(integer));
      GetMem(P1,j);
      Stream.ReadBuffer(P1^,j);
      S:=StrPas(P1);
      THS.FFieldName:=S;
      FreeMem(P1,j);
      dec(i);
    end;
    FreeMem(y);
  end else
(*  if i=-2 then
  begin
    Stream.ReadBuffer(i,SizeOf(integer));
    GetMem(P,SizeOf(TSectionData));
    while i>0 do
    begin
      Stream.ReadBuffer(p^,SizeOf(TSectionData));
      THS:=AddSection;
      THS.FVisible:=P^.FVisible;
      THS.FStyle  :=P^.FStyle;
      THS.FWidth  :=P^.FWidth;
      THS.FMinWidth:=P^.FMinWidth;
      THS.FMaxWidth:=P^.FMaxWidth;
      THS.FSortMode:=P^.FSortMode;
      THS.FAllowClick:=P^.FAllowClick;
      THS.FAlignment:=P^.FAlignment;
      THS.FTag      := P^.FTag;
      THS.FFieldType:= P^.FFieldType;
      THS.FEditable := P^.FEditable;
      THS.FImageIndex:=P^.FImageIndex;
      THS.FPicAlign  :=P^.FPicAlign;
      // read caption
      Stream.ReadBuffer(J,SizeOf(integer));
      GetMem(P1,j);
      Stream.ReadBuffer(P1^,j);
      S:=StrPas(P1);
      THS.FText:=S;
      FreeMem(P1,j);
      // read field name
      Stream.ReadBuffer(J,SizeOf(integer));
      GetMem(P1,j);
      Stream.ReadBuffer(P1^,j);
      S:=StrPas(P1);
      THS.FFieldName:=S;
      FreeMem(P1);
      dec(i);
    end;
    FreeMem(P);
  end
  else
*)  if i=-1 then // 1.4 format
  begin
    Stream.ReadBuffer(i,SizeOf(integer));
    GetMem(P,SizeOf(T4SectionData));
    while i>0 do
    begin
      Stream.ReadBuffer(p^,SizeOf(T4SectionData));
      THS:=AddSection;
      THS.FVisible:=P^.FVisible;
      THS.FStyle  :=P^.FStyle;
      THS.FWidth  :=P^.FWidth;
      THS.FMinWidth:=P^.FMinWidth;
      THS.FMaxWidth:=P^.FMaxWidth;
      THS.FSortMode:=P^.FSortMode;
      THS.FAllowClick:=P^.FAllowClick;
      THS.FAlignment:=P^.FAlignment;
      THS.FTag      := P^.FTag;
      THS.FFieldType:= P^.FFieldType;
      THS.FEditable := P^.FEditable;
      THS.FImageIndex:=P^.FImageIndex;
      THS.FPicAlign  :=P^.FPicAlign;
      Stream.ReadBuffer(J,SizeOf(integer));
      GetMem(P1,j);
      Stream.ReadBuffer(P1^,j);
      S:=StrPas(P1);
      THS.FText:=S;
      FreeMem(P1,j);
      dec(i);
    end;
    FreeMem(P);
  end else
  begin
    GetMem(Z, sizeof(T3SectionData));
    k:=0;
    while i>0 do
    begin
      Stream.ReadBuffer(Z^,SizeOf(T3SectionData));
      THS:=AddSection;
      THS.FVisible :=Z^.FVisible;
      THS.FStyle   :=Z^.FStyle;
      THS.FWidth   :=Z^.FWidth;
      THS.FMinWidth:=Z^.FMinWidth;
      THS.FMaxWidth:=Z^.FMaxWidth;
      THS.FSortMode:=Z^.FSortMode;
      THS.FAllowClick:=Z^.FAllowClick;
      THS.FAlignment:= Z^.FAlignment;
      THS.FTag      := k;
      THS.FFieldType:= sftText;
      THS.FEditable := false;
      THS.FImageIndex:=-1;
      THS.FPicAlign  :=hsaCenter;
      if Z^.FStyle=ElhsText then
      begin
        Stream.ReadBuffer(J,SizeOf(integer));
        GetMem(P1,j);
        Stream.ReadBuffer(P1^,j);
        THS.FText:=StrPas(P1);
        FreeMem(P1,j);
      end;
      dec(i);
      inc(k);
    end;
    FreeMem(Z);
  end;
  FOwner.Repaint;
end;

procedure TElHeaderSections.LoadFromStream(Stream:TStream);
begin
  ReadData(Stream);
  if FOwner<>nil then FOwner.Invalidate;
end;

procedure TElHeaderSections.SaveToStream(Stream:TStream);
begin
  WriteData(Stream);
end;

procedure TElHeaderSections.SaveToFile(FileName:String);
var T:TStream;
begin
  T:=nil;
  try
    T:=TFileStream.Create(FileName, fmCreate or fmShareDenyWrite);
    SaveToStream(T);
  finally
    T.Free;
  end;
end;

procedure TElHeaderSections.LoadFromFile(FileName:String);
var T:TStream;
begin
  T:=nil;
  try
    T:=TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
    LoadFromStream(T);
  finally
    T.Free;
  end;
end;

function TElHeaderSections.LastVisibleSection:TElHeaderSection;
begin
  result:=TElHeaderSection(FList[Count-1]);
  if not Result.Visible then result:=GetPrevVisibleSection(Result);
end;

function TElHeaderSections.GetPrevVisibleSection;
var I:Integer;

begin
  result:=nil;
  I:=FList.IndexOf(Section)-1;
  if i<0 then exit;
  while i>=0 do
  begin
    if TElHeaderSection(FList[i]).Visible then
    begin
      result:=TElHeaderSection(FList[i]);
      break;
    end;
    dec(i);
  end;    // while
end;

procedure TElHeaderSections.Assign;
var IT: TElHeaderSections;
    j : integer;
begin
  if source is TElHeaderSections then
  begin
    IT:=TElHeaderSections(source);
    // delete old sections
    for j := 0 to FList.Count-1 do DeleteSection(Item[j]);
    // add new sections
    for j := 0 to IT.Count-1 do CreateSection.Assign(IT.Item[j]);
    FOwner.MarkStickySections;
  end else inherited;
end;

function TElHeaderSections.FindSection (tag : integer) : TElHeaderSection;  { protected }
var i: integer;
begin
  for i:=0 to FList.Count-1 do
    if TElHeaderSection(FList[i]).FTag=tag then
    begin
      result:=TElHeaderSection(FList[i]);
      exit;
    end;
    result := nil;
end;  { FindSection }

// =============================================================================
constructor TCustomElHeader.Create(AOwner:TComponent);
begin
  inherited;
{$IFNDEF VER90}
  ControlStyle:=[csClickEvents, csDoubleClicks, csCaptureMouse,
                 csDesignInteractive, csReplicatable, csOpaque, csReflector];
{$ELSE}
  ControlStyle:=[csClickEvents, csDoubleClicks, csCaptureMouse,
                 csDesignInteractive, csReplicatable, csOpaque];
{$ENDIF}
  FSections:=CreateSections;
  FIsNT:=IsWinNT;
  TabStop:=false;
  Font.OnChange:=OnFontChange;
  Height := abs(Font.Height) + 6 + GetSystemMetrics(SM_CYBORDER) * 2;
  DragMode:=dmManual;
  FPressed:=false;
  DragCursor:=Cursor;
  FAscBmp:=TBitmap.Create;
  FDescBmp:=TBitmap.Create;
  FLeftBmp:=TBitmap.Create;
  FRightBmp:=TBitmap.Create;
  FPointBmp:=TBitmap.Create;
  FPressed:=false;
  FTracking:=true;
  FAllowDrag:=true;
  FResizeOnDrag:=false;
  FImageChangeLink := TChangeLink.Create;
  FImageChangeLink.OnChange := OnImageListChange;
  FAscBmp.LoadFromResourceName(HInstance,  'ELHEADERASCBMP');
  FDescBmp.LoadFromResourceName(HInstance, 'ELHEADERDESCBMP');
  FLeftBmp.LoadFromResourceName(HInstance, 'ELHEADERLEFTBMP');
  FRightBmp.LoadFromResourceName(HInstance,'ELHEADERRIGHTBMP');
  FPointBmp.LoadFromResourceName(HInstance,'ELHEADERPOINTBMP');
  Color:=clBtnFace;
  Align := alTop;
end;

destructor TCustomElHeader.Destroy;
begin
  Destroying;
  FSections.Free;
  FPointBmp.Free;
  FAscBmp.Free;
  FDescBmp.Free;
  FLeftBmp.Free;
  FRightBmp.Free;
  if FImages<>nil then FImages.UnregisterChanges(FImageChangeLink);
  FImageChangeLink.Free;
  inherited Destroy;
end;

function TCustomElHeader.CreateSections;
begin
  result:=TElHeaderSections.Create(self);
end;

function TCustomElHeader.GetSectionAt(X,Y:integer):TElHeaderSection;
var i:integer;
    j:integer;
begin
  j:=0;
  result:=nil;
  for i:=0 to FSections.FList.Count-1 do
  begin
    if TElHeaderSection(FSections.FList[i]).Visible then j := j + TElHeaderSection(FSections.FList[i]).Width;
    if j>X then
    begin
      result:=TElHeaderSection(FSections.FList[i]);
      exit;
    end;
  end;
end;

procedure TCustomElHeader.InvalidateRight(value:integer);
var R:TRect;
begin
  if not inRange(0,Sections.Count-1,value) then Raise EElError.Create('Index out of bounds');
  R:=Rect(TElHeaderSection(Sections[value]).Left, 0, ClientWidth, ClientHeight);
  InvalidateRect(Canvas.Handle,@R,false);
  Update;
end;

function TCustomElHeader.GetColumnsWidth: integer;
var i:integer;
begin
  result:=0;
  i:=FSections.Count-1;
  while i>=0 do
  begin
    if TElHeaderSection(FSections.FList[i]).Visible then
    begin
      result:=TElHeaderSection(FSections.FList[i]).Right;
      exit;
    end;
    dec(i);
  end;
end;

function TCustomElHeader.InResizeArea(X:integer; var HitSection:TElHeaderSection):boolean;
var THS : TElHeaderSection;
    l : integer;
begin
  THS:=GetSectionAt(X,0);
  HitSection:=THS;
  if THS = nil then
  begin
    if Sections.Count>0 then THS:=Sections.LastVisibleSection;
    if THS<>nil then
    begin
      l:=THS.Right;
      if L+ResizeWidth>X then
      begin
        result:=true;
        HitSection:=THS;
        exit;
      end;
    end;
    result:=false;
    HitSection:=nil;
    exit;
  end
  else
  begin
    if THS.Width<ResizeWidth*3 then result:=false else
    begin
      l:=THS.Left;
      if L+ResizeWidth>X then
      begin
        if Sections.GetPrevVisibleSection(THS)=nil then
        begin
          result:=false;
          HitSection:=THS;
          exit;
        end;
        result:=true;
        HitSection:=Sections.GetPrevVisibleSection(THS);
        exit;
      end;
      if L+THS.FWidth-ResizeWidth<=X then
      begin
        result:=true;
        exit;
      end;
      result:=false;
    end;
  end;
end;

procedure TCustomElHeader.WMLButtonDown(var Message: TWMLButtonDown);
var THS,THS1 : TElHeaderSection;
    b  : boolean;
    rm : integer;
    R : TRect;
begin
  if FDoingLookup then
  begin
    FDoingLookup := false;
    FLookup.Visible := false;
    DoSectionLookupDoneEvent(FLookupSection, FLookup.Text, false);
    Windows.SetFocus(FFocusedCtl);
    inherited;
    exit;
  end;
  if not (csNoStdEvents in ControlStyle) then
  begin
    if InResizeArea(Message.XPos, THS) and (THS.FResizable) then
    begin
      FResizing:=true;
      FResizeSection:=THS;
      THS1:=FTrackSection;
      FTrackSection:=nil;
      if THS1<>nil then THS1.UpdateSection;
      Screen.Cursor:=crHSplit;
      SetCapture(Handle);
      FHeaderLine:=FResizeSection.Right;
      FLineTab:=Message.XPos - FHeaderLine;
      if (not ResizeOnDrag) then
      begin
        AllocateLineDC;
        DrawLine;
      end;
      DoSectionResizing(THS, trsBegin, THS.Width);
    end // Resizing
    else
    begin
      FPressCoord:=Point(Message.XPos,Message.YPos);
      FPressed:=true;
      if FPressedItem<>nil then
      begin
        THS:=FPressedItem;
        FPressedItem:=nil;
        THS.UpdateSection;
      end;
      FPressedItem:=GetSectionAt(Message.XPos, Message.YPos);
      THS := FPressedItem;
      if FPressedItem <> nil then
      begin
        b := true;
        rm := FPressedItem.Right - 4;
        if FPressedItem.Expandable then
        begin
          if ((rm + 13 - FPressedItem.Width) + (ResizeWidth - 4) < RM)
            and InRange(rm - 6 - (ResizeWidth - 4), RM, Message.XPos)
            and InRange(2, 9, Message.YPos) then
          begin
            THS := FPressedItem;
            FPressedItem :=nil;
            THS.Expanded := not THS.Expanded;
            if THS.Expanded then DoSectionExpandEvent(THS) else DoSectionCollapseEvent(THS);
            b := false;
          end;
        end;
        if THS.FLookupEnabled then
        begin
          if ((rm + 13 - THS.Width) + (ResizeWidth - 4) < RM)
            and InRange(rm - 6 - (ResizeWidth - 4), RM, Message.XPos)
            and InRange(12, 16, Message.YPos) then
          begin
            if Flookup <>nil then FLookup.Text := ''
            else
            begin
              Flookup := TComboBox.Create(nil);
              FLookup.Visible := false;
              FLookup.Style := csDropDown;
              //FLookup.BorderStyle := bsNone;
              //FLookup.AutoSize := false;
            end;
            FLookup.Parent := Self;
            FLookup.OnExit := EditExit;
            FLookup.OnKeyDown := EditKeyDown;
            FLookup.OnChange := EditChange;
            FLookup.OnKeyUp := EditKeyUp;
            FLookup.Items.Assign(THS.FLookupHist);
            FLookupSection := THS;
            R := GetSectionRect(THS.Index);
            FLookup.BoundsRect := R;
            FLookup.Font.Assign(Font);
            FLookup.Font.Height:=Abs(Font.Height)-3;
            FLookup.Height := Max(FLookup.Height, R.Bottom - R.Top - 2);
            FDoingLookup := true;
            FLookup.Visible := true;
            FFocusedCtl := GetFocus;
            FLookup.SetFocus;
            b := false;
          end;
        end;
        if b and (FPressedItem.FAllowClick {or FPressed}) then FPressedItem.UpdateSection else FPressedItem:=nil;
      end;
    end; // not Resizing
  end;
  inherited;
end;

procedure TCustomElHeader.WMMouseMove(var Message: TWMMouseMove);
var THS : TElHeaderSection;
    NewPos, NewWidth,L : integer;
begin
  if (FPressed) and (FAllowDrag) then
  begin
    if (FPressedItem<>nil) and (FDropSrc = nil)
       and (sqrt(sqr(FPressCoord.X-Message.XPos)+sqr(FPressCoord.Y-Message.YPos))>=5) then
    begin
      FDropSrc:=FPressedItem;
      FPressed:=false;
      GetDragImage(FPressCoord.X);
      SetCapture(Handle);
      FDropTrg := nil;
      FDragCoord := FPressCoord;
      DragRect := GetSectionRect(FDropSrc.Index);
    end;
  end;
  if (FAllowDrag) and (not FPressed) and (FDropSrc<>nil) then
  begin
    FDropTrg:=GetSectionAt(Message.XPos, Message.YPos);
    if FMoveOnDrag then
    begin
      if (FDropSrc<>FDropTrg) then
      begin
        if FDropTrg=nil then
        begin
          NewPos:=Sections.Count-1;
          FSections.MoveSection(FDropSrc, NewPos);
        end
        else
        if ((FDropSrc.Position < FDropTrg.Position) and (FDropTrg.Right - FDropSrc.Width < Message.XPos))
           or ((FDropSrc.Position > FDropTrg.Position) and (FDropTrg.Left + FDropSrc.Width > Message.XPos)) then
        begin
          NewPos:=FSections.FList.IndexOf(FDropTrg);
          FSections.MoveSection(FDropSrc, NewPos);
        end;
      end;
    end;
    InvalidateRect(Canvas.Handle, @DragRect, false);
    Paint;
    dec(DragRect.Left, FDragCoord.x - Message.XPos);
    dec(DragRect.Right, FDragCoord.x - Message.XPos);
    InvalidateRect(Canvas.Handle, @DragRect, false);
{$IFNDEF VER90}
    if FIsNT then
       MaskBlt(Canvas.Handle, DragRect.Left, DragRect.Top, DragBmp.Width-1, DragBmp.Height-1, DragBmp.Canvas.Handle,
            0, 0, DragBmpMask.MaskHandle, 0, 0,((SRCCOPY SHL 8) and $FF000000) or SRCAND)
    else
{$ENDIF}
       BitBlt(Canvas.Handle, DragRect.Left, DragRect.Top, DragBmp.Width-1, DragBmp.Height-1, DragBmp.Canvas.Handle,
            0, 0, SRCCOPY);
    FDragCoord.X := Message.XPos;

  end;
  if (FTracking) and (not FResizing) and (FDropSrc=nil) and (not (csNoStdEvents in ControlStyle)) then
  begin
    THS:=FTrackSection;
    FTrackSection:=GetSectionAt(Message.XPos, Message.YPos);
    if THS<>FTrackSection then
    begin
      if THS<>nil then THS.UpdateSection;
      if FTrackSection<>nil then FTrackSection.UpdateSection;
      Update;
    end;
  end;
  if (not FPressed) and (not FResizing) and (FDropSrc=nil) then
  begin
    if InResizeArea(Message.XPos, THS) and (not (csDesigning in ComponentState)) and (THS.FResizable) then
    begin
      if Screen.Cursor<>crHSplit then Screen.Cursor:=crHSplit;
    end
    else if Screen.Cursor=crHSplit then Screen.Cursor:=Cursor;
  end;
  if FResizing then
  begin
    L:=FResizeSection.Left;
    if (Message.XPos-FLineTab>=L) { and ((Message.XPos<Width) or ResizeRight) }then
    begin
      NewWidth:=Message.XPos-L-FLineTab;
      with FResizeSection do
        if InRange(MinWidth,MaxWidth,NewWidth) then
        begin
          if FResizeOnDrag then
          begin
            Width:=NewWidth;
            InvalidateRight(FResizeSection.Index);
            Update;
            DoSectionResizing(FResizeSection, trsMove, NewWidth);
          end
          else
          begin
            DrawLine;
            FHeaderLine:=L+NewWidth;
            DrawLine;
            DoSectionResizing(FResizeSection, trsMove, NewWidth);
          end;
        end;
    end; // really in client width
  end;
  inherited;
end;

procedure TCustomElHeader.WMLButtonUp(var Message: TWMLButtonUp);
var THS : TElHeaderSection;
    NewPos : integer;
begin
  if not (csNoStdEvents in ControlStyle) then
  begin
    if FResizing then
    begin
      FResizing:=false;
      Screen.Cursor:=Cursor;
      ReleaseCapture;
      if (not FResizeOnDrag) then
      begin
        if (FHeaderLineVis) then DrawLine;
        ReleaseLineDC;
        FResizeSection.Width:=FHeaderLine-FResizeSection.Left;
      end;
      DoSectionResizing(FResizeSection, trsEnd, FResizeSection.Width);
      DoSectionResize(FResizeSection);
      FResizeSection:=nil;
      Invalidate;
      Update;
    end else
    if FDropSrc <>nil then
    begin
      FDropTrg:=GetSectionAt(Message.XPos, Message.YPos);
      if (FDropSrc<>FDropTrg) then
      begin
        if FDropTrg=nil then NewPos:=Sections.Count-1 else NewPos:=FSections.FList.IndexOf(FDropTrg);
        FSections.MoveSection(FDropSrc, NewPos);
      end;
      THS:=FPressedItem;
      FPressedItem:=nil;
      if THS<>nil then THS.UpdateSection;
      FDropSrc:=nil;
      FDropTrg:=nil;
      Repaint;
      DragBmp.Free;
      DragBmp := nil;
      DragBmpMask.Free;
      ReleaseCapture;
    end else
    begin
      FPressed:=false;
      THS:=FPressedItem;
      FPressedItem:=nil;
      if THS<>nil then
      begin
        THS.UpdateSection;
        if (GetSectionAt(FPressCoord.X,FPressCoord.Y) =
            GetSectionAt(Message.XPos,Message.YPos)) then DoSectionClick(THS);
      end;
    end;
  end;
  inherited;
end;

procedure TCustomElHeader.CMMouseLeave;
var THS:TElHeaderSection;
begin
  FDropTrg:=nil;             
  FPressed:=false;
  if FPressedItem<>nil then
  begin
    THS:=FPressedItem;
    THS.UpdateSection;
  end;
  if FTrackSection<>nil then
  begin
    THS:=FTrackSection;
    FTrackSection:=nil;
    THS.UpdateSection;
  end;
  if (not FResizing) and (Screen.Cursor=crHSplit) then Screen.Cursor:=Cursor;    
end;

procedure TCustomElHeader.CMMouseEnter;
begin
  inherited;
  if (FPressedItem<>nil) and (FDropSrc=nil) then
  begin
    FPressed:=true;
    FPressedItem.UpdateSection;
  end;  
end;

procedure TCustomElHeader.CMDrag(var Message: TCMDrag);
var TSI : TElHeaderSection;
begin
  inherited;
  with Message, DragRec^ do
    case DragMessage of
//      dmDragMove: with ScreenToClient(Pos) do DoDragOver(Source, X, Y, Message.Result<>0);
      dmDragLeave:
        begin
          TSI:=FDropTrg;
          FDropTrg := nil;
          if TSI<>nil then
          begin
            TSI.UpdateSection;
            Update;
          end;
        end;
      dmDragDrop:
        begin
          TSI:=FDropTrg;
          FDropTrg := nil;
          if TSI<>nil then
          begin
            TSI.UpdateSection;
            Update;
          end;
        end;
      dmDragCancel:
        begin
          TSI:=FDropTrg;
          FDropTrg := nil;
          if TSI<>nil then
          begin
            TSI.UpdateSection;
            Update;
          end;
          FDropSrc:=nil;
        end;
    end;
end;

function TCustomElHeader.GetSectionRect(SectionNum:integer):TRect;
var THS:TElHeaderSection;
    i:integer;
begin
  THS:=Sections.FindSection(SectionNum);
  i:=THS.Left;
  result:=Rect(i,0,i+THS.Width,ClientHeight);
end;

function TCustomElHeader.DoGetPicture;
begin
  result:=-1;
  if Assigned(FOnPictureNeeded) then FOnPictureNeeded(self, Section, result);
end;

procedure TCustomElHeader.RedrawSection(Canvas : TCanvas; Section:TElHeaderSection; R:TRect);
var TS:TElHeaderSection;
    R1,R2:TRect;
    w:integer;
    DC:THandle;
    s:string;
    SaveCol : TColor;
    SaveStyle : TFontStyles;
    StImIndex : integer;
    BMP       : TBitmap;
    AL: integer;
begin
  TS:=Section;
  if TS.FParentColor
     then Canvas.Brush.Color := Color
     else Canvas.Brush.Color := TS.FColor;
  Canvas.FillRect(R);
  DC := Canvas.Handle;
  if (TS=FPressedItem) and (FPressed=true) then
  begin
    DrawEdge(DC, R, BDR_SUNKENINNER, BF_TOPLEFT);              { black     }
    DrawEdge(DC, R, BDR_SUNKENOUTER, BF_BOTTOMRIGHT);          { btnhilite }
    InflateRect(R, -1, -1);
    OffsetRect(R, 1, 1);
  end else
  begin
    DrawEdge(DC, R, BDR_RAISEDOUTER, BF_BOTTOMRIGHT);          { black }
    Dec(R.Bottom);
    Dec(R.Right);
    DrawEdge(DC, R, BDR_RAISEDINNER, BF_TOPLEFT);              { btnhilite }
    Inc(R.Top);
    Inc(R.Left);
    DrawEdge(DC, R, BDR_RAISEDINNER, BF_BOTTOMRIGHT); { btnshadow }
  end;
  inc(R.Top); inc(R.Left); dec(R.Bottom); dec(R.Right, 4);
  if (TS.FExpandable or TS.FLookupEnabled) and (R.Left + 6 + (ResizeWidth - 4) < R.Right) then
  begin
    if TS.Expandable then
    begin
      if TS.Expanded then BMP := FLeftBmp else BMP := FRightBmp;
      R2 := Rect(R.Right - 5, R.Top, R.Right, R.Top + 7);
      Canvas.BrushCopy(R2, Bmp, Rect(0,0, R2.Right - R2.Left, R2.Bottom - R2.Top), BMP.TransparentColor)
    end;
    if TS.FLookupEnabled then
    begin
      BMP := FPointBmp;
      R2 := Rect(R.Right - 5, R.Top + 10, R.Right, Min(R.Top + 14, R.Bottom));
      Canvas.BrushCopy(R2, Bmp, Rect(0,0, R2.Right - R2.Left, R2.Bottom - R2.Top), BMP.TransparentColor)
    end;
    Dec(R.Right, 8);
  end;
  R1:=R;
  //if (TS=FPressedItem) and (FPressed=true) then inc(R1.Left,2);
  if TS.FSortMode<>hsmNone then
  begin
    if R1.Right-8 < R1.Left+ResizeWidth then
    begin
      w:=0;
      R1.Right:=R1.Left+ResizeWidth;
    end else
    begin
      w:=8;
      dec(R1.Right, 9);
    end;
    R2:=Rect(R1.Right, ((R1.Bottom - R.Top) div 2 + R.Top) - 2, R1.Right+ W +1, ((R1.Bottom - R.Top) div 2 + R.Top) + 3);
    if TS.FParentColor
       then Canvas.Brush.Color := Color
       else Canvas.Brush.Color := TS.FColor;
    if w>0 then
    begin
      if TS.FSortMode=hsmAscend then BMP := FAscBmp else BMP := FDescBmp;
      Canvas.BrushCopy(R2,Bmp,Rect(0,0,9,6),FDescBmp.TransparentColor);
      dec(R1.Right, 3);
    end;
  end ;
  if TS.FResizable then inc(R1.Left,ResizeWidth);
  if R1.Right<R1.Left then exit;
  if TS.FStyle=ElhsOwnerDraw then
  begin
    DoSectionDraw(TS,R1,(TS=FPressedItem) and (FPressed=true));
  end
  else
  begin
    if TS.FStyle = elhsPictureOnly then
    begin
      if (FImages<>nil) then
      begin
        StImIndex:=TS.FImageIndex;
        if StImIndex=-1 then StImIndex:=DoGetPicture(TS);
        if InRange(0,FImages.Count-1,StImIndex) then
        begin
          BMP:=TBitmap.Create;
          BMP.Width:=FImages.Width;
          BMP.Height:=FImages.Height;
          if R.Left<R.Right then
          begin
            BMP.Canvas.Brush.Color:=Canvas.Brush.Color;
            BMP.Canvas.FillRect(Rect(0,0, FImages.Width, FImages.Height));
            FImages.Draw(BMP.Canvas,0,0,StImIndex);
            R2:=Rect(Max(R1.Left + ((R1.Right-R1.Left+1) div 2)-(FImages.Width div 2), R1.Left),
                     R1.Top+((R1.Bottom-R1.Top+1) div 2)-(FImages.Height div 2),
                     Min(R1.Right, R1.Left + ((R1.Right-R1.Left+1) div 2)-(FImages.Width div 2)+FImages.Width),
                     R1.Top+FImages.Height);
            Canvas.CopyRect(R2, BMP.Canvas, Rect(0, 0, min(FImages.Width, Min(R1.Right, R1.Left+FImages.Width-1)- R1.Left+1), FImages.Height));
          end;
          inc(R1.Left,FImages.Width);
          BMP.Free;
        end;
      end;
      exit;
    end;
    if (FImages<>nil) and (TS.FPicAlign=hsaLeft) then
    begin
      StImIndex:=TS.FImageIndex;
      if StImIndex=-1 then StImIndex:=DoGetPicture(TS);
      if InRange(0,FImages.Count-1,StImIndex) then
      begin
        BMP:=TBitmap.Create;
        BMP.Width:=FImages.Width;
        BMP.Height:=FImages.Height;
        if R.Left<R.Right then
        begin
          BMP.Canvas.Brush.Color:=Canvas.Brush.Color;
          BMP.Canvas.FillRect(Rect(0,0, FImages.Width, FImages.Height));
          FImages.Draw(BMP.Canvas,0,0,StImIndex);
{$IFNDEF VER90}
          BMP.TransparentMode:=tmAuto;
{$ENDIF}
          R2:=Rect(R1.Left, R1.Top+((R1.Bottom-R1.Top+1) div 2)-(FImages.Height div 2),Min(R1.Right,R1.Left+FImages.Width), R1.Top+FImages.Height);
          Canvas.CopyRect(R2, BMP.Canvas, Rect(0, 0, min(FImages.Width, Min(R1.Right, R1.Left+FImages.Width-1)-R1.Left+1), FImages.Height));
        end;
        inc(R1.Left,FImages.Width);
        BMP.Free;
      end;
    end;
    if R1.Right<R1.Left then exit;
    if TS.FStyle = elhsText then
    begin
      s:=TS.Text;
      if FRightAlignedText then
      begin
        Al:=DT_RIGHT;
        case TS.FAlignment of
          hsaCenter: Al:=DT_CENTER;
          hsaRight:  Al:=DT_LEFT;
        end;    // case
      end else
      begin
        Al:=DT_LEFT;
        case TS.FAlignment of
          hsaCenter: Al:=DT_CENTER;
          hsaRight:  Al:=DT_RIGHT;
        end;    // case
      end;
      if RightAlignedText then AL := AL or DT_RTLREADING;
      if TS=FTrackSection then
      begin
        SaveCol:=Canvas.Font.Color;
        SaveStyle:=Canvas.Font.Style;
        Canvas.Font.Color:= clHighlight;
        Canvas.Font.Style:= Canvas.Font.Style+[fsUnderline];
        DrawText(Canvas.Handle, PChar(S), -1, R1,
            DT_NOPREFIX or DT_SINGLELINE or DT_TOP or DT_END_ELLIPSIS or AL);
        Canvas.Font.Color:=SaveCol;
        Canvas.Font.Style:=SaveStyle;
      end else
      begin
        if TS.ParentColor
           then Canvas.Font.Color := Font.Color
           else Canvas.Font.Color := TS.FontColor;
        DrawText(Canvas.Handle, PChar(S), -1, R1,
              DT_NOPREFIX or DT_SINGLELINE or DT_TOP or DT_END_ELLIPSIS or AL);
      end;
    end;
    inc(R1.Left, Canvas.TextWidth(S)+3);
    if R1.Right<R1.Left then exit;
  end;
  if (FImages<>nil) and (TS.FPicAlign = hsaRight) then
  begin
    StImIndex:=TS.FImageIndex;
    if StImIndex=-1 then StImIndex:=DoGetPicture(TS);
    if InRange(0,FImages.Count-1,StImIndex) then
    begin
      BMP:=TBitmap.Create;
      BMP.Width:=FImages.Width;
      BMP.Height:=FImages.Height;
      if R.Left<R.Right then
      begin
        BMP.Canvas.Brush.Color:=Canvas.Brush.Color;
        BMP.Canvas.FillRect(Rect(0,0, FImages.Width, FImages.Height));
        FImages.Draw(BMP.Canvas,0,0,StImIndex);
{$IFNDEF VER90}
        BMP.TransparentMode:=tmAuto;
{$ENDIF}
        R2:=Rect(R1.Left, R1.Top+((R1.Bottom-R1.Top+1) div 2)-(FImages.Height div 2),Min(R1.Right,R1.Left+FImages.Width), R1.Top+FImages.Height);
        Canvas.CopyRect(R2, BMP.Canvas, Rect(0, 0, min(FImages.Width, Min(R1.Right, R1.Left+FImages.Width-1)-R1.Left+1), FImages.Height));
      end;
      inc(R1.Left,FImages.Width);
      BMP.Free;
    end;
  end;
end;

procedure TCustomElHeader.RedrawSections;
var R:TRect;
    R1:TRect;
    i:integer;

begin
  for i:=0 to Sections.Count-1 do
  begin
    R:=GetSectionRect(i);
    if IntersectRect(R1, R, Canvas.ClipRect) and (Sections[i].Visible)
       then RedrawSection(Canvas, Sections[i],R);
  end;
  R:=ClientRect;
  R.Left:=SectionsWidth;
  Canvas.Brush.Color:=Color;
  Canvas.FillRect(R);
  DrawEdge(Canvas.Handle, R, BDR_RAISEDINNER, BF_RECT{BF_BOTTOMRIGHT});
end;

procedure TCustomElHeader.AdjustStickySize(Caller : TElHeaderSection);
var i, cw, sw : integer;
    HS : TElHeaderSection;

begin
  i := 0;
  while (SectionsWidth >= ClientWidth) and (ClientWidth > 0) and (i > -1) do
  begin
    i := FSections.Count - 1;
    while i >= 0 do
    begin
      HS := FSections.ItemByPos[i];
      cw := ClientWidth;
      sw := SectionsWidth;
      if HS.Visible and (HS.Width > 0) and (HS.Resizable) and (HS <> Caller) then
      begin
        if HS.Width > (sw - cw - 1)
           then Hs.FWidth := hs.Width - (sw - cw) - 1
           else HS.FWidth := 0;
        //HS.FStickySize := HS.Width / cw;
        break;
      end;
      dec(i);
    end;    // while
  end;
  i := 0;
  while (SectionsWidth >= ClientWidth) and (ClientWidth > 0) and (i > -1) do // we probably have non-resizable sections
  begin
    i := FSections.Count - 1;
    while i >= 0 do
    begin
      HS := FSections.ItemByPos[i];
      cw := ClientWidth;
      sw := SectionsWidth;
      if HS.Visible and (HS.Width > 0) and ((not HS.Resizable) or (HS = Caller)) then
      begin
        if HS.FSaveSize = -1 then HS.FSaveSize := HS.Width; // this is the size we have to save, cause later we'll have to restore it
        if HS.Width > (sw - cw - 1)
           then HS.FWidth := hs.Width - (sw - cw) - 1
           else HS.FWidth := 0;
        break;
      end;
      dec(i);
    end;
  end;

  // first restore "non-resizable" sections
  i := 0;
  while (SectionsWidth < ClientWidth - 1) and (i > -1) do
  begin
    i := FSections.Count - 1;
    while i >= 0 do
    begin
      HS := FSections.ItemByPos[i];
      cw := ClientWidth;
      sw := SectionsWidth;
      if HS.Visible and (not HS.Resizable) and (HS.Width < HS.FSaveSize) and (HS.FSaveSize <> -1) then
      begin
        HS.FWidth := min(hs.Width + (cw - sw) - 1, HS.FSaveSize);
        if HS.FWidth = HS.FSaveSize then HS.FSaveSize := -1 else break;
      end;
      dec(i);
    end;    // while
  end;

  i := FSections.Count - 1;
  while i >= 0 do
  begin
    HS := FSections.ItemByPos[i];
    cw := ClientWidth;
    sw := SectionsWidth;
    if HS.Visible and (HS.Resizable) and (HS <> Caller) then
    begin
      HS.FWidth := hs.Width + (cw - sw) - 1;
      //HS.FStickySize := HS.Width / cw;
      break;
    end;
    dec(i);
  end;    // while
end;

function TCustomElHeader.GetResizableWidth : integer;
var i, w : integer;
    HS : TElHeaderSection;
begin
  w := Width;
  for i := 0 to FSections.Count - 1 do    // Iterate
  begin
    HS := FSections[i];
    //if HS.Visible and (HS.Resizable) then w := w + HS.Width;
    if HS.Visible and (not HS.Resizable) then w := w - HS.Width;
  end;    // for
  result := w;
end;

procedure TCustomElHeader.MarkStickySections;
var w, i : integer;
    HS : TElHeaderSection;
begin
  W := GetResizableWidth;
  for i := 0 to FSections.Count - 1 do
  begin
    HS := FSections.ItemByPos[i];
    if HS.Resizable and HS.Visible then HS.FStickySize := HS.FWidth / w;
  end;
end;

procedure TCustomElHeader.Resize;
var i, w : integer;
    HS : TElHeaderSection;
begin
  if Images <>nil then
    ClientHeight:=Max(Max(Images.Height, ClientHeight), Abs(Font.Height)+6)
  else
    ClientHeight:=Max(ClientHeight, Abs(Font.Height)+6);
  if StickySections then
  begin
    FInStick := true;
    if FOldWidth < ClientWidth then
      for i := FSections.Count - 1 downto 0 do    // Iterate
      begin
        HS := FSections.ItemByPos[i];
        if (not HS.Resizable) and (HS.Visible) and (HS.FSaveSize <> -1) then
        begin
          HS.Width := min (HS.FSaveSize, ClientWidth);
          if HS.Width = HS.FSaveSize then HS.FSaveSize := -1;
        end;
      end;
    W := GetResizableWidth;
    for i := 0 to FSections.Count - 1 do    // Iterate
    begin
      HS := FSections.ItemByPos[i];
      if HS.Resizable and HS.Visible then HS.Width := Round(w * HS.FStickySize);
    end;    // for
    AdjustStickySize(nil);
    FOldWidth := ClientWidth;
    FInStick := false;
  end;
  if Assigned(FOnResize) then FOnResize(Self);
end;

procedure TCustomElHeader.AllocateLineDC;
begin
  FLineDC := GetDCEx(Handle, 0, DCX_CACHE or DCX_CLIPSIBLINGS
    or DCX_LOCKWINDOWUPDATE);
end;

procedure TCustomElHeader.ReleaseLineDC;
begin
  ReleaseDC(Handle, FLineDC);
end;

procedure TCustomElHeader.DrawLine;
begin
  FHeaderLineVis:= not FHeaderLineVis;
  PatBlt(FLineDC, FHeaderLine, 0, 1, Height, PATINVERT);
end;

procedure TCustomElHeader.Paint;
begin
  FPainting:=true;
  RedrawSections;
  FPainting:=false;
end;

procedure TCustomElHeader.SetSections;
begin
  FSections.Assign(value);
end;

{$IFNDEF LITE}
procedure TCustomElHeader.SetImages(newValue : TElImageList);
{$ELSE}
procedure TCustomElHeader.SetImages(newValue : TImageList);
{$ENDIF}
{ Sets data member FImages to newValue. }
begin
  if (FImages <> newValue) then
  begin
    if FImages <> nil then
      FImages.UnRegisterChanges(FImageChangeLink);
    FImages:=newValue;
    if FImages<>nil then
    begin
     FImages.RegisterChanges(FImageChangeLink);
     ClientHeight := Max(Max(Images.Height+6, ClientHeight), Abs(Font.Height) + 6);
    end else ClientHeight := Abs(Font.Height) + 6;
    if FPainting then exit;
    Repaint;
  end;  { if }
end;  { SetImages }

procedure TCustomElHeader.OnImageListChange(Sender : TObject);  { private }
begin
  if FPainting then exit;
  Repaint;
end;  { OnImageListChange }

procedure TCustomElHeader.DoVisChanged(Section : TElHeaderSection);  { protected }
begin
  if Assigned (FOnVisibleChange) then FOnVisibleChange(Self, Section);
end;  { DoVisChanged }

procedure TCustomElHeader.DoSectionDelete(Section : TElHeaderSection);  { protected }
begin
  if assigned(FOnSectionDelete) then FOnSectionDelete(Self, Section);
end;  { DoSectionDelete }

procedure TCustomElHeader.DoSectionMove(Section:TElHeaderSection; OldPos, NewPos : integer);  { protected }
begin
  if assigned(FOnSectionMove) then FOnSectionMove(Self, Section, OldPos, NewPos);
end;  { DoSectionMove }

procedure TCustomElHeader.DoSectionResizing(Section:TElHeaderSection; State: TElHResizingStates; NewWidth : integer);  { protected }
begin
  if Assigned(FOnSectionResizing) then FOnSectionResizing(Self, Section, State, NewWidth);
end;  { DoSectionResizing }

procedure TCustomElHeader.DoSectionResize(Section : TElHeaderSection);  { protected }
begin
  if Assigned(FOnSectionResize) then FOnSectionResize(Self, Section);
end;  { DoSectionResize }

procedure TCustomElHeader.DoSectionClick(Section : TElHeaderSection);  { protected }
begin
  if (Assigned(FOnSectionClick)) then FOnSectionClick(Self, Section);
end;  { DoSectionClick }

procedure TCustomElHeader.DoSectionDraw(Section : TElHeaderSection; R:TRect; Pressed:boolean);  { protected }
begin
  if Assigned(FOnSectionDraw) then FOnSectionDraw(Self,Section, R, Pressed);
end;  { DoSectionDraw }

procedure TCustomElHeader.DoNotifySectionChange(Section : TElHeaderSection; Change : TSectionChangeMode);
begin
  if (assigned(FOnSectionChange)) then
    FOnSectionChange(Self, Section , Change );
end;  { TriggerSectionChangeEvent }

procedure TCustomElHeader.OnFontChange(Sender : TObject);  { protected }
begin
  if Images <>nil then
    ClientHeight:=Max(Max(Images.Height, ClientHeight), Abs(Font.Height)+6)
  else
    ClientHeight:=Max(ClientHeight, Abs(Font.Height)+6);
  Canvas.Font.Assign(Font);
  Repaint;
end;  { OnFontChange }

procedure TCustomElHeader.DoSectionCreate(Section : TElHeaderSection);
begin
  if (assigned(FOnSectionCreate)) then FOnSectionCreate(Self, Section );
end;  { TriggerSectionCreateEvent }

procedure TCustomElHeader.GetDragImage;  { private }
var //i : integer;
    R : TRect;
    S : TElHeaderSection;

begin
  S:=GetSectionAt(XPos, 0);
  if S = nil then exit;
  R :=GetSectionRect(S.Index);
  DragRect := R;
  DragBmp := TBitmap.Create;
  DragBmp.Height := R.Bottom-R.Top+1;
  DragBmp.Width  := R.Right - R.Left+1;
  DragBmp.Canvas.Brush.Color := Color;
  R:= Rect(0, 0, DragBmp.Width-1, DragBmp.Height-1);
  DragBmp.Canvas.FillRect(R);
  DragBmpMask:=TBitmap.Create;
  DragBmp.Canvas.Font.Assign(Font);
  RedrawSection(DragBmp.Canvas, S, R);
  DragBmpMask.Width:=DragBmp.Width;
  DragBmpMask.Height:=DragBmp.Height;
  DragBmpMask.Canvas.CopyRect(R, DragBmp.Canvas, R);
{$IFNDEF VER90}
  DragBmpMask.Mask(Color);
{$ENDIF}
end;  { GetDragImage }

procedure TCustomElHeader.SetTracking (newValue : boolean);
begin
  if FTracking <> newValue then
  begin
    FTracking := newValue;
    if not FTracking then
    begin
      FTrackSection := nil;
      Repaint;
    end;
  end;
end;

procedure TCustomElHeader.WMEraseBkgnd(var Msg : TWMEraseBkgnd);  { private }
begin
  Msg.Result := 1;
end;  { WMEraseBkgnd }

{$IFNDEF LITE}

procedure TCustomElHeader.Save;
var SaveKey, CKey : string;
    i : integer;
begin
  if Assigned(FStorage) then
  begin
    SaveKey := FStorage.CurrentKey;
    if FStorage.OpenKey(StoragePath + FStorage.Delimiter + 'ElHeader', true) then
    begin
      FStorage.WriteInteger(CKey, 'Width', Width);
      for i := 0 to FSections.Count - 1 do    // Iterate
      begin
        CKey := 'Section' + IntToStr(i);
        FStorage.WriteBool(CKey, 'Visible', FSections[i].FVisible);
        FStorage.WriteBool(CKey, 'Expanded', FSections[i].Expanded);
        FStorage.WriteInteger(CKey, 'Width', FSections[i].Width);
        FStorage.WriteInteger(CKey, 'SortOrder', Integer(FSections[i].SortMode));
      end;    // for
      FStorage.OpenKey(SaveKey, false);
    end;
  end;
end;  {Save}

procedure TCustomElHeader.Restore;
var SaveKey, CKey : string;
    i, j, k : integer;
    b : boolean;
begin
  inc(LoadingCount);
  if Assigned(FStorage) then
  begin
    SaveKey := FStorage.CurrentKey;
    if FStorage.OpenKey(StoragePath + FStorage.Delimiter + 'ElHeader', false) then
    begin
      FStorage.ReadInteger(CKey, 'Width', Width, k);
      for i := 0 to FSections.Count - 1 do    // Iterate
      begin
        CKey := 'Section' + IntToStr(i);
        FStorage.ReadBool(CKey, 'Visible', FSections[i].FVisible, FSections[i].FVisible);
        if FStorage.ReadBool(CKey, 'Expanded', FSections[i].Expanded, b) then FSections[i].Expanded := b;
        if FStorage.ReadInteger(CKey, 'Width', FSections[i].Width, j) then
          if ((FSections[i].Resizable) and (StickySections)) then FSections[i].Width := Trunc(j * (Width/k)) else FSections[i].Width := j;
        if FStorage.ReadInteger(CKey, 'SortOrder', Integer(FSections[i].SortMode), j) then FSections[i].SortMode := TElSSortMode(j);
      end;    // for
      FStorage.OpenKey(SaveKey, false);
    end;
  end;
  dec(LoadingCount);
  b := FStickySections;
  FStickySections := false;
  SetStickySections(b);
end;  {Restore}
{$ENDIF}

procedure TCustomElHeader.EditExit(Sender : TObject);
begin
  if FDoingLookup then
  begin
    FLookup.Visible := false;
    FDoingLookup := false;
    DoSectionLookupDoneEvent(FLookupSection, FLookup.Text, false);
  end;
end;

procedure TCustomElHeader.EditChange;
var S : string;
begin
  if not FIgnoreLookupChange then
  begin
    S := FLookup.Text;
    DoSectionLookupEvent(FLookupSection, S);
    if S <> FLookup.Text then
    begin
      FIgnoreLookupChange := true;
      FLookup.Text := S;
      FIgnoreLookupChange := false;
    end;
  end;
end;

procedure TCustomElHeader.EditKeyUp(Sender : TObject; var Key: Word; Shift: TShiftState);
begin
  if (Shift = []) and ((Key = VK_RETURN) or (Key = VK_ESCAPE)) then Key := 0;
end;

procedure TCustomElHeader.EditKeyDown(Sender : TObject; var Key: Word; Shift: TShiftState);
var i : integer;
begin
  if (Shift = []) then
  begin
    if Key = VK_RETURN then
    begin
      FDoingLookup := false;
      DoSectionLookupDoneEvent(FLookupSection, FLookup.Text, true);
      i := FLookupSection.FLookupHist.IndexOf(FLookup.Text);
      if i <> -1 then FLookupSection.FLookupHist.Delete(i);
      FLookupSection.FLookupHist.Insert(0, FLookup.Text);
      FLookup.Visible := false;
      Windows.SetFocus(FFocusedCtl);
      Key := 0;
    end;
    if Key = VK_ESCAPE then
    begin
      FDoingLookup := false;
      DoSectionLookupDoneEvent(FLookupSection, FLookup.Text, false);
      FLookup.Visible := false;
      Windows.SetFocus(FFocusedCtl);
      Key := 0;
    end;
  end;
end;
{$IFNDEF LITE}

procedure TCustomElHeader.SetStorage(newValue : TElIniFile);
begin
  if newValue <> FStorage then
  begin
    FStorage := newValue;
    if FStorage <> nil then FStorage.FreeNotification(Self);
  end;
end;

{$ENDIF}
procedure TCustomElHeader.Notification(AComponent: TComponent; Operation: TOperation);
var i : integer;
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
{$IFNDEF LITE}
//    if AComponent = FStorage then FStorage := nil;
{$ENDIF}
    if AComponent = Images then Images := nil;
    if AComponent is TPopupMenu then
    for i := 0 to FSections.Count -1 do
      if TElHeaderSection(FSections.FList[i]).FPopupMenu = AComponent
         then TElHeaderSection(FSections.Item[i]).FPopupMenu := nil;
  end;
end;

procedure TCustomElHeader.DoSectionExpandEvent(Section : TElHeaderSection);
{CDK: Call as needed to trigger event.}
begin
  if (assigned(FOnSectionExpand)) then FOnSectionExpand(Self, Section );
end;

procedure TCustomElHeader.DoSectionCollapseEvent(Section :
  TElHeaderSection);
{CDK: Call as needed to trigger event.}
begin
  if (assigned(FOnSectionCollapse)) then FOnSectionCollapse(Self, Section );
end;

procedure TCustomElHeader.DoSectionLookupEvent;
begin
  if (assigned(FOnHeaderLookup)) then FOnHeaderLookup(Self, Section, Text);
end;  {TriggerHeaderLookupEvent}

procedure TCustomElHeader.DoSectionLookupDoneEvent;
begin
  if (assigned(FOnHeaderLookupDone)) then FOnHeaderLookupDone(Self, Section, Text, Accepted);
end;  {TriggerHeaderLookupDoneEvent}

procedure TCustomElHeader.WMRButtonUp(var Msg : TWMRButtonUp);
begin
  if FResizing then
  begin
    Msg.Result := 0;
    exit;
  end;
  inherited;
end;  {WMRButtonUp}

procedure TCustomElHeader.WMRButtonDown(var Msg: TWMRButtonDown);
var S : TElHeaderSection;
    P : TPoint;
begin
  if FResizing then
  begin
    Msg.Result := 0;
    exit;
  end;
  S := Self.GetSectionAt(Msg.XPos, Msg.YPos);
  P := ClientToScreen(Point(Msg.XPos, Msg.YPos));
  if (S = nil) or (S.PopupMenu = nil) then
  begin
    if Assigned(PopupMenu) and (PopupMenu.AutoPopup) then PopupMenu.Popup(P.X, P.Y);
  end else
    if S.FPopupMenu.AutoPopup then S.FPopupMenu.Popup(P.X, P.Y);
  Msg.Result := 1;
end;

procedure TCustomElHeader.Loaded;
var i : integer;
    O : TComponent;
    Form : TForm;
    b : boolean;
begin
  inherited Loaded;
  Form := ElVCLUtils.GetOwnerForm(Self);
  if Form = nil then exit;
  for i := 0 to FSections.Count -1 do
  begin
    if FSections[i].FPopupName <> '' then
    begin
      O := Form.FindComponent(FSections[i].FPopupName);
      if (O <> nil) and (O is TPopupMenu) then FSections[i].FPopupMenu := TPopupMenu(O);
    end;
  end;
  b := FStickySections;
  FStickySections := false;
  SetStickySections(b);
  if RightAlignedOrder then
  begin
    FRightAlignedOrder := false;
    RightAlignedOrder  := true; 
  end;
end;

procedure TCustomElHeader.WMSize(var Msg: TWMSize);
begin
  inherited;
{$IFNDEF D_4_UP}
  Resize;
{$ENDIF}
end;

procedure TCustomElHeader.SetStickySections(newValue : Boolean);
var i, w : integer;
    HS : TElHeaderSection;
begin
  if IsLoading then
  begin
    FStickySections := newValue;
    exit;
  end;
  //if (FStickySections <> newValue) then
  begin
    FStickySections := newValue;
    if newValue then
    begin
      w := GetResizableWidth;
      for i := 0 to FSections.Count - 1 do    // Iterate
      begin
        HS := FSections[i];
        if HS.Visible and (HS.Resizable) and (not IsLoading)
           then HS.FStickySize := HS.Width / w else HS.FStickySize := 0;
      end;    // for
    end;
  end;  {if}
end;  {SetStickySections}

function TCustomElHeader.IsLoading : Boolean;
begin
  if Owner <> nil then
  begin
    result := (csLoading in ComponentState) or (LoadingCount > 0);
    if not result then result := csLoading in Owner.ComponentState;
  end else
  begin
    result := csLoading in ComponentState;
  end;
end;  {IsLoading}

procedure TCustomElHeader.SetRightAlignedText(newValue: Boolean);
begin
  if (FRightAlignedText <> newValue) then
  begin
    FRightAlignedText := newValue;
    RecreateWnd;
  end;  {if}
end;

procedure TCustomElHeader.CreateParams(var Params: TCreateParams);
begin
  inherited;
  if RightAlignedText then
    begin
      Params.ExStyle := Params.ExStyle or WS_EX_RTLREADING or WS_EX_LEFTSCROLLBAR or WS_EX_RIGHT;
    end;
end;

procedure TCustomElHeader.SetRightAlignedOrder(newValue: Boolean);
var FList : TElList;
    i : integer;
begin
  if (FRightAlignedOrder <> newValue) then
  begin
    FRightAlignedOrder := newValue;
    if csLoading in ComponentState then exit;
    FList := TElList.Create;
    for i := 0 to FSections.FList.Count -1 do
      FList.Add(FSections.FList[FSections.FList.Count - i - 1]);
    FSections.FList.Clear;
    for i := 0 to FList.Count -1 do
      FSections.FList.Add(FList[i]);
    FList.Free;
  end;  {if}
end;

function TElHeaderSections.GetSectionsOrder : String;
var i : integer;
    Section : TElHeaderSection;

begin
  result := '';
  for i := 0 to FList.Count - 1 do
  begin
    Section := TElHeaderSection(FList[i]);
    result := result + 'i'+IntToStr(Section.Index)+':w'+IntToStr(Section.Width)+':v';
    if Section.Visible then result := result + 't,' else result := result + 'f,';
  end;
end;  { GetSectionsOrder }

procedure TElHeaderSections.SetSectionsOrder(newValue : String);
var i : integer;
    Section : TElHeaderSection;
    s : string;
begin
  i := 0;
  while newValue <> '' do
  begin
    s := copy (newValue, 2, pos(':', newValue)-2);
    delete(newValue, 1, pos(':', newValue));
    Section := Item[StrToInt(s)];
    MoveSection(Section, i);
    s := copy (newValue, 2, pos(':', newValue)-2);
    delete(newValue, 1, pos(':', newValue));
    Section.Width := StrToInt(s);
    if (pos('vf,', newValue) = 1) or (pos('vf;', newValue) = 1)
       then Section.Visible := false
       else Section.Visible := true;
    delete(newValue, 1, 3);
    inc(i);
  end;
end;  { SetSectionsOrder }

end.

