unit ElVCLUtils;

interface

uses Classes, Windows, Controls, Forms, Graphics, ElTools;

type TElBkGndType = (bgtTileBitmap, bgtStretchBitmap, bgtColorFill, bgtCenterBitmap);

function GetOwnerForm (Component : TComponent) : TForm;

function GetShiftState : TShiftState;

function HitTest(R : TRect; Pt : TPoint; CornerSize, BorderSize : integer) : integer;

procedure TiledPaint(Canvas : TCanvas; Bitmap : TBitmap; Rect : TRect);
// destination rectangle is meant.

implementation

procedure TiledPaint(Canvas : TCanvas; Bitmap : TBitmap; Rect : TRect);
var CurRect : TRect;
    CurLeft,
    CurTop  : integer;
begin
  if not IsWinNT then
  begin
    CurTop  := Rect.Top;
    while CurTop < Rect.Bottom do
    begin
      CurLeft := Rect.Left;
      while CurLeft < Rect.Right do
      begin
        CurRect := Classes.Rect(CurLeft, CurTop, Min(CurLeft + Bitmap.Width, Rect.Right), Min(CurTop + Bitmap.Height, Rect.Bottom));
        BitBlt(Canvas.Handle, CurLeft, CurTop, CurRect.Right - CurRect.Left, CurRect.Bottom - CurRect.Top, Bitmap.Canvas.Handle, 0, 0, SRCCOPY);
        Inc(CurLeft, Bitmap.Width);
      end;
      Inc(CurTop, Bitmap.Height);
    end;
  end else
  begin
    Canvas.Brush.Bitmap := TBitmap.Create;
    Canvas.Brush.Bitmap.Assign(Bitmap);
    Canvas.FillRect(Rect);
    Canvas.Brush.Bitmap.Free;
    Canvas.Brush.Bitmap := nil;
  end;
end;

function HitTest(R : TRect; Pt : TPoint; CornerSize, BorderSize : integer) : integer;
begin
  if not PtInRect(R, Pt) then
  begin
    result := HTNOWHERE;
    exit;
  end;
  if InRange(R.Left, R.Left + BorderSize - 1, Pt.X) then
  begin
    if Pt.Y <= (R.Top + CornerSize) then
    begin
      result := HTTOPLEFT;
      exit;
    end;
    if Pt.Y >= (R.Bottom - CornerSize -1) then
    begin
      result := HTBOTTOMLEFT;
      exit;
    end;
    result := HTLEFT;
    exit;
  end else
  if InRange(R.Top, R.Top + BorderSize - 1, Pt.Y) then
  begin
    if PT.X <= (R.Left + CornerSize) then
    begin
      result := HTTOPLEFT;
      exit;
    end;
    if PT.x >= (R.Right - CornerSize -1)  then
    begin
      result := HTTOPRIGHT;
      exit;
    end;
    result := HTTOP;
    exit;
  end
  else
  if InRange(R.Bottom - BorderSize, R.Bottom - 1, Pt.Y) then
  begin
    if PT.X <= (R.Left + CornerSize) then
    begin
      result := HTBOTTOMLEFT;
      exit;
    end;
    if PT.x >= (R.Right - CornerSize -1) then
    begin
      result := HTBOTTOMRIGHT;
      exit;
    end;
    result := HTBOTTOM;
    exit;
  end else
  if InRange(R.Right - BorderSize, R.Right - 1, Pt.X) then
  begin
    if Pt.Y <= (R.Top + CornerSize) then
    begin
      result := HTTOPRIGHT;
      exit;
    end;
    if Pt.Y >= (R.Bottom - CornerSize -1) then
    begin
      result := HTBOTTOMRIGHT;
      exit;
    end;
    result := HTRIGHT;
    exit;
  end;
  result := HTCLIENT;
end;

function GetShiftState : TShiftState;
begin
  result := [];
  if GetKeyState(VK_MENU) < 0 then Include(Result, ssAlt);
  if GetKeyState(VK_CONTROL) < 0 then Include(Result, ssCtrl);
  if GetKeyState(VK_Shift) < 0 then Include(Result, ssShift);  
end;

function GetOwnerForm (Component : TComponent) : TForm;
begin
  result := nil;
  while (not (Component is TForm)) and (Component.Owner <> nil) do
    Component := Component.Owner;
  if Component is TForm then result := (Component as TForm);    
end;

end.
 
