unit frmSectProp;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DsgnIntf, ElHeader, StdCtrls, frmSectEdit;

type
  TElSectionsPropDlg = class(TForm)
    SecList: TListBox;
    AddBtn: TButton;
    DeleteBtn: TButton;
    UpBtn: TButton;
    DownBtn: TButton;
    LoadBtn: TButton;
    SaveBtn: TButton;
    OpenDlg: TOpenDialog;
    SaveDlg: TSaveDialog;
    EditBtn: TButton;
    procedure LoadBtnClick(Sender: TObject);
    procedure SaveBtnClick(Sender: TObject);
    procedure EditBtnClick(Sender: TObject);
    procedure AddBtnClick(Sender: TObject);
    procedure DeleteBtnClick(Sender: TObject);
    procedure UpBtnClick(Sender: TObject);
    procedure DownBtnClick(Sender: TObject);
    procedure SecListClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure SecListKeyPress(Sender: TObject; var Key: Char);
  protected
  public
    DSect : TElHeaderSections;
  end;

  TElSectionsProperty = class(TClassProperty)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
    function GetValue:string; override;
  end;

var
  ElSectionsPropDlg: TElSectionsPropDlg;

implementation

{$R *.DFM}

procedure TElSectionsProperty.Edit;
var Editor: TElSectionsPropDlg;
    j : integer;
begin
  Editor:=nil;
  try
    Editor:=TElSectionsPropDlg.Create(Application);
    Editor.DSect:=TElHeaderSections(GetOrdValue);
    for j := 0 to Editor.DSect.Count-1 do    // Iterate
      Editor.SecList.Items.AddObject(Editor.DSect.ItemByPos[j].Text +
        '  (index: '+IntToStr(Editor.DSect.ItemByPos[j].Index)+')', Editor.DSect.ItemByPos[j]);
    Editor.ShowModal;
  finally
    Editor.Free;
  end;
end;

function TElSectionsProperty.GetAttributes;
begin
  Result := [paDialog, paReadOnly];
end;

function TElSectionsProperty.GetValue;
begin
  FmtStr(Result, '(%s)', [GetPropType^.Name]);
end;

procedure TElSectionsPropDlg.LoadBtnClick(Sender: TObject);
var j:integer;
    Form : TCustomForm;
begin
  if not OpenDlg.Execute then exit;
  DSect.LoadFromFile(OpenDlg.FileName);
  SecList.Items.Clear;
  for j := 0 to DSect.Count-1 do    // Iterate
    SecList.Items.AddObject(DSect.ItemByPos[j].Text + '  (index: '+IntToStr(DSect.ItemByPos[j].Index)+')', DSect.ItemByPos[j]);
  Form := GetParentForm(DSect.Owner);
  if (Form <> nil) and (Form.Designer <> nil)
     then Form.Designer.Modified;
end;

procedure TElSectionsPropDlg.SaveBtnClick(Sender: TObject);
begin
  if not SaveDlg.Execute then exit;
  DSect.SaveToFile(SaveDlg.FileName);
end;

procedure TElSectionsPropDlg.EditBtnClick(Sender: TObject);
var SectDlg : TSectEdit;
    THS     : TElHeaderSection;
       Form : TCustomForm;
begin
  SectDlg:=nil;
  if SecList.ItemIndex=-1 then exit;
  try
    SectDlg:=TSectEdit.Create(self);
    THS:=TElHeaderSection(SecList.Items.Objects[SecList.ItemIndex]);
    SectDlg.Item := THS;
    SectDlg.Items:= DSect;
    SectDlg.Form := TForm(GetParentForm(DSect.Owner));
    SectDlg.SetData;
    if SectDlg.ShowModal=mrOk then
    begin
      SectDlg.GetData;
      SecList.Items[SecList.ItemIndex]:=THS.Text+ '  (index: '+IntToStr(THS.Index)+')';
      Form := GetParentForm(DSect.Owner);
      if (Form <> nil) and (Form.Designer <> nil)
         then Form.Designer.Modified;
    end;
  finally
    SectDlg.Free;
  end;
end;

procedure TElSectionsPropDlg.AddBtnClick(Sender: TObject);
var THS: TElHeaderSection;
    Form : TCustomForm;
begin
  THS:=DSect.AddSection;
  THS.Text:='New section';
  SecList.Items.AddObject(THS.Text+ '  (index: '+IntToStr(THS.Index)+')',THS);
  Form := GetParentForm(DSect.Owner);
  if (Form <> nil) and (Form.Designer <> nil)
     then Form.Designer.Modified;
  SecListClick(Sender);
end;

procedure TElSectionsPropDlg.DeleteBtnClick(Sender: TObject);
var THS: TElHeaderSection;
    Form : TCustomForm;
    j : integer;
begin
  if SecList.ItemIndex=-1 then exit;
  THS:=TElHeaderSection(SecList.Items.Objects[SecList.ItemIndex]);
  DSect.DeleteSection(THS);
  SecList.Items.Clear;
  for j := 0 to DSect.Count-1 do    // Iterate
    SecList.Items.AddObject(DSect.ItemByPos[j].Text + '  (index: '+IntToStr(DSect.ItemByPos[j].Index)+')', DSect.ItemByPos[j]);
  Form := GetParentForm(DSect.Owner);
  if (Form <> nil) and (Form.Designer <> nil)
     then Form.Designer.Modified;
  SecListClick(Sender);
end;

procedure TElSectionsPropDlg.UpBtnClick(Sender: TObject);
var i  : integer;
    THS: TElHeaderSection;
  Form : TCustomForm;

begin
  if SecList.ItemIndex=-1 then exit;
  i:=SecList.ItemIndex;
  THS:=TElHeaderSection(SecList.Items.Objects[i]);
  SecList.Items.Exchange(i,i-1);
  DSect.MoveSection(THS,i-1);
  SecList.ItemIndex := i-1; SecListClick(Self);
  Form := GetParentForm(DSect.Owner);
  if (Form <> nil) and (Form.Designer <> nil)
     then Form.Designer.Modified;
end;

procedure TElSectionsPropDlg.DownBtnClick(Sender: TObject);
var i  : integer;
    THS: TElHeaderSection;
    Form : TCustomForm;
begin
  if SecList.ItemIndex=-1 then exit;
  i:=SecList.ItemIndex;
  THS:=TElHeaderSection(SecList.Items.Objects[i]);
  SecList.Items.Exchange(i,i+1);
  DSect.MoveSection(THS,i+1);
  SecList.ItemIndex := i+1; SecListClick(Self);
  Form := GetParentForm(DSect.Owner);
  if (Form <> nil) and (Form.Designer <> nil)
     then Form.Designer.Modified;
end;

procedure TElSectionsPropDlg.SecListClick(Sender: TObject);
begin
  if (SecList.ItemIndex=0) or (SecList.ItemIndex=-1) then
    UpBtn.Enabled:=false
    else UpBtn.Enabled:=true;
  if SecList.ItemIndex=SecList.Items.Count-1 then
    DownBtn.Enabled:=false
    else DownBtn.Enabled:=true;
  DeleteBtn.Enabled:=SecList.ItemIndex<>-1;
  EditBtn.Enabled:=SecList.ItemIndex<>-1;
end;

procedure TElSectionsPropDlg.FormCreate(Sender: TObject);
begin
  SecListClick(sender);
end;

procedure TElSectionsPropDlg.SecListKeyPress(Sender: TObject;
  var Key: Char);
begin
  SecListClick(sender);
end;

end.
