{*******************************************************}
{                                                       }
{         Vladimir Gaitanoff Delphi VCL Library         }
{         Multi-tier: interface factories               }
{                                                       }
{         Copyright (c) 1997, 1999                      }
{                                                       }
{*******************************************************}

{$I VG.INC }
{$D-,L-}

unit ProvFact;

interface
uses StdVcl, Classes, ActiveX, ComObj, vgI0_TLB {$IFDEF _D5_}, Midas{$ENDIF};

type
  TIProviderApplyUpdatesEvent = procedure (Sender: TObject; Provider: IProviderDisp;
    Delta: OleVariant; MaxErrors: Integer; var ErrorCount: Integer;
    var Data: OleVariant) of object;
  TIProviderGet_DataEvent = procedure (Sender: TObject; Provider: IProviderDisp;
    var Data: OleVariant) of object;
  TIProviderGetMetaDataEvent = TIProviderGet_DataEvent;
  TIProviderGetRecordsEvent = procedure (Sender: TObject; Provider: IProviderDisp;
    Count: Integer; var RecsOut: Integer; var Data: OleVariant) of object;
  TIProviderDataRequestEvent = procedure (Sender: TObject; Provider: IProviderDisp;
    Input: OleVariant; var Data: OleVariant) of object;
  TIProviderSetParamsEvent = procedure (Sender: TObject; Provider: IProviderDisp;
    Values: OleVariant) of object;

{ TProviderFactory }
  TProviderFactory = class(TComponent)
  private
    FOnGet_Data: TIProviderGet_DataEvent;
    FOnApplyUpdates: TIProviderApplyUpdatesEvent;
    FOnGetMetaData: TIProviderGetMetaDataEvent;
    FOnGetRecords: TIProviderGetRecordsEvent;
    FOnDataRequest: TIProviderDataRequestEvent;
    FOnSetParams: TIProviderSetParamsEvent;
  protected
    { IProvider events }
    function Get_Data(Provider: IProviderDisp): OleVariant; virtual;
    function ApplyUpdates(Provider: IProviderDisp; Delta: OleVariant; MaxErrors: Integer; out ErrorCount: Integer): OleVariant; virtual;
    function GetMetaData(Provider: IProviderDisp): OleVariant; virtual;
    function GetRecords(Provider: IProviderDisp; Count: Integer; out RecsOut: Integer): OleVariant; virtual;
    function DataRequest(Provider: IProviderDisp; Input: OleVariant): OleVariant; virtual;
    procedure SetParams(Provider: IProviderDisp; Values: OleVariant); virtual;
  public
    function CreateProvider(const Provider: IProviderDisp): IProvider;
  published
    property OnGet_Data: TIProviderGet_DataEvent read FOnGet_Data write FOnGet_Data;
    property OnApplyUpdates: TIProviderApplyUpdatesEvent read FOnApplyUpdates write FOnApplyUpdates;
    property OnGetMetaData: TIProviderGetMetaDataEvent read FOnGetMetaData write FOnGetMetaData;
    property OnGetRecords: TIProviderGetRecordsEvent read FOnGetRecords write FOnGetRecords;
    property OnDataRequest: TIProviderDataRequestEvent read FOnDataRequest write FOnDataRequest;
    property OnSetParams: TIProviderSetParamsEvent read FOnSetParams write FOnSetParams;
  end;

  TIAppServerAS_ApplyUpdatesEvent = procedure(Sender: TObject; const AppServer: IAppServerDisp;
    const ProviderName: WideString; Delta: OleVariant; MaxErrors: Integer;
    out ErrorCount: Integer; var OwnerData: OleVariant; var Data: OleVariant) of object;

  TIAppServerAS_GetRecordsEvent = procedure(Sender: TObject; const AppServer: IAppServerDisp;
    const ProviderName: WideString; Count: Integer; out RecsOut: Integer;
    Options: Integer; const CommandText: WideString; var Params: OleVariant;
    var OwnerData: OleVariant; var Data: OleVariant) of object;

  TIAppServerAS_DataRequestEvent = procedure(Sender: TObject; const AppServer: IAppServerDisp;
    const ProviderName: WideString; Input: OleVariant; var Data: OleVariant) of object;

  TIAppServerAS_GetProviderNamesEvent = procedure(Sender: TObject; const AppServer: IAppServerDisp;
    var Data: OleVariant) of object;

  TIAppServerAS_GetParamsEvent = procedure(Sender: TObject; const AppServer: IAppServerDisp;
    const ProviderName: WideString; var OwnerData: OleVariant; var Data: OleVariant) of object;

  TIAppServerAS_RowRequestEvent = procedure(Sender: TObject; const AppServer: IAppServerDisp;
    const ProviderName: WideString; Row: OleVariant; RequestType: Integer;
    var OwnerData: OleVariant; var Data: OleVariant) of object;

  TIAppServerAS_ExecuteEvent = procedure(Sender: TObject; const AppServer: IAppServerDisp; const ProviderName: WideString; const CommandText: WideString;
    var Params: OleVariant; var OwnerData: OleVariant) of object;

  TAppServerFactory = class(TComponent)
  private
    FOnAS_ApplyUpdates: TIAppServerAS_ApplyUpdatesEvent;
    FOnAS_GetRecords: TIAppServerAS_GetRecordsEvent;
    FOnAS_DataRequest: TIAppServerAS_DataRequestEvent;
    FOnAS_GetProviderNames: TIAppServerAS_GetProviderNamesEvent;
    FOnAS_GetParams: TIAppServerAS_GetParamsEvent;
    FOnAS_RowRequest: TIAppServerAS_RowRequestEvent;
    FOnAS_Execute: TIAppServerAS_ExecuteEvent;
  protected
    function AS_ApplyUpdates(const AppServer: IAppServerDisp; const ProviderName: WideString; Delta: OleVariant;
      MaxErrors: Integer; out ErrorCount: Integer; var OwnerData: OleVariant): OleVariant; virtual;
    function AS_GetRecords(const AppServer: IAppServerDisp; const ProviderName: WideString; Count: Integer; out RecsOut: Integer;
      Options: Integer; const CommandText: WideString;
      var Params: OleVariant; var OwnerData: OleVariant): OleVariant; virtual;
    function AS_DataRequest(const AppServer: IAppServerDisp; const ProviderName: WideString;
      Data: OleVariant): OleVariant; virtual;
    function AS_GetProviderNames(const AppServer: IAppServerDisp): OleVariant; virtual;
    function AS_GetParams(const AppServer: IAppServerDisp; const ProviderName: WideString;
      var OwnerData: OleVariant): OleVariant; virtual;
    function AS_RowRequest(const AppServer: IAppServerDisp; const ProviderName: WideString;
      Row: OleVariant; RequestType: Integer; var OwnerData: OleVariant): OleVariant; virtual;
    procedure AS_Execute(const AppServer: IAppServerDisp; const ProviderName: WideString;
      const CommandText: WideString; var Params: OleVariant; var OwnerData: OleVariant); virtual;
  public
    function CreateAppServer(const AppServer: IAppServerDisp): IAppServer;
  published
    property OnAS_ApplyUpdates: TIAppServerAS_ApplyUpdatesEvent read FOnAS_ApplyUpdates write FOnAS_ApplyUpdates;
    property OnAS_GetRecords: TIAppServerAS_GetRecordsEvent read FOnAS_GetRecords write FOnAS_GetRecords;
    property OnAS_DataRequest: TIAppServerAS_DataRequestEvent read FOnAS_DataRequest write FOnAS_DataRequest;
    property OnAS_GetProviderNames: TIAppServerAS_GetProviderNamesEvent read FOnAS_GetProviderNames write FOnAS_GetProviderNames;
    property OnAS_GetParams: TIAppServerAS_GetParamsEvent read FOnAS_GetParams write FOnAS_GetParams;
    property OnAS_RowRequest: TIAppServerAS_RowRequestEvent read FOnAS_RowRequest write FOnAS_RowRequest;
    property OnAS_Execute: TIAppServerAS_ExecuteEvent read FOnAS_Execute write FOnAS_Execute;
  end;

implementation
uses Windows;

type
{ TDispatchProviderHook }
  TDispatchProviderHook = class(TAutoIntfObject, IProvider)
  private
    FProvider: IProviderDisp;
    FProviderFactory: TProviderFactory;
  protected
    { IProviderDisp }
    function Get_Data: OleVariant; safecall;
    function ApplyUpdates(Delta: OleVariant; MaxErrors: Integer; out ErrorCount: Integer): OleVariant; safecall;
    function GetMetaData: OleVariant; safecall;
    function GetRecords(Count: Integer; out RecsOut: Integer): OleVariant; safecall;
    function DataRequest(Input: OleVariant): OleVariant; safecall;
    function Get_Constraints: WordBool; safecall;
    procedure Set_Constraints(Value: WordBool); safecall;
    procedure Reset(MetaData: WordBool); safecall;
    procedure SetParams(Values: OleVariant); safecall;
  public
    constructor Create(AProviderFactory: TProviderFactory; const AProvider: IProviderDisp);
  end;

{ TDispatchProviderHook }
constructor TDispatchProviderHook.Create(AProviderFactory: TProviderFactory; const AProvider: IProviderDisp);
var
  StdVcl: ITypeLib;
begin
  OleCheck(LoadRegTypeLib(LIBID_STDVCL, 4, 0, 0, StdVCL));
  inherited Create(StdVCL, IProvider);
  FProvider := AProvider;
  FProviderFactory := AProviderFactory;
end;

{ TDispatchProviderHook.IProvider }
function TDispatchProviderHook.Get_Data: OleVariant;
begin
  Result := FProviderFactory.Get_Data(FProvider);
end;

function TDispatchProviderHook.ApplyUpdates(Delta: OleVariant; MaxErrors: Integer; out ErrorCount: Integer): OleVariant;
begin
  Result := FProviderFactory.ApplyUpdates(FProvider, Delta, MaxErrors, ErrorCount);
end;

function TDispatchProviderHook.GetMetaData: OleVariant;
begin
  Result := FProviderFactory.GetMetaData(FProvider);
end;

function TDispatchProviderHook.GetRecords(Count: Integer; out RecsOut: Integer): OleVariant;
begin
  Result := FProviderFactory.GetRecords(FProvider, Count, RecsOut);
end;

function TDispatchProviderHook.DataRequest(Input: OleVariant): OleVariant;
begin
  Result := FProviderFactory.DataRequest(FProvider, Input);
end;

function TDispatchProviderHook.Get_Constraints: WordBool;
begin
  Result := FProvider.Constraints;
end;

procedure TDispatchProviderHook.Set_Constraints(Value: WordBool);
begin
  FProvider.Constraints := Value;
end;

procedure TDispatchProviderHook.Reset(MetaData: WordBool);
begin
  FProvider.Reset(MetaData);
end;

procedure TDispatchProviderHook.SetParams(Values: OleVariant);
begin
  FProviderFactory.SetParams(FProvider, Values);
end;

{ TProviderFactory }
function TProviderFactory.CreateProvider(const Provider: IProviderDisp): IProvider;
begin
  if Assigned(Provider) then
    Result := TDispatchProviderHook.Create(Self, Provider) else
    Result := nil;
end;

function TProviderFactory.Get_Data(Provider: IProviderDisp): OleVariant;
begin
  if Assigned(FOnGet_Data) then
    FOnGet_Data(Self, Provider, Result) else
    Result := Provider.Data;
end;

function TProviderFactory.ApplyUpdates(Provider: IProviderDisp; Delta: OleVariant; MaxErrors: Integer; out ErrorCount: Integer): OleVariant;
begin
  if Assigned(FOnApplyUpdates) then
    FOnApplyUpdates(Self, Provider, Delta, MaxErrors, ErrorCount, Result) else
    Result := Provider.ApplyUpdates(Delta, MaxErrors, ErrorCount);
end;

function TProviderFactory.GetMetaData(Provider: IProviderDisp): OleVariant;
begin
  if Assigned(FOnGetMetaData) then
    FOnGetMetaData(Self, Provider, Result) else
    Result := Provider.GetMetaData;
end;

function TProviderFactory.GetRecords(Provider: IProviderDisp; Count: Integer; out RecsOut: Integer): OleVariant;
begin
  if Assigned(FOnGetRecords) then
    FOnGetRecords(Self, Provider, Count, RecsOut, Result) else
    Result := Provider.GetRecords(Count, RecsOut);
end;

function TProviderFactory.DataRequest(Provider: IProviderDisp; Input: OleVariant): OleVariant;
begin
  if Assigned(FOnDataRequest) then
    FOnDataRequest(Self, Provider, Input, Result) else
    Result := Provider.DataRequest(Input);
end;

procedure TProviderFactory.SetParams(Provider: IProviderDisp; Values: OleVariant);
begin
  if Assigned(FOnSetParams) then
    FOnSetParams(Self, Provider, Values) else
    Provider.SetParams(Values);
end;

type
{ TDispatchAppServerHook }
  TDispatchAppServerHook = class(TAutoIntfObject, IAppServer)
  private
    FAppServer: IAppServerDisp;
    FAppServerFactory: TAppServerFactory;
  protected
    { IAppServerDisp }
    function AS_ApplyUpdates(const ProviderName: WideString; Delta: OleVariant;
      MaxErrors: Integer; out ErrorCount: Integer; var OwnerData: OleVariant): OleVariant; safecall;
    function AS_GetRecords(const ProviderName: WideString; Count: Integer; out RecsOut: Integer;
      Options: Integer; const CommandText: WideString;
      var Params: OleVariant; var OwnerData: OleVariant): OleVariant; safecall;
    function AS_DataRequest(const ProviderName: WideString; Data: OleVariant): OleVariant; safecall;
    function AS_GetProviderNames: OleVariant; safecall;
    function AS_GetParams(const ProviderName: WideString; var OwnerData: OleVariant): OleVariant; safecall;
    function AS_RowRequest(const ProviderName: WideString; Row: OleVariant; RequestType: Integer;
      var OwnerData: OleVariant): OleVariant; safecall;
    procedure AS_Execute(const ProviderName: WideString; const CommandText: WideString;
      var Params: OleVariant; var OwnerData: OleVariant); safecall;
  public
    constructor Create(AAppServerFactory: TAppServerFactory; const AAppServer: IAppServerDisp);
  end;


{ TDispatchAppServerHook }
constructor TDispatchAppServerHook.Create(AAppServerFactory: TAppServerFactory;
  const AAppServer: IAppServerDisp);
var
  TypeLib: ITypeLib;
begin
  OleCheck(LoadRegTypeLib(LIBID_Midas, 1, 0, 0, TypeLib));
  inherited Create(TypeLib, IAppServer);
  FAppServer := AAppServer;
  FAppServerFactory := AAppServerFactory;
end;

function TDispatchAppServerHook.AS_ApplyUpdates(
  const ProviderName: WideString; Delta: OleVariant; MaxErrors: Integer;
  out ErrorCount: Integer; var OwnerData: OleVariant): OleVariant;
begin
  Result := FAppServerFactory.AS_ApplyUpdates(FAppServer, ProviderName,
    Delta, MaxErrors, ErrorCount, OwnerData);
end;

function TDispatchAppServerHook.AS_DataRequest(
  const ProviderName: WideString; Data: OleVariant): OleVariant;
begin
  Result := FAppServerFactory.AS_DataRequest(FAppServer, ProviderName, Data);
end;

procedure TDispatchAppServerHook.AS_Execute(const ProviderName,
  CommandText: WideString; var Params, OwnerData: OleVariant);
begin
  FAppServerFactory.AS_Execute(FAppServer, ProviderName,
    CommandText, Params, OwnerData);
end;

function TDispatchAppServerHook.AS_GetParams(
  const ProviderName: WideString; var OwnerData: OleVariant): OleVariant;
begin
  Result := FAppServerFactory.AS_GetParams(FAppServer, ProviderName, OwnerData);
end;

function TDispatchAppServerHook.AS_GetProviderNames: OleVariant;
begin
  Result := FAppServerFactory.AS_GetProviderNames(FAppServer);
end;

function TDispatchAppServerHook.AS_GetRecords(
  const ProviderName: WideString; Count: Integer; out RecsOut: Integer;
  Options: Integer; const CommandText: WideString; var Params, OwnerData: OleVariant): OleVariant;
begin
  Result := FAppServerFactory.AS_GetRecords(FAppServer, ProviderName,
    Count, RecsOut, Options, CommandText, Params, OwnerData);
end;

function TDispatchAppServerHook.AS_RowRequest(
  const ProviderName: WideString; Row: OleVariant; RequestType: Integer;
  var OwnerData: OleVariant): OleVariant;
begin
  Result := FAppServerFactory.AS_RowRequest(FAppServer, ProviderName,
    Row, RequestType, OwnerData);
end;

{ TAppServerFactory }
function TAppServerFactory.CreateAppServer(const AppServer: IAppServerDisp): IAppServer;
begin
  if Assigned(AppServer) then
    Result := TDispatchAppServerHook.Create(Self, AppServer) else
    Result := nil;
end;

function TAppServerFactory.AS_ApplyUpdates(const AppServer: IAppServerDisp;
  const ProviderName: WideString; Delta: OleVariant; MaxErrors: Integer;
  out ErrorCount: Integer; var OwnerData: OleVariant): OleVariant;
begin
  if Assigned(FOnAS_ApplyUpdates) then
    FOnAS_ApplyUpdates(Self, AppServer, ProviderName, Delta, MaxErrors, ErrorCount, OwnerData, Result) else
    Result := AppServer.AS_ApplyUpdates(ProviderName, Delta, MaxErrors, ErrorCount, OwnerData);
end;

function TAppServerFactory.AS_DataRequest(const AppServer: IAppServerDisp;
  const ProviderName: WideString; Data: OleVariant): OleVariant;
begin
  if Assigned(FOnAS_DataRequest) then
    FOnAS_DataRequest(Self, AppServer, ProviderName, Data, Result) else
    Result := AppServer.AS_DataRequest(ProviderName, Data);
end;

procedure TAppServerFactory.AS_Execute(const AppServer: IAppServerDisp;
  const ProviderName, CommandText: WideString; var Params,
  OwnerData: OleVariant);
begin
  if Assigned(FOnAS_Execute) then
    FOnAS_Execute(Self, AppServer, ProviderName, CommandText, Params, OwnerData) else
    AppServer.AS_Execute(ProviderName, CommandText, Params, OwnerData);
end;

function TAppServerFactory.AS_GetParams(const AppServer: IAppServerDisp;
  const ProviderName: WideString; var OwnerData: OleVariant): OleVariant;
begin
  if Assigned(FOnAS_GetParams) then
    FOnAS_GetParams(Self, AppServer, ProviderName, OwnerData, Result) else
    Result := AppServer.AS_GetParams(ProviderName, OwnerData);
end;

function TAppServerFactory.AS_GetProviderNames(const AppServer: IAppServerDisp): OleVariant;
begin
  if Assigned(FOnAS_GetProviderNames) then
    FOnAS_GetProviderNames(Self, AppServer, Result) else
    Result := AppServer.AS_GetProviderNames;
end;

function TAppServerFactory.AS_GetRecords(const AppServer: IAppServerDisp;
  const ProviderName: WideString; Count: Integer; out RecsOut: Integer;
  Options: Integer; const CommandText: WideString; var Params, OwnerData: OleVariant): OleVariant;
begin
  if Assigned(FOnAS_GetRecords) then
    FOnAS_GetRecords(Self, AppServer, ProviderName, Count, RecsOut, Options, CommandText, Params, OwnerData, Result) else
    Result := AppServer.AS_GetRecords(ProviderName, Count, RecsOut, Options, CommandText, Params, OwnerData);
end;

function TAppServerFactory.AS_RowRequest(const AppServer: IAppServerDisp;
  const ProviderName: WideString; Row: OleVariant; RequestType: Integer;
  var OwnerData: OleVariant): OleVariant;
begin
  if Assigned(FOnAS_RowRequest) then
    FOnAS_RowRequest(Self, AppServer, ProviderName, Row, RequestType, OwnerData, Result) else
    Result := AppServer.AS_RowRequest(ProviderName, Row, RequestType, OwnerData);
end;

end.

