{*******************************************************}
{                                                       }
{         Vladimir Gaitanoff Delphi VCL Library         }
{         Multi-tier: client components for InfoPower   }
{                                                       }
{         Copyright (c) 1997, 1999                      }
{                                                       }
{*******************************************************}

{$I VG.INC }
{$D-,L-}

unit vgCDSIP;

interface
uses Classes, wwTypes, vgCDS;

type
{ TwwvgClientDataSet }
  TwwvgClientDataSet = class(TvgClientDataSet)
  private
    { Private declarations }
    FControlType: TStrings;
    FPictureMasks: TStrings;
    FUsePictureMask: Boolean;
    FOnInvalidValue: TwwInvalidValueEvent;
    procedure SetControlType(AControlType: TStrings);
   procedure SetPictureMasks(APictureMasks: TStrings);
  protected
    { Protected declarations }
    procedure DoBeforePost; override;  //for picture support
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property ControlType: TStrings read FControlType write SetControlType;
    property PictureMasks: TStrings read FPictureMasks write SetPictureMasks;
    property ValidateWithMask: Boolean read FUsePictureMask write FUsePictureMask;
    property OnInvalidValue: TwwInvalidValueEvent read FOnInvalidValue write FOnInvalidValue;
  end;

{ TwwvgClientTable }
  TwwvgClientTable = class(TvgClientTable)
  private
    { Private declarations }
    FControlType: TStrings;
    FPictureMasks: TStrings;
    FUsePictureMask: Boolean;
    FOnInvalidValue: TwwInvalidValueEvent;
    procedure SetControlType(AControlType: TStrings);
   procedure SetPictureMasks(APictureMasks: TStrings);
  protected
    { Protected declarations }
    procedure DoBeforePost; override;  //for picture support
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property ControlType: TStrings read FControlType write SetControlType;
    property PictureMasks: TStrings read FPictureMasks write SetPictureMasks;
    property ValidateWithMask: Boolean read FUsePictureMask write FUsePictureMask;
    property OnInvalidValue: TwwInvalidValueEvent read FOnInvalidValue write FOnInvalidValue;
  end;

{ TwwvgClientQuery }
  TwwvgClientQuery = class(TvgClientQuery)
  private
    { Private declarations }
    FControlType: TStrings;
    FPictureMasks: TStrings;
    FUsePictureMask: Boolean;
    FOnInvalidValue: TwwInvalidValueEvent;
    procedure SetControlType(AControlType: TStrings);
   procedure SetPictureMasks(APictureMasks: TStrings);
  protected
    { Protected declarations }
    procedure DoBeforePost; override;  //for picture support
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property ControlType: TStrings read FControlType write SetControlType;
    property PictureMasks: TStrings read FPictureMasks write SetPictureMasks;
    property ValidateWithMask: Boolean read FUsePictureMask write FUsePictureMask;
    property OnInvalidValue: TwwInvalidValueEvent read FOnInvalidValue write FOnInvalidValue;
  end;

{ TwwvgClientStoredProc }
  TwwvgClientStoredProc = class(TvgClientStoredProc)
  private
    { Private declarations }
    FControlType: TStrings;
    FPictureMasks: TStrings;
    FUsePictureMask: Boolean;
    FOnInvalidValue: TwwInvalidValueEvent;
    procedure SetControlType(AControlType: TStrings);
   procedure SetPictureMasks(APictureMasks: TStrings);
  protected
    { Protected declarations }
    procedure DoBeforePost; override;  //for picture support
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property ControlType: TStrings read FControlType write SetControlType;
    property PictureMasks: TStrings read FPictureMasks write SetPictureMasks;
    property ValidateWithMask: Boolean read FUsePictureMask write FUsePictureMask;
    property OnInvalidValue: TwwInvalidValueEvent read FOnInvalidValue write FOnInvalidValue;
  end;

implementation
uses wwCommon;

constructor TwwvgClientDataSet.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FControlType:= TStringList.Create;
  FPictureMasks:= TStringList.Create;
  FUsePictureMask:= True;
end;

destructor TwwvgClientDataSet.Destroy;
begin
  FControlType.Free;
  FPictureMasks.Free;
  inherited Destroy;
end;

procedure TwwvgClientDataSet.SetControlType(AControlType: TStrings);
begin
  FControlType.Assign(AControlType);
end;

procedure TwwvgClientDataSet.SetPictureMasks(APictureMasks: TStrings);
begin
  FPictureMasks.Assign(APictureMasks);
end;

procedure TwwvgClientDataSet.DoBeforePost;
begin
  inherited DoBeforePost;
  if FUsePictureMask then
    wwValidatePictureFields(Self, FOnInvalidValue);
end;

{ TwwvgClientTable }
constructor TwwvgClientTable.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FControlType:= TStringList.Create;
  FPictureMasks:= TStringList.Create;
  FUsePictureMask:= True;
end;

destructor TwwvgClientTable.Destroy;
begin
  FControlType.Free;
  FPictureMasks.Free;
  inherited Destroy;
end;

procedure TwwvgClientTable.SetControlType(AControlType: TStrings);
begin
  FControlType.Assign(AControlType);
end;

procedure TwwvgClientTable.SetPictureMasks(APictureMasks: TStrings);
begin
  FPictureMasks.Assign(APictureMasks);
end;

procedure TwwvgClientTable.DoBeforePost;
begin
  inherited DoBeforePost;
  if FUsePictureMask then
    wwValidatePictureFields(Self, FOnInvalidValue);
end;

{ TwwvgClientQuery }
constructor TwwvgClientQuery.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FControlType:= TStringList.Create;
  FPictureMasks:= TStringList.Create;
  FUsePictureMask:= True;
end;

destructor TwwvgClientQuery.Destroy;
begin
  FControlType.Free;
  FPictureMasks.Free;
  inherited Destroy;
end;

procedure TwwvgClientQuery.SetControlType(AControlType: TStrings);
begin
  FControlType.Assign(AControlType);
end;

procedure TwwvgClientQuery.SetPictureMasks(APictureMasks: TStrings);
begin
  FPictureMasks.Assign(APictureMasks);
end;

procedure TwwvgClientQuery.DoBeforePost;
begin
  inherited DoBeforePost;
  if FUsePictureMask then
    wwValidatePictureFields(Self, FOnInvalidValue);
end;

{ TwwvgClientStoredProc }
constructor TwwvgClientStoredProc.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FControlType:= TStringList.Create;
  FPictureMasks:= TStringList.Create;
  FUsePictureMask:= True;
end;

destructor TwwvgClientStoredProc.Destroy;
begin
  FControlType.Free;
  FPictureMasks.Free;
  inherited Destroy;
end;

procedure TwwvgClientStoredProc.SetControlType(AControlType: TStrings);
begin
  FControlType.Assign(AControlType);
end;

procedure TwwvgClientStoredProc.SetPictureMasks(APictureMasks: TStrings);
begin
  FPictureMasks.Assign(APictureMasks);
end;

procedure TwwvgClientStoredProc.DoBeforePost;
begin
  inherited DoBeforePost;
  if FUsePictureMask then
    wwValidatePictureFields(Self, FOnInvalidValue);
end;

end.
