{*******************************************************}
{                                                       }
{         Vladimir Gaitanoff Delphi VCL Library         }
{         Midas registration                            }
{                                                       }
{         Copyright (c) 1997, 1999                      }
{                                                       }
{*******************************************************}

{$I VG.INC }
{$D-,L-}

unit vgMIDRg;

interface

procedure Register;

implementation
uses Classes, DsgnIntf, vgCDS, DBConsts, TypInfo, SysUtils, ActiveX, DB, vgVCLRes, vgUtils, vgDBUtl,
  vgComObj, ProvFact, DBClient, vgWPCDS, vgProvid, vgBDEPrv, vgCOMCon, vgProvEd, vgSckCon,
  vgUpdSQL
  {$IFDEF _D4_}, DsnDBCst, vgCorba{$ENDIF}
  {$IFDEF _D5_}, DbReg, DmDesigner{$ENDIF};

{$R vgDMID.dcr}

type
{ TListNamesProperty }
  TListNamesProperty = class(TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
  end;

{ TProviderNameProperty }
  TProviderNameProperty = class(TListNamesProperty)
  public
    procedure GetValues(Proc: TGetStrProc); override;
 end;

{ TAliasNameRemoteProperty }
  TAliasNameRemoteProperty = class(TListNamesProperty)
  public
    procedure GetValues(Proc: TGetStrProc); override;
  end;

{ TClientDatabaseNameProperty }
  TClientDatabaseNameProperty = class(TListNamesProperty)
  public
    procedure GetValues(Proc: TGetStrProc); override;
  end;

{ TRemoteBDENameListProperty }
  TRemoteBDENameListProperty = class(TListNamesProperty)
  protected
    procedure EnumValues(Database: TvgClientDatabase; List: TStrings); virtual;
  public
    procedure GetValues(Proc: TGetStrProc); override;
  end;

{ TRemoteTableNameProperty }
  TRemoteTableNameProperty = class(TRemoteBDENameListProperty)
  protected
    procedure EnumValues(Database: TvgClientDatabase; List: TStrings); override;
  end;

{ TRemoteStoredProcNameProperty }
  TRemoteStoredProcNameProperty = class(TRemoteBDENameListProperty)
  protected
    procedure EnumValues(Database: TvgClientDatabase; List: TStrings); override;
  end;

{ TFieldNameProperty }
  TFieldNameProperty = class(TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

  TRemoteServerHack = class(TCustomRemoteServer);

{ TvgClientDataSetEditor }
{$IFDEF _D5_}
  TvgClientDataSetEditor = class(TDataSetEditor)
  public
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;
{$ENDIF}

{ TvgClientUpdateSQLObjectEditor }
  TvgClientUpdateSQLObjectEditor = class(TComponentEditor)
  public
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

{ TListNamesProperty }
function TListNamesProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paSortList, paMultiSelect];
end;

{ TProviderNameProperty }
procedure TProviderNameProperty.GetValues(Proc: TGetStrProc);
var
  DataSet: TClientDataSet;
begin
  DataSet := TClientDataSet(GetComponent(0));
  if Assigned(DataSet.RemoteServer) then
  begin
    TRemoteServerHack(DataSet.RemoteServer).Connected := True;
    TRemoteServerHack(DataSet.RemoteServer).GetProviderNames(Proc);
  end;
end;

{ TAliasNameRemoteProperty }
procedure TAliasNameRemoteProperty.GetValues(Proc: TGetStrProc);
var
  Database: TvgClientDatabase;
  Server: TvgCustomRemoteServer;
  List: TStrings;
begin
  Database := GetComponent(0) as TvgClientDatabase;
  Server := Database.RemoteServer;
  if Assigned(Server) then
  begin
    Server.Connected := True;
    List := TStringList.Create;
    try
      Server.GetDatabaseNames(List);
      EnumStrings(List, Proc);
    finally
      List.Free;
    end;
  end;
end;

{ TClientDatabaseNameProperty }
procedure TClientDatabaseNameProperty.GetValues(Proc: TGetStrProc);
var
  I: Integer;
  Server: TComponent;
begin
  Server := TComponent(GetOrdProp(GetComponent(0), TypInfo.GetPropInfo(GetComponent(0).ClassInfo, 'RemoteServer')));
  if (Server is TvgCustomRemoteServer) then
    with TvgCustomRemoteServer(Server) do
      for I := 0 to DatabaseCount - 1 do
        Proc(Database[I].DatabaseName);
end;

{ TRemoteBDENameListProperty }
procedure TRemoteBDENameListProperty.EnumValues(Database: TvgClientDatabase; List: TStrings);
begin
end;

procedure TRemoteBDENameListProperty.GetValues(Proc: TGetStrProc);
var
  List: TStrings;
  Server: TComponent;
  DatabaseName: string;
  Database: TvgClientDatabase;
begin
  Server := TComponent(GetOrdProp(GetComponent(0), TypInfo.GetPropInfo(GetComponent(0).ClassInfo, 'RemoteServer')));
  DatabaseName := GetStrProp(GetComponent(0), TypInfo.GetPropInfo(GetComponent(0).ClassInfo, 'DatabaseName'));
  if (Server is TvgCustomRemoteServer) then
  begin
    Database := TvgCustomRemoteServer(Server).DatabaseByName(DatabaseName);

    List := TStringList.Create;
    try
      EnumValues(Database, List);
      EnumStrings(List, Proc);
    finally
      List.Free;
    end;
  end;
end;

{ TRemoteTableNameProperty }
procedure TRemoteTableNameProperty.EnumValues(Database: TvgClientDatabase; List: TStrings);
begin
  Database.GetTableNames('*.*', True, False, List);
end;

{ TRemoteStoredProcNameProperty }
procedure TRemoteStoredProcNameProperty.EnumValues(Database: TvgClientDatabase; List: TStrings);
begin
  Database.GetStoredProcNames(List);
end;

{ TFieldNameProperty }
function TFieldNameProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paMultiSelect];
end;

procedure TFieldNameProperty.GetValues(Proc: TGetStrProc);
var
  I: Integer;
  Values: TStringList;
begin
  Values := TStringList.Create;
  try
    with GetComponent(0) as TDataSet do GetFieldNames(Values);
    for I := 0 to Values.Count - 1 do Proc(Values[I]);
  finally
    Values.Free;
  end;
end;

type
{ TDatabaseProviderEditor }
  TDatabaseProviderEditor = class(TComponentEditor)
  public
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

{ TDatabaseProviderEditor }
{$IFNDEF _D4_}
function InterfaceMemberText(Component: TComponent): string;
begin
  Result := Format('property %s: IDatabase; readonly;', [Component.Name]);
end;
{$ENDIF}

function InterfaceMemberMenuText(Component: TComponent): string;
begin
  Result := Format(SExportProvider, [Component.Name]);
end;

procedure TDatabaseProviderEditor.ExecuteVerb(Index: Integer);
begin
  case Index of
    0:
    {$IFDEF _D4_}
      Designer.AddToInterface(INVOKE_PROPERTYGET, Component.Name, VT_USERDEFINED, 'IDatabase');
    {$ELSE}
      {$IFDEF CBUilder}
        Designer.AddToInterface(INVOKE_PROPERTYGET, Component.Name, VT_USERDEFINED, 'IDatabase');
      {$ELSE}
        Designer.AddInterfaceMember(InterfaceMemberText(Component));
      {$ENDIF}
    {$ENDIF}
  end;
end;

function TDatabaseProviderEditor.GetVerb(Index: Integer): string;
begin
  case Index of
    0: Result := InterfaceMemberMenuText(Component);
  end;
end;

function TDatabaseProviderEditor.GetVerbCount: Integer;
begin
  Result := Ord(Designer.HasInterface and not Designer.HasInterfaceMember(Component.Name));
end;

{ TvgClientDataSetEditor }
{$IFDEF _D5_}
procedure TvgClientDataSetEditor.ExecuteVerb(Index: Integer);
begin
  if Index <= inherited GetVerbCount - 1 then
    inherited ExecuteVerb(Index) else
  begin
    Dec(Index, inherited GetVerbCount);
    case Index of
      0: begin
           TvgClientDataSet(Component).FetchDBParams;
           Designer.Modified;
         end;
      1: begin
           TvgClientDataSet(Component).FetchDBMacros;
           Designer.Modified;
         end;
    end;
  end;
end;

function TvgClientDataSetEditor.GetVerb(Index: Integer): string;
begin
  if Index <= inherited GetVerbCount - 1 then
    Result := inherited GetVerb(Index) else
  begin
    Dec(Index, inherited GetVerbCount);
    case Index of
      0: Result := 'Fetch DBParams';
      1: Result := 'Fetch DBMacros';
    end;
  end;
end;

function TvgClientDataSetEditor.GetVerbCount: Integer;
begin
  Result := inherited GetVerbCount + 2;
end;
{$ENDIF}

{ TvgClientUpdateSQLObjectEditor }
procedure TvgClientUpdateSQLObjectEditor.ExecuteVerb(Index: Integer);
begin
  case Index of
    0: if EditUpdateSQL(TvgClientUpdateSQLObject(Component)) then
      Designer.Modified;
  end;
end;

function TvgClientUpdateSQLObjectEditor.GetVerb(Index: Integer): string;
begin
  case Index of
    0: Result := 'UpdateSQL Editor...';
  end;
end;

function TvgClientUpdateSQLObjectEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;

procedure Register;
begin
  RegisterComponents(LoadStr(SRegMultiTier), [
    TvgClientDataSet, TClientDataSetHook,
    TSinglePointServer, TvgRemoteServer, TvgSocketConnection,
    {$IFDEF _D4_}TvgCORBARemoteServer, {$ENDIF}
    TvgSocketDispatcher,
    TProviderList,
    {$IFDEF _D4_} TDataSetProviderList, {$ENDIF}
    {$IFDEF _D5_} TRequestProvider, {$ENDIF}
    TDatabaseProvider,
    TProviderFactory, TAppServerFactory
  ]);

  RegisterComponents(LoadStr(SRegVGServer), [
    TvgClientDatabase,
    TvgClientTable, TvgClientQuery, TvgClientStoredProc,
    TvgClientUpdateSQL, TvgClientBatchMove,
    TvgClientSQLScript, TvgClientUpdateScript
  ]);

  RegisterComponents(LoadStr(SRegReports), [TvgCDSWordPrint]);

  RegisterNoIcon([TProviderItem{$IFDEF _D4_}, TDataSetProviderItem{$ENDIF}]);

  {$IFDEF _D5_}
  RegisterCustomModule(TvgRemoteDataModule, TDataModuleDesignerCustomModule);
  {$ELSE}
  RegisterCustomModule(TvgRemoteDataModule, TCustomModule);
  {$ENDIF}

{$IFDEF _D5_}
  RegisterComponentEditor(TvgClientDataSet, TvgClientDataSetEditor);
{$ENDIF}

  RegisterComponentEditor(TCustomProviderList, TProviderListEditor);

  RegisterComponentEditor(TCustomDatabaseProvider, TDatabaseProviderEditor);

  RegisterComponentEditor(TvgClientUpdateSQLObject, TvgClientUpdateSQLObjectEditor);

  RegisterPropertyEditor(TypeInfo(string), TClientDataSet, 'ProviderName', TProviderNameProperty);
  RegisterPropertyEditor(TypeInfo(string), TvgClientDataSet, 'IndexFieldNamesEx', TFieldNameProperty);
  RegisterPropertyEditor(TypeInfo(string), TvgClientBDEDataSet, 'DatabaseName', TClientDatabaseNameProperty);
  RegisterPropertyEditor(TypeInfo(string), TvgClientSQLScript, 'DatabaseName', TClientDatabaseNameProperty);
  RegisterPropertyEditor(TypeInfo(string), TvgClientDatabase, 'AliasName', TAliasNameRemoteProperty);
  RegisterPropertyEditor(TypeInfo(string), TvgClientBDEDataSet, 'TableName', TRemoteTableNameProperty);
  RegisterPropertyEditor(TypeInfo(string), TvgClientStoredProc, 'StoredProcName', TRemoteStoredProcNameProperty);
end;

initialization
  InitFieldAutoSizers([TClientDataSet, TvgClientDataSet,
    TvgClientBDEDataSet, TvgClientTable, TvgClientQuery, TvgClientStoredProc]);

finalization
  DoneFieldAutoSizers([TClientDataSet, TvgClientDataSet,
    TvgClientBDEDataSet, TvgClientTable, TvgClientQuery, TvgClientStoredProc]);

end.
