{*******************************************************}
{                                                       }
{         Vladimir Gaitanoff Delphi VCL Library         }
{         Multi-tier: Socket connection                 }
{         Most of code is copied from SConnect, MSrvr   }
{                                                       }
{         Copyright (c) 1997, 1999                      }
{                                                       }
{*******************************************************}

{$I VG.INC }
{$D-,L-}

unit vgSckCon;

interface

uses
{$IFDEF _D4_}
  Windows, SysUtils, Classes, MConnect, ScktComp, vgUtils, SConnect;
{$ELSE}
  Windows, Messages, Classes, SysUtils, DBClient, ScktComp, vgUtils;
{$ENDIF}

{$IFNDEF _D4_}
type
  { IDataBlock }
  IDataBlock = interface(IUnknown)
  ['{CA6564C2-4683-11D1-88D4-00A0248E5091}']
    function GetBytesReserved: Integer; stdcall;
    function GetMemory: Pointer; stdcall;
    function GetSize: Integer; stdcall;
    procedure SetSize(Value: Integer); stdcall;
    function GetSignature: Integer; stdcall;
    procedure SetSignature(Value: Integer); stdcall;
    procedure Clear; stdcall;
    function Write(const Buffer; Count: Integer): Integer; stdcall;
    function Read(var Buffer; Count: Integer): Integer; stdcall;
    property BytesReserved: Integer read GetBytesReserved;
    property Memory: Pointer read GetMemory;
    property Signature: Integer read GetSignature write SetSignature;
    property Size: Integer read GetSize write SetSize;
  end;

  { ISendDataBlock }
  ISendDataBlock = interface
  ['{87AD1043-470E-11D1-88D5-00A0248E5091}']
    function Send(const Data: IDataBlock; WaitForResult: Boolean): IDataBlock; stdcall;
  end;

  { ITransport }
  ITransport = interface(IUnknown)
  ['{CA6564C1-4683-11D1-88D4-00A0248E5091}']
    function GetWaitEvent: THandle; stdcall;
    function GetConnected: Boolean; stdcall;
    procedure SetConnected(Value: Boolean); stdcall;
    function Receive(WaitForInput: Boolean): IDataBlock; stdcall;
    procedure Send(const Data: IDataBlock); stdcall;
    property Connected: Boolean read GetConnected write SetConnected;
  end;

  { IDataIntercept }
  IDataIntercept = interface
  ['{B249776B-E429-11D1-AAA4-00C04FA35CFA}']
    procedure DataIn(const Data: IDataBlock); stdcall;
    procedure DataOut(const Data: IDataBlock); stdcall;
  end;

  { TDataBlock }
  TDataBlock = class(TInterfacedObject, IDataBlock)
  private
    FStream: TMemoryStream;
    FReadPos: Integer;
    FWritePos: Integer;
  protected
    { IDataBlock }
    function GetBytesReserved: Integer; stdcall;
    function GetMemory: Pointer; stdcall;
    function GetSize: Integer; stdcall;
    procedure SetSize(Value: Integer); stdcall;
    function GetSignature: Integer; stdcall;
    procedure SetSignature(Value: Integer); stdcall;
    procedure Clear; stdcall;
    function Write(const Buffer; Count: Integer): Integer; stdcall;
    function Read(var Buffer; Count: Integer): Integer; stdcall;
    property BytesReserved: Integer read GetBytesReserved;
    property Memory: Pointer read GetMemory;
    property Signature: Integer read GetSignature write SetSignature;
    property Size: Integer read GetSize write SetSize;
  public
    constructor Create;
    destructor Destroy; override;
  end;
{$ENDIF}

type
  TvgDataDispatch = class;

{ TvgDataBlockInterpreter }
  TvgDataBlockInterpreter = class
  private
    FDispatchList: TList;
    FDispList: OleVariant;
    FSendDataBlock: ISendDataBlock;
    function GetVariantPointer(const Value: Variant): Pointer;
    procedure CopyDataByRef(Source: TVarData; var Dest: TVarData);
    function ReadArray(VType: Integer; const Data: IDataBlock): OleVariant;
    procedure WriteArray(const Value: OleVariant; const Data: IDataBlock);
    function CallGetServerList: OleVariant;
    procedure CheckData(const Data: IDataBlock);
    procedure DoException(const Data: IDataBlock);
    procedure DoGetServerList(const Data: IDataBlock);
  protected
    procedure AddDispatch(Value: TvgDataDispatch);
    procedure RemoveDispatch(Value: TvgDataDispatch);
    { Sending Calls }
    procedure CallFreeObject(DispatchIndex: Integer);
    function CallGetIDsOfNames(DispatchIndex: Integer; const IID: TGUID; Names: Pointer;
      NameCount, LocaleID: Integer; DispIDs: Pointer): HResult; stdcall;
    function CallInvoke(DispatchIndex, DispID: Integer; const IID: TGUID; LocaleID: Integer;
      Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult; stdcall;
    { Receiving Calls }
    procedure DoCreateObject(const Data: IDataBlock); virtual;
    procedure DoFreeObject(const Data: IDataBlock); virtual;
    procedure DoGetIDsOfNames(const Data: IDataBlock); virtual;
    procedure DoInvoke(const Data: IDataBlock); virtual;
    function DoCustomAction(Action: Integer; const Data: IDataBlock): Boolean; virtual;
  public
    constructor Create(SendDataBlock: ISendDataBlock); virtual;
    destructor Destroy; override;
    function CreateObject(Name: string): OleVariant;
    procedure InterpretData(const Data: IDataBlock);
    function ReadVariant(out Flags: TVarFlags; const Data: IDataBlock): OleVariant;
    procedure WriteVariant(const Value: OleVariant; const Data: IDataBlock);
  end;

{ TvgDataDispatch }
  TvgDataDispatch = class(TInterfacedObject, IDispatch)
  private
    FDispatchIndex: Integer;
    FInterpreter: TvgDataBlockInterpreter;
  protected
    property DispatchIndex: Integer read FDispatchIndex;
    { IDispatch }
    function GetTypeInfoCount(out Count: Integer): HResult; stdcall;
    function GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): HResult; stdcall;
    function GetIDsOfNames(const IID: TGUID; Names: Pointer;
      NameCount, LocaleID: Integer; DispIDs: Pointer): HResult; stdcall;
    function Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer;
      Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult; stdcall;
  public
    constructor Create(Interpreter: TvgDataBlockInterpreter; DispatchIndex: Integer);
    destructor Destroy; override;
  end;

  { TvgStreamedConnection }
  TvgStreamedConnection = class({$IFDEF _D4_}TDispatchConnection{$ELSE}TRemoteServer{$ENDIF}, ISendDataBlock)
  private
    FLock: TRTLCriticalSection;
    FRefCount: Integer;
    FInterpreter: TvgDataBlockInterpreter;
    FTransport: ITransport;
  protected
    { IUnknown }
    function QueryInterface(const IID: TGUID; out Obj): HResult; {$IFDEF _D4_} override; {$ENDIF} stdcall; 
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    { ISendDataBlock }
    function Send(const Data: IDataBlock; WaitForResult: Boolean): IDataBlock; stdcall;

    function CreateTransport: ITransport; virtual;
    procedure DoConnect; override;
    procedure DoDisconnect; override;
    procedure DoError(E: Exception); virtual;
    function GetServerList: OleVariant; {$IFDEF _D5_}override{$ELSE}virtual{$ENDIF};
    property Interpreter: TvgDataBlockInterpreter read FInterpreter;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Lock;
    procedure Unlock;
  end;

{$IFNDEF _D4_}
{ TSocketTransport }
  TSocketTransport = class(TInterfacedObject, ITransport)
  private
    FEvent: THandle;
    FAddress: string;
    FHost: string;
    FPort: Integer;
    FClientSocket: TClientSocket;
    FSocket: TCustomWinSocket;
    FInterceptGUID: string;
    FInterceptor: IDataIntercept;
    FCreateAttempted: Boolean;
    function CheckInterceptor: Boolean;
    procedure InterceptIncoming(const Data: IDataBlock);
    procedure InterceptOutgoing(const Data: IDataBlock);
  protected
    { ITransport }
    function GetWaitEvent: THandle; stdcall;
    function GetConnected: Boolean; stdcall;
    procedure SetConnected(Value: Boolean); stdcall;
    function Receive(WaitForInput: Boolean): IDataBlock; stdcall;
    procedure Send(const Data: IDataBlock); stdcall;
  public
    destructor Destroy; override;
    property Host: string read FHost write FHost;
    property Address: string read FAddress write FAddress;
    property Port: Integer read FPort write FPort;
    property Socket: TCustomWinSocket read FSocket write FSocket;
    property InterceptGUID: string read FInterceptGUID write FInterceptGUID;
  end;
{$ENDIF}

{ TvgSocketConnection }
  TvgSocketConnection = class(TvgStreamedConnection)
  private
    FAddress: string;
    FHost: string;
    FPort: Integer;
    FInterceptGUID: string;
    procedure SetAddress(Value: string);
    procedure SetHost(Value: string);
    function IsHostStored: Boolean;
    function IsAddressStored: Boolean;
  protected
    function CreateTransport: ITransport; override;
    procedure DoConnect; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Address: string read FAddress write SetAddress stored IsAddressStored;
    property Host: string read FHost write SetHost stored IsHostStored;
    property InterceptGUID: string read FInterceptGUID write FInterceptGUID;
    property Port: Integer read FPort write FPort default 211;
  end;

  TGetDataBlockInterpreterEvent = procedure (Sender: TObject; SendDataBlock: ISendDataBlock; var Interpreter: TvgDataBlockInterpreter) of object;

{ TvgSocketDispatcher }
  TvgSocketDispatcher = class(TServerSocket)
  private
    FInterceptGUID: string;
    FTimeout: Integer;
    procedure GetThread(Sender: TObject; ClientSocket: TServerClientWinSocket;
      var SocketThread: TServerClientThread);
  public
    constructor Create(AOwner: TComponent); override;
  published
    property InterceptGUID: string read FInterceptGUID write FInterceptGUID;
    property Timeout: Integer read FTimeout write FTimeout;
  end;

  TvgSocketDispatcherThread = class(TServerClientThread, ISendDataBlock)
  private
    FRefCount: Integer;
    FInterpreter: TvgDataBlockInterpreter;
    FTransport: ITransport;
    FInterceptGUID: string;
    FLastActivity: TDateTime;
    FTimeout: TDateTime;
  protected
    function CreateServerTransport: ITransport; virtual;
    { IUnknown }
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    { ISendDataBlock }
    function Send(const Data: IDataBlock; WaitForResult: Boolean): IDataBlock; stdcall;
  public
    constructor Create(CreateSuspended: Boolean; ASocket: TServerClientWinSocket;
      const InterceptGUID: string; Timeout: Integer);
    procedure ClientExecute; override;
    property LastActivity: TDateTime read FLastActivity;
  end;

{$IFNDEF _D4_}
type
  EInterpreterError = class(Exception);
  ESocketConnectionError = class(Exception);

const
  { Action Signatures }
  CallSig         = $DA00; // Call signature
  ResultSig       = $DB00; // Result signature
  asError         = $01;   // Specify an exception was raised
  asInvoke        = $02;   // Specify a call to Invoke
  asGetID         = $03;   // Specify a call to GetIdsOfNames
  asCreateObject  = $04;   // Specify a com object to create
  asFreeObject    = $05;   // Specify a dispatch to free
  asGetServers    = $10;   // Get classname list
  asGetGUID       = $11;   // Get GUID for ClassName
  asMask          = $FF;   // Mask for action
{$ENDIF}

implementation
{$IFNDEF _D4_}
uses WinSock, ActiveX, ComObj, DBConsts, vgCDSUtl, vgSock, Forms;

resourcestring
  SNoWinSock2  = 'WinSock 2 must be installed to use the socket connection';

{$ELSE}
uses Messages, MidConst, ActiveX, ComObj, vgCDSUtl;
{$ENDIF}

procedure CheckLoadWinSock2;
begin
  if not LoadWinSock2 then
    raise Exception.Create(SNoWinSock2);
end;

{$IFNDEF _D4_}
{ TDataBlock }
constructor TDataBlock.Create;
begin
  FStream := TMemoryStream.Create;
  Clear;
end;

destructor TDataBlock.Destroy;
begin
  FStream.Free;
  inherited Destroy;
end;

{ TDataBlock.IDataBlock }
function TDataBlock.GetBytesReserved: Integer;
begin
  Result := SizeOf(Integer) * 2;
end;

function TDataBlock.GetMemory: Pointer;
var
  DataSize: Integer;
begin
  FStream.Position := 4;
  DataSize := FStream.Size - BytesReserved;
  FStream.Write(DataSize, SizeOf(DataSize));
  Result := FStream.Memory;
end;

function TDataBlock.GetSize: Integer;
begin
  Result := FStream.Size - BytesReserved;
end;

procedure TDataBlock.SetSize(Value: Integer);
begin
  FStream.Size := Value + BytesReserved;
end;

function TDataBlock.GetSignature: Integer;
begin
  FStream.Position := 0;
  FStream.Read(Result, SizeOf(Result));
end;

procedure TDataBlock.SetSignature(Value: Integer);
begin
  FStream.Position := 0;
  FStream.Write(Value, SizeOf(Value));
end;

procedure TDataBlock.Clear;
begin
  FStream.Size := BytesReserved;
  FReadPos := BytesReserved;
  FWritePos := BytesReserved;
end;

function TDataBlock.Write(const Buffer; Count: Integer): Integer;
begin
  FStream.Position := FWritePos;
  Result := FStream.Write(Buffer, Count);
  FWritePos := FStream.Position;
end;

function TDataBlock.Read(var Buffer; Count: Integer): Integer;
begin
  FStream.Position := FReadPos;
  Result := FStream.Read(Buffer, Count);
  FReadPos := FStream.Position;
end;
{$ENDIF}

{ TvgDataBlockInterpreter }
constructor TvgDataBlockInterpreter.Create(SendDataBlock: ISendDataBlock);
begin
  inherited Create;
  FSendDataBlock := SendDataBlock;
end;

destructor TvgDataBlockInterpreter.Destroy;
var
  i: Integer;
begin
  for i := ListCount(FDispatchList) - 1 downto 0 do
    TvgDataDispatch(FDispatchList[i]).FInterpreter := nil;
  FDispatchList.Free;
  FSendDataBlock := nil;
  inherited Destroy;
end;

procedure TvgDataBlockInterpreter.AddDispatch(Value: TvgDataDispatch);
begin
  ListAdd(FDispatchList, Value);
end;

procedure TvgDataBlockInterpreter.RemoveDispatch(Value: TvgDataDispatch);
begin
  ListRemove(FDispatchList, Value);
end;

{ Variant conversion methods }
function TvgDataBlockInterpreter.GetVariantPointer(const Value: Variant): Pointer;
begin
  case VarType(Value) of
    varEmpty, varNull: Result := nil;
    varDispatch: Result := TVarData(Value).VDispatch;
    varVariant: Result := @Value;
    varUnknown: Result := TVarData(Value).VUnknown;
  else
    Result := @TVarData(Value).VPointer;
  end;
end;

procedure TvgDataBlockInterpreter.CopyDataByRef(Source: TVarData; var Dest: TVarData);
var
  VType: Integer;
begin
  VType := Source.VType;
  if Source.VType and varArray = varArray then
  begin
    VarClear(OleVariant(Dest));
    SafeArrayCopy(PSafeArray(Source.VArray), PSafeArray(Dest.VArray));
  end else
    case Source.VType and varTypeMask of
      varEmpty, varNull: ;
      varOleStr:
      begin
        if (Dest.VType and varTypeMask) <> varOleStr then
          Dest.VOleStr := SysAllocString(Source.VOleStr) else
          SysReallocString(PBStr(Dest.VOleStr)^,Source.VOleStr);
      end;
      varDispatch: Dest.VDispatch := Source.VDispatch;
      varVariant: CopyDataByRef(PVarData(Source.VPointer)^, Dest);
      varUnknown: Dest.VUnknown := Source.VUnknown;
    else
      if Dest.VType = 0 then
        OleVariant(Dest) := OleVariant(Source) else
      if Dest.VType and varByRef = varByRef then
      begin
        VType := VType or varByRef;
        Move(Source.VPointer, Dest.VPointer^, VariantSize[Source.VType and varTypeMask]);
      end else
        Move(Source.VPointer, Dest.VPointer, VariantSize[Source.VType and varTypeMask]);
    end;
  Dest.VType := VType;
end;

function TvgDataBlockInterpreter.ReadArray(VType: Integer;
  const Data: IDataBlock): OleVariant;
var
  Flags: TVarFlags;
  LoDim, HiDim, Indices, Bounds: PIntArray;
  DimCount, VSize, i: Integer;
  P: Pointer;
  V: Variant;
  VarArrayPtr: PSafeArray;
begin
  VarClear(Result);
  Data.Read(DimCount, SizeOf(DimCount));
  VSize := DimCount * SizeOf(Integer);
  GetMem(LoDim, VSize);
  try
    GetMem(HiDim, VSize);
    try
      Data.Read(LoDim^, VSize);
      Data.Read(HiDim^, VSize);
      GetMem(Bounds, VSize * 2);
      try
        for i := 0 to DimCount - 1 do
        begin
          Bounds[i * 2] := LoDim[i];
          Bounds[i * 2 + 1] := HiDim[i];
        end;
        Result := VarArrayCreate(Slice(Bounds^,DimCount * 2), VType and varTypeMask);
      finally
        FreeMem(Bounds);
      end;
      VarArrayPtr := PSafeArray(TVarData(Result).VArray);
      if VType and varTypeMask in EasyArrayTypes then
      begin
        Data.Read(VSize, SizeOf(VSize));
        P := VarArrayLock(Result);
        try
          Data.Read(P^, VSize);
        finally
          VarArrayUnlock(Result);
        end;
      end else
      begin
        GetMem(Indices, VSize);
        try
          ZeroMem(Indices, VSize);
          for I := 0 to DimCount - 1 do
            Indices[I] := LoDim[I];
          while True do
          begin
            V := ReadVariant(Flags, Data);
            if VType and varTypeMask = varVariant then
              OleCheck(SafeArrayPutElement(VarArrayPtr, Indices^, V)) else
              OleCheck(SafeArrayPutElement(VarArrayPtr, Indices^, TVarData(V).VPointer^));
            Inc(Indices[DimCount - 1]);
            if Indices[DimCount - 1] > HiDim[DimCount - 1] then
              for i := DimCount - 1 downto 0 do
                if Indices[i] > HiDim[i] then
                begin
                  if i = 0 then Exit;
                  Inc(Indices[i - 1]);
                  Indices[i] := LoDim[i];
                end;
          end;
        finally
          FreeMem(Indices);
        end;
      end;
    finally
      FreeMem(HiDim);
    end;
  finally
    FreeMem(LoDim);
  end;
end;

procedure TvgDataBlockInterpreter.WriteArray(const Value: OleVariant;
  const Data: IDataBlock);
var
  VType, VSize, i, DimCount, ElemSize: Integer;
  VarArrayPtr: PSafeArray;
  LoDim, HiDim, Indices: PIntArray;
  V: Variant;
  P: Pointer;
begin
  VType := VarType(Value);
  Data.Write(VType, SizeOf(Integer));
  DimCount := VarArrayDimCount(Value);
  Data.Write(DimCount, SizeOf(DimCount));
  VarArrayPtr := PSafeArray(TVarData(Value).VArray);
  VSize := SizeOf(Integer) * DimCount;
  GetMem(LoDim, VSize);
  try
    GetMem(HiDim, VSize);
    try
      for i := 1 to DimCount do
      begin
        LoDim[i - 1] := VarArrayLowBound(Value, i);
        HiDim[i - 1] := VarArrayHighBound(Value, i);
      end;
      Data.Write(LoDim^,VSize);
      Data.Write(HiDim^,VSize);
      if VType and varTypeMask in EasyArrayTypes then
      begin
        ElemSize := SafeArrayGetElemSize(VarArrayPtr);
        VSize := 1;
        for i := 0 to DimCount - 1 do
          VSize := (HiDim[i] - LoDim[i] + 1) * VSize;
        VSize := VSize * ElemSize;
        P := VarArrayLock(Value);
        try
          Data.Write(VSize, SizeOf(VSize));
          Data.Write(P^,VSize);
        finally
          VarArrayUnlock(Value);
        end;
      end else
      begin
        GetMem(Indices, VSize);
        try
          for I := 0 to DimCount - 1 do
            Indices[I] := LoDim[I];
          while True do
          begin
            if VType and varTypeMask <> varVariant then
            begin
              OleCheck(SafeArrayGetElement(VarArrayPtr, Indices^, TVarData(V).VPointer));
              TVarData(V).VType := VType and varTypeMask;
            end else
              OleCheck(SafeArrayGetElement(VarArrayPtr, Indices^, V));
            WriteVariant(V, Data);
            Inc(Indices[DimCount - 1]);
            if Indices[DimCount - 1] > HiDim[DimCount - 1] then
              for i := DimCount - 1 downto 0 do
                if Indices[i] > HiDim[i] then
                begin
                  if i = 0 then Exit;
                  Inc(Indices[i - 1]);
                  Indices[i] := LoDim[i];
                end;
          end;
        finally
          FreeMem(Indices);
        end;
      end;
    finally
      FreeMem(HiDim);
    end;
  finally
    FreeMem(LoDim);
  end;
end;

function TvgDataBlockInterpreter.ReadVariant(out Flags: TVarFlags;
  const Data: IDataBlock): OleVariant;
var
  I, VType: Integer;
  W: WideString;
  TmpFlags: TVarFlags;
begin
  VarClear(Result);
  Flags := [];
  Data.Read(VType, SizeOf(VType));
  if VType and varByRef = varByRef then Include(Flags, vfByRef);
  if VType = varByRef then
  begin
    Include(Flags, vfVariant);
    Result := ReadVariant(TmpFlags, Data);
    Exit;
  end;
  if vfByRef in Flags then VType := VType xor varByRef;
  if (VType and varArray) = varArray then
    Result := ReadArray(VType, Data) else
  case VType and varTypeMask of
    varEmpty: VarClear(Result);
    varNull: Result := Null;
    varOleStr:
    begin
      Data.Read(I, SizeOf(Integer));
      SetLength(W, I);
      Data.Read(W[1], I * 2);
      Result := W;
    end;
    varDispatch:
    begin
      Data.Read(I, SizeOf(Integer));
      Result := TvgDataDispatch.Create(Self, I) as IDispatch;
    end;
    varUnknown:
      raise EInterpreterError.CreateFmt(SBadVariantType, [IntToHex(VType, 4)]);
  else
    TVarData(Result).VType := VType;
    Data.Read(TVarData(Result).VPointer, VariantSize[VType and varTypeMask]);
  end;
end;

procedure TvgDataBlockInterpreter.WriteVariant(const Value: OleVariant;
  const Data: IDataBlock);
var
  I, VType: Integer;
  W: WideString;
begin
  VType := VarType(Value);
  if VarIsArray(Value) then
    WriteArray(Value, Data) else
  case (VType and varTypeMask) of
    varEmpty, varNull: Data.Write(VType, SizeOf(Integer));
    varOleStr:
    begin
      W := WideString(Value);
      I := Length(W);
      Data.Write(VType, SizeOf(Integer));
      Data.Write(I,SizeOf(Integer));
      Data.Write(W[1], I * 2);
    end;
    varDispatch:
    begin
      if not VarIsArray(FDispList) then
        FDispList := VarArrayCreate([0,10], varVariant);
      I := 0;
      while I <= VarArrayHighBound(FDispList, 1) do
        if VarIsEmpty(FDispList[I]) then break else Inc(I);
      if I > VarArrayHighBound(FDispList, 1) then
        VarArrayRedim(FDispList, I + 10);
      FDispList[I] := Value;
      Data.Write(VType, SizeOf(Integer));
      Data.Write(I, SizeOf(Integer));
    end;
    varVariant:
    begin
      if VType and varByRef <> varByRef then
        raise EInterpreterError.CreateFmt(SBadVariantType, [IntToHex(VType, 4)]);
      I := varByRef;
      Data.Write(I, SizeOf(Integer));
      WriteVariant(Variant(TVarData(Value).VPointer^), Data);
    end;
    varUnknown:
      raise EInterpreterError.CreateFmt(SBadVariantType, [IntToHex(VType, 4)]);
  else
    Data.Write(VType, SizeOf(Integer));
    if VType and varByRef = varByRef then
      Data.Write(TVarData(Value).VPointer^, VariantSize[VType and varTypeMask]) else
      Data.Write(TVarData(Value).VPointer, VariantSize[VType and varTypeMask]);
  end;
end;

{ Sending Calls }
function TvgDataBlockInterpreter.CallGetServerList: OleVariant;
var
  Flags: TVarFlags;
  Data: IDataBlock;
begin
  Data := TDataBlock.Create as IDataBlock;
  Data.Signature := CallSig or asGetServers;
  Data := FSendDataBlock.Send(Data, True);
  CheckData(Data);
  Result := ReadVariant(Flags, Data);
end;

function TvgDataBlockInterpreter.CreateObject(Name: string): OleVariant;
var
  Flags: TVarFlags;
  Data: IDataBlock;
begin
  Data := TDataBlock.Create as IDataBlock;
  WriteVariant(Name, Data);
  Data.Signature := CallSig or asCreateObject;
  Data := FSendDataBlock.Send(Data, True);
  CheckData(Data);
  Result := VarAsType(ReadVariant(Flags, Data), varDispatch);
end;

procedure TvgDataBlockInterpreter.CallFreeObject(DispatchIndex: Integer);
var
  Data: IDataBlock;
begin
  Data := TDataBlock.Create as IDataBlock;
  WriteVariant(DispatchIndex, Data);
  Data.Signature := CallSig or asFreeObject;
  FSendDataBlock.Send(Data, False);
end;

function TvgDataBlockInterpreter.CallGetIDsOfNames(DispatchIndex: Integer;
  const IID: TGUID; Names: Pointer; NameCount, LocaleID: Integer;
  DispIDs: Pointer): HResult; stdcall;
var
  Flags: TVarFlags;
  Data: IDataBlock;
begin
  if NameCount <> 1 then
    Result := E_NOTIMPL else
  begin
    Data := TDataBlock.Create as IDataBlock;
    WriteVariant(DispatchIndex, Data);
    WriteVariant(WideString(POleStrList(Names)^[0]), Data);
    Data.Signature := CallSig or asGetID;
    Data := FSendDataBlock.Send(Data, True);
    CheckData(Data);
    Result := ReadVariant(Flags, Data);
    if Result = S_OK then
      PDispIdList(DispIDs)^[0] := ReadVariant(Flags, Data);
  end;
end;

function TvgDataBlockInterpreter.CallInvoke(DispatchIndex, DispID: Integer; const IID: TGUID; LocaleID: Integer;
  Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult; stdcall;
var
  VarFlags: TVarFlags;
  PDest: PVarData;
  I: Integer;
  Data: IDataBlock;
begin
  Data := TDataBlock.Create as IDataBlock;
  WriteVariant(DispatchIndex, Data);
  WriteVariant(DispID, Data);
  WriteVariant(Flags, Data);
  WriteVariant(VarResult <> nil, Data);
  WriteVariant(PDispParams(@Params).cArgs, Data);
  WriteVariant(PDispParams(@Params).cNamedArgs, Data);
  for I := 0 to PDispParams(@Params).cNamedArgs - 1 do
    WriteVariant(PDispParams(@Params).rgdispidNamedArgs[I], Data);
  for I := 0 to PDispParams(@Params).cArgs - 1 do
    WriteVariant(OleVariant(PDispParams(@Params).rgvarg^[I]), Data);
  Data.Signature := CallSig or asInvoke;
  Data := FSendDataBlock.Send(Data, True);
  CheckData(Data);
  Result := ReadVariant(VarFlags, Data);
  if (Result = DISP_E_EXCEPTION) then
  begin
    PExcepInfo(ExcepInfo).scode := ReadVariant(VarFlags, Data);
    PExcepInfo(ExcepInfo).bstrDescription := ReadVariant(VarFlags, Data);
  end;
  for I := 0 to PDispParams(@Params).cArgs - 1 do
    with PDispParams(@Params)^ do
      if rgvarg^[I].vt and varByRef = varByRef then
      begin
        if rgvarg^[I].vt = (varByRef or varVariant) then
          PDest := @TVarData(TVarData(rgvarg^[I]).VPointer^) else
          PDest := @TVarData(rgvarg^[I]);
        CopyDataByRef(TVarData(ReadVariant(VarFlags, Data)), PDest^);
      end;
  if VarResult <> nil then
    PVariant(VarResult)^ := ReadVariant(VarFlags, Data);
end;

{ Receiving Calls }
procedure TvgDataBlockInterpreter.InterpretData(const Data: IDataBlock);
var
  Action: Integer;
begin
  CheckData(Data);
  Action := Data.Signature;
  try
    case (Action and asMask) of
      asInvoke:
        DoInvoke(Data);
      asGetID:
        DoGetIDsOfNames(Data);
      asCreateObject:
        DoCreateObject(Data);
      asFreeObject:
        DoFreeObject(Data);
      asGetServers:
        DoGetServerList(Data);
    else
      if not DoCustomAction(Action and asMask, Data) then
        raise EInterpreterError.CreateFmt(SInvalidAction, [Action and asMask]);
    end;
  except
    on E: Exception do
    begin
      Data.Clear;
      WriteVariant(E.Message, Data);
      Data.Signature := ResultSig or asError;
      FSendDataBlock.Send(Data, False);
    end;
  end;
end;

procedure TvgDataBlockInterpreter.CheckData(const Data: IDataBlock);
begin
  if (Data.Signature and asMask) = asError then DoException(Data);
end;

procedure TvgDataBlockInterpreter.DoException(const Data: IDataBlock);
var
  VarFlags: TVarFlags;
begin
  raise Exception.Create(ReadVariant(VarFlags, Data));
end;

procedure TvgDataBlockInterpreter.DoGetServerList(const Data: IDataBlock);
var
  VList: OleVariant;
  List: TStringList;
begin
  Data.Clear;
  List := TStringList.Create;
  try
    GetDataBrokers(List);
    VList := VarArrayFromStrings(List);
  finally
    List.Free;
  end;
  WriteVariant(VList, Data);
  Data.Signature := ResultSig or asGetServers;
  FSendDataBlock.Send(Data, False);
end;

procedure TvgDataBlockInterpreter.DoCreateObject(const Data: IDataBlock);
var
  W: WideString;
  V: Variant;
  VarFlags: TVarFlags;
begin
  W := ReadVariant(VarFlags, Data);
  Data.Clear;
  if (W[1] = '{') and (W[Length(W)] = '}') then
    W := ClassIDToProgID(StringToGUID(W));
  { Original SConnect.pas CreateAnyComObject procedure was eliminated }
  { due errors in socket dispatcher in Win95 environment. }
  V := CreateOleObject(W);
  WriteVariant(V, Data);
  Data.Signature := ResultSig or asCreateObject;

  FSendDataBlock.Send(Data, False);
end;

procedure TvgDataBlockInterpreter.DoFreeObject(const Data: IDataBlock);
var
  DispatchIndex: Integer;
  VarFlags: TVarFlags;
begin
  try
    DispatchIndex := ReadVariant(VarFlags, Data);
    if (DispatchIndex >= 0) and (VarIsArray(FDispList)) and
       (DispatchIndex < VarArrayHighBound(FDispList, 1)) then
      FDispList[DispatchIndex] := UNASSIGNED;
  except
    { Don't return any exceptions }
  end;
end;

procedure TvgDataBlockInterpreter.DoGetIDsOfNames(const Data: IDataBlock);
var
  DispatchIndex, RetVal, DispID: Integer;
  Disp: IDispatch;
  W: WideString;
  VarFlags: TVarFlags;
begin
  DispatchIndex := ReadVariant(VarFlags, Data);
  W := ReadVariant(VarFlags, Data);
  Data.Clear;
  Disp := IDispatch(FDispList[DispatchIndex]);
  RetVal := Disp.GetIDsOfNames(GUID_NULL, @W, 1, 0, @DispID);
  WriteVariant(RetVal, Data);
  if RetVal = S_OK then
    WriteVariant(DispID, Data);
  Data.Signature := ResultSig or asGetID;
  FSendDataBlock.Send(Data, False);
end;

procedure TvgDataBlockInterpreter.DoInvoke(const Data: IDataBlock);
var
  ExcepInfo: TExcepInfo;
  DispParams: TDispParams;
  DispatchIndex, DispID, Flags, I: Integer;
  RetVal: HRESULT;
  ExpectResult: Boolean;
  VarFlags: TVarFlags;
  Disp: IDispatch;
  VarList: PVariantArray;
  V: OleVariant;
begin
  VarList := nil;
  ZeroMem(@ExcepInfo, SizeOf(ExcepInfo));
  ZeroMem(@DispParams, SizeOf(DispParams));
  DispatchIndex := ReadVariant(VarFlags, Data);
  DispID := ReadVariant(VarFlags, Data);
  Flags := ReadVariant(VarFlags, Data);
  ExpectResult := ReadVariant(VarFlags, Data);
  DispParams.cArgs := ReadVariant(VarFlags, Data);
  DispParams.cNamedArgs := ReadVariant(VarFlags, Data);
  Disp := IDispatch(FDispList[DispatchIndex]);
  try
    DispParams.rgdispidNamedArgs := nil;
    if DispParams.cNamedArgs > 0 then
    begin
      GetMem(DispParams.rgdispidNamedArgs, DispParams.cNamedArgs * SizeOf(Integer));
      for I := 0 to DispParams.cNamedArgs - 1 do
        DispParams.rgdispidNamedArgs[I] := ReadVariant(VarFlags, Data);
    end;
    if DispParams.cArgs > 0 then
    begin
      GetMem(DispParams.rgvarg, DispParams.cArgs * SizeOf(TVariantArg));
      GetMem(VarList, DispParams.cArgs * SizeOf(OleVariant));
      Initialize(VarList^, DispParams.cArgs);
      for I := 0 to DispParams.cArgs - 1 do
      begin
        VarList[I] := ReadVariant(VarFlags, Data);
        if vfByRef in VarFlags then
        begin
          if vfVariant in VarFlags then
          begin
            DispParams.rgvarg[I].vt := varVariant or varByRef;
            TVarData(DispParams.rgvarg[I]).VPointer := @VarList[I];
          end else
          begin
            DispParams.rgvarg[I].vt := VarType(VarList[I]) or varByRef;
            TVarData(DispParams.rgvarg[I]).VPointer := GetVariantPointer(VarList[I]);
          end;
        end else
          DispParams.rgvarg[I] := TVariantArg(VarList[I]);
      end;
    end;
    Data.Clear;
    RetVal := Disp.Invoke(DispID, GUID_NULL, 0, Flags, DispParams, @V, @ExcepInfo, nil);
    WriteVariant(RetVal, Data);
    if RetVal = DISP_E_EXCEPTION then
    begin
      WriteVariant(ExcepInfo.scode, Data);
      WriteVariant(ExcepInfo.bstrDescription, Data);
    end;
    if DispParams.rgvarg <> nil then
    begin
      for I := 0 to DispParams.cArgs - 1 do
        if DispParams.rgvarg[I].vt and varByRef = varByRef then
          WriteVariant(OleVariant(DispParams.rgvarg[I]), Data);
    end;
    if ExpectResult then WriteVariant(V, Data);
    Data.Signature := ResultSig or asInvoke;
    FSendDataBlock.Send(Data, False);
  finally
    if DispParams.rgdispidNamedArgs <> nil then
      FreeMem(DispParams.rgdispidNamedArgs);
    if VarList <> nil then
    begin
      Finalize(VarList^, DispParams.cArgs);
      FreeMem(VarList);
    end;
    if DispParams.rgvarg <> nil then
      FreeMem(DispParams.rgvarg);
  end;
end;

function TvgDataBlockInterpreter.DoCustomAction(Action: Integer;
  const Data: IDataBlock): Boolean;
begin
  Result := False;
end;

{ TvgDataDispatch }
constructor TvgDataDispatch.Create(Interpreter: TvgDataBlockInterpreter; DispatchIndex: Integer);
begin
  inherited Create;
  FDispatchIndex := DispatchIndex;
  FInterpreter := Interpreter;
  Interpreter.AddDispatch(Self);
end;

destructor TvgDataDispatch.Destroy;
begin
  if Assigned(FInterpreter) then
  begin
    FInterpreter.CallFreeObject(FDispatchIndex);
    FInterpreter.RemoveDispatch(Self);
  end;
  inherited Destroy;
end;

{ TvgDataDispatch.IDispatch }
function TvgDataDispatch.GetTypeInfoCount(out Count: Integer): HResult; stdcall;
begin
  Count := 0;
  Result := S_OK;
end;

function TvgDataDispatch.GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): HResult; stdcall;
begin
  Result := E_NOTIMPL;
end;

function TvgDataDispatch.GetIDsOfNames(const IID: TGUID; Names: Pointer;
  NameCount, LocaleID: Integer; DispIDs: Pointer): HResult; stdcall;
begin
  Result := FInterpreter.CallGetIDsOfNames(FDispatchIndex, IID, Names, NameCount,
    LocaleID, DispIDs);
end;

function TvgDataDispatch.Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer;
  Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult; stdcall;
begin
  Result := FInterpreter.CallInvoke(FDispatchIndex, DispID, IID, LocaleID, Flags,
    Params, VarResult, ExcepInfo, ArgErr);
end;

{ TvgStreamedConnection }
constructor TvgStreamedConnection.Create(AOwner: TComponent);
var
  Obj: ISendDataBlock;
begin
  inherited Create(AOwner);
  InitializeCriticalSection(FLock);
  GetInterface(ISendDataBlock, Obj);
  FInterpreter := TvgDataBlockInterpreter.Create(Obj);
end;

destructor TvgStreamedConnection.Destroy;
begin
  SetConnected(False);
  FInterpreter.Free;
  inherited Destroy;
  DeleteCriticalSection(FLock);
end;

function TvgStreamedConnection.GetServerList: OleVariant;
var
  TransportCreated: Boolean;
begin
  Lock;
  try
    TransportCreated := not Assigned(FTransport);
    if TransportCreated then
      FTransport := CreateTransport;
    try
      FTransport.SetConnected(True);
      Result := Interpreter.CallGetServerList;
    finally
      if TransportCreated then
        FTransport := nil;
    end;
  finally
    Unlock;
  end;
end;

procedure TvgStreamedConnection.DoError(E: Exception);
begin
  raise E;
end;

procedure TvgStreamedConnection.DoConnect;
var
  TempStr: string;
begin
  Lock;
  try
    FTransport := CreateTransport;
    try
      FTransport.SetConnected(True);
      if ServerGUID <> '' then
        TempStr := ServerGUID else
        TempStr := ServerName;
      if TempStr = '' then
        raise Exception.CreateFmt(SServerNameBlank, [Name]);
      SetAppServer(Interpreter.CreateObject(TempStr));
    except
      FTransport := nil;
      raise;
    end;
  finally
    Unlock;
  end;
end;

procedure TvgStreamedConnection.DoDisconnect;
begin
  Lock;
  try
    inherited DoDisconnect;
    FTransport := nil;
  finally
    Unlock;
  end;
end;

function TvgStreamedConnection.CreateTransport: ITransport;
begin
  Result := nil;
end;

procedure TvgStreamedConnection.Lock;
begin
  EnterCriticalSection(FLock);
end;

procedure TvgStreamedConnection.Unlock;
begin
  LeaveCriticalSection(FLock);
end;

{ TvgStreamedConnection.IUnknown }
function TvgStreamedConnection.QueryInterface(const IID: TGUID; out Obj): HResult;
begin
  if GetInterface(IID, Obj) then Result := 0 else Result := E_NOINTERFACE;
end;

function TvgStreamedConnection._AddRef: Integer;
begin
  Inc(FRefCount);
  Result := FRefCount;
end;

function TvgStreamedConnection._Release: Integer;
begin
  Dec(FRefCount);
  Result := FRefCount;
end;

{ TvgStreamedConnection.ISendDataBlock }
function TvgStreamedConnection.Send(const Data: IDataBlock; WaitForResult: Boolean): IDataBlock;
begin
  Lock;
  try
    { There is possible situations when IDispatchs calls FreeObject  }
    { when transport is no longer exists. }
    if Assigned(FTransport) then
      FTransport.Send(Data)
    else
      WaitForResult := False;

    Result := nil;
    while WaitForResult do
    begin
      Result := FTransport.Receive(True{$IFDEF _D5_}, 0{$ENDIF});
      if Assigned(Result) then
        if ((Result.Signature and ResultSig) <> ResultSig) then
          FInterpreter.InterpretData(Result)
        else
          WaitForResult := False
      else
        WaitForResult := False;
    end;
  finally
    Unlock;
  end;
end;

{$IFNDEF _D4_}
{ TSocketTransport }
destructor TSocketTransport.Destroy;
begin
  FInterceptor := nil;
  FClientSocket.Free;
  WSACloseEvent(FEvent);
  inherited Destroy;
end;

function TSocketTransport.GetWaitEvent: THandle;
begin
  FEvent := WSACreateEvent;
  WSAEventSelect(FSocket.SocketHandle, FEvent, FD_READ or FD_CLOSE);
  Result := FEvent;
end;

function TSocketTransport.GetConnected: Boolean;
begin
  Result := (FSocket <> nil) and (FSocket.Connected);
end;

procedure TSocketTransport.SetConnected(Value: Boolean);
begin
  if GetConnected = Value then Exit;
  if Value then
  begin
    FClientSocket := TClientSocket.Create(nil);
    FClientSocket.ClientType := ctBlocking;
    FSocket := FClientSocket.Socket;
    FClientSocket.Port := FPort;
    if FAddress <> '' then
      FClientSocket.Address := FAddress else
      FClientSocket.Host := FHost;
    FClientSocket.Open;
  end else
    FSocket.Close;
end;

function TSocketTransport.Receive(WaitForInput: Boolean): IDataBlock;
var
  RetLen, Sig, StreamLen: Integer;
  P: Pointer;
  FDSet: TFDSet;
  TimeVal: PTimeVal;
  RetVal: Integer;
begin
  Result := nil;
  TimeVal := nil;
  FD_ZERO(FDSet);
  FD_SET(FSocket.SocketHandle, FDSet);
  if not WaitForInput then
  begin
    New(TimeVal);
    TimeVal.tv_sec := 0;
    TimeVal.tv_usec := 1;
  end;
  RetVal := select(0, @FDSet, nil, nil, TimeVal);

  if Assigned(TimeVal) then
    FreeMem(TimeVal);
  if RetVal = SOCKET_ERROR then
    raise ESocketConnectionError.Create(SysErrorMessage(WSAGetLastError));
  if (RetVal = 0) then Exit;
  RetLen := FSocket.ReceiveBuf(Sig, SizeOf(Sig));
  if RetLen <> SizeOf(Sig) then
    raise ESocketConnectionError.Create(SSocketReadError);
  if (Sig and CallSig <> CallSig) and
     (Sig and ResultSig <> ResultSig) then
    raise Exception.Create(SInvalidDataPacket);
  RetLen := FSocket.ReceiveBuf(StreamLen, SizeOf(StreamLen));
  if RetLen = 0 then
    raise ESocketConnectionError.Create(SSocketReadError);
  if RetLen <> SizeOf(StreamLen) then
    raise ESocketConnectionError.Create(SSocketReadError);
  Result := TDataBlock.Create as IDataBlock;
  Result.Size := StreamLen;
  Result.Signature := Sig;
  P := Result.Memory;
  Inc(Integer(P), Result.BytesReserved);
  while StreamLen > 0 do
  begin
    RetLen := FSocket.ReceiveBuf(P^, StreamLen);
    if RetLen = 0 then
      raise ESocketConnectionError.Create(SSocketReadError);
    if RetLen > 0 then
    begin
      Dec(StreamLen, RetLen);
      Inc(Integer(P), RetLen);
    end;
  end;
  if StreamLen <> 0 then
    raise ESocketConnectionError.Create(SInvalidDataPacket);
  InterceptIncoming(Result);
end;

procedure TSocketTransport.Send(const Data: IDataBlock);
var
  P: Pointer;
begin
  InterceptOutgoing(Data);
  P := Data.Memory;
  FSocket.SendBuf(P^, Data.Size + Data.BytesReserved);
end;

function TSocketTransport.CheckInterceptor: Boolean;
var
  GUID: TGUID;
begin
  if not Assigned(FInterceptor) and (FInterceptGUID <> '') then
    if not FCreateAttempted then
    try
      FCreateAttempted := True;
      Guid := StringToGuid(FInterceptGUID);
      FInterceptor := CreateComObject(Guid) as IDataIntercept;
    except
      { raise no exception if the creating failed }
    end;
  Result := Assigned(FInterceptor);
end;

procedure TSocketTransport.InterceptIncoming(const Data: IDataBlock);
begin
  if CheckInterceptor then
    FInterceptor.DataIn(Data);
end;

procedure TSocketTransport.InterceptOutgoing(const Data: IDataBlock);
begin
  if CheckInterceptor then
    FInterceptor.DataOut(Data);
end;
{$ENDIF}

{ TSocketConnection }
constructor TvgSocketConnection.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPort := 211;
end;

function TvgSocketConnection.IsAddressStored: Boolean;
begin
  Result := (Address <> '');
end;

procedure TvgSocketConnection.SetAddress(Value: string);
begin
  if Value <> '' then FHost := '';
  FAddress := Value;
end;

function TvgSocketConnection.IsHostStored: Boolean;
begin
  Result := (Host <> '');
end;

procedure TvgSocketConnection.SetHost(Value: string);
begin
  if Value <> '' then FAddress := '';
  FHost := Value;
end;

function TvgSocketConnection.CreateTransport: ITransport;
var
  SocketTransport: TSocketTransport;
begin
  SocketTransport := TSocketTransport.Create;
  SocketTransport.Host := FHost;
  SocketTransport.Address := FAddress;
  SocketTransport.Port := FPort;
  SocketTransport.InterceptGUID := InterceptGUID;
  Result := SocketTransport as ITransport;
end;

procedure TvgSocketConnection.DoConnect;
begin
  CheckLoadWinSock2;
  inherited DoConnect;
end;

{ TvgSocketDispatcher }
constructor TvgSocketDispatcher.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  CheckLoadWinSock2;
  ServerType := stThreadBlocking;
  OnGetThread := GetThread;
end;

procedure TvgSocketDispatcher.GetThread(Sender: TObject;
  ClientSocket: TServerClientWinSocket;
  var SocketThread: TServerClientThread);
begin
  SocketThread := TvgSocketDispatcherThread.Create(False, ClientSocket,
    InterceptGUID, Timeout);
end;

{ TvgSocketDispatcherThread }
constructor TvgSocketDispatcherThread.Create(CreateSuspended: Boolean;
  ASocket: TServerClientWinSocket; const InterceptGUID: string; Timeout: Integer);
begin
  FInterceptGUID := InterceptGUID;
  FTimeout := EncodeTime(Timeout div 60, Timeout mod 60, 0, 0);
  FLastActivity := Now;
  inherited Create(CreateSuspended, ASocket);
end;

function TvgSocketDispatcherThread.CreateServerTransport: ITransport;
var
  SocketTransport: TSocketTransport;
begin
  SocketTransport := TSocketTransport.Create;
  SocketTransport.Socket := ClientSocket;
  SocketTransport.InterceptGUID := FInterceptGUID;
  Result := SocketTransport as ITransport;
end;

{ TvgSocketDispatcherThread.IUnknown }
function TvgSocketDispatcherThread.QueryInterface(const IID: TGUID; out Obj): HResult;
begin
  if GetInterface(IID, Obj) then Result := 0 else Result := E_NOINTERFACE;
end;

function TvgSocketDispatcherThread._AddRef: Integer;
begin
  Inc(FRefCount);
  Result := FRefCount;
end;

function TvgSocketDispatcherThread._Release: Integer;
begin
  Dec(FRefCount);
  Result := FRefCount;
end;

{ TvgSocketDispatcherThread.ISendDataBlock }
function TvgSocketDispatcherThread.Send(const Data: IDataBlock; WaitForResult: Boolean): IDataBlock;
begin
  FTransport.Send(Data);
  if WaitForResult then
    while True do
    begin
      Result := FTransport.Receive(True{$IFDEF _D5_}, 0{$ENDIF});
      if Result = nil then Break;
      if (Result.Signature and ResultSig) = ResultSig then
        Break else
        FInterpreter.InterpretData(Result);
    end;
end;

procedure TvgSocketDispatcherThread.ClientExecute;
var
  Data: IDataBlock;
  Obj: ISendDataBlock;
  Event: THandle;
  WaitTime: DWord;
begin
  CoInitialize(nil);
  try
    FTransport := CreateServerTransport;
    try
      Event := FTransport.GetWaitEvent;
      GetInterface(ISendDataBlock, Obj);
      FInterpreter := TvgDataBlockInterpreter.Create(Obj);
      try
        Obj := nil;
        if FTimeout = 0 then
          WaitTime := INFINITE else
          WaitTime := 1000;
        while not Terminated and FTransport.Connected do
        try
          case WaitForSingleObject(Event, WaitTime) of
            WAIT_OBJECT_0:
            begin
              WSAResetEvent(Event);
              Data := FTransport.Receive(False{$IFDEF _D5_}, 0{$ENDIF});
              if Assigned(Data) then
              begin
                FLastActivity := Now;
                FInterpreter.InterpretData(Data);
                Data := nil;
                FLastActivity := Now;
              end;
            end;
            WAIT_TIMEOUT:
              begin
                if (FTimeout > 0) and ((Now - FLastActivity) > FTimeout) then
                  FTransport.Connected := False;
              end;
          end;
        except
          FTransport.Connected := False;
        end;
      finally
        FInterpreter.Free;
        FInterpreter := nil;
      end;
    finally
      FTransport := nil;
    end;
  finally
    CoUninitialize;
  end;
end;

end.


