{*******************************************************}
{                                                       }
{         Vladimir Gaitanoff Delphi VCL Library         }
{         WinSock utility routines                      }
{                                                       }
{         Copyright (c) 1997, 1999                      }
{                                                       }
{*******************************************************}

{$I VG.INC }
{$D-,L- }

unit vgSock;

interface
uses Windows, Winsock, ScktComp;

{ --- IP Address parsing routunes }
function IsAddress(const ComputerName: string): Boolean;
{ Returns true of ComputerName may be assigned to Address property }

function IsHost(const ComputerName: string): Boolean;
{ Returns true of ComputerName may be assigned to Host property }

{ --- TFDSet routines }
function SocketSet(Socket: TCustomWinSocket): TFDSet;
{ Returs FDSet structure initialized with Socket.SocketHandle }

{ --- TTimeVal routines }
procedure SetTimeVal(var TimeVal: TTimeVal; Secs, MicroSecs: Integer);
{ Initializes TimeVal structure }

function GetTimeVal(Secs, MicroSecs: Integer): TTimeVal;
{ Returns initialized TTimeVal structure }

procedure SetMSecTimeVal(var TimeVal: TTimeVal; MSecs: Integer);
{ Initializes TTimeVal structure }

function GetMSecTimeVal(var TimeVal: TTimeVal; MSecs: Integer): TTimeVal;
{ Returns initialized TTimeVal structure }

{ --- Send and receive routines }
function ReceiveString(Socket: TCustomWinSocket): string;
{ Receives string that exists in recieve buffer }

function LoadWinSock2: Boolean;


var
  WSACreateEvent: function: THandle stdcall;
  WSAResetEvent: function(hEvent: THandle): Boolean stdcall;
  WSACloseEvent: function(hEvent: THandle): Boolean stdcall;
  WSAEventSelect: function(s: TSocket; hEventObject: THandle; lNetworkEvents: Integer): Integer stdcall;

const
  AddrLoopback = '127.0.0.1';

implementation

function IsAddress(const ComputerName: string): Boolean;
var
  I, Points: Integer;
begin
  Points := 0;
  for I := 1 to Length(ComputerName) do
  begin
    if (ComputerName[I] = '.') then
    begin
      Inc(Points);
      if Points > 3 then
      begin
        Result := False;
        Exit;
      end else
        Continue;
    end;

    if not (ComputerName[I] in ['0'..'9']) then
    begin
      Result := False;
      Exit;
    end;
  end;
  Result := True;
end;

function IsHost(const ComputerName: string): Boolean;
begin
  Result := not IsAddress(ComputerName);
end;

function SocketSet(Socket: TCustomWinSocket): TFDSet;
begin
  FD_ZERO(Result);
  FD_SET(Socket.SocketHandle, Result);
end;

procedure SetTimeVal(var TimeVal: TTimeVal; Secs, MicroSecs: Integer);
begin
  TimeVal.tv_sec := Secs;
  TimeVal.tv_usec := MicroSecs;
end;

function GetTimeVal(Secs, MicroSecs: Integer): TTimeVal;
begin
  SetTimeVal(Result, Secs, MicroSecs);
end;

procedure SetMSecTimeVal(var TimeVal: TTimeVal; MSecs: Integer);
begin
  TimeVal.tv_sec := MSecs div 1000;
  TimeVal.tv_usec := (MSecs mod 1000) * 1000;
end;

function GetMSecTimeVal(var TimeVal: TTimeVal; MSecs: Integer): TTimeVal;
begin
 SetMSecTimeVal(Result, MSecs);
end;

function ReceiveString(Socket: TCustomWinSocket): string;
var
  I: Integer;
begin
  Socket.Lock;
  try
    I := Socket.ReceiveLength;
    SetLength(Result, I);
    Socket.ReceiveBuf(Result[1], I);
  finally
    Socket.Unlock;
  end;
end;

var
  hWinSock2: THandle;

function LoadWinSock2: Boolean;
const
  DLLName = 'ws2_32.dll';
begin
  Result := hWinSock2 > HINSTANCE_ERROR;
  if Result then Exit;
  hWinSock2 := LoadLibrary(PChar(DLLName));
  Result := hWinSock2 > HINSTANCE_ERROR;
  if Result then
  begin
    WSACreateEvent := GetProcAddress(hWinSock2, 'WSACreateEvent');
    WSAResetEvent := GetProcAddress(hWinSock2, 'WSAResetEvent');
    WSACloseEvent := GetProcAddress(hWinSock2, 'WSACloseEvent');
    WSAEventSelect := GetProcAddress(hWinSock2, 'WSAEventSelect');
  end;
end;

procedure FreeWinSock2;
begin
  if hWinSock2 > HINSTANCE_ERROR then
  begin
    WSACreateEvent := nil;
    WSAResetEvent := nil;
    WSACloseEvent := nil;
    WSAEventSelect := nil;
    FreeLibrary(hWinSock2);
  end;
  hWinSock2 := 0;
end;

initialization

finalization
  FreeWinSock2;

end.
