{*******************************************************}
{                                                       }
{         Vladimir Gaitanoff Delphi VCL Library         }
{                                                       }
{         vgUDFLib unit                                 }
{                                                       }
{         Copyright (c) 1997, 1998                      }
{                                                       }
{*******************************************************}

{$I VG.INC }
{$D-,L-}

unit vgUDFLib;

interface
uses SysUtils, Classes, ibase, ib_externals;

{.$DEFINE Debug}

{ NVL functions }
function NvlFloat(var Value: Double): Double; cdecl; export;
function NvlInteger(var Value: Integer): Integer; cdecl; export;
function NvlString(Value: PChar; NullTo: PChar): PChar; cdecl; export;

{ Case functions }
{ CaseMask = val1,Case1,val2,Case2,CaseElse }
function CaseString(Value: PChar; CaseMask: PChar): PChar; cdecl; export;

{ Math functions }
function RoundFloat(var Value: Double; var Digits: Integer): Double; cdecl; export;
function CompareFloat(var Value1, Value2: Double; var Digits: Integer): Integer; cdecl; export;
function IsEqualFloat(var Value1, Value2: Double; var Digits: Integer): Integer; cdecl; export;
function IsAboveFloat(var Value1, Value2: Double; var Digits: Integer): Integer; cdecl; export;
function IsBehindFloat(var Value1, Value2: Double; var Digits: Integer): Integer; cdecl; export;
function IsAboveEqualFloat(var Value1, Value2: Double; var Digits: Integer): Integer; cdecl; export;
function IsBehindEqualFloat(var Value1, Value2: Double; var Digits: Integer): Integer; cdecl; export;

function CompareInteger(var Value1, Value2: Integer): Integer; cdecl; export;
function IsEqualInteger(var Value1, Value2: Integer): Integer; cdecl; export;
function IsAboveInteger(var Value1, Value2: Integer): Integer; cdecl; export;
function IsBehindInteger(var Value1, Value2: Integer): Integer; cdecl; export;
function IsAboveEqualInteger(var Value1, Value2: Integer): Integer; cdecl; export;
function IsBehindEqualInteger(var Value1, Value2: Integer): Integer; cdecl; export;

{ DateTime functions }
function EncodeDate(var Day, Month, Year: Integer): PISC_QUAD; cdecl; export;

{ Log }
function WriteLog(Msg: PChar): Integer; cdecl; export;

implementation
uses Windows, vgUtils, vgSystem, vgMSVCRT;

var
  FBuffer: TTlsBuffer;

function Buffer: PChar;
begin
  Result := FBuffer.Memory;
end;

procedure HandleException(const Func: String; E: TObject);
begin
  AppWriteLog(Format('Module vgUDF.DLL: exception %s in function %s', [Exception(E).Message, Func]));
end;

function NvlFloat(var Value: Double): Double;
begin
{$IFDEF Debug }
  AppWriteLog('NvlFloat enter');
{$ENDIF}
  Result := Value;
{$IFDEF Debug }
  AppWriteLog('NvlFloat leave');
{$ENDIF}
end;

function NvlInteger(var Value: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('NvlInteger enter');
{$ENDIF}
  Result := Value;
{$IFDEF Debug }
  AppWriteLog('NvlInteger leave');
{$ENDIF}
end;

function NvlString(Value: PChar; NullTo: PChar): PChar;
var
  Tmp: PChar;
begin
{$IFDEF Debug }
  AppWriteLog('NvlString enter');
{$ENDIF}
  try
    if StrLen(Value) = 0 then Tmp := NullTo else Tmp := Value;
    Result := StrCopy(Buffer, Tmp);
  except
    HandleException('NvlString', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('NvlString leave');
{$ENDIF}
end;

function CaseString(Value: PChar; CaseMask: PChar): PChar;
var
  Strings: TStrings;
  I, Cases: Integer;
  Found: Boolean;
  Tmp: String;
begin
{$IFDEF Debug }
  AppWriteLog('CaseString enter');
{$ENDIF}
  try
    Strings := TStringList.Create;
    try
      Strings.CommaText := Trim(StrPas(CaseMask));
      Cases := Strings.Count div 2;
      Tmp := '';
      Found := False;
      for I := 0 to Cases - 1 do
      begin
        { comparing }
        if AnsiCompareStr(StrPas(Value), Strings[I * 2]) = 0 then
        begin
          Tmp := Strings[I * 2 + 1];
          Found := True;
          Break;
        end;
      end;
      { else }
      if not Found and (Strings.Count mod 2 = 1) then
        Tmp := Strings[Strings.Count - 1];
    finally
      Strings.Free;
    end;
    Result := StrCopy(Buffer, PChar(Tmp));
  except
    HandleException('CaseString', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('CaseString leave');
{$ENDIF}
end;

function RoundFloat(var Value: Double; var Digits: Integer): Double;
begin
{$IFDEF Debug }
  AppWriteLog('RoundString enter');
{$ENDIF}
  try
    Result := vgUtils.RoundFloat(Value, Digits);
  except
    HandleException('RoundFloat', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('RoundString leave');
{$ENDIF}
end;

function CompareFloat(var Value1, Value2: Double; var Digits: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('CompareFloat enter');
{$ENDIF}
  try
    Result := vgUtils.CompareFloat(Value1, Value2, Digits);
  except
    HandleException('CompareFloat', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('CompareFloat leave');
{$ENDIF}
end;

function IsEqualFloat(var Value1, Value2: Double; var Digits: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('IsEqualFloat enter');
{$ENDIF}
  try
    Result := Ord(CompareFloat(Value1, Value2, Digits) = 0);
  except
    HandleException('IsEqualFloat', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('IsEqualFloat leave');
{$ENDIF}
end;

function IsAboveFloat(var Value1, Value2: Double; var Digits: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('IsAboveFloat enter');
{$ENDIF}
  try
    Result := Ord(CompareFloat(Value1, Value2, Digits) = 1);
  except
    HandleException('IsAboveFloat', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('IsAboveFloat leave');
{$ENDIF}
end;

function IsBehindFloat(var Value1, Value2: Double; var Digits: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('IsBehindFloat enter');
{$ENDIF}
  try
    Result := Ord(CompareFloat(Value1, Value2, Digits) = -1);
  except
    HandleException('IsBehindFloat', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('IsBehindFloat leave');
{$ENDIF}
end;

function IsAboveEqualFloat(var Value1, Value2: Double; var Digits: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('IsAboveEqualFloat enter');
{$ENDIF}
  try
    Result := Ord(CompareFloat(Value1, Value2, Digits) >= 0);
  except
    HandleException('IsAboveEqualFloat', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('IsAboveEqualFloat leave');
{$ENDIF}
end;

function IsBehindEqualFloat(var Value1, Value2: Double; var Digits: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('IsBehindEqualFloat enter');
{$ENDIF}
  try
    Result := Ord(CompareFloat(Value1, Value2, Digits) <= 0);
  except
    HandleException('IsBehindEqualFloat', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('IsBehindEqualFloat leave');
{$ENDIF}
end;

function CompareInteger(var Value1, Value2: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('CompareInteger enter');
{$ENDIF}
  try
    if Value1 > Value2 then Result := 1 else
    if Value1 < Value2 then Result := -1 else Result := 0;
  except
    HandleException('CompareInteger', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('CompareInteger leave');
{$ENDIF}
end;

function IsEqualInteger(var Value1, Value2: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('IsEqualInteger enter');
{$ENDIF}
  try
    Result := Ord(CompareInteger(Value1, Value2) = 0);
  except
    HandleException('IsEqualInteger', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('IsEqualInteger leave');
{$ENDIF}
end;

function IsAboveInteger(var Value1, Value2: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('IsAboveInteger enter');
{$ENDIF}
  try
    Result := Ord(CompareInteger(Value1, Value2) = 1);
  except
    HandleException('IsAboveInteger', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('IsAboveInteger leave');
{$ENDIF}
end;

function IsBehindInteger(var Value1, Value2: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('IsBehindInteger enter');
{$ENDIF}
  try
    Result := Ord(CompareInteger(Value1, Value2) = -1);
  except
    HandleException('IsBehindInteger', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('IsBehindInteger leave');
{$ENDIF}
end;

function IsAboveEqualInteger(var Value1, Value2: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('IsAboveEqualInteger enter');
{$ENDIF}
  try
    Result := Ord(CompareInteger(Value1, Value2) >= 0);
  except
    HandleException('IsAboveEqualInteger', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('IsAboveEqualInteger leave');
{$ENDIF}
end;

function IsBehindEqualInteger(var Value1, Value2: Integer): Integer;
begin
{$IFDEF Debug }
  AppWriteLog('IsBehindEqualInteger enter');
{$ENDIF}
  try
    Result := Ord(CompareInteger(Value1, Value2) <= 0);
  except
    HandleException('IsBehindEqualInteger', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('IsBehindEqualInteger leave');
{$ENDIF}
end;

{ DateTime functions }
function EncodeDate(var Day, Month, Year: Integer): PISC_QUAD;
var
  res_date: tm;
begin
{$IFDEF Debug }
  AppWriteLog('EncodeDate enter');
{$ENDIF}
  try
    InitializeTCTimeStructure(res_date);
    with res_date do begin
      tm_mday := Day;
      tm_mon := Month + 1;
      tm_year := Year;
    end;
    Result := Pointer(Buffer);
    isc_encode_date(@res_date, Result);
  except
    HandleException('EncodeDate', ExceptObject);
    raise;
  end;
{$IFDEF Debug }
  AppWriteLog('EncodeDate leave');
{$ENDIF}
end;

function WriteLog(Msg: PChar): Integer; cdecl; export;
begin
{$IFDEF Debug }
  AppWriteLog('WriteLog enter');
{$ENDIF}
  AppWriteLog(StrPas(Msg));
  Result := 0;
{$IFDEF Debug }
  AppWriteLog('WriteLog leave');
{$ENDIF}
end;

procedure LibEntry(Reason: Integer);
begin
  case Reason of
    DLL_THREAD_DETACH:
    begin
      TTlsBuffer.DoDetachThread;
    end;
  end;
end;

initialization
  DLLProc := @LibEntry;
  FBuffer := TTlsBuffer.Create(1024);

finalization
  FBuffer.Free;

end.
