unit ElTree;

{$define LITE}

{$I ElPack.inc}
{.$DEFINE EL_DEMO}
{$B-}                      
{$RANGECHECKS OFF}
{$ALIGN ON}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  StdCtrls, ExtCtrls, ElTools, ElHeader, ElList,
{$IFNDEF LITE}
  ElImgLst,
  ElIni,
  DragDrop,
{$ENDIF}
  ElVCLUtils,
{$IFNDEF VER90}
  ActiveX,
{$ENDIF}
{$IFDEF D_4_UP}
  ImgList,
{$ENDIF}
  Menus;

{$R *.res}

const stsFocused    = 1;  // TElTreeItem states
      stsSelected   = 2;
      stsCut        = 3;
      stsUnderlined = 4;
      stsBold       = 5;
      stsItalic     = 6;
      stsExpanded   = 7;
      stsStrikeOut  = 8;
      stiMaxState   = 8;

type TSTIStates = set of 1 .. stiMaxState;

type EElTreeError = class(Exception)
     end;

const STExOutOfBounds  = 2356; // error message ID 1 - Tree index out of bounds
      STexInvItem      = 2357; // error message ID 2 - Tree item not specified
      STexRecursiveMove= 2358; // error message ID 3 - moving to subitem

type TItemChangeMode = (icmText, icmState, icmCheckState);

type TSTItemPart = (ipButton, ipMainText, ipColumn, ipInside, ipPicture, ipCheckBox);
     TSTSelModes = (smSimple, smUsual);
     TSortDirs   = (sdAscend, sdDescend);
     TSortModes  = (smNone, smAdd, smClick, smAddClick);
     TSortTypes  = (stCustom, stText, stNumber, stFloating, stDateTime, stDate, stTime);
     THintModes  = (shmNone,shmLong,shmAll);
     TDragImgMode= (dimNever, dimOne, dimAll);
     TNodeAttachMode = (naAdd, naAddFirst, naAddChild, naAddChildFirst, naInsert);
     TElCheckBoxType = (ectCheckBox, ect3SCheckBox, ectRadioButton);

type TElDragType = (dtOLE, dtDelphi, dtBoth);

type TDragTargetDraw = (ColorFrame, ColorRect, SelColorRect);
     // ColorFrame - usual background, color frame
     // ColorRect  - green rect (drop allowed) or red rect (drop not allowed)
     // SelColorRect - usual selected color rect

// *****************************************************************************

type TCustomElTree = class;
     TElTreeItem = class;
     TElTreeItems = class;

     TElCellStyle = class
       private
         FTag: Integer;
         FOwner: TElTreeItem;
         FCellBkColor: TColor;
         FTextBkColor: TColor;
         FTextColor: TColor;
         FTextFlags: DWORD;
         FPicture: TBitmap;
         FCellType: TFieldTypes;
         FStyle: TElSectionStyle;
         FOwnerProps: Boolean;
         FFontSize: integer;
         FFontStyles: TFontStyles;
         FFontName: TFontName;
         procedure SetFontSize(newValue: integer);
         procedure SetFontStyles(newValue: TFontStyles);
         procedure SetFontName(newValue: TFontName);
         procedure SetOwnerColors(newValue: Boolean);
         procedure SetStyle(newValue: TElSectionStyle);
         procedure SetCellBkColor(newValue: TColor);
         procedure SetTextBkColor(newValue: TColor);
         procedure SetTextColor(newValue: TColor);
         procedure SetTextFlags(newValue: DWORD);
         procedure SetPicture(newValue: TBitmap);
         procedure SetCellType(newValue: TFieldTypes);
       protected
         procedure Assign(Source: TElCellStyle);
       public
         constructor Create(Owner : TElTreeItem);

         property Tag: Integer read FTag write FTag;
         property CellBkColor: TColor read FCellBkColor write SetCellBkColor;
         property TextBkColor: TColor read FTextBkColor write SetTextBkColor;
         property TextColor: TColor read FTextColor write SetTextColor;
         property TextFlags: DWORD read FTextFlags write SetTextFlags;
         property Picture: TBitmap read FPicture write SetPicture;
         property CellType: TFieldTypes read FCellType write SetCellType;
         property Style: TElSectionStyle read FStyle write SetStyle;
         property OwnerProps: Boolean read FOwnerProps write SetOwnerColors;
         property FontSize: integer read FFontSize write SetFontSize;
         property FontStyles: TFontStyles read FFontStyles write SetFontStyles;
         property FontName: TFontName read FFontName write SetFontName;
     end;

     TElTreeItem = class (TPersistent)
      private
       FFullyVisible: Boolean;
       FHidden: Boolean;
       FEnabled : Boolean;
       FSuppressButtons: Boolean;
       FCheckBoxEnabled : Boolean;
       FCheckBoxState : TCheckBoxState;
       FShowCheckBox : Boolean;
       FCheckBoxType : TElCheckBoxType;
       FMainStyle: TElCellStyle;
       FUseStyles: Boolean;
       FState       : TSTIStates;
       FText        : string; // Main text
       FColText     : TStringList; // text for additional columns
       FChildren    : TElList;
       FOwner       : TCustomElTree;
       FList        : TElTreeItems;
       FData        : pointer; // generic data
       FColor,
       FBkColor     : TColor;
       FParentColors: boolean;
       FTextLeft,
       FTextRight   : integer;
       FParentStyle : boolean;
       FImageIndex,
       FStImageIndex: integer;
       FImageIndex2,
       FStImageIndex2: integer;
       FImageDrawn  : boolean;
       FImageDrawn2 : boolean;
       FForceButtons: boolean;
       FStyles      : TElList;

       function GetParent: TElTreeItem;
       function GetLevel : integer;
       procedure SetColor(index:integer;value:TColor);
       function EqualColumns(cols:TStringList):boolean;
       function GetHasChildren:boolean;
       function GetHasVisibleChildren: Boolean;
       procedure SetExpanded(value:boolean);
       procedure SetParentColors(value:boolean);
       procedure SetParentStyle(value:boolean);
       function GetIndex:integer;
       function GetAbsIndex:integer;
       function GetVisIndex:integer;
       function GetChildIndex(Child : TElTreeItem):integer;
       function IsExpanded:boolean;
       function GetFullExpand : boolean;
       procedure MakeFullyExpanded(value:boolean);
       procedure OnColTextChange(Sender:TObject);
       procedure SetImageIndex(value : integer);
       procedure SetStImageIndex(value : integer);
       procedure SetImageIndex2(value : integer);
       procedure SetStImageIndex2(value : integer);
       procedure SetForceButtons(newValue: Boolean);
       function GetChildrenCount: Integer;
       function GetCount : Integer;
       function GetItems(Index : integer) : TElTreeItem;
       procedure SetUseStyles(newValue: Boolean);
       procedure OnStyleDelete(Sender : TObject; Item : pointer);
       function GetStyles(index : integer): TElCellStyle;
       procedure SetStyles(index : integer; newValue: TElCellStyle);
       function GetStylesCount: Integer;
       procedure SetCheckBoxState(newValue : TCheckBoxState);
       procedure SetChecked(newValue : Boolean);
       function  GetChecked : boolean;
       procedure SetShowCheckBox(newValue : Boolean);
       procedure SetCheckBoxType(newValue : TElCheckBoxType);
       procedure SetCheckBoxEnabled(newValue : Boolean);
       procedure SetSuppressButtons(newValue: Boolean);
       procedure SetEnabled(newValue : Boolean);
       function GetHidden: Boolean;
       procedure SetHidden(newValue: Boolean);
       function GetFullyVisible : Boolean;
       procedure SetFullyVisible(newValue: Boolean);
      protected
       FParent      : TElTreeItem;
       FRoot        : TElTreeItem;
       procedure UpdateItem;
       procedure SetText(Value:string); virtual;
       function GetState(index:integer):boolean;
       procedure SetState(index:integer; value:boolean);
       procedure RemoveChild(Child : TElTreeItem);
       procedure DeleteChild(Child : TElTreeItem);
       function  AddChild(Child : TElTreeItem):integer;
       function InsertChild(index:integer;Child : TElTreeItem):integer;
       function IsEqual(Item:TElTreeItem):boolean;
       procedure ReadData(Stream:TStream); virtual;
       procedure WriteData(Stream:TStream); virtual;
       procedure ExchangeItems(I,J:integer);
       procedure QuickSort(recursive:boolean;L, R: Integer);
      public
       constructor Create(AOwner:TCustomElTree);
       destructor Destroy; override;
       function IsUnder(Item:TElTreeItem):boolean;
       function GetFullName(separator:string):string;
       procedure Expand(recursive:boolean);
       procedure Collapse(recursive:boolean);
       procedure Sort(recursive:boolean);
       procedure MoveTo(NewParent : TElTreeItem);
       // Moves the item to the new parent, adding it to the NewParent's
       // children list.
       procedure MoveToIns(NewParent : TElTreeItem; Index:integer);
       // Moves the item to the new parent, inserting it to the NewParent's
       // children list at Index position

       procedure Clear;
       function GetFirstVisibleChild:TElTreeItem;
       function GetFirstChild:TElTreeItem;
       function GetLastChild:TElTreeItem;
       function GetNextChild(Child : TElTreeItem):TElTreeItem;
       function GetPrevChild(Child : TElTreeItem):TElTreeItem;
       function GetFirstSibling:TElTreeItem;
       function GetLastSibling:TElTreeItem;
       function GetNextSibling:TElTreeItem;
       function GetPrevSibling:TElTreeItem;
       function GetChildByIndex(index:integer):TElTreeItem;
       // Get child item, which is on "index" position in the children list
       procedure EditText;
       // If inline editing is allowed, the EditBox appears and the user
       // can change the Text property of the item

       procedure Assign(Source:TPersistent); override;

       procedure Delete; virtual;
       property TreeView : TCustomElTree read FOwner;

       function IsVisible : Boolean;
       function GetNextVisible : TElTreeItem;
       function GetPrevVisible : TElTreeItem;
       function GetPrev : TElTreeItem; virtual;
       function GetNext : TElTreeItem; virtual;
       procedure MoveToItem(Item : TElTreeItem; Mode : TNodeAttachMode); virtual;
       function AddStyle : TElCellStyle; virtual;
       procedure RemoveStyle(Style : TElCellStyle); virtual;
       procedure RedrawItem(DoCheck:boolean);
       procedure RedrawItemPart(DoCheck:boolean; Left, Right : integer);

       property Focused    : boolean index 1 read GetState write SetState;
       // The item currently has focus
       property Selected   : boolean index 2 read GetState write SetState;
       // The item is marked as selected
       property Cut        : boolean index 3 read GetState write SetState;
       // The item is marked as cut
       property Underlined : boolean index 4 read GetState write SetState;
       // This property sets the item's main text font style
       // This doesn't affect columns text style
       property Bold       : boolean index 5 read GetState write SetState;
       // This property sets the item's main text font style
       // This doesn't affect columns text style
       property Italic     : boolean index 6 read GetState write SetState;
       // This property sets the item's main text font style
       // This doesn't affect columns text style
       property StrikeOut  : boolean index 8 read GetState write SetState;
       // This property sets the item's main text font style
       // This doesn't affect columns text style
       property ParentStyle : boolean read FParentStyle write SetParentStyle;
       // if ParentStyle=true, then item's font styles are ignored, and tree's
       // are used
       property Text : string read FText write SetText;
       // Main tree text
       property ColumnText : TStringList read FColText;
       // Text for additional columns
       property Data : pointer read FData write FData;
       // User defined data. The ElTree doesn't free this memory when
       // the item is deleted
       property Level: integer read GetLevel;
       // Tells, on which level the item is
       // Root items have Level 0 (zero)
       property Owner: TCustomElTree read FOwner;
       // Points to the item's owner ElTree component
       property Parent: TElTreeItem read GetParent write MoveTo;
       // Points to the item's parent item

       property HasVisibleChildren : boolean read GetHasVisibleChildren;
       // Shows, if the item has child items, that are not hidden

       property HasChildren : boolean read GetHasChildren;
       // Shows, if the item has child items
       property Index : integer read GetIndex;
       // This is the index of the item in the parent's children list

       property AbsoluteIndex : integer read GetAbsIndex;
       // This is the index of the item among all items in the tree

       property VisIndex : integer read GetVisIndex;
       // This is the index of the item among all expanded items in the tree

       property Expanded : boolean read IsExpanded write SetExpanded;
       // The item is expanded

       property FullyExpanded : boolean read GetFullExpand write MakeFullyExpanded;
       // The item and all its parent are expanded

       property Color : TColor index 1 read FColor write SetColor;
       // Color of item's text

       property BkColor : TColor index 2 read FBkColor write SetColor;
       // Color of item's text background

       property ParentColors : boolean read FParentColors write SetParentColors;
       // if ParentColors=True, then Color and BkColor properties are ignored

       property ImageIndex : integer read FImageIndex write SetImageIndex;
       property StateImageIndex : integer read FStImageIndex write SetStImageIndex;

       property ImageIndex2 : integer read FImageIndex2 write SetImageIndex2;
       property StateImageIndex2 : integer read FStImageIndex2 write SetStImageIndex2;

       property ForceButtons: Boolean read FForceButtons write SetForceButtons default False;
       property SuppressButtons: Boolean read FSuppressButtons write SetSuppressButtons default False;

       property ChildrenCount : integer read GetChildrenCount;
       property Count : Integer read GetCount;  { Public }
       property Children[Index : integer] : TElTreeItem read GetItems;  { Public }
       // the same but for TTreeView compatibility
// Added 10/27/98 by EM
       property Item[Index : integer] : TElTreeItem read GetItems;  { Public }
// Added 12/26/98 by EM
       property UseStyles: Boolean read FUseStyles write SetUseStyles;
       property MainStyle: TElCellStyle read FMainStyle;
       property Styles[index : integer]: TElCellStyle read GetStyles write SetStyles;
       property StylesCount: Integer read GetStylesCount;
       property CheckBoxState : TCheckBoxState read FCheckBoxState write SetCheckBoxState;
       property Checked : Boolean read GetChecked write SetChecked default False;
       property ShowCheckBox : Boolean read FShowCheckBox write SetShowCheckBox default True;
       property CheckBoxType : TElCheckBoxType read FCheckBoxType write SetCheckBoxType;
       property CheckBoxEnabled : Boolean read FCheckBoxEnabled write SetCheckBoxEnabled;
       property Enabled : Boolean read FEnabled write SetEnabled default True;
       property Hidden  : Boolean read GetHidden write SetHidden;
       property FullyVisible: Boolean read GetFullyVisible write SetFullyVisible;
     end;

// *****************************************************************************

  TElLookupCompareProc = function (Item : TElTreeItem; SearchDetails :Pointer) : boolean;
  TIterateProc = procedure (Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                            IterateData:pointer; Tree:TCustomElTree);

  TElTreeItems = class (TPersistent)
   private
    FOwner    : TCustomElTree;
    FCount    : integer;

    function  GetItem(index:integer):TElTreeItem;
    function  GetVisItem(index:integer):TElTreeItem;
    function  GetVisCount:integer;
   protected
    FRoot     : TElTreeItem;
    procedure ReadData(Stream:TStream); virtual;
    procedure WriteData(Stream:TStream); virtual;
    procedure DefineProperties(Filer: TFiler); override;
    function CreateItem(FOwner:TCustomElTree):TElTreeItem; virtual;
   public
    procedure Assign(Source:TPersistent); override;
    constructor Create(AOwner:TCustomElTree);
    destructor Destroy; override;
    procedure LoadFromStream(Stream:TStream);
    procedure SaveToStream(Stream:TStream);
    procedure SaveToFile(FileName:String);
    procedure LoadFromFile(FileName:String);
    procedure DeleteItem(Child:TElTreeItem);
    procedure RemoveItem(Child:TElTreeItem);
    function  GetAbsIndex(Child:TElTreeItem):integer;
    function  GetVisIndex(Child:TElTreeItem):integer;
    function  AddItem(Parent:TElTreeItem):TElTreeItem;
    procedure SetItem(Index: Integer; Value: TElTreeItem);
    function  InsertItem(Index:integer; Parent:TElTreeItem):TElTreeItem;
    procedure AddExistingItem(Item, Parent : TElTreeItem);
    procedure InsertExistingItem(Item, Parent : TElTreeItem; Index : integer);

    function Add(Item : TElTreeItem; Text : string) : TElTreeItem; virtual;
    function AddChild(Item : TElTreeItem; Text : string) : TElTreeItem; virtual;
    function AddChildFirst(Item : TElTreeItem; Text : string) : TElTreeItem; virtual;
    function AddChildObject(Item : TElTreeItem; Text : string; Ptr : pointer) : TElTreeItem; virtual;
    function AddChildObjectFirst(Item : TElTreeItem; Text : string; Ptr : pointer) : TElTreeItem; virtual;
    function AddFirst(Item : TElTreeItem; Text : string) : TElTreeItem; virtual;
    function AddObject(Item : TElTreeItem; Text : string; Ptr : pointer) : TElTreeItem; virtual;
    function AddObjectFirst(Item : TElTreeItem; Text : string; Ptr : pointer) : TElTreeItem; virtual;
    function Insert(Item : TElTreeItem; Text : string) : TElTreeItem; virtual;
    function InsertObject(Item : TElTreeItem; Text : string; Ptr : pointer) : TElTreeItem; virtual;
    function InsertAfter(Item : TElTreeItem; Text : String) : TElTreeItem; virtual;
    function InsertAfterObject(Item : TElTreeItem; Text : String; Ptr : Pointer) : TElTreeItem; virtual;

    procedure Delete(Item : TElTreeItem); virtual;
    function GetFirstNode : TElTreeItem;

    procedure Clear;
    procedure IterateFrom(VisibleOnly:boolean; IterateProc:TIterateProc; IterateData:pointer; StartFrom:TElTreeItem);
    procedure IterateBackFrom(VisibleOnly:boolean; IterateProc:TIterateProc; IterateData:pointer; StartFrom:TElTreeItem);
    procedure Iterate(VisibleOnly:boolean; IterateProc:TIterateProc; IterateData:pointer);
    procedure IterateBack(VisibleOnly:boolean; IterateProc:TIterateProc; IterateData:pointer);
    procedure BeginUpdate; virtual;
    procedure EndUpdate; virtual;

    function LookForItem(StartItem : TElTreeItem;
                         TextToFind : string;
                         DataToFind : pointer;
                         ColumnNum : integer;
                         LookForData,
                         CheckStartItem,
                         SubItemsOnly,
                         VisibleOnly,
                         NoCase : boolean) : TElTreeItem;

    function LookForItemEx(StartItem : TElTreeItem; ColumnNum : integer;
                           CheckStartItem, SubItemsOnly, VisibleOnly : boolean;
                           SearchDetails : pointer;
                           CompareProc : TElLookupCompareProc) : TElTreeItem;

    property Owner : TCustomElTree read FOwner;
    property Item[Index:integer] : TElTreeItem read GetItem; default;
    property ItemAsVis[Index:integer] : TElTreeItem read GetVisItem;
    property Count : integer read FCount;
    property VisCount : integer read GetVisCount;
  end;

// *****************************************************************************

{$IFNDEF VER90}
{$IFNDEF LITE}
  TOleDragStartEvent = procedure(Sender : TObject; var dataObj: IDataObject;
                                 var dropSource: IDropSource; var dwOKEffects: TDragTypes) of object;
  TOleDragFinishEvent= procedure(Sender : TObject; dwEffect : TDragType; Result : HResult) of object;
{$ENDIF}
{$ENDIF}

  TMeasureItemPartEvent = procedure(Sender : TObject; Item : TElTreeItem; PartIndex : integer; var Size : TPoint) of object;
  THeaderSectionEvent= procedure (Sender : TObject; Section : TElHeaderSection) of object;
  TColumnNotifyEvent = procedure(Sender: TObject; SectionIndex:integer) of object;
  TOnItemDrawEvent   = procedure(Sender: TObject; Item:TElTreeItem; Surface:TCanvas;
                                 R:TRect; SectionIndex:integer) of object;
  TOnTextNeededEvent = procedure(Sender: TObject; Item:TElTreeItem; var Text:String) of object;
  TOnShowHintEvent   = procedure(Sender: TObject; Item:TElTreeItem; var Text:String;
                                 HintWindow: THintWindow;
                                 MousePos : TPoint;
                                 var DoShowHint : boolean) of object;
  TApplyVisFilterEvent = procedure(Sender: TObject; Item : TElTreeItem; var Hidden : boolean) of object;

  TOnItemExpandEvent = procedure(Sender: TObject; Item:TElTreeItem) of object;
  TItemSelChangeEvent = procedure(Sender : TObject; Item: TElTreeItem) of object;
  TOnItemChangeEvent = procedure(Sender:TObject; Item:TElTreeItem;
                                 ItemChangeMode:TItemChangeMode) of object;

  TOnCompareItems    = procedure(Sender:TObject; Item1, Item2:TElTreeItem;
                                 var res:integer) of object;

  TOnItemExpanding   = procedure(Sender:TObject; Item:TElTreeItem;
                                 var CanProcess:boolean) of object;

  TOnPicDrawEvent    = procedure(Sender:TObject; Item:TElTreeItem;
                                 var ImageIndex:integer) of object;

  THotTrackEvent     = procedure(Sender: TObject; OldItem, NewItem : TElTreeItem) of object;

  TOnValidateEvent   = procedure(Sender: TObject; Item:TElTreeItem;
                                 Section : TElHeaderSection;
                                 var Text:string; var Accept:boolean) of object;

  TTryEditEvent      = procedure(Sender: TObject; Item:TElTreeItem;
                                 Section : TElHeaderSection; var CellType : TFieldTypes; var CanEdit : boolean) of object;

  TEditRequestEvent  = procedure(Sender : TObject; Item:TElTreeItem;
                                 Section : TElHeaderSection) of object;

  TComboEditShowEvent= procedure(Sender : TObject; Item : TElTreeItem;
                                 Section : TElHeaderSection;
                                 Combobox : TCombobox) of object;

  TValidateComboEvent = procedure(Sender : TObject; Item : TElTreeItem;
                        Section : TElHeaderSection;
                        Combo: TComboBox;
                        var Accept : boolean) of object;

  TElScrollEvent      = procedure(Sender : TObject;
                        ScrollBarKind : TScrollBarKind;
                        ScrollCode : integer) of object;

  TElColumnMoveEvent = procedure (Sender:TCustomElTree;
                        Section : TElHeaderSection;
                        OldPos, NewPos : integer) of object;

  TItemSaveEvent     = procedure(Sender: TObject; Stream : TStream;
                                 Item : TElTreeItem) of object;

  TCellStyleSaveEvent = procedure(Sender: TObject; Stream : TStream; Style : TElCellStyle) of object;

  TCustomElTree = class(TCustomControl)
  private
    FFilteredVisibility: Boolean;
    FOnApplyVisFilter: TApplyVisFilterEvent;
    FRightAlignedText: Boolean;
    FFlat : Boolean;
    FRightAlignedTree: Boolean;
    FPathSeparator : Char;
    FLinesStyle : TPenStyle;
    FLinesColor : TColor;
    FDeselectChildrenOnCollapse : Boolean;
    FDrawFocusRect : Boolean;
    FBarStyle : Boolean;
    FAlwaysKeepSelection: Boolean;
    FFullRowSelect: Boolean;
    FDragType   : TElDragType;
{$IFNDEF VER90}
{$IFNDEF LITE}
    FDropTarget : TElDropTarget;
{$ENDIF}
{$ENDIF}
    FAutoLookup: Boolean;
    FSelectColumn : Integer;
    FAutoExpand : Boolean;
    FPlusPicture : TBitmap;
    FMinusPicture : TBitmap;
    FCustomPlusMinus : Boolean;
    FShowCheckboxes : Boolean;
{$IFNDEF LITE}
    FStorage : TElIniFile;
    FStoragePath : String;
{$ENDIF}
    FDragImageMode : TDragImgMode;
    FHideHorzScrollBar : Boolean;
    FHideVertScrollBar : Boolean;
    FExpandOnDblClick : Boolean;
    FHideHintOnMove: Boolean;
    FSortSection: Integer;
    FSortMode    : TSortModes;
    FSortType    : TSortTypes;
    FDragAllowed : Boolean;
    PlusBmp,
    MinusBmp     : TBitmap;
    FBkColor     : TColor;
    FTextColor   : TColor;
    FShowButtons : boolean;
    FShowLines   : boolean;
    FShowImages  : boolean;
    FShowRoot    : boolean;
    FShowHintMode: THintModes;
    BMP          : TBitmap;
    FBorderStyle : TBorderStyle;

    FCanEdit     : boolean;
    FInpEdit     : TWinControl;
    FFakePopup   : TPopupMenu;
    FEditing     : boolean;
    FEditingItem : TElTreeItem;
    FEditingSect : integer;
    FPressCoord  : TPoint;
    FPressed     : boolean;
    FClickCoord  : TPoint;
    FClicked     : boolean;
    FClickItem   : TElTreeItem;

    FHLines,
    FVLines      : boolean;

    FScrollTracking,
    FTracking    : boolean;
    FTrackItem   : TElTreeItem;
    FHeaderHotTrack : boolean;
    FHeaderLine  : integer;
    FHeaderLineVis: boolean;
    FLineDC      : HDC;
    FODFollowCol : boolean;
    FODMask      : string;

    FVisible     : TElList;

    FFocused,
    FSelected,
    FDropTrg     : TElTreeItem;
    FDropAcc     : boolean;

{$IFNDEF LITE}
    FImages      : TElImageList;
    FImages2     : TElImageList;
{$ELSE}
    FImages      : TImageList;
    FImages2     : TImageList;
{$ENDIF}

{$IFDEF D_4_UP}
    FDragImages  : TDragImageList;
{$ELSE}
{$IFNDEF LITE}
    FDragImages  : TElImageList;
{$ELSE}
    FDragImages  : TImageList;
{$ENDIF}
{$ENDIF}
    FImageChangeLink: TChangeLink;
    FDragTrgDraw : TDragTargetDraw;

    FTopIndex,
    FBottomIndex : integer; // visible items
    FChStateImage: boolean;

    FMainTreeCol : integer;
    FMultiSelect : boolean;
    FRowSelect   : boolean;
    FHideSelect  : boolean;
    FLineHeight  : integer;
    FAutoLineHeight : boolean;
    FHRange      : integer;
    FHPos        : integer;
    FVScrollVisible,
    FHScrollVisible : boolean;
    FSelMode     : TSTSelModes;
    FSortDir     : TSortDirs;

    FUpdating,     // UpdateCount > 0
    FPainting,     // already in Paint
    FClearVis,     // do update visibles list
    FClearAll,     // clear the whole ClientRect
    FVisUpdated,   // visibles list updated
    FProcUpdate,   // already in SetIsUpdating
    FRangeUpdate,  // all items should be updated
    FUpdated     : boolean;
    FUpdateCount : integer;
    FDDY         : integer;{vertical offset of the image being dragged}

    FHintCoord   : TPoint;
    FHintWnd     : THintWindow;
    FHintTimer   : TTimer;
    FHintHide    : boolean;
    FIgnoreClick,
    FIgnoreClick2: boolean;
    IgnoreResize : boolean;
    FCurBkColor  : TColor;
    FCurTextColor: TColor;
    FOverColors  : boolean;
    FDelOnEdit   : boolean;
    FOldHide     : boolean;
    FSelCount    : integer;
    FItems          : TElTreeItems;

    FOnColumnResize : TColumnNotifyEvent;
    FOnColumnClick  : TColumnNotifyEvent;
    FOnColumnDraw   : TElSectionRedrawEvent;
    FOnResize       : TNotifyEvent;
    FOnValidate     : TOnValidateEvent;
    FOnItemChange   : TOnItemChangeEvent;
    FOnItemDraw     : TOnItemDrawEvent;
    FOnItemExpand   : TOnItemExpandEvent;
    FOnItemCollapse : TOnItemExpandEvent;
    FOnItemExpanding: TOnItemExpanding;
    FOnItemCollapsing: TOnItemExpanding;
    FOnItemDelete   : TOnItemExpandEvent;
    FOnItemFocused  : TNotifyEvent;
    FOnShowHint     : TOnShowHintEvent;
    FOnCompareItems : TOnCompareItems;
    FOnItemPicDraw  : TOnPicDrawEvent;
    FOnItemPicDraw2 : TOnPicDrawEvent;    
    FOnHotTrack     : THotTrackEvent;
    FOnComboEditShow : TComboEditShowEvent;
    FOnEditRequest : TEditRequestEvent;
    FOnValidateCombo : TValidateComboEvent;
    FOnScroll : TElScrollEvent;
    FOnItemSave: TItemSaveEvent;
    FOnItemLoad: TItemSaveEvent;
    FOnTryEdit: TTryEditEvent;
    FOnHeaderColumnMove : TElColumnMoveEvent;
    FOnSave : TCellStyleSaveEvent;
    FOnLoad : TCellStyleSaveEvent;
    FOnItemSelectedChange : TItemSelChangeEvent;
    FOnHeaderLookup : TElHeaderLookupEvent;
    FOnHeaderLookupDone : TElHeaderLookupDoneEvent;
    FOnHeaderResize: TNotifyEvent;
    FOnHeaderSectionExpand: THeaderSectionEvent;
    FOnHeaderSectionCollapse: THeaderSectionEvent;
    FOnMeasureItemPart : TMeasureItemPartEvent;
{$IFNDEF VER90}
{$IFNDEF LITE}
    FOnOleDragFinish : TOleDragFinishEvent;
    FOnOleDragStart : TOleDragStartEvent;
    FOnOleTargetDrag : TTargetDragEvent;
    FOnOleTargetDrop : TTargetDropEvent;
{$ENDIF}
{$ENDIF}
    procedure DrawLine;
    procedure AllocateLineDC;
    procedure ReleaseLineDC;
    procedure DoRedrawItem(Item:TElTreeItem; ItemRect, SurfRect : TRect);
    procedure DoRedrawItemTree(Item:TElTreeItem; ItemRect, SurfRect : TRect);
    procedure OnHeaderSectionResize(Header: TCustomElHeader; Section: TElHeaderSection);
    procedure OnHeaderSectionClick(Header: TCustomElHeader; Section: TElHeaderSection);
    procedure OnHeaderSectionDelete(Header: TCustomElHeader; Section: TElHeaderSection);
    procedure OnHeaderSectionTrack(Header: TCustomElHeader; Section: TElHeaderSection;
              State: TElHResizingStates; Width: Integer);
    procedure OnEditExit(Sender: TObject);
    procedure OnEditKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure OnHintTimer(Sender: TObject);
    procedure DoHeaderResize(Sender:TObject);
    procedure OnHeaderMouseDown(Sender: TObject; Button: TMouseButton;
              Shift: TShiftState; X, Y: Integer);
    procedure OnHeaderSectionLookup(Sender : TObject; Section : TElHeaderSection; var Text : string);
    procedure OnHeaderSectionLookupDone(Sender : TObject; Section : TElHeaderSection; Text : string; Accepted : boolean);
    procedure OnHeaderExpand(Sender : TCustomElHeader; Section : TElHeaderSection);
    procedure OnHeaderCollapse(Sender : TCustomElHeader; Section : TElHeaderSection);
    procedure OnFontChange(Sender:TObject);
    procedure OnSignChange(Sender:TObject);
    procedure ImageListChange(Sender: TObject);
    procedure DoHideLineHint;
    procedure DoShowLineHint(Item:TElTreeItem);

    procedure SetTextColor(value:TColor);
    procedure SetBkColor(value:TColor);

    function  GetShowHeader:boolean;
    procedure SetHeaderHotTrack(value:boolean);
    procedure SetHeaderHeight(value:integer);
{$IFNDEF LITE}
    procedure SetImages(Value: TElImageList);
    procedure SetImages2(newValue: TElImageList);
{$ELSE}
    procedure SetImages(Value: TImageList);
    procedure SetImages2(newValue: TImageList);
{$ENDIF}

    procedure SetLineStyle(Value: Boolean);
    procedure SetRootStyle(Value: Boolean);
    procedure SetImagesStyle(Value: Boolean);
    procedure DoSetTopIndex(Value: integer);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetButtonStyle(Value: Boolean);
    procedure SetUpdating(value:boolean);
    procedure SetHLines(value:boolean);
    procedure SetVLines(value:boolean);
    procedure SetRowSelect(value:boolean);
    procedure SetMultiSelect(value:boolean);
    procedure SetFocused(value:TElTreeItem);
    procedure SetHideSelect(value:boolean);
    function GetHeaderSections:TElHeaderSections;
    procedure SetHeaderSections(value:TElHeaderSections);
    procedure SetChStateImage(value:boolean);

    procedure DoSetSelected(value:TElTreeItem);
    procedure SetLineHeight(value:integer);
    procedure SetAutoLineHeight(value:boolean);
    function  GetHeaderHeight:integer;
    procedure SetMainTreeCol(value:integer);
    procedure SetItems(value:TElTreeItems);
    function  GetTotalVisCount:integer;
    procedure FillVisFwd(StartIndex:integer; ToGet: integer; var Got: integer);
    procedure UpdateElTree;
    procedure DefineHRange;
    procedure FillDragImage;
    function GetDraggableSections : Boolean ;
    procedure SetDraggableSections( newValue : Boolean ) ;
    procedure SetSortMode(newValue: TSortModes);
    procedure SetSortSection(newValue: Integer);
    procedure OnHeaderSectionVisChange(Sender:TCustomElHeader; Section : TElHeaderSection);
    function GetMoveColumnOnDrag : Boolean;
    procedure SetMoveColumnOnDrag(newValue : Boolean);
    procedure SetHideHorzScrollBar(newValue : Boolean);
    procedure SetHideVertScrollBar(newValue : Boolean);
{$IFNDEF LITE}
    function GetHeaderImages : TElImageList;
    procedure SetHeaderImages(newValue : TElImageList);
{$ELSE}
    function GetHeaderImages : TImageList;
    procedure SetHeaderImages(newValue : TImageList);
{$ENDIF}

    function GetSelCount : integer;
    function GetSelected : TElTreeItem;
    procedure SetSelected(newValue : TElTreeItem);
{$IFNDEF LITE}
    procedure SetStorage(newValue : TElIniFile);
{$ENDIF}
    procedure SetShowCheckboxes(newValue : Boolean);
    procedure SetPlusPicture(newValue : TBitmap);
    procedure SetMinusPicture(newValue : TBitmap);
    procedure SetCustomPlusMinus(newValue : Boolean);
    procedure SetSelectColumn(newValue : Integer);
    procedure SetDragType(newValue : TElDragType);
{$IFNDEF VER90}
{$IFNDEF LITE}
    procedure TargetDrag_FDropTargetHandler(Sender : TObject; State : TDragState; Source : TOleDragObject; Shift : TShiftState; X : Integer; Y : Integer; var DragType : TDragType);
    procedure TargetDrop_FDropTargetHandler(Sender : TObject; Source : TOleDragObject; Shift : TShiftState; X : Integer; Y : Integer; var DragType : TDragType);
{$ENDIF}
{$ENDIF}
    procedure HeaderResizeTransfer(Sender: TObject);
    procedure HeaderResizeHandler(Sender: TObject);
    function GetStickyHeaderSections : Boolean;
    procedure SetStickyHeaderSections(newValue : Boolean);
    procedure SetBarStyle(newValue : Boolean);
    procedure SetDrawFocusRect(newValue : Boolean);
    procedure SetLinesColor(newValue : TColor);
    procedure SetLinesStyle(newValue : TPenStyle);
    procedure SetRightAlignedTree(newValue: Boolean);
    procedure SetFlat(newValue : Boolean);
    procedure WMNCPaint(var Msg : TMessage); message WM_NCPAINT;
    procedure SetRightAlignedText(newValue: Boolean);
    procedure SetFilteredVisibility(newValue: Boolean);
  protected
    { Protected declarations }
    Destroying : boolean; // used in ElDataTree to prevent repainting
    TotalHiddenCount,
    TotalVisCount : integer;
    FHeader : TElHeader;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DoEditItem(Item:TElTreeItem; SectionNum : integer); virtual;
    function CompareItems(Item1, Item2 : TElTreeItem):integer; virtual;
    function  GetVisCount:integer;
    procedure SetCanEdit(value:boolean); virtual;
    procedure SetShowHeader(value:boolean); virtual;
    procedure WndProc(var Message: TMessage); override;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    procedure CMMouseLeave(var Message: TMessage); message CM_MouseLeave;
    procedure CMDrag(var Message: TCMDrag); message CM_DRAG;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMHScroll(var Message: TWMHScroll); message WM_HSCROLL;
    procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LButtonDblClk;
    procedure WMRButtonDown(var Message: TWMRButtonDown); message WM_RBUTTONDOWN;
    procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
{$IFDEF D_4_UP}
    procedure Resize; override;
{$ELSE}
    procedure Resize; dynamic;
{$ENDIF}
    procedure DoStartDrag(var DragObject: TDragObject); override;
    procedure DoEndDrag(Target: TObject; X, Y: Integer); override;
    procedure DoDragOver(Source: TDragObject; X, Y: Integer; CanDrop: Boolean); virtual;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure DoEnter; override;
    procedure DoExit; override;
    function DoGetPicture (Item:TElTreeItem):integer; virtual;
    function DoGetPicture2(Item:TElTreeItem):integer; virtual;
    function  DefineLineHeight:integer; virtual;
    procedure RedrawTree;
    procedure UpdateScrollBars;
    procedure CreateParams(var Params: TCreateParams); override;
{$IFDEF D_4_UP}
    function GetDragImages: TDragImageList; override;
{$ELSE}
    function GetDragImages: TCustomImageList; override;
{$ENDIF}
    function CreateItems: TElTreeItems; virtual;
    function CreateHeader : TElHeader; virtual;
    procedure DoOnColumnResize(SectionIndex : integer); virtual;
    procedure DoColumnClick(SectionIndex : integer); virtual;
    procedure DoItemFocused; virtual;
    procedure DoItemDraw(Item:TElTreeItem; Surface:TCanvas; R:TRect; SectionIndex:integer); virtual;
    procedure DoValidate(Item:TElTreeItem; Section : TElHeaderSection; var Text:string; var Accept:boolean); virtual;
    procedure DoItemChange(Item:TElTreeItem; ItemChangeMode:TItemChangeMode); virtual;
    procedure DoItemExpanding(Item : TElTreeItem; var CanProcess: boolean); virtual;
    procedure DoItemCollapsing(Item : TElTreeItem; var CanProcess: boolean); virtual;
    procedure DoItemExpand(Item : TElTreeItem); virtual;
    procedure DoItemCollapse(Item : TElTreeItem); virtual;
    procedure DoItemDelete(Item : TElTreeItem); virtual;
    procedure DoCompareItems(Item1, Item2:TElTreeItem; var res:integer); virtual;
    procedure DoHeaderDraw(Header: TCustomElHeader; Section: TElHeaderSection;
              Rect: TRect; Pressed: Boolean); virtual;
    procedure OnHeaderSectionChange(Sender : TCustomElHeader; Section : TElHeaderSection; Change : TSectionChangeMode); virtual;
    procedure OnHeaderSectionMove(Sender:TCustomElHeader; Section : TElHeaderSection; OldPos, NewPos : integer ) ; virtual;
    procedure TriggerHotTrackEvent(OldItem, NewItem : TElTreeItem); virtual;
    procedure TriggerEditRequestEvent(Item:TElTreeItem; Section : TElHeaderSection); virtual;
    procedure TriggerComboEditShowEvent(Item : TElTreeItem; Section : TElHeaderSection; Combobox : TCombobox); virtual;
    procedure TriggerScrollEvent(ScrollBarKind : TScrollBarKind; ScrollCode : integer); virtual;
    procedure TriggerHeaderColumnMoveEvent(Section:TElHeaderSection; OldPos, NewPos : integer); virtual;
    procedure TriggerItemSaveEvent(Stream : TStream; Item : TElTreeItem); virtual;
    procedure TriggerItemLoadEvent(Stream : TStream; Item : TElTreeItem); virtual;
    procedure TriggerItemSelectedChangeEvent(Item: TElTreeItem); virtual;
    procedure DoValidateCombo(Item : TElTreeItem; Section : TElHeaderSection; Combo: TComboBox; var Accept : boolean); virtual;
    procedure DoShowHint(Item:TElTreeItem; var Text:String; HintWindow: THintWindow; MousePos : TPoint; var DoShowHint : boolean); virtual;
    procedure Paint; override;
    procedure OnHeaderSectionCreate(Header : TCustomElHeader; Section : TElHeaderSection); virtual;
    function DragScroll(Source : TDragObject; X, Y : integer) : boolean; virtual;
    procedure TriggerTryEditEvent(Item:TElTreeItem; Section : TElHeaderSection;
              var CellType : TFieldTypes; var CanEdit : boolean); virtual;
    procedure TriggerHeaderLookupEvent(Section : TElHeaderSection; var Text : string); virtual;
    procedure TriggerHeaderLookupDoneEvent(Section : TElHeaderSection; Text : string; Accepted : boolean); virtual;
    procedure TriggerHeaderSectionExpandEvent(Section  : TElHeaderSection); virtual;
    procedure TriggerHeaderSectionCollapseEvent(Section  : TElHeaderSection); virtual;
    procedure TriggerMeasureItemPartEvent(Item : TElTreeItem; PartIndex : integer; var Size : TPoint); virtual;
    procedure TriggerApplyVisFilterEvent(Item : TElTreeItem; var Hidden : boolean); virtual;

{$IFNDEF VER90}
{$IFNDEF LITE}
    procedure TriggerOleTargetDragEvent(State : TDragState; Source : TOleDragObject;
              Shift : TShiftState; X, Y : integer; var DragType : TDragType); virtual;
    procedure TriggerOleTargetDropEvent(Source : TOleDragObject; Shift : TShiftState;
              X, Y : integer; var DragType : TDragType); virtual;
    procedure TriggerOleDragStartEvent(var dataObj: IDataObject; var dropSource: IDropSource;
                                       var dwOKEffects: TDragTypes); virtual;
    procedure TriggerOleDragFinishEvent(dwEffect : TDragType; Result : HResult); virtual;
{$ENDIF}
{$ENDIF}

    function GetTopItem : TElTreeItem; virtual;
    procedure SetTopItem(Item : TElTreeItem); virtual;
    procedure Loaded; override;

    function SectionTypeToSortType(SectionType : TFieldTypes): TSortTypes;

    property TextColor : TColor read FTextColor write SetTextColor;
    property BkColor : TColor read FBkColor write SetBkColor;
    property ShowButtons: Boolean read FShowButtons write SetButtonStyle;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle;
    property ShowLines: Boolean read FShowLines write SetLineStyle;
    property ShowImages: Boolean read FShowImages write SetImagesStyle;
    property ShowRoot: boolean read FShowRoot write SetRootStyle;
    property LineHintMode: THintModes read FShowHintMode write FShowHintMode;
    property HideSelection: Boolean read FHideSelect write SetHideSelect default True;
    property HideHintOnTimer : boolean read FHintHide write FHintHide;
{$IFNDEF LITE}
    property Images : TElImageList read FImages write SetImages;
    property Images2: TElImageList read FImages2 write SetImages2;
{$ELSE}
    property Images : TImageList read FImages write SetImages;
    property Images2: TImageList read FImages2 write SetImages2;
{$ENDIF}
    property ChangeStateImage: boolean read FChStateImage write SetChStateImage;
    property ShowColumns: Boolean read GetShowHeader write SetShowHeader;
    property DragTrgDrawMode : TDragTargetDraw read FDragTrgDraw write FDragTrgDraw;
    property DraggableSections : Boolean read GetDraggableSections write SetDraggableSections ;  { Published }
    property SelectionMode : TSTSelModes read FSelMode write FSelMode;
    property DoInplaceEdit : boolean read FCanEdit write SetCanEdit;
    property VerticalLines : boolean read FVLines write SetVLines;
    property HorizontalLines : boolean read FHLines write SetHLines;
    property ScrollTracking : boolean read FScrollTracking write FScrollTracking;
    property Tracking : boolean read FTracking write FTracking;
    property RowSelect : boolean read FRowSelect write SetRowSelect;
    property MultiSelect : boolean read FMultiSelect write SetMultiSelect;
    property LineHeight : integer read FLineHeight write SetLineHeight;
    property AutoLineHeight : boolean read FAutoLineHeight write SetAutoLineHeight;
    property HeaderHotTrack : boolean read FHeaderHotTrack write SetHeaderHotTrack;
    property HeaderSections : TElHeaderSections read GetHeaderSections write SetHeaderSections;
    property HeaderHeight : integer read GetHeaderHeight write SetHeaderHeight;
    property MainTreeColumn : integer read FMainTreeCol write SetMainTreeCol;
    property OwnerDrawByColumn : boolean read FODFollowCol write FODFollowCol;
    property OwnerDrawMask : string read FODMask write FODMask;
    property DragAllowed : Boolean read FDragAllowed write FDragAllowed ;
    property SortDir : TSortDirs read FSortDir write FSortDir;
    property SortMode: TSortModes read FSortMode write SetSortMode;  { Published }
    property SortSection: Integer read FSortSection write SetSortSection;  { Published }
    property SortType : TSortTypes read FSortType write FSortType;
    property HideHintOnMove: Boolean read FHideHintOnMove write FHideHintOnMove default true;  { Protected }
    property ExpandOnDblClick : Boolean read FExpandOnDblClick write FExpandOnDblClick;  { Protected }
    property MoveColumnOnDrag : Boolean read GetMoveColumnOnDrag write SetMoveColumnOnDrag;  { Published }
    property HideHorzScrollBar : Boolean read FHideHorzScrollBar write SetHideHorzScrollBar;  { Published }
    property HideVertScrollBar : Boolean read FHideVertScrollBar write SetHideVertScrollBar;  { Published }
{$IFNDEF LITE}
    property HeaderImages : TElImageList read GetHeaderImages write SetHeaderImages;  { Protected }
{$ELSE}
    property HeaderImages : TImageList read GetHeaderImages write SetHeaderImages;  { Protected }
{$ENDIF}
    property DragImageMode : TDragImgMode read FDragImageMode write FDragImageMode;  { Protected }
{$IFNDEF LITE}
    property Storage : TElIniFile read FStorage write SetStorage;
    property StoragePath : String read FStoragePath write FStoragePath;
{$ENDIF}
    property ShowCheckboxes : Boolean read FShowCheckboxes write SetShowCheckboxes;
    property PlusPicture : TBitmap read FPlusPicture write SetPlusPicture;
    property MinusPicture : TBitmap read FMinusPicture write SetMinusPicture;
    property CustomPlusMinus : Boolean read FCustomPlusMinus write SetCustomPlusMinus default False;
    property SelectColumn : Integer read FSelectColumn write SetSelectColumn default -1;
    property AutoExpand : Boolean read FAutoExpand write FAutoExpand default False;
    property AutoLookup: Boolean read FAutoLookup write FAutoLookup;
    property DragType : TElDragType read FDragType write SetDragType;

    property OnHeaderColumnResize: TColumnNotifyEvent read FOnColumnResize write FOnColumnResize;
    property OnHeaderColumnClick: TColumnNotifyEvent read FOnColumnClick write FOnColumnClick;
    property OnHeaderColumnMove : TElColumnMoveEvent read FOnHeaderColumnMove write FOnHeaderColumnMove;

    property OnResize: TNotifyEvent read FOnResize write FOnResize;
    property OnValidateInplaceEdit : TOnValidateEvent read FOnValidate write FOnValidate;
    property OnItemChange : TOnItemChangeEvent read FOnItemChange write FOnItemChange;
    property OnHeaderColumnDraw : TElSectionRedrawEvent read FOnColumnDraw write FOnColumnDraw;
    property OnItemDraw   : TOnItemDrawEvent read FOnItemDraw write FOnItemDraw;
    property OnItemExpand   : TOnItemExpandEvent read FOnItemExpand write FOnItemExpand;
    property OnItemCollapse : TOnItemExpandEvent read FOnItemCollapse write FOnItemCollapse;
    property OnItemExpanding : TOnItemExpanding read FOnItemExpanding write FOnItemExpanding;
    property OnItemCollapsing: TOnItemExpanding read FOnItemCollapsing write FOnItemCollapsing;
    property OnScroll : TElScrollEvent read FOnScroll write FOnScroll;
    property OnItemDeletion : TOnItemExpandEvent read FOnItemDelete write FOnItemDelete;
    property OnItemFocused  : TNotifyEvent read FOnItemFocused write FOnItemFocused;
    property OnShowLineHint : TOnShowHintEvent read FOnShowHint write FOnShowHint;
    property OnCompareItems : TOnCompareItems read FOnCompareItems write FOnCompareItems;
    property OnItemPicDraw  : TOnPicDrawEvent read FOnItemPicDraw write FOnItemPicDraw;
    property OnItemPicDraw2 : TOnPicDrawEvent read FOnItemPicDraw2 write FOnItemPicDraw2;    
    property OnHotTrack     : THotTrackEvent read FOnHotTrack write FOnHotTrack;
    property OnEditRequest : TEditRequestEvent read FOnEditRequest write FOnEditRequest;
    property OnComboEditShow : TComboEditShowEvent read FOnComboEditShow write FOnComboEditShow;
    property OnValidateCombo : TValidateComboEvent read FOnValidateCombo write FOnValidateCombo;
    property OnTryEdit: TTryEditEvent read FOnTryEdit write FOnTryEdit;
    property OnItemSave: TItemSaveEvent read FOnItemSave write FOnItemSave;
    property OnItemLoad: TItemSaveEvent read FOnItemLoad write FOnItemLoad;
    property OnItemSelectedChange : TItemSelChangeEvent read FOnItemSelectedChange write FOnItemSelectedChange;
    property OnCellStyleSave: TCellStyleSaveEvent read FOnSave write FOnSave;
    property OnCellStyleLoad: TCellStyleSaveEvent read FOnLoad write FOnLoad;
    property OnHeaderResize: TNotifyEvent read FOnHeaderResize write FOnHeaderResize;
    property OnHeaderLookup : TElHeaderLookupEvent read FOnHeaderLookup write FOnHeaderLookup;
    property OnHeaderLookupDone : TElHeaderLookupDoneEvent read FOnHeaderLookupDone write FOnHeaderLookupDone;
    property OnHeaderSectionExpand: THeaderSectionEvent read FOnHeaderSectionExpand write FOnHeaderSectionExpand;
    property OnHeaderSectionCollapse: THeaderSectionEvent read FOnHeaderSectionCollapse write FOnHeaderSectionCollapse;
    property OnMeasureItemPart : TMeasureItemPartEvent read FOnMeasureItemPart write FOnMeasureItemPart;

{$IFNDEF VER90}
{$IFNDEF LITE}
    property OnOleTargetDrag : TTargetDragEvent read FOnOleTargetDrag write FOnOleTargetDrag;
    property OnOleTargetDrop : TTargetDropEvent read FOnOleTargetDrop write FOnOleTargetDrop;
    property OnOleDragStart : TOleDragStartEvent read FOnOleDragStart write FOnOleDragStart;
    property OnOleDragFinish : TOleDragFinishEvent read FOnOleDragFinish write FOnOleDragFinish;
{$ENDIF}
{$ENDIF}
    property FullRowSelect: Boolean read FFullRowSelect write FFullRowSelect;
    property AlwaysKeepSelection : Boolean read FAlwaysKeepSelection write FAlwaysKeepSelection;
    property StickyHeaderSections : Boolean read GetStickyHeaderSections write SetStickyHeaderSections;
    property BarStyle : Boolean read FBarStyle write SetBarStyle;
    property DrawFocusRect : Boolean read FDrawFocusRect write SetDrawFocusRect;
    property DeselectChildrenOnCollapse : Boolean read FDeselectChildrenOnCollapse write FDeselectChildrenOnCollapse;
    property LinesColor : TColor read FLinesColor write SetLinesColor;
    property LinesStyle : TPenStyle read FLinesStyle write SetLinesStyle;
    property PathSeparator : Char read FPathSeparator write FPathSeparator;  { Protected }
    property RightAlignedTree: Boolean read FRightAlignedTree write SetRightAlignedTree;
    property Flat : Boolean read FFlat write SetFlat;  { Protected }
    property RightAlignedText: Boolean read FRightAlignedText write 
      SetRightAlignedText;
    property FilteredVisibility: Boolean read FFilteredVisibility write
      SetFilteredVisibility;
    property OnApplyVisFilter: TApplyVisFilterEvent read FOnApplyVisFilter
      write FOnApplyVisFilter;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure Update; override;
    procedure Assign(Source:TPersistent); override;
    procedure FullCollapse; virtual;
    procedure FullExpand; virtual;
    function  GetItemRect(ItemIndex:integer):TRect; virtual;
    function GetItemAtY(y:integer):TElTreeItem; virtual;
    function  GetItemAt(X, Y: Integer; var ItemPart:TSTItemPart;
                        var HitColumn : integer): TElTreeItem; virtual;
    function  GetNextSelected(Prev:TElTreeItem):TElTreeItem; virtual;
    procedure AllSelected(SelectedItems:TElList); virtual;
    procedure SelectAll; virtual;
    procedure InvertSelection; virtual;
    procedure SelectAllEx(IncludeHidden : boolean); virtual;
    procedure InvertSelectionEx(IncludeHidden : boolean); virtual;
    procedure DeselectAll; virtual;
    procedure DeselectAllEx(IncludeHidden : boolean); virtual;
    procedure SelectRange(FromItem,ToItem: TElTreeItem); virtual;
    procedure SelectRangeEx(FromItem,ToItem: TElTreeItem; IncludeHidden : boolean); virtual;
    procedure Sort(recursive:boolean); virtual;
    procedure SetHPosition(value:integer);
    procedure SetVPosition(value:integer);
{$IFNDEF LITE}
    procedure Save; virtual;
    procedure Restore; virtual;
{$ENDIF}
    procedure EnsureVisible(Item:TElTreeItem);
    procedure EditItem(Item:TElTreeItem; SectionNum : integer); virtual;
    procedure EndEdit(ByCancel : boolean);
    function IsEditing : boolean;
    procedure SaveStringsToStream(Stream: TStream); virtual;
    function GetNodeAt(X, Y : integer) : TElTreeItem;
    procedure CreateWindowHandle(const Params : TCreateParams); override;
    function IsInView(Item : TElTreeItem): Boolean; virtual;

    property TopIndex      : integer read FTopIndex write SetVPosition;
    property BottomIndex   : integer read FBottomIndex;
    property LeftPosition  : integer read FHPos write SetHPosition;
    property IsUpdating    : boolean read FUpdating write SetUpdating;
    property Items         : TElTreeItems read FItems write SetItems;
    property ItemFocused   : TElTreeItem read FFocused write SetFocused;
    property SelectedCount : integer     read GetSelCount;

    property Selected      : TElTreeItem read GetSelected write SetSelected;
    property TopItem       : TElTreeitem read GetTopItem write SetTopItem;
    property InEditState   : Boolean     read FEditing;
  end;

  TElTree = class (TCustomElTree)
    published
      property Align;
      property AlwaysKeepSelection;
      property AutoExpand;
      property AutoLookup;
{$IFDEF D_4_UP}
      property Anchors;
      property Action;
      property Constraints;
      property DockOrientation;
      property Floating;
      property BevelKind;
      property DoubleBuffered;
      property DragKind;
{$ENDIF}
      property AutoLineHeight;
      property BarStyle;
      property BorderStyle default bsSingle;
      property ChangeStateImage;
      property Ctl3D;
      property CustomPlusMinus;
      property DeselectChildrenOnCollapse;
      property DoInplaceEdit default true;
      property DragCursor;
      property DraggableSections;
      property DragAllowed;
      property DragTrgDrawMode;
      property DragType;
      property DragImageMode;
      property DrawFocusRect;
      property Enabled;
      property ExpandOnDblClick;
      //property Flat; - doesn't work yet
      property FilteredVisibility;
      property Font stored true;
      property FullRowSelect;
      property HeaderHotTrack;
      property HeaderSections;
      property HeaderImages;
      property HideHintOnTimer;
      property HideHintOnMove;
      property HideSelection;
      property HorizontalLines;
      property HideHorzScrollBar;
      property HideVertScrollBar;
      property Images;
      property Images2;
      property Items;
      property LeftPosition;
      property LineHeight;
      property LinesColor;
      property LinesStyle;
      property LineHintMode default shmLong;
      property MainTreeColumn;
      property MinusPicture;
      property MoveColumnOnDrag;
      property MultiSelect;
      property OwnerDrawByColumn default true;
      property OwnerDrawMask;
      property ParentCtl3D;
      property ParentFont;
      property ParentShowHint;
      property PathSeparator;
      property PlusPicture;
      property PopupMenu;
      property RightAlignedText;
      property RightAlignedTree;
      property RowSelect default True;
      property ScrollTracking default false;
      property SelectColumn;
      property ShowButtons default True;
      property ShowColumns;
      property ShowCheckboxes;
      property ShowHint;
      property ShowImages default True;
      property ShowLines default True;
      property ShowRoot default false;
      property SelectionMode;
      property SortDir;
      property SortMode;
      property SortSection;
      property SortType;
{$IFNDEF LITE}
      property Storage;
      property StoragePath;
{$ENDIF}
      property StickyHeaderSections;
      property TabOrder;
      property TabStop default True;
      property Tracking;
      property VerticalLines;
      property Visible;
      property TextColor;
      property BkColor;

      property OnScroll;
      property OnHeaderColumnClick;
      property OnHeaderColumnDraw;
      property OnHeaderColumnResize;
      property OnHeaderColumnMove;
      property OnHeaderLookup;
      property OnHeaderLookupDone;
      property OnHeaderResize;
      property OnHeaderSectionExpand;
      property OnHeaderSectionCollapse;
      property OnItemChange;
      property OnItemDraw;
      property OnResize;
      property OnEditRequest;
      property OnComboEditShow;
      property OnValidateCombo;
      property OnValidateInplaceEdit;
      property OnItemExpand;
      property OnItemCollapse;
      property OnItemExpanding;
      property OnItemCollapsing;
      property OnItemDeletion;
      property OnItemFocused;
      property OnShowLineHint;
      property OnClick;
      property OnCompareItems;
      property OnEnter;
      property OnExit;
      property OnDragDrop;
      property OnDragOver;
      property OnStartDrag;
      property OnEndDrag;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnDblClick;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnItemPicDraw;
      property OnItemPicDraw2;
      property OnHotTrack;
      property OnTryEdit;
      property OnItemSave;
      property OnItemLoad;
      property OnItemSelectedChange;
      property OnCellStyleSave;
      property OnCellStyleLoad;
{$IFNDEF VER90}
{$IFNDEF LITE}
      property OnOleTargetDrag;
      property OnOleTargetDrop;
      property OnOleDragStart;
      property OnOleDragFinish;
{$ENDIF}
{$ENDIF}
    end;

implementation

uses
{$IFDEF USE_RX}
  ToolEdit,
{$ENDIF}
  ElStack;

label CryptBegin, CryptEnd;

const ItemExt = 15;
      FDivLineWidth = 1;

function GetErrorMsg(ID:integer):string;
var p:pchar;
begin
  GetMem(p,260);
  LoadString(HInstance,ID,p,260);
  result:=StrPas(p);
  FreeMem(p,260);
end;

// ****************************************************************************

// TElTreeItem methods

// ****************************************************************************

constructor TElTreeItem.Create;
begin
  inherited Create;
  FOwner:=AOwner;
  FColText:=TStringList.Create;
  FColText.OnChange:=OnColTextChange;
  FChildren:=TElList.Create;
  FParentColors:=true;
  FParentStyle:=true;
  FState:=[];
  FImageIndex := -1;
  FStImageIndex:= -1;
  FImageIndex2  := -1;
  FStImageIndex2:= -1;
  FForceButtons := False;
  FMainStyle := TElCellStyle.Create(Self);
  FStyles := TElList.Create;
  FStyles.OnDelete := OnStyleDelete;
  FMainStyle.FOwnerProps := true;
  FCheckBoxState:= cbUnchecked;
  FShowCheckBox := false;
  FCheckBoxType := ectCheckBox;
  FCheckBoxEnabled := true;
  FSuppressButtons := False;
  FEnabled := True;
end;

destructor TElTreeItem.Destroy;
begin
  if (FOwner <> nil) and Selected then dec(FOwner.FSelCount);
  FColText.Free;
  FChildren.Free;
  FMainStyle.Free;
  FStyles.Free;
  inherited Destroy;
end;

procedure TElTreeItem.OnStyleDelete(Sender : TObject; Item : pointer);
begin
  TElCellStyle(Item).Free;
end;

type

     TS26ItemInfo = record
       FState       : TSTIStates;
       FImageIndex  : integer;
       FStImageIndex: integer;
       ForceButtons : LongBool;
       FColor,
       FBkColor     : TColor;
       FParentColors    : boolean;
       FParentStyle     : boolean;
       FCheckBoxEnabled : Boolean;
       FCheckBoxState   : TCheckBoxState;
       FShowCheckBox    : Boolean;
       FCheckBoxType    : TElCheckBoxType;
       FImageIndex2     : integer;
       FStImageIndex2   : integer;
       FEnabled         : boolean;
       FHidden          : boolean;
     end;

     TSItemInfo = record
       FState       : TSTIStates;
       FImageIndex  : integer;
       FStImageIndex: integer;
       ForceButtons : LongBool;
       FColor,
       FBkColor     : TColor;
       FParentColors    : boolean;
       FParentStyle     : boolean;
       FCheckBoxEnabled : Boolean;
       FCheckBoxState   : TCheckBoxState;
       FShowCheckBox    : Boolean;
       FCheckBoxType    : TElCheckBoxType;
       FImageIndex2     : integer;
       FStImageIndex2   : integer;
       FEnabled         : boolean;
     end;

     TS22ItemInfo = record
       FState       : TSTIStates;
       FImageIndex  : integer;
       FStImageIndex: integer;
       ForceButtons : LongBool;
       b,c          : integer;
       FColor,
       FBkColor     : TColor;
       FParentColors: boolean;
       FParentStyle : boolean;
     end;

     TS20ItemInfo = record
       FState       : TSTIStates;
       FImageIndex  : integer;
       FStImageIndex: integer;
       ForceButtons : LongBool;
       FColor,
       FBkColor     : TColor;
       FParentColors: boolean;
       FParentStyle : boolean;
       FCheckBoxEnabled : Boolean;
       FCheckBoxState   : TCheckBoxState;
       FShowCheckBox    : Boolean;
       FCheckBoxType    : TElCheckBoxType;
     end;

     PStyleInfo = ^TStyleInfo;
     TStyleInfo = record
       FCellBkColor: TColor;
       FTextBkColor: TColor;
       FTextColor: TColor;
       FTextFlags: DWORD;
       FPicture: TBitmap;
       FCellType: TFieldTypes;
       FStyle: TElSectionStyle;
       FOwnerProps: Boolean;
       FFontSize: integer;
       FFontStyles: TFontStyles;
       FTag       : integer;
     end;

procedure TElTreeItem.ReadData(Stream:TStream);
var i,j,k : integer;
    s     : string;
    t     : TSItemInfo;
    t22   : TS22ItemInfo;
    t20   : TS20ItemInfo;
    T26   : TS26ItemInfo;
    ver   : integer;
    SI : TStyleInfo;
    Style : TElCellStyle;
    b : boolean;
begin
  Stream.ReadBuffer(k,SizeOf(k));
  Ver := 0;
  if k < 0 then
  begin
    ver := k;
    Stream.ReadBuffer(k,SizeOf(k));
  end;
  SetLength(s,k);
  Stream.ReadBuffer(pchar(s)^,k);
  FText:=s;
  if Ver = -3 then
  begin
    Stream.ReadBuffer(T20,SizeOf(T20));
    FState    := T20.FState - [stsFocused, stsSelected];
    FColor    := T20.FColor;
    FBkColor  := T20.FBkColor;
    FImageIndex:=T20.FImageIndex;
    FStImageIndex:=T20.FStImageIndex;
    FParentColors := T20.FParentColors;
    FParentStyle := T20.FParentStyle;
    FForceButtons := T20.ForceButtons;
    FCheckBoxEnabled := T20.FCheckBoxEnabled;
    FCheckBoxState := T20.FCheckBoxState;
    FShowCheckBox := T20.FShowCheckBox;
    FCheckBoxType := T20.FCheckBoxType;
  end else
  if Ver = -4 then
  begin
    Stream.ReadBuffer(T,SizeOf(T));
    FState    := T.FState - [stsFocused, stsSelected];
    FColor    := T.FColor;
    FBkColor  := T.FBkColor;
    FImageIndex    := T.FImageIndex;
    FStImageIndex  := T.FStImageIndex;
    FParentColors  := T.FParentColors;
    FParentStyle   := T.FParentStyle;
    FForceButtons  := T.ForceButtons;
    FCheckBoxEnabled := T.FCheckBoxEnabled;
    FCheckBoxState := T.FCheckBoxState;
    FShowCheckBox  := T.FShowCheckBox;
    FCheckBoxType  := T.FCheckBoxType;
    FImageIndex2   := T.FImageIndex2;
    FStImageIndex2 := T.FStImageIndex2;
    FEnabled       := T.FEnabled;
  end else
  if Ver = -5 then
  begin
    Stream.ReadBuffer(T26, SizeOf(T26));
    FState    := T26.FState - [stsFocused, stsSelected];
    FColor    := T26.FColor;
    FBkColor  := T26.FBkColor;
    FImageIndex    := T26.FImageIndex;
    FStImageIndex  := T26.FStImageIndex;
    FParentColors  := T26.FParentColors;
    FParentStyle   := T26.FParentStyle;
    FForceButtons  := T26.ForceButtons;
    FCheckBoxEnabled := T26.FCheckBoxEnabled;
    FCheckBoxState := T26.FCheckBoxState;
    FShowCheckBox  := T26.FShowCheckBox;
    FCheckBoxType  := T26.FCheckBoxType;
    FImageIndex2   := T26.FImageIndex2;
    FStImageIndex2 := T26.FStImageIndex2;
    FEnabled       := T26.FEnabled;
    FHidden        := T26.FHidden;
  end else
  begin
    Stream.ReadBuffer(T22,SizeOf(T22));
    FState    := T22.FState - [stsFocused, stsSelected];
    FColor    := T22.FColor;
    FBkColor  := T22.FBkColor;
    FImageIndex:=T22.FImageIndex;
    FStImageIndex:=T22.FStImageIndex;
    FParentColors := T22.FParentColors;
    FParentStyle := T22.FParentStyle;
    FForceButtons := T22.ForceButtons;
  end;
  Stream.ReadBuffer(i,SizeOf(i));
  for j:=0 to i-1 do
  begin
    Stream.ReadBuffer(k,SizeOf(k));
    SetLength(s,k);
    Stream.ReadBuffer(pchar(s)^,k);
    FColText.Add(s);
  end;
  if Ver < 0 then
  begin
    Stream.ReadBuffer(FUseStyles, sizeof(FUseStyles));
    if FUseStyles then
    begin
      if Ver < -1 then Stream.ReadBuffer(b, sizeof(boolean)) else b := false;
      if b then
      begin
        if Assigned(FOwner.FOnLoad) then FOwner.FOnLoad(Self, Stream, MainStyle);
      end
      else
      begin
        Stream.ReadBuffer(SI, sizeof(TStyleInfo));
        with MainStyle do
        begin
          FCellBkColor := SI.FCellBkColor;
          FTextBkColor := SI.FTextBkColor;
          FTextColor := SI.FTextColor;
          FTextFlags := SI.FTextFlags;
          FPicture     := SI.FPicture;
          FCellType    := SI.FCellType;
          FStyle       := SI.FStyle;
          FOwnerProps := SI.FOwnerProps;
          FFontSize    := SI.FFontSize;
          FFontStyles  := SI.FFontStyles;
          FTag         := SI.FTag;
        end;
        ReadStringFromStream(Stream, s);
        MainStyle.FFontName := s;
      end;
      Stream.ReadBuffer(i,SizeOf(i));
      for j := 1 to i do
      begin
        Style := AddStyle;
        if b then
        begin
          if Assigned(FOwner.FOnLoad) then FOwner.FOnLoad(Self, Stream, Style);
        end else
        begin
          Stream.ReadBuffer(SI, sizeof(TStyleInfo));
          with Style do
          begin
            FCellBkColor := SI.FCellBkColor;
            FTextBkColor := SI.FTextBkColor;
            FTextColor := SI.FTextColor;
            FTextFlags := SI.FTextFlags;
            FPicture     := SI.FPicture;
            FCellType    := SI.FCellType;
            FStyle       := SI.FStyle;
            FOwnerProps := SI.FOwnerProps;
            FFontSize    := SI.FFontSize;
            FFontStyles  := SI.FFontStyles;
            FTag         := SI.FTag;
          end;
          ReadStringFromStream(Stream, s);
          Style.FFontName := s;
        end;
      end; // for
    end;
  end;
  FOwner.TriggerItemLoadEvent(Stream, Self);
  Stream.ReadBuffer(i,SizeOf(i));
  for j:=0 to i-1 do FList.AddItem(self).ReadData(Stream);
end;

procedure TElTreeItem.WriteData(Stream:TStream);
var i,j,k : integer;
    s     : string;
    T20   : TS26ItemInfo;
    p     : pchar;
    SI    : TStyleInfo;
    Style : TElCellStyle;
    b : boolean;

begin
  // write version
  k := -5;
  Stream.WriteBuffer(k,SizeOf(k));
  // write text data
  s:=FText;
  k:=length(s);
  Stream.WriteBuffer(k,SizeOf(k));
  GetMem(p,k+1);
  StrPCopy(p,s);
  Stream.WriteBuffer(p^,k);
  FreeMem(p,k+1);
  // write binary data

  T20.FState:=FState;
  T20.FColor := FColor;
  T20.FBkColor := FBkColor;
  T20.FParentColors := FParentColors;
  T20.FParentStyle := FParentStyle;
  T20.FImageIndex:= FImageIndex;
  T20.FStImageIndex:= FStImageIndex;
  T20.ForceButtons := FForceButtons;
  T20.FCheckBoxEnabled := FCheckBoxEnabled;
  T20.FCheckBoxState := FCheckBoxState;
  T20.FShowCheckBox := FShowCheckBox;
  T20.FCheckBoxType := FCheckBoxType;
  T20.FEnabled := FEnabled;
  T20.FImageIndex2 := FImageIndex2;
  T20.FStImageIndex2 := FStImageIndex2;
  T20.FHidden := FHidden;
  Stream.WriteBuffer(T20,SizeOf(T20));

  // write SubStrings
  i:=FColText.Count;
  Stream.WriteBuffer(i,SizeOf(i));
  for j:=0 to i-1 do
  begin
    s:=FColText[j];
    k:=length(s);
    Stream.WriteBuffer(k,SizeOf(k));
    Stream.WriteBuffer(pchar(s)^,k);
  end;
  Stream.WriteBuffer(FUseStyles, sizeof(FUseStyles));
  if FUseStyles then
  begin
    b := Assigned(FOwner.FOnSave);
    Stream.WriteBuffer(b, sizeof(boolean));

    i := FStyles.Count;
    for j := 0 to i do
    begin
      if j = 0
         then Style := MainStyle
         else Style := TElCellStyle(FStyles[j-1]);
      if b then FOwner.FOnSave(Self, Stream, Style)
      else
      with Style do
      begin
        SI.FCellBkColor := FCellBkColor;
        SI.FTextBkColor := FTextBkColor;
        SI.FTextColor   := FTextColor;
        SI.FTextFlags   := FTextFlags;
        SI.FPicture     := FPicture;
        SI.FCellType    := FCellType;
        SI.FStyle       := FStyle;
        SI.FOwnerProps  := FOwnerProps;
        SI.FFontSize    := FFontSize;
        SI.FFontStyles  := FFontStyles;
        SI.FTag         := FTag;
      end;
      Stream.WriteBuffer(SI, sizeof(TStyleInfo));
      WriteStringToStream(Stream, Style.FFontName);
      if j = 0 then Stream.WriteBuffer(i, sizeof(i));
    end;
  end;
  FOwner.TriggerItemSaveEvent(Stream, Self);
  // write subitems
  i:=FChildren.Count;
  Stream.WriteBuffer(i,SizeOf(i));
  for j:=0 to i-1 do TElTreeItem(FChildren[j]).WriteData(Stream);
end;

procedure TElTreeItem.Assign(Source:TPersistent);
var Item, Child:TElTreeItem;
    i:integer;
    Style : TElCellStyle;
begin
  if Source is TElTreeItem then
  begin
    if FOwner<>nil then FOwner.IsUpdating:=true;
    Item:=TElTreeItem(Source);
    Clear;
    FText:=Item.Text;
    FColText.Assign(Item.FColText);
    Data          := Item.Data;
    FColor        := Item.FColor;
    FBkColor      := Item.FBkColor;
    FState        := Item.FState;
    FParentStyle  := Item.FParentStyle;
    FParentColors := Item.FParentColors;
    FImageIndex   := Item.FImageIndex;
    FStImageIndex := Item.FStImageIndex;
    FImageIndex2  := Item.FImageIndex2;
    FStImageIndex2:= Item.FStImageIndex2;
    FCheckBoxState:= Item.FCheckBoxState;
    FShowCheckBox := Item.FShowCheckBox;
    FCheckBoxType := Item.FCheckBoxType;
    FCheckBoxEnabled := Item.FCheckBoxEnabled;
    FEnabled      := Item.FEnabled;
    FHidden       := Item.FHidden;
    FForceButtons := Item.FForceButtons;
    FUseStyles    := Item.FUseStyles;
    FMainStyle.Assign(Item.FMainStyle);
    FStyles.Clear;
    for i := 0 to Item.FStyles.Count - 1 do
    begin
      Style := AddStyle;
      Style.Assign(Item.Styles[i]);
    end;
    for i:=0 to Item.FChildren.Count-1 do
    begin
      Child:=FList.CreateItem(FOwner);
      FChildren.Add(Child);
      Child.FRoot:=FRoot;
      Child.FList:=FList;
      Child.FParent:=self;
      Child.FOwner := FOwner;
      Child.Assign(TElTreeItem(Item.FChildren[i]));
    end;
    if FOwner<>nil then
    begin
      FOwner.FUpdated:=true;
      FOwner.FVisUpdated:=true;
      FOwner.FClearVis:=true;
      FOwner.FClearAll:=true;
      FOwner.IsUpdating:=false;
    end;
  end else inherited Assign(Source);
end;

procedure TElTreeItem.SetColor(index:integer;value:TColor);
begin
  if index=1 then
  begin
    if FColor=value then exit;
    FColor:=value;
  end
  else
  begin
    if FBkColor=value then exit;
    FBkColor:=value;
  end;
  UpdateItem;
  //REMOVE: RedrawItem(true);
  //REMOVE: if FOwner.IsUpdating then FOwner.FUpdated:=true else FOwner.Update;
end;

function TElTreeItem.GetParent;
begin
  if FParent=FRoot then result:=nil else result:=FParent;
end;

procedure TElTreeItem.EditText;
begin
  FOwner.DoEditItem(self, -1);
end;

function TElTreeItem.GetFullName;
begin
  result:=separator+FText;
  if FParent<>FRoot then result:=FParent.GetFullName(separator) + result;
end;

function TElTreeItem.GetChildByIndex;
begin
  if (index<0) or (index>=FChildren.Count) then
      Raise EElTreeError.Create(GetErrorMsg(STExOutOfBounds));
  result:=TElTreeItem(FChildren[index]);
end;

function TElTreeItem.GetChildrenCount;
var i:integer;
begin
  result:=FChildren.Count;
  for i:=0 to FChildren.Count-1 do
      result:=result+TElTreeItem(FChildren[i]).ChildrenCount;
end;

function TElTreeItem.GetNextChild;
var i:integer;
begin
  i:=Child.Index;
  if i = (FChildren.Count-1) then result:=nil else
     result:=TElTreeItem(FChildren[i+1]);
end;

function TElTreeItem.GetPrevChild;
var i:integer;
begin
  i:=Child.Index;
  if i=0 then result:=nil else
     result:=TElTreeItem(FChildren[i-1]);
end;

function TElTreeItem.GetFirstVisibleChild;
var i : integer;
begin
  result:=nil;
  if (not FOwner.FilteredVisibility) then
  begin
    result := GetFirstChild;
    exit;
  end;
  if FChildren.Count = 0 then exit else
  for i := 0 to FChildren.Count - 1 do
  begin
    if not TElTreeItem(FChildren[i]).Hidden then
    begin
      result := TElTreeItem(FChildren[i]);
      exit;
    end;
  end;
end;

function TElTreeItem.GetFirstChild;
begin
  if FChildren.Count = 0 then result:=nil else result:=TElTreeItem(FChildren[0]);
end;

function TElTreeItem.GetLastChild;
begin
  if FChildren.Count = 0 then result:=nil else result:=TElTreeItem(FChildren[FChildren.Count-1]);
end;

function TElTreeItem.GetFirstSibling;
begin
if FParent=nil then result:=nil else
     result:=FParent.GetFirstChild;
end;

function TElTreeItem.GetLastSibling;
begin
  if FParent=nil then result:=nil else
     result:=FParent.GetLastChild;
end;

function TElTreeItem.GetNextSibling;
begin
  if FParent=nil then result:=nil else
     result:=FParent.GetNextChild(self);
end;

function TElTreeItem.GetPrevSibling;
begin
  if FParent=nil then result:=nil else
     result:=FParent.GetPrevChild(self);
end;

function TElTreeItem.GetFullExpand;
begin
  if Parent=nil then result:=true
  else result:=(Parent.FullyExpanded and Parent.Expanded);
end;

procedure TElTreeItem.SetState;
var FOldState : TSTIStates;
    b, reqitemch : boolean;
begin
  reqitemch := true;
  FOldState := FState;
  case index of
    stsFocused: begin
                  if value=false then FOwner.ItemFocused:=nil
                                 else FOwner.ItemFocused:=self;
                  exit;
                end;
    stsSelected:
      begin
        if FOwner.FMultiSelect then
        begin
          b := (stsSelected in FState) xor value;
          if value=false then
          begin
            if stsSelected in FState then dec(FOwner.FSelCount);
            Exclude(FState, stsSelected);
          end else
          begin
            if not (stsSelected in FState) then inc(FOwner.FSelCount);
            Include (FState, stsSelected);
          end;
          if b then FOwner.TriggerItemSelectedChangeEvent(Self);
          reqitemch := false;
        end;
      end;
    stsCut:
      begin
        if value=false then Exclude(FState, stsCut)
                       else Include(FState, stsCut);
      end;
    stsUnderlined:
      begin
        if value=false then Exclude(FState, stsUnderlined)
                       else Include(FState, stsUnderlined);
      end;
    stsBold:
      begin
        if value=false then Exclude(FState, stsBold)
                       else Include(FState, stsBold);
      end;
    stsItalic:
      begin
        if value=false then Exclude(FState, stsItalic)
                       else Include(FState, stsItalic);
      end;
    stsStrikeOut:
      begin
        if value=false then Exclude(FState, stsStrikeOut)
                       else Include(FState, stsStrikeOut);
      end;
  end; // case
  if (index in [stsBold, stsItalic]) and (not FOwner.FHeader.Visible)
      and FullyExpanded then FOwner.DefineHRange;
  if FOldState <> FState then
  begin
    if reqitemch then FOwner.DoItemChange(self, icmState);
    UpdateItem;
  end;
  //REMOVE: RedrawItem(true);
  //REMOVE: if FOwner.FUpdating then FOwner.FUpdated:=true else FOwner.Update;
end;

function TElTreeItem.GetState;
begin
  result:= index in FState;
end;

function TElTreeItem.IsExpanded;
begin
  result:=stsExpanded in FState;
end;

procedure TElTreeItem.Expand;
var i:integer;
    b:boolean;
begin
  b:=true;
  FOwner.DoItemExpanding(self,b);
  if not(b) then exit;
  FOwner.IsUpdating:=true;
  FOwner.FVisUpdated:=true;
  FOwner.FRangeUpdate := true;
  FState:=FState + [stsExpanded];
  FOwner.DoItemExpand(self);
  if recursive then
    for i:=0 to FChildren.Count-1 do TElTreeItem(FChildren[i]).expand(true);
  if (not FOwner.FHeader.Visible) and FullyExpanded then FOwner.FHRange:=-1;
  FOwner.FClearVis:=true;
  FOwner.FUpdated:=true;
  FOwner.IsUpdating:=false;
end;

procedure TElTreeItem.Collapse;
var i:integer;
    b:boolean;

  procedure IterateProc (Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                          IterateData:pointer; Tree:TCustomElTree);
  begin
    if not Item.IsUnder(TElTreeItem(IterateData)) then
    begin
      ContinueIterate := false;
      exit;
    end;
    ContinueIterate := true;
    if Item <> IterateData then Item.Selected := false;
  end;

begin
  b:=true;
  FOwner.DoItemCollapsing(self,b);
  if not(b) then exit;
  FOwner.IsUpdating:=true;
  FOwner.FRangeUpdate := true;
  FOwner.FVisUpdated:=true;
  FState:=FState - [stsExpanded];
  FOwner.DoItemCollapse(self);
  if recursive then
    for i:=0 to FChildren.Count-1 do TElTreeItem(FChildren[i]).collapse(true);
  if FOwner.FDeselectChildrenOnCollapse then
  begin
    FOwner.Items.IterateFrom(false, @IterateProc, Self, Self);
  end;
  if (not FOwner.FHeader.Visible) and FullyExpanded then FOwner.FHRange:=-1;
  FOwner.FClearVis:=true;
  FOwner.FClearAll:=true;
  FOwner.FUpdated:=true;
  FOwner.IsUpdating:=false;
end;

function TElTreeItem.GetChildIndex;
begin
  result:=FChildren.IndexOf(Child);
end;

function TElTreeItem.GetIndex;
begin
  if FParent=nil then result:=-1 else result:=FParent.GetChildIndex(self);
end;

function TElTreeItem.GetVisIndex;
begin
  // no check for root - should never be called for root
  result:=FList.GetVisIndex(self);
end;

function TElTreeItem.GetAbsIndex;
begin
  // no check for root - should never be called for root
  result:=FList.GetAbsIndex(self);
end;

procedure TElTreeItem.RemoveChild;
var i:integer;
begin
  i:=FOwner.FVisible.IndexOf(Child);
  if i>=0 then FOwner.FVisible.Delete(i);
  i:=FChildren.IndexOf(Child);
  if i<>-1 then FChildren.Delete(i);
  dec(FList.FCount);
end;

procedure TElTreeItem.DeleteChild;
var i:integer;
begin
  FOwner.IsUpdating:=true;
  Child.Clear;
  Child.Selected := false;
  if Child.Hidden then
  begin
    dec(FOwner.TotalHiddenCount);
    Child.FHidden := false;
  end;
  if FOwner.FSelected=Child then FOwner.FSelected:=nil;
  // Move focus, if needed
  if Child = FOwner.ItemFocused then
  begin
    if self=FRoot then
    begin
      FOwner.ItemFocused:=GetPrevChild(Child);
      if FOwner.ItemFocused=nil then FOwner.ItemFocused:=GetNextChild(Child);
    end else FOwner.ItemFocused:=Self;
  end;
  if FOwner.FTrackItem=Child then FOwner.FTrackItem:=nil;
  if (FOwner.FInpEdit<>nil) then
  begin
    FOwner.FEditingItem := nil;
    if not FOwner.FDelOnEdit then FOwner.EndEdit(true);
  end;
  i:=FOwner.FVisible.IndexOf(Child);
  if (i>=0) then FOwner.FVisible.Delete(i);
  FOwner.DoItemDelete(Child);
  i:=FChildren.IndexOf(Child);
  if i<>-1 then FChildren.Delete(i);
  Child.Free;
  dec(FList.FCount);
  FOwner.FUpdated:=true;
  FOwner.FRangeUpdate := true;
  FOwner.IsUpdating:=false;
end;

function TElTreeItem.AddChild;
begin
  result:=FChildren.Add(Child);
  Child.FRoot:=FRoot;
  FOwner.FUpdated:=true;
end;

function TElTreeItem.InsertChild;
begin
  if Index>FChildren.Count then
     Raise EElTreeError.Create(GetErrorMsg(STExOutOfBounds));
  FChildren.Insert(Index,Child);
  result:=index;
  Child.FParent:=self;
  Child.FRoot:=FRoot;
  FOwner.FUpdated:=true;
end;

procedure TElTreeItem.MoveTo;
var TSI:TElTreeItem;
begin
  if NewParent=self then exit;
  if NewParent=nil then TSI:=FRoot else TSI:=NewParent;
  MoveToIns(NewParent,TSI.FChildren.Count);
end;

procedure TElTreeItem.MoveToIns;
var i:integer;
begin
  if NewParent=self then exit;
  if NewParent=nil then NewParent:=FRoot;
  If NewParent.IsUnder(self) then
     Raise EElTreeError.Create(GetErrorMsg(STexRecursiveMove));
  FOwner.IsUpdating:=true;
  i:=FParent.FChildren.IndexOf(self);
  if FParent<>nil then FParent.RemoveChild(self);
  if (FParent=NewParent) and (i<index) then dec(index);
  NewParent.FChildren.Insert(index,self);
  FParent:=NewParent;
  inc(FOwner.FItems.FCount);
  if (FOwner.FSortMode=smAdd) or (FOwner.FSortMode=smAddClick) then
  begin
    if (FOwner.FMainTreeCol=FOwner.FSortSection) or (FOwner.FSortSection = -1) then FParent.Sort(false);
  end;
  FOwner.FVisUpdated:=true;
  FOwner.FClearVis:=true;
  FOwner.FUpdated:=true;
  FOwner.FClearAll:=true;
  FOwner.IsUpdating:=false;
end;

procedure TElTreeItem.RedrawItemPart;
var R:TRect;
begin
  if (FOwner=nil) then exit;
  R:=FOwner.GetItemRect(FOwner.FVisible.IndexOf(self));
  inc(R.Bottom);
  if DoCheck and IsRectEmpty(R) then exit;
  R.Left := Left  - FOwner.FHPos;
  R.Right:= Right - FOwner.FHPos;
  if not FOwner.Destroying then
     InvalidateRect(FOwner.Handle,@R,true);
end;

procedure TElTreeItem.RedrawItem;
var R:TRect;
begin
  if (FOwner=nil) then exit;
  R:=FOwner.GetItemRect(FOwner.FVisible.IndexOf(self));
  inc(R.Bottom);
  if DoCheck and IsRectEmpty(R) then exit;
  if not FOwner.Destroying then
    InvalidateRect(FOwner.Handle,@R,true);
end;

procedure TElTreeItem.SetText;
begin
  FText:=Value;
  FOwner.DoItemChange(Self, icmText);
  if (not FOwner.FHeader.Visible) and FullyExpanded then FOwner.DefineHRange;
  if (FOwner.FSortMode=smAdd) or (FOwner.FSortMode=smAddClick) then
  begin
    if (FOwner.FMainTreeCol=FOwner.FSortSection) or (FOwner.FSortSection = -1) then FParent.Sort(false);
  end;
  UpdateItem;
  //REMOVE:RedrawItem(true);
  //REMOVE:if FOwner.FUpdating then FOwner.FUpdated:=true else FOwner.Update;
end;

function TElTreeItem.GetLevel : integer;
begin
  if Self=FRoot then result:=-1 else result := FParent.Level+1;
end;

function TElTreeItem.EqualColumns;
var i:integer;
begin
  result:=false;
  if FColText.Count<>Cols.Count then exit;
  for i:=0 to FColText.Count-1 do    //Iterate
    if (Cols.Count<=i) or (Cols[i]<>FColText[i]) then exit;
  result:=true;
end;

function TElTreeItem.IsEqual;
begin
  result:=(Text=Item.Text) and (FState=Item.FState) and
    (EqualColumns(Item.FColText)) and (FColor=Item.FColor) and (FBkColor=Item.FBkColor);
end;

{Returns true if the item is under then given item in the tree}

function TElTreeItem.IsUnder(Item:TElTreeItem):boolean;
begin
  if Item = nil then
  begin
    result:=false;
    exit;
  end;
  if self=item then result:=true else
    if (FParent<>FRoot) and (FParent<>nil) then result:=FParent.IsUnder(Item) else result:=false;
end;

procedure TElTreeItem.SetExpanded;
begin
  if value=Expanded then exit;
  FOwner.IsUpdating:=true;
  if value=true then Expand(false) else Collapse(false);
  FOwner.IsUpdating:=false;
end;

procedure TElTreeItem.MakeFullyExpanded;
var TSI:TElTreeItem;
begin
  if not value then exit;
  FOwner.IsUpdating:=true;
  try
    TSI := Self;
    while TSI.FParent <> nil do
    begin
      TSI.FParent.Expand(false);
      TSI:=TSI.FParent;
    end;    // while
    FOwner.FClearVis:=true;
    FOwner.FVisUpdated:=true;
    FOwner.FUpdated:=true;
    FOwner.FRangeUpdate := true;
  finally
    FOwner.IsUpdating:=false;
  end;
end;

procedure TElTreeItem.SetParentStyle(value:boolean);
begin
  if value=FParentStyle then exit;
  FParentStyle:=value;
  UpdateItem;
  //REMOVE:RedrawItem(true);
  //REMOVE:if FOwner.FUpdating then FOwner.FUpdated:=true else FOwner.Update;
end;

procedure TElTreeItem.SetParentColors(value:boolean);
begin
  if value=FParentColors then exit;
  FParentColors:=value;
  UpdateItem;
  //REMOVE:RedrawItem(true);
  //REMOVE:if FOwner.FUpdating then FOwner.FUpdated:=true else FOwner.Update;
end;

function TElTreeItem.GetHasChildren : boolean;
begin
  result:=(FChildren.Count>0);
end;

procedure TElTreeItem.Clear;
begin
  while FChildren.Count>0 do DeleteChild(TElTreeItem(FChildren[0]));
end;

procedure TElTreeItem.ExchangeItems(I,J:integer);
var TSI:TElTreeItem;
begin
  TSI:=FChildren[I];
  FChildren[I]:=FChildren[J];
  FChildren[J]:=TSI;
end;

procedure TElTreeItem.Sort(recursive:boolean);
begin
  FOwner.IsUpdating:=true;
  QuickSort(recursive,0,FChildren.Count-1);
  with FOwner do
  begin
    FClearAll:=true;
    FClearVis:=true;
    FUpdated:=true;
    IsUpdating:=false;
  end;    // with  
end;

procedure TElTreeItem.QuickSort(recursive:boolean;L, R: Integer);
var
  I, J: Integer;
  P: TElTreeItem;

begin
  if FChildren.Count=0 then exit;
  if FChildren.Count>1 then
    repeat
      I := L;
      J := R;
      P := TElTreeItem (FChildren[(L + R) shr 1]);
      repeat
        while (I<FChildren.Count) and (FOwner.CompareItems(TElTreeItem(FChildren[I]), P) < 0) do Inc(I);
        while (J>=0) and (FOwner.CompareItems(TElTreeItem(FChildren[J]), P) > 0) do Dec(J);
        if I <= J then
        begin
          ExchangeItems(I, J);
          Inc(I);
          Dec(J);
        end;
      until I > J;
      if L < J then QuickSort(recursive, L, J);
      L := I;
    until I >= R;

  if (not Recursive) then exit;

  for i := 0 to FChildren.Count-1 do    // Iterate
  begin
    P:=TElTreeItem(FChildren[i]);
    P.QuickSort(recursive, 0, P.FChildren.Count-1);
  end;    // for
end;

procedure TElTreeItem.OnColTextChange(Sender:TObject);
begin
  if FOwner<>nil then
  begin
    if (FOwner.FSortMode=smAdd) or (FOwner.FSortMode=smAddClick) then
    begin
      if InRange(-1, FOwner.FHeader.Sections.Count-1, FOwner.FSortSection) then
      begin
        if FParent <> nil then FParent.Sort(false);
      end;
    end;
    UpdateItem;
    //REMOVE:RedrawItem(true);
    //REMOVE:if not (FOwner.IsUpdating) then FOwner.Update else FOwner.FUpdated:=true;
  end;
end;

procedure TElTreeItem.SetImageIndex(value : integer);  { private }
begin
  if FImageIndex = value then exit;
  FImageIndex := value;
  if FOwner.FShowImages then UpdateItem;
end;  { SetImageIndex }

procedure TElTreeItem.SetStImageIndex(value : integer);  { private }
begin
  if FStImageIndex = value then exit;
  FStImageIndex := value;
  if FOwner.FShowImages then UpdateItem;
end;  { SetStImageIndex }

procedure TElTreeItem.SetImageIndex2(value : integer);  { private }
begin
  if FImageIndex2 = value then exit;
  FImageIndex2 := value;
  if FOwner.FShowImages then UpdateItem;
end;  { SetImageIndex }

procedure TElTreeItem.SetStImageIndex2(value : integer);  { private }
begin
  if FStImageIndex2 = value then exit;
  FStImageIndex2 := value;
  if FOwner.FShowImages then UpdateItem;
end;  { SetStImageIndex }

procedure TElTreeItem.SetForceButtons(newValue: Boolean);
{ Sets data member FForceButtons to newValue. }
begin
  if FForceButtons <> newValue then
  begin
    FForceButtons := newValue;
    UpdateItem;
  end;
end;  { SetForceButtons }

function TElTreeItem.GetCount : Integer;
begin
  result := FChildren.Count;
end;  { GetCount }

function TElTreeItem.GetItems(Index : integer) : TElTreeItem;
{ Returns the value of data member FItems[Index ]. }
begin
  result := FChildren[Index];
end;  { GetItems }

procedure TElTreeItem.Delete;  { public }
begin
  FList.DeleteItem(Self);
end;  { Delete }

function TElTreeItem.IsVisible : Boolean;  { public }
begin
  result := FullyExpanded;
end;  { IsVisible }

// Modified 11/02/98 by EM

function TElTreeItem.GetNextVisible : TElTreeItem;  { public }
begin
  Result := GetNext;
  while (Result <> nil) and (not Result.IsVisible) do Result := Result.GetNext;
end;

function TElTreeItem.GetPrevVisible : TElTreeItem;  { public }
begin
  Result := GetPrev;
  while (Result <> nil) and (not Result.IsVisible) do Result := Result.GetPrev;
end;

(*
function TElTreeItem.GetNextVisible : TElTreeItem;  { public }

type PItem= ^TItem;
     TItem = record
       Item : TElTreeItem;
       DoSet : boolean;
     end;

var Item : TItem;

   procedure IntNextVis(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                            IterateData:pointer; Tree:TCustomElTree);
   begin
     if PItem(IterateData)^.DoSet then
     begin
       PItem(IterateData)^.Item := Item;
       ContinueIterate := false;
       exit;
     end;
     PItem(IterateData)^.DoSet := true;
   end;

begin
  result := nil;
  if not IsVisible then exit else
  begin
    Item.Item := nil;
    Item.DoSet := false;
    FList.IterateFrom(true, @IntNextVis, @Item, Self);
    result := Item.Item;
  end;
end;  { GetNextVisible }


// Modified 10/29/98 by EM
function TElTreeItem.GetPrevVisible : TElTreeItem;  { public }

type PItem= ^TItem;
     TItem = record
       Item : TElTreeItem;
       DoSet : boolean;
     end;

var Item : TItem;

   procedure IntPrevVis(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                            IterateData:pointer; Tree:TCustomElTree);
   begin
     if PItem(IterateData)^.DoSet then
     begin
       PItem(IterateData)^.Item := Item;
       ContinueIterate := false;
       exit;
     end;
     PItem(IterateData)^.DoSet := true;
   end;

begin
  result := nil;
  if not IsVisible then exit else
  begin
    Item.Item := nil;
    Item.DoSet := false;
    FList.IterateBackFrom(true, @IntPrevVis, @Item, Self);
    if Item.Item <> FList.FRoot then result := Item.Item;
  end;
end;  { GetPrevVisible }
*)

// Modified 10/29/98 by EM
function TElTreeItem.GetPrev : TElTreeItem;  { public }

type PItem= ^TItem;
     TItem = record
       Item : TElTreeItem;
       DoSet : boolean;
     end;

var Item : TItem;

   procedure IntPrev(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                            IterateData:pointer; Tree:TCustomElTree);
   begin
     if PItem(IterateData)^.DoSet then
     begin
       PItem(IterateData)^.Item := Item;
       ContinueIterate := false;
       exit;
     end;
     PItem(IterateData)^.DoSet := true;
   end;

begin
  result := nil;
  Item.Item := nil;
  Item.DoSet := false;
  FList.IterateBackFrom(false, @IntPrev, @Item, Self);
  if Item.Item <> FList.FRoot then result := Item.Item;
end;  { GetPrev }

function TElTreeItem.GetNext : TElTreeItem;  { public }

type PItem= ^TItem;
     TItem = record
       Item : TElTreeItem;
       DoSet : boolean;
     end;

var Item : TItem;

   procedure IntNext(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                            IterateData:pointer; Tree:TCustomElTree);
   begin
     if PItem(IterateData)^.DoSet then
     begin
       PItem(IterateData)^.Item := Item;
       ContinueIterate := false;
       exit;
     end;
     PItem(IterateData)^.DoSet := true;
   end;

begin
  Item.Item := nil;
  Item.DoSet := false;
  FList.IterateFrom(false, @IntNext, @Item, Self);
  result := Item.Item;
end;  { GetNext }

procedure TElTreeItem.MoveToItem(Item : TElTreeItem; Mode : TNodeAttachMode);  { public }
begin
   case Mode of
      naAdd:
        begin
          if not Assigned(item) then Raise EElTreeError.Create(GetErrorMsg(STexInvItem));
          if not Assigned(item)
             then MoveTo(nil)
             else MoveToIns(item.FParent, item.FParent.Count);
        end;
      naAddFirst:
        begin
          if not Assigned(item)
             then MoveToIns(nil, 0)
             else MoveToIns(item.FParent, 0);
        end;
      naAddChild: MoveTo(item);
      naAddChildFirst: MoveToIns(item, 0);
      naInsert:
        begin
         if not Assigned(item)
            then MoveToIns(nil, 0)
            else MoveToIns(item.FParent, item.Index);
        end;
   end;
end;  { MoveToItem }

procedure TElTreeItem.SetUseStyles(newValue: Boolean);
begin
  if (FUseStyles <> newValue) then
  begin
    FUseStyles := newValue;
    UpdateItem;
  end;  {if}
end;

procedure TElTreeItem.UpdateItem;
begin
  RedrawItem(true);
  if FOwner.FUpdating then FOwner.FUpdated := true else FOwner.Update;
end;

function TElTreeItem.GetStyles(index : integer): TElCellStyle;
begin
  result := FStyles[index ];
end;

procedure TElTreeItem.SetStyles(index : integer; newValue: TElCellStyle);
begin
  if (FStyles[index] <> newValue) then
  begin
    FStyles[index] := newValue;
  end;  {if}
end;

function TElTreeItem.AddStyle: TElCellStyle;
begin
  Result := TElCellStyle.Create(Self);
  Result.FOwnerProps := true;
  Result.FFontSize := FOwner.Font.Size;
  Result.FFontStyles := FOwner.Font.Style;
  Result.FFontName := FOwner.Font.Name;
  if Assigned(FMainStyle) then Result.Assign(FMainStyle);
  FStyles.Add(Result);
end;

procedure TElTreeItem.RemoveStyle(Style : TElCellStyle);
begin
  FStyles.Remove(Style);
end;

function TElTreeItem.GetStylesCount: Integer;
begin
  result := FStyles.Count;
end;

procedure TElTreeItem.SetCheckBoxState(newValue : TCheckBoxState);
var Section : TElHeaderSection;
begin
  if (FCheckBoxState <> newValue) then
  begin
    FCheckBoxState := newValue;
    if FShowCheckBox and (FOwner.FShowCheckboxes) then
    begin
      if FOwner.FHeader.Visible then
      begin
        Section := FOwner.FHeader.Sections[FOwner.FMainTreeCol];
        RedrawItemPart(true, Section.Left, Section.Right);
      end else RedrawItem(true);
      if FOwner.IsUpdating then FOwner.FUpdated:=true else FOwner.Update;
    end;
  end;  {if}
end;  {SetCheckBoxState}

function TElTreeItem.GetChecked : boolean;
begin
  Result := FCheckBoxState = cbChecked;
end;

procedure TElTreeItem.SetChecked(newValue : Boolean);
var Section : TElHeaderSection;
    i : integer;
    Item : TElTreeItem;
begin
  if (Checked <> newValue) then
  begin
    if newValue then FCheckBoxState := cbChecked else FCheckBoxState := cbUnchecked;
    if (FCheckBoxType = ectRadioButton) and (newValue) then
    begin
      //FOwner.IsUpdating := true;
      for i := 0 to FParent.FChildren.Count - 1 do    // Iterate
      begin
        Item := FParent.Children[i];
        if (Item.FCheckBoxType = ectRadioButton) and (Item <> Self) then
        begin
          Item.Checked := false;
          //FOwner.FUpdated:=true;
        end;
      end;    // for
      //FOwner.IsUpdating := false;
    end;
    if FShowCheckBox and (FOwner.FShowCheckboxes) then
    begin
      if FOwner.FHeader.Visible then
      begin
        Section := FOwner.FHeader.Sections[FOwner.FMainTreeCol];
        RedrawItemPart(true, Section.Left, Section.Right);
      end else RedrawItem(true);
      if FOwner.IsUpdating then FOwner.FUpdated:=true else FOwner.Update;
    end;
  end;
end;  {SetChecked}

procedure TElTreeItem.SetShowCheckBox(newValue : Boolean);
begin
  if (FShowCheckBox <> newValue) then
  begin
    FShowCheckBox := newValue;
    if (FOwner.FShowCheckboxes) then UpdateItem;
  end;  {if}
end;  {SetShowCheckBox}

procedure TElTreeItem.SetCheckBoxType(newValue : TElCheckBoxType);
var Section : TElHeaderSection;
begin
  if (FCheckBoxType <> newValue) then
  begin
    FCheckBoxType := newValue;
    if FShowCheckBox and (FOwner.FShowCheckboxes) then
    begin
      if FOwner.FHeader.Visible then
      begin
        Section := FOwner.FHeader.Sections[FOwner.FMainTreeCol];
        RedrawItemPart(true, Section.Left, Section.Right);
      end else RedrawItem(true);
      if FOwner.IsUpdating then FOwner.FUpdated:=true else FOwner.Update;
    end;
  end;  {if}
end;  {SetCheckBoxType}

procedure TElTreeItem.SetCheckBoxEnabled(newValue : Boolean);
var Section : TElHeaderSection;
begin
  if (FCheckBoxEnabled <> newValue) then
  begin
    FCheckBoxEnabled := newValue;
    if FShowCheckBox and FOwner.FShowCheckboxes then
    begin
      if FOwner.FHeader.Visible then
      begin
        Section := FOwner.FHeader.Sections[FOwner.FMainTreeCol];
        RedrawItemPart(true, Section.Left, Section.Right);
      end else RedrawItem(true);
      if FOwner.IsUpdating then FOwner.FUpdated:=true else FOwner.Update;
    end;
  end;  {if}
end;  {SetCheckBoxEnabled}

procedure TElTreeItem.SetSuppressButtons(newValue: Boolean);
begin
  if (FSuppressButtons <> newValue) then
  begin
    FSuppressButtons := newValue;
    UpdateItem;
  end;  {if}
end;

procedure TElTreeItem.SetEnabled(newValue : Boolean);
begin
  if (FEnabled <> newValue) then
  begin
    FEnabled := newValue;
    UpdateItem;
  end;  {if}
end;  {SetEnabled}

function TElTreeItem.GetHidden: Boolean;
begin
  result := FHidden;
end;

procedure TElTreeItem.SetHidden(newValue: Boolean);
begin
  if (FHidden <> newValue) then
  begin
    FHidden := newValue;
    if FOwner <> nil then
    begin
      FOwner.IsUpdating := true;
      if FHidden = true then
      begin
        inc(FOwner.TotalHiddenCount);
      end else
      begin
        dec(FOwner.TotalHiddenCount);
      end;
      if FOwner.FilteredVisibility then
      with FOwner do
      begin
        FVisUpdated := true;
        FClearVis:=true;
        FClearAll:=true;
        FRangeUpdate := true;
        FUpdated := true;
      end;
      FOwner.IsUpdating := false;
    end;
  end;  {if}
end;

function TElTreeItem.GetFullyVisible : Boolean;
begin
  if Parent <> nil then result := (not hidden) and Parent.FullyVisible else result := not hidden;
end;

procedure TElTreeItem.SetFullyVisible(newValue: Boolean);
begin
  if (FFullyVisible <> newValue) then
  begin
    if newValue and (Parent <> nil) then
    begin
      Parent.FullyVisible := true;
      Hidden := false;
    end;
  end;  {if}
end;

function TElTreeItem.GetHasVisibleChildren: Boolean;
var i : integer;
begin
  result := false;
  if not FOwner.FilteredVisibility then result := FChildren.Count >0 else
  begin
    if FChildren.Count = 0 then exit else
    for i := 0 to FChildren.Count - 1 do
    begin
      if not TElTreeItem(FChildren[i]).Hidden then
      begin
        result := true;
        exit;
      end;
    end;
  end;
end;

// ****************************************************************************
// TElTreeItems methods
// ****************************************************************************

constructor TElTreeItems.Create;
begin
  inherited Create;
  FOwner:=AOwner;
  FRoot:=TElTreeItem.Create(FOwner);
  FRoot.FList:=self;
  FRoot.FParent:=nil;
  FRoot.FRoot:=FRoot;
  FRoot.FState:=[stsExpanded];
end;

destructor TElTreeItems.Destroy;
begin
  FOwner.IsUpdating:=true;
  FRoot.Clear;
  FRoot.Free;
  inherited Destroy;
end;

function  TElTreeItems.GetVisCount:integer;
begin
  result:=FOwner.TotalVisCount;
end;

procedure TElTreeItems.DefineProperties(Filer: TFiler);

function WriteNodes: Boolean;
  begin
    Result := Count > 0;
  end;

begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('Data', ReadData, WriteData, WriteNodes);
end;

procedure TElTreeItems.SetItem(Index: Integer; Value: TElTreeItem);
begin
  GetItem(Index).Assign(Value);
end;

procedure TElTreeItems.Iterate;
var j:integer;
    DoContinue: boolean;

  procedure IntIterate(VisibleOnly : boolean; Item:TElTreeItem);
  var i:integer;
  begin
    inc(j);
    if (j >= 0) and ((not VisibleOnly) or ((not Item.Hidden) or (not FOwner.FilteredVisibility)))
      then IterateProc(Item, j, DoContinue, IterateData, FOwner);
    if not (DoContinue) then exit;
    if (not (VisibleOnly)) or (Item.Expanded) then
      for i:=0 to Item.FChildren.Count-1 do
      begin
        if (not VisibleOnly) or ((not TElTreeItem(Item.FChildren[i]).Hidden) or (not FOwner.FilteredVisibility)) then
           IntIterate(VisibleOnly, TElTreeItem(Item.FChildren[i]));
        if not (DoContinue) then exit;
      end;
  end;

begin
  j:=-2;
  DoContinue:=true;
  IntIterate(VisibleOnly, FRoot);
end;

procedure TElTreeItems.IterateFrom;
var j:integer;
    DoContinue: boolean;
    Stack : TElStack;
    TSI : TElTreeItem;

  procedure IntIterate(VisibleOnly : boolean; Item:TElTreeItem; StartFrom:TElTreeItem; var Stack : TElStack);
  var i:integer;
      k:integer;
  begin
    inc(j);
    if (Stack<>nil) and (Stack.Count>0)then
    begin
      k:=Item.FChildren.IndexOf(Stack.Pop);
      if Stack.Count = 0 then begin Stack.Free; Stack:=nil; end;
    end
    else
    begin
      if (j >= 0) and ((not VisibleOnly) or ((not Item.Hidden) or (not FOwner.FilteredVisibility)))
         then IterateProc(Item, j, DoContinue, IterateData, FOwner);
      if not (DoContinue) then exit;
      k:=0;
    end;
    if (not (VisibleOnly)) or (Item.Expanded) then
    for i:=k to Item.FChildren.Count-1 do
    begin
      if (not VisibleOnly) or ((not TElTreeItem(Item.FChildren[i]).Hidden) or (not FOwner.FilteredVisibility)) then
         IntIterate(VisibleOnly, TElTreeItem(Item.FChildren[i]), StartFrom, Stack);
      if not (DoContinue) then exit;
    end;
  end;

begin             
  j:=-2;
  if StartFrom<>nil then
  begin
    Stack:=TElStack.Create;
    TSI:=StartFrom;
    Stack.Push(TSI);
    while TSI.Parent<>nil do
    begin
      Stack.Push(TSI.Parent);
      TSI:=TSI.Parent;
    end; // while
  end else Stack:=nil;
  DoContinue:=true;
  IntIterate(VisibleOnly, FRoot, StartFrom, Stack);
  if Stack<>nil then Stack.Free;
end;

procedure TElTreeItems.IterateBack;
var j:integer;
    DoContinue: boolean;

  procedure IntIterate(VisibleOnly : boolean; Item:TElTreeItem);
  var i:integer;
  begin
    if not (VisibleOnly) or (Item.Expanded) then
      for i:=Item.FChildren.Count-1 downto 0 do
      begin
        if (not VisibleOnly) or ((not TElTreeItem(Item.FChildren[i]).Hidden) or (not FOwner.FilteredVisibility))
           then IntIterate(VisibleOnly, TElTreeItem(Item.FChildren[i]));
        if not (DoContinue) then exit;
      end;
      dec(j);
    if (j>=0) and ((not VisibleOnly) or ((not Item.Hidden) or (not FOwner.FilteredVisibility)))
       then IterateProc(Item,j,DoContinue, IterateData,FOwner);
    if not (DoContinue) then exit;
  end;

begin
  if VisibleOnly then j:=FOwner.TotalVisCount else j:=FCount;
  DoContinue := true;
  IntIterate(VisibleOnly, FRoot);
end;

// Changed 10/28/98 by EM
procedure TElTreeItems.IterateBackFrom;
var j:integer;
    DoContinue: boolean;
    Stack : TElStack;
    TSI : TElTreeItem;

  procedure IntIterate(VisibleOnly : boolean; Item:TElTreeItem; StartFrom:TElTreeItem; var Stack : TElStack);
  var i,k:integer;
  begin
    if (Stack<>nil) and (Stack.Count>0)then
    begin
      k:=Item.FChildren.IndexOf(Stack.Pop);
      if Stack.Count=0 then begin Stack.Free; Stack:=nil; end;
    end
    else
      k:=Item.FChildren.Count-1;

    if (not (VisibleOnly) or (Item.Expanded)) and (Item <> StartFrom) then
    for i:=k downto 0 do
      begin
        if (not VisibleOnly) or ((not TElTreeItem(Item.FChildren[i]).Hidden) or (not FOwner.FilteredVisibility))
           then IntIterate(VisibleOnly, TElTreeItem(Item.FChildren[i]), StartFrom, Stack);
        if not (DoContinue) then exit;
      end;
    dec(j);
    if (j>=0) and (not VisibleOnly) or ((not Item.Hidden) or (not FOwner.FilteredVisibility))
       then IterateProc(Item, j, DoContinue, IterateData, FOwner);
    if (not DoContinue) then exit;
  end;

begin
  if VisibleOnly then j:=FOwner.TotalVisCount else j:=FCount;
  if StartFrom<>nil then
  begin
    Stack:=TElStack.Create;
    TSI:=StartFrom;
    Stack.Push(TSI);
    while TSI.Parent<>nil do
    begin
      Stack.Push(TSI.Parent);
      TSI:=TSI.Parent;
    end; // while
  end else Stack:=nil;
  DoContinue:=true;
  IntIterate(VisibleOnly, FRoot, StartFrom, Stack);
  if Stack<>nil then Stack.Free;
end;

function TElTreeItems.GetVisItem;
type PGIRec = ^TGIRec;
     TGIRec = record
       j:integer;
       TSI:TElTreeItem;
     end;
var GIRec : TGIRec;

  procedure IntGetItem(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                       IterateData:pointer; Tree:TCustomElTree);
  begin
    if Index=PGIRec(IterateData)^.j then
    begin
      PGIRec(IterateData)^.TSI:=Item;
      ContinueIterate:=false;
    end;
  end;

begin
  if (index<0) or (index>=FOwner.TotalVisCount) then
  begin
    result:=nil;
    exit;
  end;
  GIRec.TSI:=nil;
  GIRec.j:=index;
  if index<=(FOwner.TotalVisCount div 2)
     then Iterate(true, @IntGetItem, @GIRec)
     else IterateBack(true, @IntGetItem, @GIRec);
  result:=GIRec.TSI;
end;

function TElTreeItems.GetItem;
type PGIRec = ^TGIRec;
     TGIRec = record
       j:integer;
       TSI:TElTreeItem;
     end;
var GIRec : TGIRec;

  procedure IntGetItem(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                       IterateData:pointer; Tree:TCustomElTree);
  begin
    if Index=PGIRec(IterateData)^.j then
    begin
      PGIRec(IterateData)^.TSI:=Item;
      ContinueIterate:=false;
    end;
  end;

begin
  if (index<0) or (index>=FCount) then
  begin
    result:=nil;
    exit;
  end;
  GIRec.TSI:=nil;
  GIRec.j:=index;
  if index<=(FCount div 2)
     then Iterate(false, @IntGetItem, @GIRec)
     else IterateBack(false, @IntGetItem, @GIRec);
  result:=GIRec.TSI;
end;

function TElTreeItems.GetVisIndex;
type PGIRec = ^TGIRec;
     TGIRec = record
       j:integer;
       TSI:TElTreeItem;
     end;
var GIRec : TGIRec;

  procedure IntGetIndex(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                          IterateData:pointer; Tree:TCustomElTree);
  begin
    if PGIRec(IterateData)^.TSI=Item then
    begin
      PGIRec(IterateData)^.j:=index;
      ContinueIterate:=false;
    end;
  end;

begin
  if Child=nil then begin result:=-1; exit; end;
  GIRec.j:=-1;
  GIRec.TSI:=Child;
  Iterate(true, @IntGetIndex,@GIRec);
  result:=GIRec.j;
end;

function TElTreeItems.GetAbsIndex;
type PGIRec = ^TGIRec;
     TGIRec = record
       j:integer;
       TSI:TElTreeItem;
     end;
var GIRec : TGIRec;

  procedure IntGetIndex(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                          IterateData:pointer; Tree:TCustomElTree);
  begin
    if PGIRec(IterateData)^.TSI=Item then
    begin
      PGIRec(IterateData)^.j:=index;
      ContinueIterate:=false;
    end;
  end;

begin
  if Child=nil then begin result:=-1; exit; end;
  GIRec.j:=-1;
  GIRec.TSI:=Child;
  Iterate(false, @IntGetIndex,@GIRec);
  result:=GIRec.j;
end;

procedure TElTreeItems.DeleteItem;
var TSI:TElTreeItem;
begin
  if (Child = nil) or (child = FRoot) then exit;
  FOwner.IsUpdating:=true;
  try
    TSI:=Child.FParent;
    TSI.DeleteChild(Child);
    if TSI.FullyExpanded and TSI.Expanded then
    begin
      FOwner.FClearVis:=true;
      FOwner.FClearAll:=true;
      FOwner.FVisUpdated:=true;
    end;
    FOwner.FUpdated:=true;
  finally
    FOwner.IsUpdating:=false;
  end;
end;

procedure TElTreeItems.RemoveItem;
var TSI:TElTreeItem;
begin
  if (Child = nil) or (child = FRoot) then exit;
  FOwner.IsUpdating:=true;
  try
    TSI:=Child.FParent;
    // Move focus, if needed
    if Child = FOwner.ItemFocused then
    begin
      if TSI=FRoot then
      begin
        FOwner.ItemFocused:=TSI.GetPrevChild(Child);
        if FOwner.ItemFocused=nil then FOwner.ItemFocused:=TSI.GetNextChild(Child);
      end else
      FOwner.ItemFocused:=TSI;
    end;
    // actually remove child
    TSI.RemoveChild(Child);
    Child.FParent:=nil;
    if TSI.FullyExpanded and TSI.Expanded then
    begin
      FOwner.FClearVis:=true;
      FOwner.FVisUpdated:=true;
    end;
    FOwner.FUpdated:=true;
    FOwner.FRangeUpdate := true;
  finally
    FOwner.IsUpdating:=false;
  end;
end;

procedure TElTreeItems.AddExistingItem(Item, Parent : TElTreeItem);  { public }
var
    Child :TElTreeItem;
begin
  FOwner.IsUpdating := true;
  Child := Item;
  Child.FList := self;
  Child.FParent := Parent;
  if Child.FParent = nil then Child.FParent := FRoot;
  Child.FParent.AddChild(Child);
  inc(FCount);
  if Child.FParent.Expanded and Child.FParent.FullyExpanded then
  begin
    FOwner.FClearVis:=true;
    FOwner.FVisUpdated:=true;
  end;
  FOwner.FUpdated:=true;
  FOwner.FRangeUpdate := true;
  FOwner.IsUpdating := false;
end;

procedure TElTreeItems.InsertExistingItem(Item, Parent : TElTreeItem; Index : integer);  { public }
var
    Child :TElTreeItem;
begin
  FOwner.IsUpdating:=true;
  Child := Item;
  Child.FList := self;
  Child.FParent := Parent;
  if Child.FParent = nil then Child.FParent:=FRoot;
  if (index>Child.FParent.FChildren.Count) or (index<0) then
    Raise EElTreeError.Create(GetErrorMsg(STExOutOfBounds));
  Child.FParent.InsertChild(Index,Child);
  inc(FCount);
  if Child.FParent.FullyExpanded and Child.FParent.Expanded then
  begin
    FOwner.FClearVis:=true;
    FOwner.FClearAll:=true;
    FOwner.FVisUpdated:=true;
  end;
  FOwner.FUpdated:=true;
  FOwner.FRangeUpdate := true;
  FOwner.IsUpdating:=false;
end;

function TElTreeItems.InsertItem;
var
    Child :TElTreeItem;
begin
  FOwner.IsUpdating:=true;
  Child:=CreateItem(FOwner);
  Child.FList:=self;
  Child.FParent:=Parent;
  if Child.FParent=nil then Child.FParent:=FRoot;
  if (index>Child.FParent.FChildren.Count) or (index<0) then
    Raise EElTreeError.Create(GetErrorMsg(STExOutOfBounds));
  Child.FParent.InsertChild(Index,Child);
  inc(FCount);
//  Child.FRoot:=FRoot;
  result:=Child;
  if Child.FParent.FullyExpanded and Child.FParent.Expanded then
  begin
    FOwner.FClearVis:=true;
    FOwner.FClearAll:=true;
    FOwner.FVisUpdated:=true;
  end;
  FOwner.FUpdated:=true;
  FOwner.FRangeUpdate := true;
  FOwner.IsUpdating:=false;
end;

function TElTreeItems.AddItem;
var
    Child :TElTreeItem;
begin
  FOwner.IsUpdating:=true;
  Child:=CreateItem(FOwner);
  Child.FList:=self;
  Child.FParent:=Parent;
  if Child.FParent=nil then Child.FParent:=FRoot;
  Child.FParent.AddChild(Child);
  inc(FCount);
  result:=Child;
  if Child.FParent.Expanded and Child.FParent.FullyExpanded then
  begin
    FOwner.FClearVis:=true;
    FOwner.FVisUpdated:=true;
  end;
  FOwner.FUpdated:=true;
  FOwner.FRangeUpdate := true;
  FOwner.IsUpdating := false;
end;

procedure TElTreeItems.Assign;
begin
  if Source is TElTreeItems then
  begin
    if FOwner<>nil then FOwner.IsUpdating:=true;
    FRoot.Assign(TElTreeItems(Source).FRoot);
    FCount:=TElTreeItems(Source).FCount;
    if FOwner<>nil then with FOwner do
    begin
      FHRange:=-1;
      FUpdated:=true;
      FVisUpdated:=true;
      FClearVis:=true;
      FClearAll:=true;
      IsUpdating:=false;
    end;
  end else inherited;
end;

procedure TElTreeItems.Clear;
begin
  if FOwner<>nil then FOwner.IsUpdating:=true;
  FRoot.Clear;
  if FOwner<>nil then with FOwner do
  begin
    FHRange:=-1;
    FVisUpdated:=true;
    FClearVis:=true;
    FClearAll:=true;
    FUpdated:=true;
    IsUpdating:=false;
  end;
end;

function TElTreeItems.CreateItem;
begin
  result:=TElTreeItem.Create(FOwner);
end;

procedure TElTreeItems.ReadData;
begin
  if FOwner<>nil then FOwner.IsUpdating:=true;
  FRoot.Clear;
  FRoot.ReadData(Stream);
  if FOwner<>nil then with FOwner do
  begin
    FHRange:=-1;
    FVisUpdated:=true;
    FClearVis:=true;
    FClearAll:=true;
    FUpdated:=true;
    IsUpdating:=false;
  end;
end;

procedure TElTreeItems.WriteData;
var TSI:TElTreeItem;
begin
  TSI:=nil;
  if FOwner<>nil then
  begin
    TSI:=FOwner.ItemFocused;
    FOwner.ItemFocused:=nil;
  end;
  FRoot.WriteData(Stream);
  if FOwner<>nil then
  begin
    FOwner.ItemFocused:=TSI;
  end;
end;

procedure TElTreeItems.SaveToStream;
begin
  WriteData(Stream);
end;

procedure TElTreeItems.LoadFromStream;
begin
  ReadData(Stream);
end;

procedure TElTreeItems.SaveToFile;
var T:TFileStream;
begin
  T:=nil;
  try
    T:=TFileStream.Create(FileName,fmCreate or fmShareDenyWrite);
    SaveToStream(T);
  finally
    T.Free;
  end;
end;

procedure TElTreeItems.LoadFromFile;
var T:TFileStream;
begin
  T:=nil;
  try
    T:=TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
    LoadFromStream(T);
  finally
    T.Free;
  end;
end;

// Changed 10/27/98 by EM
function TElTreeItems.Add(Item : TElTreeItem; Text : string) : TElTreeItem;  { public }
begin
  if Item = nil
    then result := InsertItem(0, nil)
    else result := AddItem(Item.FParent);
  Result.Text := Text;
end;  { Add }

function TElTreeItems.AddChild(Item : TElTreeItem; Text : string) : TElTreeItem;  { public }
begin
  result := AddItem(Item);
  Result.Text := Text;
end;  { AddChild }

function TElTreeItems.AddChildFirst(Item : TElTreeItem; Text : string) : TElTreeItem;  { public }
begin
  result := InsertItem(0, Item);
  Result.Text := Text;
end;  { AddChildFirst }

function TElTreeItems.AddChildObject(Item : TElTreeItem; Text : string; Ptr : pointer) : TElTreeItem;  { public }
begin
  Result := AddItem(Item);
  Result.Text := Text;
  Result.Data := Ptr;
end;  { AddChildObject }

function TElTreeItems.AddChildObjectFirst(Item : TElTreeItem; Text : string; Ptr : pointer) : TElTreeItem;  { public }
begin
  result := InsertItem(0, Item);
  Result.Text := Text;
  Result.Data := Ptr;
end;  { AddChildObjectFirst }

// Changed 10/27/98 by EM
function TElTreeItems.AddFirst(Item : TElTreeItem; Text : string) : TElTreeItem;  { public }
begin
  if Item = nil
    then result := InsertItem(0, nil)
    else result := InsertItem(0, Item.FParent);
  Result.Text := Text;
end;  { AddFirst }

// Changed 10/27/98 by EM
function TElTreeItems.AddObject(Item : TElTreeItem; Text : string; Ptr : pointer) : TElTreeItem;  { public }
begin
  if Item = nil
    then result := InsertItem(0, nil)
    else result := AddItem(Item.FParent);
  Result.Text := Text;
  Result.Data := Ptr;
end;  { AddObject }

// Changed 10/27/98 by EM
function TElTreeItems.AddObjectFirst(Item : TElTreeItem; Text : string; Ptr : pointer) : TElTreeItem;  { public }
begin
  if Item = nil
    then result := InsertItem(0, nil)
    else result := InsertItem(0, Item.FParent);
  Result.Text := Text;
  Result.Data := Ptr;
end;  { AddObjectFirst }

// Changed 10/28/98 by EM
function TElTreeItems.Insert(Item : TElTreeItem; Text : string) : TElTreeItem;  { public }
var i : integer;
    FParent : TElTreeItem;
begin
  if Item = nil then
  begin
    i := 0;
    FParent := nil;
  end else
  begin
    i := Item.Index;
    FParent := Item.FParent;
  end;
  result := InsertItem(i, FParent);
  Result.Text := Text;
end;  { Insert }

// Changed 10/28/98 by EM
function TElTreeItems.InsertObject(Item : TElTreeItem; Text : string; Ptr : pointer) : TElTreeItem;  { public }
var i : integer;
    FParent : TElTreeItem;
begin
  if Item = nil then
  begin
    i := 0;
    FParent := nil;
  end else
  begin
    i := Item.Index;
    FParent := Item.FParent;
  end;
  result := InsertItem(i, FParent);
  Result.Text := Text;
  Result.Data := Ptr;
end;  { InsertObject }

procedure TElTreeItems.Delete;  { public }
begin
  DeleteItem(Item);
end;  { Delete }

function TElTreeItems.GetFirstNode : TElTreeItem;  { public }
begin
  Result := Item[0];
end;  { GetFirstNode }

// Changed 10/27/98 by EM
procedure TElTreeItems.BeginUpdate;  { public }
begin
  FOwner.IsUpdating := true;
end;  { BeginUpdate }

// Changed 10/27/98 by EM
procedure TElTreeItems.EndUpdate;  { public }
begin
  FOwner.IsUpdating := false;
end;  { EndUpdate }


// Changed 10/28/98 by EM
function TElTreeItems.InsertAfterObject(Item : TElTreeItem; Text : String; Ptr : Pointer) : TElTreeItem;  { public }
var i : integer;
    FParent : TElTreeItem;
begin
  if Item = nil then
  begin
    i := 0;
    FParent := nil;
  end else
  begin
    i := Item.Index + 1;
    FParent := Item.FParent;
  end;
  result := InsertItem(i, FParent);
  Result.Text := Text;
  Result.Data := Ptr;
end;  { InsertAfterObject }

// Changed 10/28/98 by EM
function TElTreeItems.InsertAfter(Item : TElTreeItem; Text : String) : TElTreeItem;  { public }
var i : integer;
    FParent : TElTreeItem;
begin
  if Item = nil then
  begin
    i := 0;
    FParent := nil;
  end else
  begin
    i := Item.Index + 1;
    FParent := Item.FParent;
  end;
  result := InsertItem(i, FParent);
  Result.Text := Text;
end;  { InsertAfter }

function TElTreeItems.LookForItem(StartItem : TElTreeItem;
                         TextToFind : string;
                         DataToFind : pointer;
                         ColumnNum : integer;
                         LookForData,
                         CheckStartItem,
                         SubItemsOnly,
                         VisibleOnly,
                         NoCase: boolean) : TElTreeItem;

type TSRec = record
       TextToFind : String;
       ColumnNum  : integer;
       StartItem  : TElTreeItem;
       DataToFind : pointer;
       CheckStartItem,
       LookForData,
       SubItemsOnly : boolean;
       NoCase     : boolean;
       Result     : TElTreeItem;
     end;
     PSRec = ^TSRec;

             procedure IntFind(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean; 
                                          IterateData:pointer; Tree:TCustomElTree);
             var i : integer;
                 s : string;
             begin
               if (Item = PSRec(IterateData).StartItem) and (not PSRec(IterateData).CheckStartItem) then exit;
               if (PSRec(IterateData).SubItemsOnly) and (not Item.IsUnder(PSRec(IterateData).StartItem)) then
               begin
                 ContinueIterate := false;
                 exit;
               end;
               if PSRec(IterateData).LookForData then
               begin
                 if Item.FData = PSRec(IterateData).DataToFind then
                 begin
                   PSRec(IterateData).result := Item;
                   ContinueIterate := false;
                   exit;
                 end;
               end else
               begin
                 if PSRec(IterateData).ColumnNum = -1 then s := Item.Text else
                 begin
                   i := PSRec(IterateData).ColumnNum;
                   if i > Tree.MainTreeColumn then dec(i);
                   if Item.ColumnText.Count <= i
                      then s := ''
                      else s := Item.ColumnText[i];
                 end;
                 if PSRec(IterateData).NoCase then
                 begin
                   if CompareText(S, PSRec(IterateData).TextToFind) = 0 then
                   begin
                     PSRec(IterateData).result := Item;
                     ContinueIterate := false;
                     exit;
                   end;
                 end else
                 begin
                   if S = PSRec(IterateData).TextToFind then
                   begin
                     PSRec(IterateData).result := Item;
                     ContinueIterate := false;
                     exit;
                   end;
                 end;
               end;
             end;

var SRec : TSRec;
begin
  SRec.TextToFind := TextToFind;
  SRec.ColumnNum  := ColumnNum;
  SRec.StartItem  := StartItem;
  SRec.LookForData := LookForData;
  SRec.DataToFind := DataToFind;
  SRec.NoCase := NoCase;
  SRec.CheckStartItem := CheckStartItem;
  SRec.SubItemsOnly := SubItemsOnly;
  SRec.Result := nil;
  try
    IterateFrom(VisibleOnly, @IntFind, @SRec, StartItem);
  finally
    Result := SRec.Result;
  end;
end;

function TElTreeItems.LookForItemEx(StartItem : TElTreeItem; ColumnNum : integer; CheckStartItem, SubItemsOnly,
                                    VisibleOnly : boolean; SearchDetails : pointer; CompareProc : TElLookupCompareProc) : TElTreeItem;

type TSRec = record
       ColumnNum  : integer;
       StartItem  : TElTreeItem;
       CheckStartItem,
       SubItemsOnly : boolean;
       Result     : TElTreeItem;
       CompProc   : TElLookupCompareProc;
       SearchDetails: pointer;
     end;
     PSRec = ^TSRec;

             procedure IntFind(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                          IterateData:pointer; Tree:TCustomElTree);
             var b : boolean;

             begin
               if (Item = PSRec(IterateData).StartItem) and (not PSRec(IterateData).CheckStartItem) then exit;
               if (PSRec(IterateData).SubItemsOnly) and (not Item.IsUnder(PSRec(IterateData).StartItem)) then
               begin
                 ContinueIterate := false;
                 exit;
               end;
               b := PSRec(IterateData).CompProc(Item, PSRec(IterateData).SearchDetails);
               if b then
               begin
                 PSRec(IterateData).result := Item;
                 ContinueIterate := false;
                 exit;
               end;
             end;

var SRec : TSRec;
begin
  SRec.ColumnNum  := ColumnNum;
  SRec.StartItem  := StartItem;
  SRec.CheckStartItem := CheckStartItem;
  SRec.SubItemsOnly := SubItemsOnly;
  SRec.CompProc := CompareProc;
  SRec.SearchDetails := SearchDetails;
  SRec.Result := nil;
  try
    IterateFrom(VisibleOnly, @IntFind, @SRec, StartItem);
  finally
    Result := SRec.Result;
  end;
end;

// ****************************************************************************
// TCustomElTree methods

// ****************************************************************************
constructor TCustomElTree.Create;
begin
  inherited;
  ControlStyle :=
{$IFDEF VER90}
  [csClickEvents, csOpaque, csDoubleClicks, csCaptureMouse,
   csDisplayDragImage, csReplicatable];
{$ELSE}
  [csClickEvents, csOpaque, csDoubleClicks, csCaptureMouse,
   csDisplayDragImage, csReplicatable, csReflector];
{$ENDIF}
  FVisible:=TElList.Create;
  FItems:=CreateItems;
  BMP:=TBitmap.Create;
  BoundsRect:=Rect(100,100,200,200);
  FHeader:=CreateHeader;
  with FHeader do
  begin
    OnSectionDraw:=DoHeaderDraw;
    OnResize:=HeaderResizeHandler;
    OnSectionResize:=OnHeaderSectionResize;
    OnSectionClick:=OnHeaderSectionClick;
    OnSectionDelete:=OnHeaderSectionDelete;
    OnMouseDown:=OnHeaderMouseDown;
    OnSectionResizing:=OnHeaderSectionTrack;
    OnSectionMove := OnHeaderSectionMove;
    OnSectionShowHide := OnHeaderSectionVisChange;
    OnSectionChange := OnHeaderSectionChange;
    OnSectionCreate := OnHeaderSectionCreate;
    OnHeaderLookup := OnHeaderSectionLookup;
    OnHeaderLookupDone := OnHeaderSectionLookupDone;
    OnSectionExpand := Self.OnHeaderExpand;
    OnSectionCollapse := Self.OnHeaderCollapse;
    Tracking:=FHeaderHotTrack;
    Parent:=self;
    Align:=alNone;
    Visible:=true;
    AllowDrag:=false;
    ControlStyle:=ControlStyle + [csNoDesignVisible];
  end;    //with
  FHeader.Font.Assign(Font);
  FHeader.Visible := False;
  Font.OnChange:=OnFontChange;

  PlusBmp:=TBitmap.Create;
  MinusBmp:=TBitmap.Create;
  PlusBmp.LoadFromResourceName(HInstance,'PlusBmp');
  MinusBmp.LoadFromResourceName(HInstance,'MinusBmp');

  FBorderStyle := bsSingle;
  FBkColor:= clWindow;
  FTextColor:=clWindowText;
  FCanEdit := true;
  FHLines := false;
  FVLines := false;
  FRowSelect := false;
  FFullRowSelect := true;
  FSelMode := smUsual;
  FShowHintMode := shmLong;
  FMultiSelect := true;
  FHideSelect := false;
  FODFollowCol := true;
  FODMask := '~~@~~';
  FRowSelect := True;
  FTracking:=true;
  FShowButtons := True;
  FShowImages := True;
  FShowLines := True;
  FDrawFocusRect := true;
  TabStop := True;
  FAutoLineHeight:=true;
  FLineHeight:=DefineLineHeight;
  FDragTrgDraw:=SelColorRect;
  FHRange := -1;
  FPathSeparator := '\';
  FSortMode := smNone;
  FImageChangeLink := TChangeLink.Create;
  FImageChangeLink.OnChange := ImageListChange;
  DragMode:=dmManual;
  FHideHintOnMove := true;
  FPlusPicture := TBitmap.Create;
  FPlusPicture.OnChange := OnSignChange;
  FMinusPicture := TBitmap.Create;
  FMinusPicture.OnChange := OnSignChange;
  FSelectColumn := -1;
  FDragType := dtDelphi;
  FLinesColor := clBtnFace;
  FLinesStyle := psSolid;
{$IFNDEF LITE}
  FStoragePath := 'Tree';
{$ENDIF}  
end;

destructor TCustomElTree.Destroy;
begin
  Destroying:=true;
  if FInpEdit<>nil then
  begin
    EndEdit(true);
    if FFakePopup <> nil then FFakePopup.Free; FFakePopup := nil;
  end;
  FHeader.Free;
  MinusBmp.Free;
  PlusBmp.Free;
  FPlusPicture.Free;
  FMinusPicture.Free;
  FItems.Free;
  FVisible.Free;
  if Font<>nil then Font.OnChange:=nil;
  if FImages<>nil then FImages.UnregisterChanges(FImageChangeLink);
  if FImages2<>nil then FImages2.UnregisterChanges(FImageChangeLink);
  FImageChangeLink.Free;
  BMP.Free;
  inherited Destroy;
end;

procedure TCustomElTree.CreateWindowHandle(const Params : TCreateParams);
begin
  inherited CreateWindowHandle(Params);
{$IFNDEF VER90}
{$IFNDEF LITE}
  if Assigned(FDropTarget) then FDropTarget.Target := Self;
{$ENDIF}
{$ENDIF}
end;  {CreateWindowHandle}

procedure TCustomElTree.DefineHRange;
    procedure CalcWidth(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                          IterateData:pointer; Tree:TCustomElTree);
    var w:integer;
        function GetWidth(Item:TElTreeItem):integer;
        var SText : string;
        begin
          SText:=Item.FText;
          result:=Tree.Canvas.TextWidth(SText)+(ItemExt div 3 + 3);
          if (Tree.FShowRoot and Tree.FShowLines) or Tree.FShowButtons then inc(Result, ItemExt);
          if Tree.FShowCheckBoxes and Item.FShowCheckBox then inc(result, ItemExt);
          inc(result, (Item.Level) * ItemExt);
          if (Tree.FShowImages) then
          begin
            if (Tree.FImages<>nil) then inc(result, Tree.FImages.Width);
            inc(result, ItemExt div 3);
            if (Tree.FImages2<>nil) then inc(result, Tree.FImages2.Width) else
               if (Tree.FImages<>nil) then inc(result, Tree.FImages.Width);
          end;
        end;

    begin
      w := GetWidth(Item);
      if w > Tree.FHRange then Tree.FHRange:=w;
    end;

begin
  FItems.Iterate(true, @CalcWidth, nil);
end;

function TCustomElTree.GetSelCount;
var L : TElList;
begin
  result := FSelCount;
  exit;
  if MultiSelect then
  begin
    L := TElList.Create;
    AllSelected(L);
    result := L.Count;
    L.Free;
  end else
  begin
    if ItemFocused = nil then result :=0 else result := 1;
  end;
end;

procedure TCustomElTree.CreateParams;
const
{$IFNDEF D_4_UP}
  BorderStyles: array[TBorderStyle] of Longint = (0, WS_BORDER);
{$ELSE}
  BorderStyles: array[TBorderStyle] of DWord = (0, WS_BORDER);
{$ENDIF}
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or BorderStyles[FBorderStyle];
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
    with Params.WindowClass do
      style := style and not (CS_HREDRAW or CS_VREDRAW);
    if RightAlignedText then
    begin
      ExStyle := ExStyle or WS_EX_RTLREADING or WS_EX_LEFTSCROLLBAR or WS_EX_RIGHT;
    end;
  end;
end;

procedure TCustomElTree.CMCtl3DChanged(var Message: TMessage);
begin
  if NewStyleControls and (FBorderStyle = bsSingle) then RecreateWnd;
  inherited;
end;

function TCustomElTree.GetItemAt;
var dX, i : integer;
begin
  result:=GetItemAtY(Y);
  ItemPart:=ipInside;
  if Result = nil then exit;
  if RightAlignedTree then
  begin
    dX:=X + FHPos;
    if FHeader.Visible then
      for i:=0 to FHeader.Sections.Count-1 do
      begin
        if (dX>=FHeader.Sections[i].Left) and (dX<FHeader.Sections[i].Right) then
        begin
          HitColumn:=i;
          if HitColumn<>FMainTreeCol then
          begin
            ItemPart:=ipColumn;
            exit;
          end else
          begin
  //          dec(dX, FHeader.Sections[i].Left);
            break;
          end;
        end; // if
      end; // for
    if (dX>=Result.FTextLeft) and (dX<=Result.FTextRight) then
    begin
      ItemPart:=ipMainText;
      HitColumn := MainTreeColumn;
      exit;
    end; // else ItemPart := ipInside;
    if Result.FImageDrawn then
    begin
      i:=Result.FTextRight + FImages.Width + (ItemExt div 3);
      if (dX >= i - FImages.Width) and ( dX < i) then
      begin
        ItemPart:=ipPicture;
        HitColumn := MainTreeColumn;
        exit;
      end;
    end else i:= Result.FTextRight + (ItemExt div 3);
    if Result.FImageDrawn2 then
    begin
      if Result.FImageDrawn then inc(i, ItemExt div 3);
      if FImages2 <> nil then
      begin
        inc(i, FImages2.Width);
        if (dX >= i - FImages2.Width) and ( dX < i) then
        begin
          ItemPart  := ipPicture;
          HitColumn := MainTreeColumn;
          exit;
        end;
      end else
      begin
        inc(i, FImages.Width);
        if (dX >= i - FImages.Width) and ( dX < i) then
        begin
          ItemPart  := ipPicture;
          HitColumn := MainTreeColumn;
          exit;
        end;
      end;
    end;
    if FShowCheckBoxes and (Result.FShowCheckBox) then
    begin
      if (dX >= i) and ( dX < i + ItemExt) then
      begin
        ItemPart:=ipCheckBox;
        HitColumn := MainTreeColumn;
        exit;
      end;
      inc(i, ItemExt);
    end;
    if not (ShowButtons) or (not (Result.HasVisibleChildren or Result.ForceButtons)) then exit;
    if (dX >= i) and ( dX < i + ItemExt) then
    begin
      ItemPart:=ipButton;
      HitColumn := MainTreeColumn;
      exit;
    end;
  end else
  begin
    dX:=X+FHPos;
    if FHeader.Visible then
      for i:=0 to FHeader.Sections.Count-1 do
      begin
        if (dX>=FHeader.Sections[i].Left) and (dX<FHeader.Sections[i].Right) then
        begin
          HitColumn:=i;
          if HitColumn<>FMainTreeCol then
          begin
            ItemPart:=ipColumn;
            exit;
          end else
          begin
  //          dec(dX, FHeader.Sections[i].Left);
            break;
          end;
        end; // if
      end; // for
    if (dX>=Result.FTextLeft) and (dX<=Result.FTextRight) then
    begin
      ItemPart:=ipMainText;
      HitColumn := MainTreeColumn;
      exit;
    end; // else ItemPart := ipInside;
    if Result.FImageDrawn then
    begin
      i:=Result.FTextLeft - FImages.Width -(ItemExt div 3);
      if (dX< i + FImages.Width) and ( dX >= i) then
      begin
        ItemPart:=ipPicture;
        HitColumn := MainTreeColumn;
        exit;
      end;
    end else i:=Result.FTextLeft-(ItemExt div 3);
    if Result.FImageDrawn2 then
    begin
      if Result.FImageDrawn then dec(i, ItemExt div 3);
      if FImages2 <> nil then
      begin
        dec(i, FImages2.Width);
        if (dX < i + FImages2.Width) and ( dX>=i) then
        begin
          ItemPart  := ipPicture;
          HitColumn := MainTreeColumn;
          exit;
        end;
      end else
      begin
        dec(i, FImages.Width);
        if (dX < i + FImages.Width) and ( dX>=i) then
        begin
          ItemPart  := ipPicture;
          HitColumn := MainTreeColumn;
          exit;
        end;
      end;
    end;
    if FShowCheckBoxes and (Result.FShowCheckBox) then
    begin
      if (dX<i) and ( dX>=i-ItemExt) then
      begin
        ItemPart:=ipCheckBox;
        HitColumn := MainTreeColumn;
        exit;
      end;
      dec(i, ItemExt);
    end;
    if not (ShowButtons) or (not (Result.HasVisibleChildren or Result.ForceButtons)) then exit;
    if (dX<i) and ( dX>=i-ItemExt) then
    begin
      ItemPart:=ipButton;
      HitColumn := MainTreeColumn;
      exit;
    end;
  end;
  ItemPart:=ipInside;
end;

procedure TCustomElTree.SetImages;
var i:integer;
begin
  if FImages=value then exit;
  IsUpdating:=true;
  if FImages <> nil then FImages.UnRegisterChanges(FImageChangeLink);
  FImages:=Value;
  if FImages<>nil then FImages.RegisterChanges(FImageChangeLink);
  i:=DefineLineHeight;
  if FAutoLineHeight and (i<>FLineHeight) then
  begin
    FLineHeight:=i;
    FClearVis:=true;
    FClearAll:=true;
  end;
  if not FHeader.Visible then
  begin
    FHRange := -1;
    DefineHRange;
    if RightAlignedTree then
    begin
      FRightAlignedTree := false;
      RightAlignedTree := true;
    end;
  end;
  FRangeUpdate := true;
  FUpdated:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetLineStyle;
begin
  if FShowLines=value then exit;
  IsUpdating:=true;
  FShowLines:=value;
  if not FHeader.Visible then FHRange := -1;
  FRangeUpdate := true;
  FUpdated:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetRootStyle;
begin
  if FShowRoot=value then exit;
  IsUpdating:=true;
  FShowRoot:=value;
  if not FHeader.Visible then FHRange := -1;
  FUpdated:=true;
  FRangeUpdate := true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetImagesStyle;
var i:integer;
begin
  if FShowImages=value then exit;
  IsUpdating:=true;
  FShowImages:=value;
  if not FHeader.Visible then FHRange := -1;
  i:=DefineLineHeight;
  if FAutoLineHeight and (i<>FLineHeight) then
  begin
    FLineHeight:=i;
    FClearVis:=true;
    FClearAll:=true;
  end;
  FRangeUpdate := true;
  FUpdated:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.DoSetTopIndex;
var
    ItemsDrawn : integer;

begin
  if ((value>=FItems.Count) and (value>0)) or (value<0) then
    // Raise EElTreeError.Create(GetErrorMsg(STExOutOfBounds));
    value:=0;
  if TotalVisCount < GetVisCount then value:=0;
  FTopIndex:=value;
  if (FTopIndex=0) and (FItems.Count=0) then FBottomIndex:=0
  else
  begin
    ItemsDrawn:=0;
    FillVisFwd(FTopIndex, GetVisCount, ItemsDrawn);
    FBottomIndex:=FTopIndex+FVisible.Count;
  end;
end;

procedure TCustomElTree.ImageListChange;
var i:integer;
begin
  if FPainting then exit;
  IsUpdating:=true;
  if AutoLineHeight then
  begin
    i:=DefineLineHeight;
    if i<>FLineHeight then
    begin
      FClearVis:=true;
      if i<FLineHeight then FClearAll:=true;
      FLineHeight:=i;
    end;
  end;
  FUpdated:=true;
  FRangeUpdate := true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetBorderStyle;
begin
  if FBorderStyle=value then exit;
  FBorderStyle:=value;
  RecreateWnd;
end;

procedure TCustomElTree.SetButtonStyle;
begin
  if FShowButtons=value then exit;
  IsUpdating:=true;
  FShowButtons:=value;
  if not FHeader.Visible then FHRange := -1;
  FUpdated:=true;
  FRangeUpdate := true;
  IsUpdating:=false;
end;
                    
procedure TCustomElTree.SetUpdating;
var i:integer;
begin
  if FProcUpdate then exit;
  if value=true then
  begin
    FUpdating:=value;
    inc(FUpdateCount);
  end else if FUpdateCount>0 then dec(FUpdateCount);
  if FUpdateCount=0 then FUpdating:=false;
  if FUpdating or Destroying then exit;
  FProcUpdate:=true;
  if FAutoLineHeight then
  begin
    i:=DefineLineHeight;
    if i<>FLineHeight then
    begin
      FClearVis:=true;
      FLineHeight:=i;
    end;
  end;
  if (FHRange=-1) and (not FHeader.Visible) then DefineHRange;
  if FUpdated then
  begin
    if FVisUpdated then // recount visibles
    begin
      TotalVisCount := GetTotalVisCount;
      FVisUpdated := false;
    end;
    IgnoreResize:=true;
    if (not (csLoading in ComponentState)) and (HandleAllocated) then
    begin
      UpdateScrollBars;
      if FHeader.Visible and (FHeader.Width<>FHPos+ClientWidth+1)
         then FHeader.Width:=FHPos+ClientWidth+1;
    end;
    IgnoreResize:=false;
    if FClearVis then // refresh FVisible
    begin
      DoSetTopIndex(FTopIndex);
      FClearVis:=false;
    end;
    IgnoreResize:=true;
    if (not (csLoading in ComponentState)) and (HandleAllocated) then UpdateScrollBars;
    IgnoreResize:=false;
    UpdateElTree;
    FUpdated:=false;
  end;
  FProcUpdate:=false;
end;

function TCustomElTree.GetTotalVisCount;
var IntCount:integer;
type pinteger = ^integer;

  procedure IntTotalVisCount(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                               IterateData : pointer; Tree:TCustomElTree);
  begin
    pinteger(IterateData)^ := Index;
  end;

begin
  IntCount:=-1;
  FItems.Iterate(true, @IntTotalVisCount, @IntCount);
  result:=IntCount + 1;
end;
{
function TCustomElTree.VisToIndex;
type PGIRec = ^TGIRec;
     TGIRec = record
       VisIndex,
       Visibles:integer;
       j:integer;
     end;
var GIRec : TGIRec;

    procedure IntVisToIndex(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                              IterateData:pointer; Tree:TCustomElTree);
    var i : integer;
    begin
      if Item.FullyExpanded then inc(PGIRec(IterateData)^.Visibles);
      if PGIRec(IterateData)^.Visibles=PGIRec(IterateData)^.VisIndex then
      begin
        PGIRec(IterateData)^.j:=Index;
        ContinueIterate:=false;
      end;
    end;

begin
  GIRec.VisIndex:=Index;
  GIRec.Visibles:=-1;
  GIRec.j:=-1;
  FItems.Iterate(false,@IntVisToIndex,@GIRec);
  result:=GIRec.j;
end;
}
{function TCustomElTree.IndexToVis;
type PGIRec = ^TGIRec;
     TGIRec = record
       VisIndex,
       Visibles:integer;
       j:integer;
     end;
var GIRec : TGIRec;

    procedure IntIndexToVis(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                              IterateData:pointer; Tree:TCustomElTree);
    var i : integer;
    begin
      if Item.FullyExpanded then inc(PGIRec(IterateData)^.Visibles);
      if Index=PGIRec(IterateData)^.VisIndex then
      begin
        PGIRec(IterateData)^.j:=PGIRec(IterateData)^.Visibles;
        ContinueIterate:=false;
      end;
    end;

begin
  GIRec.VisIndex:=Index;
  GIRec.Visibles:=-1;
  GIRec.j:=-1;
  FItems.Iterate(false,@IntIndexToVis,@GIRec);
  result:=GIRec.j;
end;
}
procedure TCustomElTree.UpdateScrollBars;
var b: boolean;

    function UpdHBar:boolean;
    var b1:boolean;
        ScrollInfo : TScrollInfo;
        CurPos : integer;

    begin
      b1:=FHScrollVisible;
      GetScrollInfo(Handle,SB_Horz,ScrollInfo);
      if FHeader.Visible then
      begin
        if FHeader.Sections.Count>0 then
           ScrollInfo.nMax:=FHeader.SectionsWidth
           else ScrollInfo.nMax:=0;
        FHRange:=ScrollInfo.nMax;
      end
      else ScrollInfo.nMax:=FHRange;
      if ScrollInfo.nMax<0 then ScrollInfo.nMax:=0;
      FHScrollVisible:=(ScrollInfo.nMax>ClientWidth) and (not FHideHorzScrollBar);
      if (not FHScrollVisible) then
      begin
        ScrollInfo.nMax:=0;
        FHPos:=0;
        FHeader.Left:=0;
      end;
      result:=FHScrollVisible<>b1;
      CurPos:=FHPos;
      ScrollInfo.fMask:=SIF_ALL;
      ScrollInfo.nMin:=0;
      ScrollInfo.cbSize:=SizeOf(ScrollInfo);
      ScrollInfo.nPos:=CurPos;
      ScrollInfo.nTrackPos:=CurPos;

      if FHScrollVisible then ScrollInfo.nPage:=ClientWidth else ScrollInfo.nPage:=1;
      SetScrollInfo(Handle,SB_HORZ,ScrollInfo,true);
    end;

    function UpdVBar:boolean;
    var b : boolean;
        CHeight,i  : integer;
        ScrollInfo : TScrollInfo;

    begin
      ScrollInfo.nMax:= TotalVisCount-1;
      ScrollInfo.fMask:=SIF_ALL;
      ScrollInfo.nMin:=0;
      ScrollInfo.cbSize:=SizeOf(ScrollInfo);
      ScrollInfo.nPos:=FTopIndex;
      ScrollInfo.nTrackPos:=FTopIndex;
      B:=FVScrollVisible;
      I:=FLineHeight;
      CHeight:=ClientHeight;
      if FHeader.Visible then dec(CHeight, GetHeaderHeight);
      FVScrollVisible := (ScrollInfo.nMax >= (CHeight div i)) and (not FHideVertScrollBar);
      if (FTopIndex>0) and (not FHideVertScrollBar) then FVScrollVisible:=true;
      ScrollInfo.nPage:= (GetVisCount-1);
      if not (FVScrollVisible) then
      begin
        ScrollInfo.nMax:=0;
        ScrollInfo.nPage:=1;
      end;

      SetScrollInfo(Handle,SB_VERT,ScrollInfo,true);

      result:=FVScrollVisible<>b;
    end;

begin
  b:=true;
  while b do
    if UpdHBar or UpdVBar then b:=true else b:=false;
end;

procedure TCustomElTree.SetHLines;
var i:integer;
begin
  if FHLines=value then exit;
  IsUpdating:=true;
  FHLines:=value;
  if FAutoLineHeight then
  begin
    i:=DefineLineHeight;
    if i<>FLineHeight then
    begin
      FLineHeight:=i;
      FClearVis:=true;
      FClearAll:=true;
    end;
  end;
  FRangeUpdate := true;
  FUpdated:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetVLines;
begin
  if FVLines=value then exit;
  IsUpdating:=true;
  FVLines:=value;
  FClearAll:=true;
  FUpdated:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.OnHeaderSectionResize;
begin
  IsUpdating:=true;
  DoOnColumnResize(Section.Index);
  FUpdated:=true;
  FClearAll:=true;
  IsUpdating:=false;
  if FHeader.SectionsWidth<ClientWidth then FHeader.Left:=0;
end;

procedure TCustomElTree.OnHeaderSectionLookup(Sender : TObject; Section : TElHeaderSection; var Text : string);

type TSRec = record
       Text   : PChar;
       ColNum : integer;
     end;
     PSRec = ^TSRec;

var SRec : TSrec;
    TI : TElTreeItem;

   function IntCompare(Item : TElTreeItem; SearchDetails :Pointer) : boolean;
   var i  : integer;
       AT : String;
   begin
     i := PSRec(SearchDetails).ColNum;
     if Item.FOwner.MainTreeColumn < i then dec(i);
     if Item.ColumnText.Count <= i then AT := '' else
     begin
       if I = Item.FOwner.MainTreeColumn
          then AT := Uppercase(Item.Text)
          else AT := Uppercase(Item.ColumnText[i]);
     end;
     result := Pos(Uppercase(StrPas(PSRec(SearchDetails).Text)), AT) = 1;
   end;

begin
  TriggerHeaderLookupEvent(Section, Text);
  if AutoLookup then
  begin
    SRec.Text := PChar(Text);
    SRec.ColNum := Section.Index;
    TI := Items.LookForItemEx(nil, Section.Index, false, false, false, @SRec, @IntCompare);
    if TI <> nil then
    begin
      EnsureVisible(TI);
      ItemFocused := TI;
    end;
  end;
end;

procedure TCustomElTree.OnHeaderExpand;
begin
  TriggerHeaderSectionExpandEvent(Section);
  try
    IsUpdating := true;
    FUpdated := true;
    FClearAll:=true;
    if FHeader.Visible then FHeader.Repaint;
  finally
    IsUpdating := false;
  end;
end;

procedure TCustomElTree.OnHeaderCollapse;
begin
  TriggerHeaderSectionCollapseEvent(Section);
  try
    IsUpdating := true;
    FUpdated := true;
    FClearAll:=true;
    if FHeader.Visible then FHeader.Repaint;
  finally
    IsUpdating := false;
  end;
end;

procedure TCustomElTree.OnHeaderSectionLookupDone(Sender : TObject; Section : TElHeaderSection; Text : string; Accepted : boolean);
begin
  TriggerHeaderLookupDoneEvent(Section, Text, Accepted);
end;

procedure TCustomElTree.OnHeaderMouseDown;
begin
  if FInpEdit<>nil then EndEdit(true);
end;

procedure TCustomElTree.OnHeaderSectionClick;
begin
  if (FSortMode=smClick) or (FSortMode=smAddClick) then
  begin
    IsUpdating:=true;
    SortSection:=Section.Index;
    if Section.SortMode=hsmNone then
      if SortDir=sdAscend then Section.SortMode:=hsmAscend else Section.SortMode:=hsmDescend;
    if Section.SortMode=hsmAscend then Section.SortMode:=hsmDescend else Section.SortMode:=hsmAscend;
    case Section.FieldType of
      sftText:   SortType:=stText;
      sftNumber: SortType:=stNumber;
      sftFloating: SortType:=stFloating;
      sftDateTime: SortType:=stDateTime;
      sftDate:   SortType:=stDate;
      sftTime:   SortType:=stTime;
      else       SortType:=stCustom;
    end;
    FItems.FRoot.Sort(true);
    FUpdated:=true;
    FRangeUpdate := true;
    FVisUpdated:=true;
    IsUpdating:=false;
  end;
  DoColumnClick(Section.Index);
end;

procedure TCustomElTree.OnSignChange(Sender:TObject);
var i : integer;
begin
  if not FCustomPlusMinus then exit;
  IsUpdating := true;
  if AutoLineHeight then
  begin
    i := DefineLineHeight;
    if i<>FLineHeight then
    begin
      FClearVis:=true;
      if i<FLineHeight then FClearAll:=true;
      FLineHeight:=i;
    end;
    FVisUpdated := true;
  end;
  FRangeUpdate := true;
  FUpdated:=true;
  IsUpdating := false;
end;

procedure TCustomElTree.OnFontChange;
var i:integer;
begin
  if FPainting then exit;
  IsUpdating := true;
  ParentFont := false;
  Canvas.Font.Assign(Font);
  FHeader.Font.Assign(Font);
  if AutoLineHeight then
  begin
    i:=DefineLineHeight;
    if i<>FLineHeight then
    begin
      FClearVis:=true;
      if i<FLineHeight then FClearAll:=true;
      FLineHeight:=i;
    end;
  end;
  FUpdated:=true;
  FRangeUpdate := true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetHeaderHotTrack;
begin
  FHeaderHotTrack:=value;
  FHeader.Tracking:=value;
end;

procedure TCustomElTree.SetRowSelect;
begin
  IsUpdating:=true;
  FRowSelect:=value;
  if FRowSelect = false then FullRowSelect := false;
  FClearAll:=true;
  FClearVis:=true;
  FUpdated:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetMultiSelect;
var i:integer;
    TSI: TElTreeItem;
    SIList : TElList;

begin
  IsUpdating:=true;
  FUpdated:=true;
  FMultiSelect := value;
  TSI:=ItemFocused;
  if Value=true then if FSelected<>nil then begin FSelected.Selected:=true; end else
  begin
    SIList:=TElList.Create;
    AllSelected(SIList);
    for i:=0 to SIList.Count-1 do TElTreeItem(SIList[i]).Selected:=false;
    SIList.Free;
    if TSI<>nil then
    begin
      TSI.Focused:=true;
      TSI.Selected:=true;
    end;
  end;
  IsUpdating:=false;
end;

procedure TCustomElTree.DoSetSelected;
begin
  if FMultiSelect then
  begin
    if Value <> nil then with Value do
    begin
      if stsSelected in FState then
      begin
        dec(FSelCount);
        Exclude(FState, stsSelected);
      end else
      begin
        inc(FSelCount);
        Include(FState, stsSelected);
      end;
      TriggerItemSelectedChangeEvent(Value);
      Value.UpdateItem;
    end
  end
  else
  begin
    if FSelected<>nil then with FSelected do
    begin
      if stsSelected in FState then dec(FSelCount);
      Exclude(FState, stsSelected);
      TriggerItemSelectedChangeEvent(FSelected);
      FSelected.UpdateItem;
      FSelected:=nil;
    end;
    if Value<>nil then with value do
    begin
      if not (stsSelected in FState) then inc(FSelCount);
      Include(FState, stsSelected);
      TriggerItemSelectedChangeEvent(Value);
      Value.UpdateItem;
      Self.FSelected:=Value;
    end;
  end;
end;

procedure TCustomElTree.SetFocused;
var FOld : TElTreeItem;
    modu : integer;
begin
  if FFocused <> Value then
  begin
    IsUpdating := true;
    if FFocused<>nil then with FFocused do
    begin
      Exclude(FState, stsFocused);
      RedrawItem(true);
      if FAutoExpand and
         (value <> nil) then
         begin
           FOld := FFocused;
           while (FOld <> nil) and (not Value.IsUnder(FOld)) do
           begin
             FOld.Collapse(false);
             FOld := FOld.Parent;
           end;    // while
           FUpdated := true;
         end;
    end;
    FFocused:=value;
    DoSetSelected(FFocused);
    if FFocused<>nil then
    begin
      Include(FFocused.FState, stsFocused);
      FFocused.RedrawItem(true);
      if FAutoExpand then FFocused.Expand(false);
      if not (csDestroying in ComponentState) then
      begin
        if FHeader.Visible
           then modu := (ClientHeight - FHeader.Height) mod FLineHeight
           else modu := ClientHeight mod FLineHeight;
        if (FVisible.IndexOf(FFocused) = FVisible.Count - 1) and (modu > 0) then
          TopIndex := TopIndex + 1;
      end;
    end;
    DoItemFocused;
    FUpdated:=true;
    IsUpdating := false;
  end;
end;

function TCustomElTree.GetNextSelected;
type PGIRec = ^TGIRec;
     TGIRec = record
       PrevSel,CurSel:TElTreeItem;
       b:boolean;
     end;
var GIRec : TGIRec;
    procedure IntNextSel(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                           IterateData:pointer; Tree:TCustomElTree);
    begin
      if PGIRec(IterateData)^.b and (Item.Selected) then
      begin
        PGIRec(IterateData)^.CurSel:=Item;
        ContinueIterate:=false;
        exit;
      end;
      if PGIRec(IterateData)^.PrevSel=Item then PGIRec(IterateData)^.b:=true;
    end;

begin
  if not (FMultiSelect) then 
  begin
    result:=nil;
    exit;
  end;
  GIRec.CurSel:=nil;
  GIRec.PrevSel:=Prev;
  GIRec.b:= (Prev=nil);
  if FMultiSelect then FItems.Iterate(false,@IntNextSel,@GIRec);
  result:=GIRec.CurSel;
end;

procedure TCustomElTree.AllSelected(SelectedItems:TElList);
  procedure IntNextSel(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                         IterateData:pointer; Tree:TCustomElTree);
    begin
      if Item.Selected then TElList(IterateData).Add(Item);
    end;
begin
  FItems.Iterate(false,@IntNextSel,SelectedItems);
end;

procedure TCustomElTree.EnsureVisible;
var j:integer;
begin
 if Item = nil then exit;
 j:=FVisible.IndexOf(Item);
 if j<>-1 then exit;
 if not Item.FullyExpanded then Item.MakeFullyExpanded(true);
 IsUpdating:=true;
 DoSetTopIndex(Item.VisIndex);
 FUpdated:=true;
 FRangeUpdate:=true;
 FClearAll := true;
 IsUpdating:=false;
end;

procedure TCustomElTree.SetAutoLineHeight;
begin
  if FAutoLineHeight=value then exit;
  IsUpdating:=true;
  FAutoLineHeight:=value;
  if value=true then FLineHeight:=DefineLineHeight;
  FUpdated:=true;
  FClearAll:=true;
  FUpdated:=true;
  FClearVis:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetLineHeight;
begin
  if (FLineHeight=value) or FAutoLineHeight then exit;
  IsUpdating:=true;
  FUpdated:=true;
  FLineHeight:=value;
  FClearVis:=true;
  FClearAll:=true;
  if value=0 then FLineHeight:=DefineLineHeight;
  IsUpdating:=false;
end;

function TCustomElTree.DefineLineHeight;
var m: integer;
begin
  m:=0;
  if FShowImages then
  begin
    if (Images <>  nil) then m:=Images.Height;
    if (Images2 <> nil) then m := Max(m, Images2.Height);
  end;
  if ItemExt > m then m:=ItemExt;
  if Font.Height < 0
     then m := max(m, Abs(Font.Height) + 4)
     else m := max(m, Abs(Font.Height) + 2);
  result:=m + 2;
  if FBarStyle then inc(result, 2);
  if FHLines and not (FBarStyle) then inc(result, FDivLineWidth + 1);
end;

procedure TCustomElTree.WMSize;
begin
  inherited;
  if not (csLoading in ComponentState) and (not IgnoreResize) then Resize;
end;

procedure TCustomElTree.Resize;
begin
 if FHeader.Visible and (FHeader.Width<>FHPos+ClientWidth+1)
    then FHeader.Width:=FHPos+ClientWidth+1;
{$IFNDEF D_4_UP}
  if Assigned(FOnResize) then FOnResize(Self);
{$ENDIF}
  IsUpdating:=true;
  FVisUpdated:=true;
  FClearAll:=true;
  FClearVis:=true;
  FUpdated:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.RedrawTree;
var i,y, h,
    oy        : integer;
    R,r1,r4   : TRect;
    Item      : TElTreeItem;
    Sect      : TElHeaderSection;

    //TickCount : integer;
begin
 //TickCount := GetTickCount;

 if FHeader.Visible and (FHeader.Width<>FHPos+ClientWidth+1)
    then FHeader.Width:=FHPos+ClientWidth+1;

 if FHeader.Sections.Count<=MainTreeColumn then MainTreeColumn:=0;

 // redraw items
 for i:=0 to FVisible.Count-1 do
 begin
     R:=GetItemRect(i);
     if FHLines then dec(R.Bottom, FDivLineWidth-1);
     if IntersectRect(R4,R,Canvas.ClipRect) then
     begin
       r1:=r4;
       R1.Bottom:=r1.Top+LineHeight;
       if FHLines then dec(R1.Bottom, FDivLineWidth-1);
       if FHeader.Visible
         then r1.right:=FHeader.SectionsWidth else
         begin
           R.Right := Max(FHRange, FHPos+ClientWidth);
           r1.right:= Max(FHRange, FHPos+ClientWidth);
         end;

       Item:=TElTreeItem(FVisible[i]);
       if not Item.FUseStyles then
       begin
         Canvas.Font.Color:=FTextColor;
         Canvas.Pen.Color:=FTextColor;
       end;
       FCurBkColor := FBkColor;
       FCurTextColor := FTextColor;
       if FBarStyle then
       begin
         InflateRect(R, -1, -1);
         InflateRect(R1, -1, -1);
       end;
       DoRedrawItem(Item, R, R1);
       if FBarStyle then
       begin
         InflateRect(R, 1, 1);
         inc(R.Bottom);
         if FHeader.Visible then R.Right := -FHPos + FHeader.SectionsWidth;
         if not Item.Selected
            then DrawEdge(Canvas.Handle, R, BDR_RAISEDINNER, BF_RECT)
            else DrawEdge(Canvas.Handle, R, BDR_SUNKENOUTER, BF_RECT);
       end;
     end;
 end;
 // fill the rest of the window
 if IntersectRect(R4,ClientRect, Canvas.ClipRect) then
 begin
   Canvas.Brush.Color:=FBkColor;
   oy:=R4.top - 1;
   Y:=FVisible.Count*FLineHeight;
   if (FHeader.Visible) then inc(Y,GetHeaderHeight);
   if (R4.Top<Y) then R4.Top:=Y;
   Canvas.FillRect(R4);
   if FHeader.Visible then
   begin
     if FHeader.Sections.Count > 0 then
        r4.Left:= FHeader.SectionsWidth else r4.left:=0;
     dec(R4.Left,FHPos);
     R4.Top:=oy;
     Canvas.FillRect(R4);
   end;
 end;

 Canvas.Pen.Color:=clBtnFace;

 // Draw horizontal lines
 if not FBarStyle then
 begin
   if FHLines then
   begin
     if FHeader.Visible then
     begin
       y := GetHeaderHeight;
       h := FHeader.SectionsWidth - FHPos;
      end else
      begin
        Y:=0;
        h := ClientWidth;
      end;
     i:=0;
     if FVisible.Count>0 then
     begin
       Inc(Y, FLineHeight - 1);
       while Y<ClientHeight do
       begin
         if IntersectRect(R,Rect(0,Y, h, Y + FDivLineWidth), Canvas.ClipRect) then
         begin
            Canvas.MoveTo(R.Left, R.Top);
            Canvas.LineTo(R.Right, R.Top);
         end;
         Inc(Y, FLineHeight);
         inc(i);
         if i>=FVisible.Count then break;
       end;
     end;
   end;
  
   // Draw vertical lines
   if FVLines and (FHeader.Visible) and (FHeader.Sections.Count>0) then
   begin
     for i := 0 to FHeader.Sections.Count-1 do
     begin
       Sect := FHeader.Sections[i];
       if (not Sect.Visible) or (Sect.Width <= 0) then continue;
       if IntersectRect(R, Rect(Sect.Right-FHPos-FDivLineWidth, GetHeaderHeight,
                           Sect.Right-FHPos, ClientHeight), Canvas.ClipRect) then
       begin
         Canvas.MoveTo(R.Left, R.Top);
         Canvas.LineTo(R.Left, R.Bottom);
       end;
     end;
   end;
 end;

 //TickCount := GetTickCount - TickCount;
 //SendDebug('Redrawing time: ' + IntToStr(TickCount) + ' msec');
 // note for users: I get 10 msec with multiselect and less than 1 msec without multiselect
end;

function TCustomElTree.GetItemRect;
var y:integer;

begin
  if (not InRange(0,FVisible.Count-1,ItemIndex)) or Destroying{(csDestroying in ComponentState)} then
  begin
    SetRectEmpty(result);
    exit;
  end;
  if FHeader.Visible then Y:=GetHeaderHeight else Y:=0;
  Result:=Rect(0, FLineHeight*ItemIndex+Y, ClientWidth, FLineHeight*(ItemIndex+1)+Y-1);
end;

procedure TCustomElTree.OnHeaderSectionVisChange(Sender:TCustomElHeader; Section : TElHeaderSection);  { private }
begin
  IsUpdating:=true;
  FClearAll:=true;
  FHRange:=-1;
  if FHeader.SectionsWidth<=ClientWidth then
  begin
    FHeader.Invalidate;
    SetHPosition(0);
  end;
  FUpdated:=true;
  IsUpdating:=false;
end;  { OnHeaderSectionVisChange }

procedure TCustomElTree.DoOnColumnResize(SectionIndex : integer);  { protected }
begin
  if Assigned (FOnColumnResize) then FOnColumnResize(self, SectionIndex);
end;  { DoOnColumnResize }

procedure TCustomElTree.DoColumnClick(SectionIndex : integer);  { protected }
begin
  if Assigned (FOnColumnClick) then FOnColumnClick(self, SectionIndex);
end;  { DoColumnClick }

procedure TCustomElTree.DoItemFocused;  { protected }
begin
  if Assigned(FOnItemFocused) and (not Destroying) then FOnItemFocused(self);
end;  { DoItemFocused }

procedure TCustomElTree.DoItemDraw(Item:TElTreeItem; Surface:TCanvas; R:TRect; SectionIndex:integer);  { protected }
begin
  if Assigned(FOnItemDraw) then FOnItemDraw(Self, Item, Surface, R, SectionIndex);
end;  { DoItemDraw }

procedure TCustomElTree.DoValidate(Item:TElTreeItem; Section : TElHeaderSection; var Text:string; var Accept:boolean);  { protected }
begin
  FDelOnEdit := true;
  If assigned(FOnValidate) then FOnValidate(self, Item, Section, Text, Accept);
  FDelOnEdit := false;
end;  { DoValidate }

procedure TCustomElTree.DoShowHint(Item:TElTreeItem; var Text:String; HintWindow: THintWindow; MousePos : TPoint; var DoShowHint : boolean);  { protected }
begin
  if assigned(FOnShowHint) then FOnShowHint(Self, Item, Text, HintWindow, MousePos, DoShowHint);
end;  { DoShowHint }

procedure TCustomElTree.DoItemChange(Item:TElTreeItem; ItemChangeMode:TItemChangeMode);  { protected }
begin
  if Assigned(FOnItemChange) then FOnItemChange(Self, Item, ItemChangeMode);
end;  { DoOnItemChange }

procedure TCustomElTree.DoItemExpanding(Item : TElTreeItem; var CanProcess: boolean);  { protected }
begin
  if assigned(FOnItemExpanding) then FOnItemExpanding(Self, Item, CanProcess);
end;  { DoItemExpanding }

procedure TCustomElTree.DoItemCollapsing(Item : TElTreeItem; var CanProcess: boolean);  { protected }
begin
  if assigned(FOnItemCollapsing) then FOnItemCollapsing(Self,Item, CanProcess);
end;  { DoItemCollapsing }

procedure TCustomElTree.DoItemExpand(Item : TElTreeItem);  { protected }
begin
  if assigned(FOnItemExpand) then FOnItemExpand(Self, Item);
end;  { DoItemExpand }

procedure TCustomElTree.DoItemCollapse(Item : TElTreeItem);  { protected }
begin
  if assigned(FOnItemCollapse) then FOnItemCollapse(Self, Item);
end;  { DoItemCollapse }

procedure TCustomElTree.DoItemDelete(Item : TElTreeItem);  { protected }
begin
  if assigned(FOnItemDelete) then FOnItemDelete(Self, Item);
end;  { DoItemDelete }

procedure TCustomElTree.DoCompareItems(Item1, Item2:TElTreeItem; var res:integer);  { protected }
begin
  if assigned(FOnCompareItems) then FOnCompareItems(Self, Item1, Item2, res)
    else
      res := 0;
end;  { DoCompareItems }

procedure TCustomElTree.TriggerEditRequestEvent(Item:TElTreeItem; Section : TElHeaderSection);
begin
  if (assigned(FOnEditRequest)) then
    FOnEditRequest(Self, Item, Section );
end;  { TriggerEditRequestEvent }

procedure TCustomElTree.TriggerComboEditShowEvent(Item : TElTreeItem; Section : TElHeaderSection; Combobox : TCombobox);
begin
  if (assigned(FOnComboEditShow)) then
    FOnComboEditShow(Self, Item , Section , Combobox );
end;  { TriggerComboEditShowEvent }

procedure TCustomElTree.DoValidateCombo(Item : TElTreeItem; Section : TElHeaderSection; Combo: TComboBox; var Accept : boolean);
begin
  FDelOnEdit := true;
  if (assigned(FOnValidateCombo)) then
    FOnValidateCombo(Self, Item , Section , Combo, Accept );
  FDelOnEdit := false;
end;  { TriggerValidateComboEvent }

procedure TCustomElTree.OnHeaderSectionChange(Sender : TCustomElHeader; Section : TElHeaderSection; Change : TSectionChangeMode);  { protected }
begin
  case Change of
    scmAlign : if Section.Visible then
      begin
        IsUpdating:=true;
        FClearAll:=true;
        IsUpdating:=false;
      end;
    scmStyle : if FODFollowCol and Section.Visible then
                begin
                  IsUpdating:=true;
                  FClearAll:=true;
                  IsUpdating:=false;
                end;
    scmPassword : if Section.Visible then Repaint;
  end;
end;  { OnHeaderSectionChange }

function TCustomElTree.CreateHeader : TElHeader;  { protected }
begin
  result:=TElHeader.Create(self);
end;  { CreateHeader }

procedure TCustomElTree.SetCanEdit(value:boolean);
begin
  FCanEdit:=value;
end;

procedure TCustomElTree.OnHeaderSectionCreate(Header : TCustomElHeader; Section : TElHeaderSection);  { protected }
begin
  IsUpdating:=true;
  FClearAll:=true;
  FUpdated:=true;
  IsUpdating:=false;
end;  { OnHeaderSectionCreate }

procedure TCustomElTree.Paint;  { protected }
begin
  FPainting:=true;
  RedrawTree;
  FPainting:=false;
end;  { Paint }

function TCustomElTree.GetMoveColumnOnDrag : Boolean;
{ Returns the value of data member FMoveColumnOnDrag. }
begin
  result := FHeader.MoveOnDrag;
end;  { GetMoveColumnOnDrag }

procedure TCustomElTree.SetMoveColumnOnDrag(newValue : Boolean);
{ Sets data member FMoveColumnOnDrag to newValue. }
begin
  if (FHeader.MoveOnDrag <> newValue) then FHeader.MoveOnDrag := newValue;
end;  { SetMoveColumnOnDrag }

procedure TCustomElTree.SetHideHorzScrollBar(newValue : Boolean);
begin
  if (FHideHorzScrollBar <> newValue) then
  begin
    FHideHorzScrollBar := newValue;
    IsUpdating:=true;
    FVisUpdated:=true;
    FClearAll:=true;
    FClearVis:=true;
    FUpdated:=true;
    IsUpdating:=false;
  end;  { if }
end;  { SetHideHorzScrollBar }

procedure TCustomElTree.SetHideVertScrollBar(newValue : Boolean);
begin
  if (FHideVertScrollBar <> newValue) then
  begin
    FHideVertScrollBar := newValue;
    IsUpdating:=true;
    FVisUpdated:=true;
    FClearAll:=true;
    FClearVis:=true;
    FUpdated:=true;
    IsUpdating:=false;
  end;  { if }
end;  { SetHideVertScrollBar }

{$IFNDEF LITE}
function TCustomElTree.GetHeaderImages : TElImageList;
{$ELSE}
function TCustomElTree.GetHeaderImages : TImageList;
{$ENDIF}
{ Returns the value of data member FHeaderImages. }
begin
  result := FHeader.Images;
end;  { GetHeaderImages }

{$IFNDEF LITE}
procedure TCustomElTree.SetHeaderImages(newValue : TElImageList);
{$else}
procedure TCustomElTree.SetHeaderImages(newValue : TImageList);
{$ENDIF}
{ Sets data member FHeaderImages to newValue. }
begin
  if (FHeader.Images <> newValue) then FHeader.Images := newValue;
end;  { SetHeaderImages }

procedure TCustomElTree.TriggerScrollEvent(ScrollBarKind : TScrollBarKind; ScrollCode : integer);
begin
  if (assigned(FOnScroll)) then
    FOnScroll(Self, ScrollBarKind, ScrollCode);
end;  { TriggerScrollEvent }

procedure TCustomElTree.DeSelectAllEx;  { public }

  procedure IntSelAll(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                          IterateData:pointer; Tree:TCustomElTree);
  begin
    Item.Selected := false;
    if Tree.FSelCount = 0 then ContinueIterate := false;
  end;

begin
  if FMultiSelect = false then exit;
  IsUpdating:=true;
  FItems.Iterate(not IncludeHidden, @IntSelAll, nil);
  IsUpdating:=false;
end;  { DeSelectAll }

procedure TCustomElTree.DeSelectAll;  { public }

  procedure IntSelAll(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                          IterateData:pointer; Tree:TCustomElTree);
  begin
    Item.Selected := false;
    if Tree.FSelCount = 0 then ContinueIterate := false;
  end;

begin
  if FMultiSelect = false then exit;
  IsUpdating:=true;
  FItems.Iterate(FilteredVisibility, @IntSelAll, nil);
  IsUpdating:=false;
end;  { DeSelectAll }

procedure TCustomElTree.SelectAllEx;  { public }

  procedure IntSelAll(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                          IterateData:pointer; Tree:TCustomElTree);
  begin
    Item.Selected := true;
  end;

begin
  if FMultiSelect=false then exit;
  IsUpdating:=true;
  FItems.Iterate(not IncludeHidden, @IntSelAll, nil);
  IsUpdating:=false;
end;  { SelectAll }

procedure TCustomElTree.SelectAll;  { public }

  procedure IntSelAll(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                          IterateData:pointer; Tree:TCustomElTree);
  begin
    Item.Selected := true;
  end;

begin
  if FMultiSelect=false then exit;
  IsUpdating:=true;
  FItems.Iterate(FilteredVisibility, @IntSelAll, nil);
  IsUpdating:=false;
end;  { SelectAll }

procedure TCustomElTree.InvertSelectionEx;  { public }

  procedure IntInvSel(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                        IterateData:pointer; Tree:TCustomElTree);
  begin
    Item.Selected := not Item.Selected;
  end;

begin
  if FMultiSelect=false then exit;
  IsUpdating:=true;
  FItems.Iterate(not IncludeHidden, @IntInvSel, nil);
  IsUpdating:=false;
end;

procedure TCustomElTree.InvertSelection;  { public }

  procedure IntInvSel(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                        IterateData:pointer; Tree:TCustomElTree);
  begin
    Item.Selected := not Item.Selected;
  end;

begin
  if FMultiSelect=false then exit;
  IsUpdating:=true;
  FItems.Iterate(FilteredVisibility, @IntInvSel, nil);
  IsUpdating:=false;
end;

function TCustomElTree.GetSelected : TElTreeItem;
begin
  result := FFocused;
end;  { GetSelected }

procedure TCustomElTree.SetSelected(newValue : TElTreeItem);
begin
  if not FMultiSelect then SetFocused(newValue);
end;  { SetSelected }

function TCustomElTree.GetTopItem : TElTreeItem;  { public }
begin
  if FVisible.Count = 0 then result := nil else Result := FVisible[0];
end;  { GetTopItem }

procedure TCustomElTree.SetTopItem(Item : TElTreeItem);  { public }
begin
  if Item = nil then raise EElTreeError.Create(GetErrorMsg(STexInvItem));
  TopIndex := Item.VisIndex;
end;  { SetTopItem }

procedure TCustomElTree.TriggerHeaderColumnMoveEvent;
begin
  if (assigned(FOnHeaderColumnMove)) then
    FOnHeaderColumnMove(Self, Section, OldPos, NewPos);
end;  { TriggerHeaderColumnMoveEvent }

procedure TCustomElTree.TriggerTryEditEvent(Item:TElTreeItem; Section : 
  TElHeaderSection; var CellType : TFieldTypes; var CanEdit : boolean);
begin
  if (assigned(FOnTryEdit)) and (not Destroying) then
    FOnTryEdit(Self, Item, Section , CellType, CanEdit );
end;

procedure TCustomElTree.TriggerItemSaveEvent(Stream : TStream; Item : TElTreeItem);

begin
  if (assigned(FOnItemSave)) then FOnItemSave(Self, Stream , Item );
end;

procedure TCustomElTree.TriggerItemLoadEvent(Stream : TStream; Item :
  TElTreeItem);
begin
  if (assigned(FOnItemLoad)) then FOnItemLoad(Self, Stream , Item );
end;

{$IFNDEF LITE}
procedure TCustomElTree.Save;
var FSaveKey : string;
    F        : TForm;
    S        : String;
begin
  if Assigned(FStorage) then
  begin
    FHeader.Storage := FStorage;
    FHeader.StoragePath := FStorage.Delimiter + FStoragePath;
    FHeader.Save;
    FSaveKey := FStorage.CurrentKey;
    F := GetOwnerForm(Self);
    if F <> nil then S := F.ClassName else S := '';
    if FStorage.OpenKey(FStorage.Delimiter + S + FStoragePath, true) then
    begin
      FStorage.WriteInteger('', 'FontSize', Font.Size);
      FStorage.WriteInteger('', 'FontColor', Font.Color);
      FStorage.WriteString('', 'FontName', Font.Name);
      FStorage.WriteBool('', 'FontBold', fsBold in Font.Style);
      FStorage.WriteBool('', 'FontItalic', fsItalic in Font.Style);
      FStorage.WriteBool('', 'FontUnderline', fsUnderline in Font.Style);
      FStorage.WriteBool('', 'FontStrikeout', fsStrikeout in Font.Style);
      FStorage.WriteInteger('', 'SortType', integer(FSortType));
      FStorage.WriteInteger('', 'SortMode', integer(FSortMode));
      FStorage.WriteInteger('', 'SortDir', integer(FSortDir));
      FStorage.WriteInteger('', 'SortSection', integer(FSortSection));
    end;
    FStorage.OpenKey(FSaveKey, false);
  end;
end;  {Save}

procedure TCustomElTree.Restore;
var FSaveKey : string;
    i : integer;
    b : Boolean;
    s : string;
    F : TForm;
begin
  IsUpdating := true;
  if Assigned(FStorage) then
  begin
    FHeader.Storage := FStorage;
    FHeader.StoragePath := FStorage.Delimiter + FStoragePath;
    FHeader.Restore;
    FSaveKey := FStorage.CurrentKey;
    F := GetOwnerForm(Self);
    if F <> nil then S := F.ClassName else S := '';
    if FStorage.OpenKey(FStorage.Delimiter + S + FStoragePath, false) then
    begin
      FStorage.ReadInteger('', 'FontSize', Font.Size, i);
      Font.Size := i;
      FStorage.ReadInteger('', 'FontColor', Font.Color, i);
      Font.Color := i;
      FStorage.ReadString('', 'FontName', Font.Name, s);
      Font.Name := s;         
      FStorage.ReadBool('', 'FontBold', fsBold in Font.Style, b);
      if b then Font.Style := Font.Style + [fsBold] else Font.Style := Font.Style - [fsBold];
      FStorage.ReadBool('', 'FontItalic', fsItalic in Font.Style, b);
      if b then Font.Style := Font.Style + [fsItalic] else Font.Style := Font.Style - [fsItalic];
      FStorage.ReadBool('', 'FontUnderline', fsUnderline in Font.Style, b);
      if b then Font.Style := Font.Style + [fsUnderline] else Font.Style := Font.Style - [fsUnderline];
      FStorage.ReadBool('', 'FontStrikeout', fsStrikeout in Font.Style, b);
      if b then Font.Style := Font.Style + [fsStrikeout] else Font.Style := Font.Style - [fsStrikeout];
      FStorage.ReadInteger('', 'SortType', integer(FSortType), i);
      FSortType := TSortTypes(i);
      FStorage.ReadInteger('', 'SortMode', integer(FSortMode), i);
      FSortMode := TSortModes(i);
      FStorage.ReadInteger('', 'SortDir', integer(FSortDir), i);
      FSortDir := TSortDirs(i);
      FStorage.ReadInteger('', 'SortSection', FSortSection, i);
      FSortSection := i;
    end;
    FStorage.OpenKey(FSaveKey, false);
  end;
  IsUpdating := false;
end;  {Restore}
{$ENDIF}

procedure TCustomElTree.TriggerItemSelectedChangeEvent(Item: TElTreeItem);
begin
  if (assigned(FOnItemSelectedChange)) and (not Destroying) then
    FOnItemSelectedChange(Self, Item);
end;  {TriggerItemSelectedChangeEvent}

procedure TCustomElTree.FillDragImage;
var Bmp:TBitmap;
{$IFNDEF LITE}
    TIL:TElImageList;
{$ELSE}
    TIL:TImageList;
{$ENDIF}
    i : integer;
    b : boolean;
    tt,
    bb: integer;

    procedure IntRedrawItemTree(Index:integer; Surface:TBitmap);
    var R:TRect;
        SText:string;
        Item : TElTreeItem;
        R1 : TRect;
        SaveStyle : TFontStyles;
        ImDrawStyle: TDrawingStyle;
        StImIndex : integer;

    begin
      Item:=FVisible[index];
      SText:=Item.Text;

      R:=GetItemRect(index);
      dec(R.Top, FDDY);
      dec(R.Bottom, FDDY);

      if RightAlignedTree then
      begin
        dec(R.Right, (Item.Level-1)*ItemExt);
      end else
      begin
        inc(R.Left,(Item.Level-1)*ItemExt);
      end;
      if FShowImages then
      begin
          if (FImages<>nil) then
          begin
            ImDrawStyle:=FImages.DrawingStyle;
            FImages.DrawingStyle:=dsFocus;
            if Item.FState=[] then StImIndex:=Item.FImageIndex else StImIndex:=Item.FStImageIndex;
            if StImIndex=-1 then
               STImIndex:=DoGetPicture(Item);
            if InRange(0,FImages.Count-1,StImIndex) then
            begin
              if RightAlignedTree then
              begin
                FImages.Draw(Surface.Canvas, R.Right - FImages.Width,(R.Top+(R.Bottom-R.Top+1) div 2)-(FImages.Height div 2),StImIndex);
                dec(R.Right, FImages.Width);
              end else
              begin
                FImages.Draw(Surface.Canvas,R.Left,(R.Top+(R.Bottom-R.Top+1) div 2)-(FImages.Height div 2),StImIndex);
                inc(R.Left,FImages.Width);
              end;
            end;
            FImages.DrawingStyle:=ImDrawStyle;
          end;
      end; // Show images
      if RightAlignedTree then
      begin
        dec(R.Right, (ItemExt div 3));
      end else
      begin
        inc(R.Left, (ItemExt div 3));
      end;

      if (FODFollowCol and (FHeader.Sections.Count>0) and (FHeader.Sections[FMainTreeCol].Style=ElhsOwnerDraw))
          or ((not (FODFollowCol)) and (SText=FODMask)) then
        begin
          DoItemDraw(Item, Canvas, R, FMainTreeCol);
        end
        else
        begin
          with Surface.Canvas do
          begin
            if Item.Bold then Font.Style:=Font.Style+[fsBold] else Font.Style:=Font.Style-[fsBold];
            if Item.Italic then Font.Style:=Font.Style+[fsItalic] else Font.Style:=Font.Style-[fsItalic];
            if Item.Underlined then Font.Style:=Font.Style+[fsUnderline] else Font.Style:=Font.Style-[fsUnderline];
            if Item.StrikeOut then Font.Style:=Font.Style+[fsStrikeOut] else Font.Style:=Font.Style-[fsStrikeOut];
          end;
          R1:=R;
          if RightAlignedTree then
          begin
            R1.Left := R1.Right - Surface.Canvas.TextWidth(SText);
          end else
          begin
            R1.Right := R1.Left + Surface.Canvas.TextWidth(SText);
          end;
          SaveStyle:=Surface.Canvas.Font.Style;
          if RightAlignedText then
          begin
            DrawText(Surface.Canvas.Handle, PChar(SText), -1, R1, DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER or DT_RIGHT or DT_RTLREADING);
          end else
          begin
            DrawText(Surface.Canvas.Handle, PChar(SText), -1, R1, DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER);
          end;
          Font.Style:=SaveStyle;
        end; // if ... else
    end;

  procedure DrawDragImages;
  var
      i :integer;
      R,
      R1:TRect;

  begin
    for i:=0 to FVisible.Count-1 do
      if (TElTreeItem(FVisible[i]).Selected) then
      begin
        R:=GetItemRect(i);
        dec(R.Top, FDDY);
        dec(R.Bottom, FDDY);
        if FHLines then dec(R.Bottom, FDivLineWidth-1);
        R.Left:=0;
        r1:=r;
        if FHeader.Visible
          then r1.right:=FHeader.Sections[FMainTreeCol].Width
          else r1.right:=Max(FHRange, FHPos + ClientWidth);
        Bmp.Canvas.Font:=Canvas.Font;
        Bmp.Canvas.Font.Color:=FTextColor;
        Bmp.Canvas.Pen.Color:=FTextColor;
        IntRedrawItemTree(i,Bmp);
      end;
  end;

begin
  if DragImageMode = dimNever then exit;
  b := false;
  tt := -1;
  bb := -1;

  if DragImageMode = dimOne then
    for i:=0 to FVisible.Count-1 do
    begin
      if (TElTreeItem(FVisible[i]).Selected) then
      begin
        if b then exit else
        begin
          b := true;
          tt := i * LineHeight;
          bb := (i+1) * LineHeight;
        end;
      end;
    end
  else
  begin
    for i:=0 to FVisible.Count-1 do
    begin
      if (TElTreeItem(FVisible[i]).Selected) then
      begin
        bb := (i + 1) * LineHeight;
        if not b then
        begin
          tt := i * LineHeight;
          b := true;
        end;
      end;
    end;
  end;
{$IFNDEF LITE}
  TIL := TElImageList.Create(self);
{$ELSE}
  TIL := TImageList.Create(self);
{$ENDIF}
  TIL.Height:= bb - tt + 1;
  if FHeader.Visible then inc(tt, FHeader.Height);
  FDDY := tt;

  if FMainTreeCol>=FHeader.Sections.Count then FMainTreeCol:=0;
  if FHeader.Visible
     then TIL.Width:=FHeader.Sections[FMainTreeCol].Width
     else TIL.Width:=Max(FHRange,FHPos+ClientWidth);

  Bmp:=TBitmap.Create;
  Bmp.Width := TIL.Width;
  Bmp.Height := TIL.Height;
{$IFNDEF VER90}
  Bmp.TransparentMode:=tmFixed;
  Bmp.TransparentColor:=FBkColor;
{$ENDIF}
  Bmp.Canvas.FillRect(Rect(0,0,BMP.Width,BMP.Height));
{$IFNDEF VER90}
  Bmp.Canvas.Lock;
{$ENDIF}
  try
    DrawDragImages;
  finally
{$IFNDEF VER90}
    Bmp.Canvas.UnLock;
{$ENDIF}
  end;
  TIL.AddMasked(Bmp, clDefault);
  FDragImages:=TIL;
end;

procedure TCustomElTree.DoRedrawItemTree;
var SText: string;
    TSI,
    TSI1 : TElTreeItem;
    R,R1,R2   : TRect;
    SavePen   : TPenStyle;
    SavePenCol: TColor;
    ImDrawStyle: TDrawingStyle;
    Stack     : TElStack;
    StImIndex : integer;
    AL,W,H    : integer;
    CurStyle : TElCellStyle;
    FCCanvas : TCanvas;
    i : integer;
    DC: HDC;
    FontStyle: TFontStyles;
{$IFNDEF LITE}
    FTImages : TElImageList;
{$ELSE}
    FTImages : TImageList;
{$ENDIF}
    function GetNextVisChild(Parent, Item : TElTreeItem):TElTreeItem;
    var III : TElTreeItem;
    begin
      if (Item.FParent=FItems.FRoot) then
      begin
        result := nil;
        exit;
      end;
      III := Parent.GetNextChild(Item);
      if FilteredVisibility then
         while (III<>nil) and (III.Hidden) do III := Parent.GetNextChild(III);
      result := III;
    end;

begin
  SText:=Item.Text;
{$IFDEF EL_DEMO}
   if ((length(SText) mod 3) = 0) then SText:='ElTree DEMO';
{$ENDIF}
  if FHeader.Visible then
  begin
    R:=Rect(FHeader.Sections[FMainTreeCol].Left{+1}-FHPos, ItemRect.Top,
            FHeader.Sections[FMainTreeCol].Right{-1}-FHPos, ItemRect.Bottom);
    if FVLines and (not FBarStyle) then Dec(R.Right, FDivLineWidth);
  end
  else
    R:=Rect(ItemRect.Left-FHPos, ItemRect.Top, ItemRect.Right-FHPos, ItemRect.Bottom);

  if Item.FUseStyles then
  begin
    CurStyle := Item.FMainStyle;
    if FOverColors or CurStyle.FOwnerProps then
    begin
      Canvas.Brush.Color := FCurBkColor;
      Canvas.Font.Color := FCurTextColor;
    end
    else
      Canvas.Brush.Color := CurStyle.FCellBkColor;
  end else
  begin
    Canvas.Brush.Color:=FCurBkColor;
    Canvas.Font.Color:=FCurTextColor;
  end;
  if Item.Selected and (not FBarStyle) and (((FSelectColumn = FMainTreeCol) and (FHeader.Visible)) or FRowSelect)
     and (((GetParentForm(self)<>nil) and (GetParentForm(self).ActiveControl=self)) or (not FHideSelect)) then
     Canvas.Brush.Color:=clHighlight;
  if (Item=FDropTrg) and (((FSelectColumn = FMainTreeCol) and (FHeader.Visible)) or FRowSelect) then
  begin
    case FDragTrgDraw of
    ColorRect:
      begin
        if FDropAcc
          then Canvas.Brush.Color:=clGreen
          else Canvas.Brush.Color:=clRed;
      end;
    SelColorRect:
        Canvas.Brush.Color:=clHighlight;
    end;    // case
  end;
  // now fill the background
  if (not FHLines) or FBarStyle then inc(R.Bottom);
  Canvas.FillRect(R);
  if (not FHLines) or FBarStyle then dec(R.Bottom);

  if RightAlignedTree then
  begin
    // Here we try to insert custom buttons
    if (FShowRoot and FShowLines) or FShowButtons then dec(R.Right, ItemExt);
  end else
  begin
    // Here we try to insert custom buttons
    if (FShowRoot and FShowLines) or FShowButtons then Inc(R.Left, ItemExt);
  end;

  // draw tree and images
  if (FShowLines) then
  begin
    Stack     := TElStack.Create;
    TSI       := Item.Parent;
    SavePen   := Canvas.Pen.Style;
    SavePenCol:= Canvas.Pen.Color;
    Canvas.Pen.Style := LinesStyle;
    Canvas.Pen.Color := LinesColor;
    while TSI<>nil do
    begin
      Stack.Push(TSI);
      TSI:=TSI.Parent;
    end;
    DC := Canvas.Handle;
    if Stack.Count > 0 then
    begin
      TSI:=TElTreeItem(Stack.Pop);
      while Stack.Count>0 do
      begin
        TSI1:=TSI;
        if Stack.Count>0 then
        begin
          TSI:=TElTreeItem(Stack.Pop);

          if RightAlignedTree then
          begin
            MoveToEx(DC, R.Right - ItemExt div 2, R.Top, nil);
            if (GetNextVisChild(TSI1, TSI)<>nil) and ((R.Right  - ItemExt div 2)> R.Left) then
               LineTo(DC, R.Right - ItemExt div 2, ItemRect.Bottom + 1);
          end else
          begin
            MoveToEx(DC, R.Left + ItemExt div 2, R.Top, nil);
            if (GetNextVisChild(TSI1,TSI)<>nil) and ((R.Left+ItemExt div 2)<R.Right) then
               LineTo(DC, R.Left+ItemExt div 2, ItemRect.Bottom + 1);
          end;
        end;
        if RightAlignedTree then
        begin
          dec(R.Right, ItemExt);
        end else
        begin
          inc(R.Left, ItemExt);
        end;
      end;
      if RightAlignedTree then
      begin
        dec(R.Right, ItemExt);
      end else
      begin
        inc(R.Left, ItemExt);
      end;
    end;
    Stack.Free;

    if (Item.FParent <> Item.FRoot) then
    begin
      if RightAlignedTree then
      begin
        inc(R.Right, ItemExt);
        if ((R.Right - ItemExt div 2)> R.Left) then
        begin
          MoveToEx(DC, R.Right - ItemExt div 2,R.Top, nil);
          LineTo(DC, R.Right - ItemExt div 2, R.Top+((R.Bottom-R.Top+1) div 2));
          LineTo(DC, Max(R.Right - ItemExt, R.Left), R.Top+((R.Bottom-R.Top+1) div 2));
        end;
      end else
      begin
        dec(R.Left,ItemExt);
        if ((R.Left+ItemExt div 2)<R.Right) then
        begin
          MoveToEx(DC, R.Left+ItemExt div 2,R.Top, nil);
          LineTo(DC, R.Left+ItemExt div 2,R.Top+((R.Bottom-R.Top+1) div 2));
          LineTo(DC, Min(R.Left+ItemExt, R.Right), R.Top+((R.Bottom-R.Top+1) div 2));
        end;
      end;

      if GetNextVisChild(Item.Parent,Item)<>nil then
      begin

        if RightAlignedTree then
        begin
          if ((R.Right - ItemExt div 2)> R.Left) then
          begin
            MoveToEx(DC, R.Right - ItemExt div 2,R.Top + ((R.Bottom - R.Top+1) div 2), nil);
            LineTo(DC, R.Right - ItemExt div 2, R.Bottom + 1);
          end;
        end else
        begin
          if ((R.Left+ItemExt div 2)<R.Right) then
          begin
            MoveToEx(DC, R.Left+ItemExt div 2,R.Top+((R.Bottom - R.Top+1) div 2), nil);
            LineTo(DC, R.Left + ItemExt div 2, R.Bottom+1);
          end;
        end;
      end; //if
    end // if
    else
    begin
      if RightAlignedTree then
      begin
        inc(R.Right,ItemExt);
      end else
      begin
        dec(R.Left,ItemExt);
      end;

      if FShowRoot then
      begin
        if RightAlignedTree then
        begin
          if (R.Right - (ItemExt div 2 - 4) > R.Left) then
          begin
            MoveToEx(DC, R.Right - (ItemExt div 2 - 4),R.Top+((R.Bottom-R.Top+1) div 2), nil);
            LineTo(DC, Max(R.Right - ItemExt, R.Left), R.Top+((R.Bottom{ItemRect.Bottom}-R.Top+1) div 2))
          end;
        end else
        begin
          if (R.Left+(ItemExt div 2-4)<R.Right) then
          begin
            MoveToEx(DC, R.Left+(ItemExt div 2-4),R.Top+((R.Bottom-R.Top+1) div 2), nil);
            LineTo(DC, Min(R.Left+ItemExt,R.Right), R.Top+((R.Bottom{ItemRect.Bottom}-R.Top+1) div 2))
          end;
        end;
      end;
    end;
    Canvas.Pen.Style:=SavePen;
    Canvas.Pen.Color:=SavePenCol;
  end // Show lines?
  else
  begin
    if RightAlignedTree then
    begin
      dec(R.Right, (Item.Level -1) * ItemExt);
    end else
    begin
      inc(R.Left, (Item.Level -1) * ItemExt);
    end;
  end;
  if FShowButtons and ((Item.HasVisibleChildren) or Item.ForceButtons) and (not Item.SuppressButtons) then
  begin
    if FCustomPlusMinus then
    begin
      CenterRects(FPlusPicture.Width, min(R.Right - R.Left + 1, ItemExt), FPlusPicture.Height, R.Bottom - R.Top + 1, R1);
      if not Item.Expanded then FCCanvas := FPlusPicture.Canvas else FCCanvas := FMinusPicture.Canvas;
    end else
    begin
      CenterRects(9, min(R.Right - R.Left + 1, ItemExt), 9, R.Bottom - R.Top + 1, R1);
      if not Item.Expanded then FCCanvas := PlusBmp.Canvas else FCCanvas := MinusBmp.Canvas;
    end;
    if RightAlignedTree then
    begin
      R1 := Rect(R.Right - R1.Right, R1.Top + R.Top, R.Right - R1.Left, R1.Bottom + R.Top);
      BitBlt(Canvas.Handle, R1.Left, R1.Top, R1.Right - R1.Left + 1, R1.Bottom- R1.Top + 1, FCCanvas.Handle, 0, 0, SRCCOPY);
    end else
    begin
      Inc(R1.Left, R.Left); Inc(R1.Right, R.Left);
      Inc(R1.Top, R.Top); Inc(R1.Bottom, R.Top);
      BitBlt(Canvas.Handle, R1.Left, R1.Top, R1.Right - R1.Left + 1, R1.Bottom- R1.Top + 1, FCCanvas.Handle, 0, 0, SRCCOPY);
    end;
  end;
  if RightAlignedTree then
  begin
    dec(R.Right, ItemExt);
  end else
  begin
    inc(R.Left, ItemExt);
  end;

  if FShowCheckBoxes then
  begin
    if Item.FShowCheckBox then
    begin
      if RightAlignedTree then
      begin
        dec(R.Right, 2);
      end else
      begin
        inc(R.Left, 2);
      end;
      i := DFCS_BUTTONCHECK or DFCS_CHECKED;
      if Item.FCheckBoxType = ectRadioButton then
      begin
        if Item.Checked then i := DFCS_BUTTONRADIO or DFCS_CHECKED
        else i := DFCS_BUTTONRADIO;
      end else
      if Item.FCheckBoxType = ectCheckBox then
      begin
        if Item.Checked then i := DFCS_BUTTONCHECK or DFCS_CHECKED
        else i := DFCS_BUTTONCHECK;
      end else
      begin
        case Item.FCheckBoxState of    //
          cbChecked: i := DFCS_BUTTONCHECK or DFCS_CHECKED;
          cbUnchecked: i := DFCS_BUTTONCHECK;
          cbGrayed: i := DFCS_BUTTON3STATE or DFCS_CHECKED;
        end;    // case
      end;
      if not Item.FCheckBoxEnabled then i := i or DFCS_INACTIVE;
      DC := CreateCompatibleDC(Canvas.Handle);
      Bmp.Height := 11; BMP.Width := 11;
      SelectObject(DC, BMP.Handle);
      R2 := Rect(0, 0, 11, 11);
      FillRect(DC, R2, Canvas.Brush.Handle);
      DrawFrameControl(DC, R2, DFC_BUTTON, i);

      if RightAlignedTree then
      begin
        R2:=Rect(Max(R.Left, R. Right - 11), R.Top+((R.Bottom-R.Top+1) div 2)-5, R.Right, R.Top+((R.Bottom-R.Top+1) div 2) + 6);
      end else
      begin
        R2:=Rect(R.Left, R.Top+((R.Bottom-R.Top+1) div 2)-5, Min(R.Right,R. Left+11), R.Top+((R.Bottom-R.Top+1) div 2) + 6);
      end;
      BitBlt(Canvas.Handle, R2.Left, R2.Top, R2.Right - R2.Left + 1, R2.Bottom- R2.Top + 1, DC, 0, 0, SRCCOPY);
      DeleteObject(DC);
      if RightAlignedTree then
      begin
        dec(R.Right, ItemExt - 2);
      end else
      begin
        inc(R.Left, ItemExt - 2);
      end;
    end;
  end;
  
  Item.FImageDrawn:=false;
  Item.FImageDrawn2:=false;

  if (FShowImages) then
  begin
    if FImages2 <> nil then FTImages := FImages2 else FTImages := FImages;
    if FTImages <> nil then
    begin
      ImDrawStyle:=FTImages.DrawingStyle;
      if FChStateImage then
      begin
        if Item.Selected then FTImages.DrawingStyle:=dsSelected else
        if Item.Focused then FTImages.DrawingStyle:=dsFocus else
        {if Item.Cut or (not Item.Enabled) then FImages.DrawingStyle := else }FTImages.DrawingStyle:=dsNormal;
      end;

      // draw 2nd image
      if (Item.Focused or Item.Selected or Item.Cut or (Item.Expanded and Item.HasVisibleChildren))
         then StImIndex:=Item.FStImageIndex2 else StImIndex:=Item.FImageIndex2;
      if StImIndex=-1 then StImIndex:=DoGetPicture2(Item);
      if InRange(0, FTImages.Count-1,StImIndex) then
      begin
        //Bmp.Assign(nil);
        BMP.Width:=FTImages.Width;
        BMP.Height:=FTImages.Height;
        BMP.Canvas.brush.Color := Canvas.Brush.Color;
        BMP.Canvas.FillRect(Rect(0, 0, FTImages.Width, FTImages.Height));
        if R.Left<R.Right then
        begin
          FTImages.Draw(BMP.Canvas, 0, 0, StImIndex);
  {$IFNDEF VER90}
          BMP.TransparentMode:=tmAuto;
  {$ENDIF}
          h := FTImages.Height; w := FTImages.Width;

          if RightAlignedTree then
          begin
            R2:=Rect(Max(R.Left, R.Right - w), R.Top+((R.Bottom-R.Top+1) div 2)-(h div 2), R.Right, R.Top+((R.Bottom-R.Top+1) div 2)- (h div 2)+h);
            // WARNING!!!
            Canvas.CopyRect(R2, BMP.Canvas, Rect(max(0, w - (R.Right - R.Left)), 0, w, h));
          end else
          begin
            R2:=Rect(R.Left, R.Top+((R.Bottom-R.Top+1) div 2)-(h div 2),Min(R.Right,R. Left+w), R.Top+((R.Bottom-R.Top+1) div 2)- (h div 2)+h);
            Canvas.CopyRect(R2, BMP.Canvas, Rect(0, 0, min(w, Min(R.Right, R.Left+w-1)-R.Left+1), h));
          end;
        end;
        if RightAlignedTree then
        begin
          dec(R.Right, FTImages.Width);
        end else
        begin
          inc(R.Left, FTImages.Width);
        end;
        Item.FImageDrawn2 := true;
      end else Item.FImageDrawn2 :=false;
      FTImages.DrawingStyle:=ImDrawStyle;
    end;
    if (FImages<>nil) then
    begin
      ImDrawStyle:=FImages.DrawingStyle;
      if FChStateImage then
      begin
        if Item.Selected then FImages.DrawingStyle:=dsSelected else
        if Item.Focused then FImages.DrawingStyle:=dsFocus else
        {if Item.Cut or (not Item.Enabled) then FImages.DrawingStyle := else }FImages.DrawingStyle:=dsNormal;
      end;

      if (Item.Focused or Item.Selected or Item.Cut or (Item.Expanded and Item.HasVisibleChildren))
         then StImIndex:=Item.FStImageIndex else StImIndex:=Item.FImageIndex;
      if StImIndex=-1 then StImIndex:=DoGetPicture(Item);
      if InRange(0,FImages.Count-1,StImIndex) then
      begin
        if RightAlignedTree then
        begin
          if Item.FImageDrawn2 then dec(R.Right, ItemExt div 3); // make the space between images
        end else
        begin
          if Item.FImageDrawn2 then inc (R.Left, ItemExt div 3); // make the space between images
        end;
        //Bmp.Assign(nil);
        BMP.Width:=FImages.Width;
        BMP.Height:=FImages.Height;
        BMP.Canvas.brush.Color := Canvas.Brush.Color;
        BMP.Canvas.FillRect(Rect(0, 0, FImages.Width, FImages.Height));
        if R.Left<R.Right then
        begin
          FImages.Draw(BMP.Canvas,0,0,StImIndex);
  {$IFNDEF VER90}
          BMP.TransparentMode:=tmAuto;
  {$ENDIF}
          h := FImages.Height; w := FImages.Width;

          if RightAlignedTree then
          begin
            R2:=Rect(Max(R.Left,R.Right - w), R.Top+((R.Bottom-R.Top+1) div 2)-(h div 2), R.Right, R.Top+((R.Bottom-R.Top+1) div 2)- (h div 2)+h);
            Canvas.CopyRect(R2, BMP.Canvas, Rect(max(0, w - (R.Right - R.Left)), 0, w, h));
          end else
          begin
            R2:=Rect(R.Left, R.Top+((R.Bottom-R.Top+1) div 2)-(h div 2),Min(R.Right,R. Left+w), R.Top+((R.Bottom-R.Top+1) div 2)- (h div 2)+h);
            Canvas.CopyRect(R2, BMP.Canvas, Rect(0, 0, min(w, Min(R.Right, R.Left+w-1)-R.Left+1), h));
          end;
        end;
        if RightAlignedTree then
        begin
          dec(R.Right, FImages.Width);
        end else
        begin
          inc(R.Left, FImages.Width);
        end;
        Item.FImageDrawn:=true;

      end else Item.FImageDrawn:=false;
      FImages.DrawingStyle:=ImDrawStyle;
    end;
  end; // Show images

  if (FODFollowCol and (FHeader.Sections.Count>0) and (FHeader.Sections[FMainTreeCol].Style=ElhsOwnerDraw))
    or ((SText=FODMask) and (not (FODFollowCol))) or (Item.FUseStyles and (Item.FMainStyle.FStyle = elhsOwnerDraw)) then
  begin
    DoItemDraw(Item, Canvas, R, FMainTreeCol);
    if RightAlignedTree then
    begin
      dec(R.Right, ItemExt div 3);
      R1:=R;
      inc(R1.Right);
      Item.FTextLeft := R1.Left + FHPos;
      Item.FTextRight:= R1.Right + FHPos;
    end else
    begin
      inc(R.Left, ItemExt div 3);
      R1:=R;
      dec(R1.Left);
      Item.FTextLeft:=R1.Left+FHPos;
      Item.FTextRight:=R1.Right+FHPos;
    end;
  end else
  begin
    // make the empty space, that relates to the text
    if RightAlignedTree then
    begin
      dec(R.Right, ItemExt div 3);
    end else
    begin
      inc (R.Left, ItemExt div 3);
    end;
    if (Item.UseStyles and (CurStyle.FStyle = elhsText)) or (not Item.FUseStyles) then
    begin
      FontStyle := [];//Canvas.Font.Style;
      if Item.FUseStyles then
      begin
        AL := CurStyle.FTextFlags;
        if FOverColors or CurStyle.FOwnerProps then
        begin
          Canvas.Font.Color := FCurTextColor;
          Canvas.Brush.Color := FCurBkColor;
        end else
        begin
          Canvas.Brush.Color := CurStyle.FTextBkColor;
          Canvas.Font.Color := CurStyle.FTextColor;
        end;
        if not CurStyle.FOwnerProps then with CurStyle do
        begin
          if Canvas.Font.Name <> FontName then Canvas.Font.Name := FontName;
          FontStyle := FontStyles; //if Canvas.Font.Style <> FontStyles then
          if Canvas.Font.Size <> FontSize then Canvas.Font.Size := FontSize;
        end else
        begin
          if Canvas.Font.Name <> Font.Name then Canvas.Font.Name := Font.Name;
          if Item.FParentStyle then FontStyle := Font.Style else
          begin
            if stsBold in Item.FState then Include(FontStyle, fsBold);
            if stsItalic in Item.FState then Include(FontStyle, fsItalic);
            if stsUnderlined in Item.FState then Include(FontStyle, fsBold);
            if stsStrikeOut in Item.FState then Include(FontStyle, fsBold);
          end;
          if (FTracking and (Item=FTrackItem)) then Include(FontStyle,fsUnderline);
          if Canvas.Font.Size <> Font.Size then Canvas.Font.Size := Font.Size;

          Al:=DT_LEFT;
          if Assigned(FHeader) and (FHeader.Visible) and (FHeader.Sections.Count>FMainTreeCol) then
            case FHeader.Sections[FMainTreeCol].Alignment of
              hsaCenter: Al:=DT_CENTER;
              hsaRight:  Al:=DT_RIGHT;
            end;    // case
          AL := DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER or AL or DT_END_ELLIPSIS;
        end;
      end else
      begin
        Canvas.Font.Color := FCurTextColor;
        Canvas.Brush.Color:= FCurBkColor;
        if Canvas.Font.Name <> Font.Name then Canvas.Font.Name := Font.Name;
        if Item.FParentStyle then FontStyle := Font.Style else
        begin
          if stsBold in Item.FState then Include(FontStyle, fsBold);
          if stsItalic in Item.FState then Include(FontStyle, fsItalic);
          if stsUnderlined in Item.FState then Include(FontStyle, fsBold);
          if stsStrikeOut in Item.FState then Include(FontStyle, fsBold);
        end;
        if not Item.FParentColors then
        begin
          Canvas.Font.Color := Item.Color;
          Canvas.Brush.Color:= Item.BkColor;
        end;
        if (FTracking and (Item=FTrackItem)) then Include(FontStyle,fsUnderline);
        if Canvas.Font.Size <> Font.Size then Canvas.Font.Size := Font.Size;

        if FRightAlignedText then al := DT_RIGHT else Al:=DT_LEFT;
        if Assigned(FHeader) and (FHeader.Visible) and (FHeader.Sections.Count>FMainTreeCol) then
        begin
          if RightAlignedText then
          begin
            case FHeader.Sections[FMainTreeCol].Alignment of
              hsaCenter: Al:=DT_CENTER;
              hsaRight:  Al:=DT_LEFT;
            end;    // case
          end else
          begin
            case FHeader.Sections[FMainTreeCol].Alignment of
              hsaCenter: Al:=DT_CENTER;
              hsaRight:  Al:=DT_RIGHT;
            end;    // case
          end;
        end;
        AL := DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER or AL or DT_END_ELLIPSIS;
        if RightAlignedText then AL := AL or DT_RTLREADING;
      end;
      Canvas.Font.Style := FontStyle;
      if (FTracking and (Item=FTrackItem)) then
         if Item.Selected and (FRowSelect or (FSelectColumn = -1) or (FHeader.Visible and (FMainTreeCol = FSelectColumn))) then Canvas.Font.Color:= clBtnHighlight else Canvas.Font.Color:=clHighlight;
      if Item.Selected and (not FBarStyle) and ((FSelectColumn = -1) or (FSelectColumn = FMainTreeCol) or FRowSelect or (not FHeader.Visible))
         and (((GetParentForm(self)<>nil)
              and (GetParentForm(self).ActiveControl=self))
         or (not FHideSelect)) then
      begin
        Canvas.Brush.Color:=clHighlight;
        Canvas.Font.Color:=clHighlightText;
      end;
      if (Item=FDropTrg) and ((FSelectColumn = -1) or (FSelectColumn = FMainTreeCol) or FRowSelect or (not FHeader.Visible)) then
      begin
        case FDragTrgDraw of
        ColorRect:
          begin
            if FDropAcc
              then Canvas.Brush.Color:=clGreen
              else Canvas.Brush.Color:=clRed;
            Canvas.Font.Color:=clBtnText;
          end;
        SelColorRect:
          begin
            Canvas.Brush.Color:=clHighlight;
            Canvas.Font.Color:=clHighlightText;
          end;
        end;    // case
      end;
      if RightAlignedTree then
      begin
        R1:=R;
        inc(R1.Right);
        R1.Left := Max(R1.Left, R1.Right - (Canvas.TextWidth(SText)+3) + FDivLineWidth);
      end else
      begin
        R1:=R;
        dec(R1.Left);
        R1.Right:=Min(R1.Left+Canvas.TextWidth(SText)+3-FDivLineWidth, R.Right);
      end;
      Item.FTextLeft:=R1.Left+FHPos;
      Item.FTextRight:=R1.Right+FHPos;
      if R1.Left<R1.Right then Canvas.FillRect(R1);
      InflateRect(R1, -1, -1);
      DrawText(Canvas.Handle, PChar(SText), -1, R1, AL or DT_EXTERNALLEADING);
    end else
    if Item.UseStyles and (CurStyle.FStyle = elhsPictureOnly) and Assigned(CurStyle.FPicture) then
    begin
      h := CurStyle.FPicture.Height;
      w := CurStyle.FPicture.Width;

      if RightAlignedTree then
      begin
        R2:=Rect(Max(R.Left, R.Right - w), R.Top + ((R.Bottom-R.Top + 1) div 2)-(h div 2), R.Right, R.Top+((R.Bottom-R.Top+1) div 2)- (h div 2) + h);
        Item.FTextLeft:=R.Left+FHPos;
        Item.FTextRight:=R.Right+FHPos;
        Canvas.CopyRect(R2, CurStyle.FPicture.Canvas, Rect(0, 0, R.Right - min(w, Max(R.Left, R.Right - w + 1) + 1), h));
      end else
      begin
        R2:=Rect(R.Left, R.Top + ((R.Bottom-R.Top + 1) div 2)-(h div 2), Min(R.Right,R. Left+w), R.Top+((R.Bottom-R.Top+1) div 2)- (h div 2) + h);
        Item.FTextLeft:=R.Left+FHPos;
        Item.FTextRight:=R.Right+FHPos;
        Canvas.CopyRect(R2, CurStyle.FPicture.Canvas, Rect(0, 0, min(w, Min(R.Right, R.Left+w-1)-R.Left+1), h));
      end;
    end;
  end; // if ... else
end;

procedure TCustomElTree.DoRedrawItem;
var
    TSI     : TElTreeItem;
    R,R1,R2 : TRect;
    SText: string;
    i,j,h,w,
    al   : integer;
    HS   : TElHeaderSection;
    flv : integer;
    CurStyle : TElCellStyle;
    
begin
  TSI:=Item;
  r1:=Rect(ItemRect.Left-FHPos,ItemRect.Top, SurfRect.Right-1-FHPos, ItemRect.Bottom);
  flv := FDivLineWidth;
  if (not VerticalLines) or FBarStyle then flv := 0;
  FOverColors := false;
  if TSI.Selected and (not FBarStyle) and FRowSelect and (((GetParentForm(self)<>nil) and (GetParentForm(self).ActiveControl=self))
  or (not FHideSelect)) then
  begin
    Canvas.Brush.Color:=clHighlight;
    Canvas.Font.Color := clHighlightText;
    FCurBkColor := clHighlight;
    FCurTextColor := clHighlightText;
    FOverColors := true;
  end; // if
  if FRowSelect then
  begin
    if (TSI=FDropTrg) then
    begin
      case FDragTrgDraw of
      ColorRect:
        begin
          if FDropAcc
            then Canvas.Brush.Color:=clGreen
            else Canvas.Brush.Color:=clRed;
          FCurBkColor := Canvas.Brush.Color;
          Canvas.Font.Color:=clBtnText;
          FCurTextColor := clBtnText;
          FOverColors := true;
        end;
      SelColorRect:
        begin
          Canvas.Brush.Color:=clHighlight;
          Canvas.Font.Color:=clHighlightText;
          FCurBkColor := clHighlight;
          FCurTextColor := clHighlightText;
          FOverColors := true;
        end;
      end;    // case
    end;
  end;

  if FHeader.Visible then
    for j := 0 to FHeader.Sections.Count-1 do    //Iterate
    begin
      HS:=FHeader.Sections[j];
      if HS.Visible then
      begin
        if j=FMainTreeCol then
            DoRedrawItemTree(Item,ItemRect, SurfRect)
          else
        begin
        // check the proper section and subtext index
          i:=j;
          if i>FMainTreeCol then dec(i);
          if i>=Item.ColumnText.count then SText:='' else
          if HS.Password then SText:='******' else SText:=Item.ColumnText[i];

          if Item.FUseStyles then
          begin
            if (i >= Item.FStyles.Count) or (HS.UseMainStyle)
               then CurStyle := Item.FMainStyle
               else CurStyle := Item.FStyles[i];
            if FOverColors or CurStyle.FOwnerProps then
            begin
              Canvas.Brush.Color:=FCurBkColor;
              Canvas.Font.Color:=FCurTextColor;
            end
            else
            begin
              Canvas.Brush.Color := CurStyle.FCellBkColor;
            end;
          end else
          begin
            Canvas.Brush.Color:=FCurBkColor;
            Canvas.Font.Color:=FCurTextColor;
          end;
          if Item.Selected and (not FBarStyle) and ((FSelectColumn = j) or FRowSelect) and (((GetParentForm(self)<>nil) and (GetParentForm(self).ActiveControl=self)) or (not FHideSelect)) then
            Canvas.Brush.Color:=clHighlight;
          if (TSI=FDropTrg) and (((not FRowSelect) and (FSelectColumn = j)) or FRowSelect) then
          begin
            case FDragTrgDraw of
              ColorRect:
                begin
                  if FDropAcc
                    then Canvas.Brush.Color:=clGreen
                    else Canvas.Brush.Color:=clRed;
                end;
              SelColorRect:
                  Canvas.Brush.Color:=clHighlight;
            end;    // case
          end;
          // now fill the background
          R:=Rect(HS.Left - FHPos, ItemRect.Top, HS.Right - FHPos - flv, ItemRect.Bottom);
          if (not FHLines) or FBarStyle then inc(R.Bottom, FDivLineWidth);
          Canvas.FillRect(R);

          if (FODFollowCol and (HS.Style = ElhsOwnerDraw)) or ((not (FODFollowCol)) and (SText=FODMask))
             or (Item.FUseStyles and (CurStyle.FStyle = elhsOwnerDraw)) then
          begin
            R:=Rect(HS.Left+1-FHPos, ItemRect.Top+1, HS.Right-(FHPos+flv+1), ItemRect.Bottom-1);
            DoItemDraw(Item, Canvas, R, j);
          end else
          begin
            if (Item.UseStyles and (CurStyle.FStyle = elhsText)) or (not Item.FUseStyles) then
            begin
              if Item.FUseStyles then
              begin
                AL := CurStyle.FTextFlags;
                if FOverColors or CurStyle.FOwnerProps then
                begin
                  Canvas.Font.Color := FCurTextColor;
                  Canvas.Brush.Color := FCurBkColor;
                end
                else
                begin
                  Canvas.Brush.Color := CurStyle.FTextBkColor;
                  Canvas.Font.Color := CurStyle.FTextColor;
                end;
                if not CurStyle.FOwnerProps then with CurStyle do
                begin
                  if Canvas.Font.Name <> FontName then Canvas.Font.Name := FontName;
                  if Canvas.Font.Style <> FontStyles then Canvas.Font.Style := FontStyles;
                  if Canvas.Font.Size <> FontSize then Canvas.Font.Size := FontSize;
                end else
                begin
                  if Canvas.Font.Name <> Font.Name then Canvas.Font.Name := Font.Name;
                  if Canvas.Font.Style <> Font.Style then Canvas.Font.Style := Font.Style;
                  if Canvas.Font.Size <> Font.Size then Canvas.Font.Size := Font.Size;
                end;
              end else
              begin
                if Canvas.Font.Name <> Font.Name then Canvas.Font.Name := Font.Name;
                if Canvas.Font.Style <> Font.Style then Canvas.Font.Style := Font.Style;
                if Canvas.Font.Size <> Font.Size then Canvas.Font.Size := Font.Size;
                if FRightAlignedText then al := DT_RIGHT else Al:=DT_LEFT;
                if FRightAlignedText then
                begin
                  case HS.Alignment of
                    hsaCenter: Al:=DT_CENTER;
                    hsaRight:  Al:=DT_RIGHT;
                  end;    // case
                  AL := AL or DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER or DT_RTLREADING or DT_END_ELLIPSIS;
                end else
                begin
                  case HS.Alignment of
                    hsaCenter: Al:=DT_CENTER;
                    hsaRight:  Al:=DT_RIGHT;
                  end;    // case
                  AL := AL or DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER or DT_END_ELLIPSIS;
                end;
              end;
              if Item.Selected and (not FBarStyle) and ((FSelectColumn = j) or FRowSelect) and (((GetParentForm(self)<>nil) and (GetParentForm(self).ActiveControl=self)) or (not FHideSelect)) then
              begin
                Canvas.Brush.Color:=clHighlight;
                Canvas.Font.Color:=clHighlightText;
              end;
              if (TSI=FDropTrg) and (((not FRowSelect) and (FSelectColumn = j)) or FRowSelect) then
              begin
                case FDragTrgDraw of
                ColorRect:
                  begin
                    if FDropAcc
                      then Canvas.Brush.Color:=clGreen
                      else Canvas.Brush.Color:=clRed;
                    Canvas.Font.Color:=clBtnText;
                  end;
                SelColorRect:
                  begin
                    Canvas.Brush.Color:=clHighlight;
                    Canvas.Font.Color:=clHighlightText;
                  end;
                end;    // case
              end;
              case (AL and (DT_LEFT or DT_RIGHT or DT_CENTER)) of
                DT_LEFT: R.Right := Min (R.Left + Canvas.TextWidth(SText) + 3, R.Right);
                DT_RIGHT: R.Left := Max (R.Right - (Canvas.TextWidth(SText) + 3), R.Left);
                DT_CENTER:
                  begin
                    i := Min(R.Right - R.Left, Canvas.TextWidth(SText) + 3);
                    R.Left := (R.Right - R.Left) div 2 - i div 2;
                    R.Right := R.Left + i;
                  end;
              end;
              // define the rectangle for the text
              R:=Rect(HS.Left - FHPos, ItemRect.Top, HS.Right - (FHPos + flv), ItemRect.Bottom);
              if FHLines then inc(R.Bottom, FDivLineWidth);

              // draw the text
              if R.Left<R.Right then Canvas.FillRect(R);
              InflateRect(R, -1, -1);
              DrawText(Canvas.Handle, PChar(SText), -1, R, AL or DT_EXTERNALLEADING);
            end else
            if Item.UseStyles and (CurStyle.FStyle = elhsPictureOnly) and Assigned(CurStyle.FPicture) then
            begin
              R:=Rect(HS.Left - FHPos, ItemRect.Top, HS.Right - (FHPos + flv), ItemRect.Bottom);
              h := CurStyle.FPicture.Height;
              w := CurStyle.FPicture.Width;
              R2:=Rect(R.Left, R.Top + ((R.Bottom-R.Top + 1) div 2)-(h div 2), Min(R.Right,R. Left+w), R.Top+((R.Bottom-R.Top+1) div 2)- (h div 2) + h);
              Canvas.CopyRect(R2, CurStyle.FPicture.Canvas, Rect(0, 0, min(w, Min(R.Right, R.Left+w-1)-R.Left+1), h));
            end;
          end; // if/else
        end;
      end;
    end    //for
  else DoRedrawItemTree(Item, ItemRect, SurfRect);

  if not (FRowSelect) then
  begin
    if (FSelectColumn = -1) or (not FHeader.Visible) or (FHeader.Sections.Count <= FSelectColumn) then
    begin
      R1.left:=TSI.FTextLeft-FHPos;
      R1.Right:=TSI.FTextRight-FHPos;
    end else
    begin
      HS := FHeader.Sections[FSelectColumn];
      R1.Left := HS.Left - FHPos;
      R1.Right := R1.Left + HS.Width;
    end;
  end else dec(R1.Right,flv -1);
  if not HorizontalLines then inc (R1.Bottom);
  if (TSI.Focused) then
  begin
    Canvas.Brush.Color:=FBkColor;
    Canvas.Pen.Color:=FTextColor;
    Canvas.Font.Color:=FTextColor;
    // this one is made to fix the colors
    Canvas.TextOut(R1.Left,R1.Top,'');
    // otherwise FocusRect won't draw right
    if FDrawFocusRect then Canvas.DrawFocusRect(R1);
  end;
  if (TSI=FDropTrg) then
  begin
    if FDragTrgDraw = ColorFrame then
    begin
      if FDropAcc
        then Canvas.Brush.Color:=clGreen
        else Canvas.Brush.Color:=clRed;
    end;
    Canvas.FrameRect(R1);
  end;
end;

procedure TCustomElTree.DoEditItem;
var s:string;
    R:TRect;
    Left,sn : integer;
    Section : TElHeaderSection;
    b : boolean;
    Style : TElCellStyle;
    i : integer;
    FT : TFieldTypes;
begin
  if Item=nil then raise EElTreeError.Create(GetErrorMsg(STexInvItem));
  if FVisible.IndexOf(Item)=-1 then EnsureVisible(Item);
  if FInpEdit<>nil then
  begin
    if FInpEdit.Visible then EndEdit(true);
    FInpEdit.Parent := nil;
    if FFakePopup <> nil then FFakePopup.Free;
    FFakePopup := nil;
    FInpEdit.Free;
    FInpEdit:=nil;
  end;
  if FHeader.Visible and (SectionNum = -1) then SectionNum := FMainTreeCol;
  FEditingItem:=Item;
  FEditingSect:=SectionNum;
  R:=GetItemRect(FVisible.IndexOf(Item));
  if (SectionNum <> -1) then
  begin
     Section := FHeader.Sections[SectionNum];
     if (not Section.Visible) then
     begin
       FEditingItem:=nil;
       exit;
     end;
  end else Section := nil;
  if Item.FUseStyles then
  begin
    if (SectionNum = FMainTreeCol) or (SectionNum = -1) then Style := Item.FMainStyle else
    begin
      i := SectionNum;
      if i > FMainTreeCol then dec(i);
      if Item.FStyles.Count > i then Style := Item.Styles[i] else Style := Item.FMainStyle;
    end;
    FT := Style.FCellType;
  end else
  begin
    if Section <> nil then FT := Section.FieldType else FT := sftText;
  end;
  b := true;
  TriggerTryEditEvent(Item, Section, FT, b);
  if not b then
  begin
    FEditingItem := nil;
    exit;
  end;
  case FT of
    sftPicture,
    sftCustom,
    sftBLOB:
      begin
        FEditing := true;
        TriggerEditRequestEvent(Item, Section);
        FEditing := false;
        FEditingItem := nil;
        exit;
      end;
    sftEnum:
      begin
        FInpEdit:=TComboBox.Create(self);
        FFakePopup := TPopupMenu.Create(nil);
        FFakePopup.AutoPopup := false;
        FInpEdit.Parent:=self;
        (FInpEdit as TComboBox).PopupMenu := FFakePopup;
        FInpEdit.Visible:=false;
        (FInpEdit as TComboBox).Style := csDropDownList;
        (FInpEdit as TComboBox).ParentCtl3D:=false;
        (FInpEdit as TComboBox).Ctl3D:=false;
        (FInpEdit as TComboBox).OnExit:=OnEditExit;
        (FInpEdit as TComboBox).OnKeyDown:=OnEditKeyDown;
        (FInpEdit as TComboBox).ItemIndex:=0;
        TriggerComboEditShowEvent(Item, Section, FInpEdit as TComboBox);
      end;
{$IFDEF USE_RX}
    sftDate:
      begin
        FInpEdit:=TDateEdit.Create(self);
        FFakePopup := TPopupMenu.Create;
        FFakePopup.AutoPopup := false;
        FInpEdit.Parent:=self;
        (FInpEdit as TDateEdit).PopupMenu := FFakePopup;
        FInpEdit.Visible:=false;
        (FInpEdit as TDateEdit).DefaultToday := true;
      end;
{$ENDIF}
    else FInpEdit:=nil;
  end; // case
  if FInpEdit = nil then
  begin
    FInpEdit:=TEdit.Create(self);
    FFakePopup := TPopupMenu.Create(nil);
    FFakePopup.AutoPopup := false;
    FInpEdit.Parent:=self;
    (FInpEdit as TEdit).PopupMenu := FFakePopup;
    FInpEdit.Visible:=false;
    (FInpEdit as TEdit).BorderStyle:=bsSingle;
    (FInpEdit as TEdit).ParentCtl3D:=false;
    (FInpEdit as TEdit).Ctl3D:=false;
    (FInpEdit as TEdit).OnExit:=OnEditExit;
    (FInpEdit as TEdit).OnKeyDown:=OnEditKeyDown;
    (FInpEdit as TEdit).AutoSize:=false;
    (FInpEdit as TEdit).BoundsRect:=Rect(0,0,0,0);
  end; //if
  if (SectionNum = -1) or (SectionNum = FMainTreeCol) then
  begin
    if (Item.FTextLeft<FHPos) or (Item.FTextLeft>FHPos+ClientWidth) then SetHPosition(Item.FTextLeft);
    R.Left:=Item.FTextLeft-FHPos;
    if assigned(FInpEdit) and (FInpEdit is TCustomEdit) then
    begin
      if Section <> nil then R.Right := Section.Right else R.Right := ClientWidth - 1;
      if R.Right>=ClientWidth then R.Right:=ClientWidth-1;
      s:=Item.Text;
    end;
    if assigned(FInpEdit) and (FInpEdit is TComboBox) then
    begin
      if Section <> nil then R.Right := Section.Right else R.Right := ClientWidth - 1;
      if R.Right>=ClientWidth then R.Right:=ClientWidth-1;
    end;
  end  //if
  else
  begin
    Left:=Section.Left;
    if (Left<FHPos) or (Left>FHPos+ClientWidth) then SetHPosition(Left);
    R.Left:=Left-FHPos;
    if assigned(FInpEdit) and (FInpEdit is TCustomEdit) then
    begin
      sn:=SectionNum;
      if sn>FMainTreeCol then dec(sn);
      if Item.FColText.Count <= sn then s := '' else s:=Item.FColText[sn];
      R.Right:=R.Left + Section.Width;
      if R.Right>=ClientWidth then R.Right:=ClientWidth-1;
    end;
    if assigned(FInpEdit) and (FInpEdit is TComboBox) then
    begin
      R.Right:=R.Left+Section.Width;
      if R.Right>=ClientWidth then R.Right:=ClientWidth-1;
    end; // if
  end; // else
  if assigned(FInpEdit) then
  begin
    FInpEdit.BoundsRect:=R;
    if FInpEdit is TCustomEdit then
    begin
      b:= (FInpEdit as TEdit).Font.Height<0;
      if b
          then (FInpEdit as TCustomEdit).Height:=-Font.Height+2
          else (FInpEdit as TCustomEdit).Height:=Font.Height+2;
      TCustomEdit(FInpEdit).Text := s;
      with (FInpEdit as TCustomEdit) do
      begin
        Modified:=false;
        SelStart:=0;
        SelLength:=Length(Text);
        Height := Max(Height, R.Bottom - R.Top);
      end;
      FOldHide := FHideSelect;
      FHideSelect := false;
      FInpEdit.Visible:=true;
      FInpEdit.SetFocus;
      TEdit(FInpEdit).Font.Assign(Font);
      FEditing := true;
    end;
    if FInpEdit is TComboBox then
    begin
      (FInpEdit as TComboBox).Font.Assign(Font);
      if (FInpEdit as TComboBox).Font.Height<0
        then (FInpEdit as TComboBox).Font.Height:=-Font.Height-3
        else (FInpEdit as TComboBox).Font.Height:=Font.Height-3;
      (FInpEdit as TComboBox).Height := Max((FInpEdit as TComboBox).Height, R.Bottom - R.Top);
      FOldHide := FHideSelect;
      FHideSelect := false;
      FInpEdit.Visible:=true;
      try
        FInpEdit.SetFocus;
        FEditing := true;
      except
      end;
    end;
  end;
end;

procedure TCustomElTree.EditItem;
begin
  if not FCanEdit then exit;
  try
    if FInpEdit<>nil then EndEdit(true);
  except
  end;
  DoEditItem(Item, SectionNum);
end;

procedure TCustomElTree.EndEdit;
var b:boolean;
    s:string;
    sn:integer;
begin
  HideSelection := FOldHide;
  if (FInpEdit=nil) or (FEditingItem=nil) or (not FInpEdit.Visible) then
  begin
    if assigned(FInpEdit) then
    begin
      FInpEdit.Parent := nil;
      FInpEdit.Free;
      if Assigned(FFakePopup) then FFakePopup.Free;
      FFakePopup := nil;
      FInpEdit := nil;
    end;
    FEditingItem:=nil;
    try
      if (not Destroying) and CanFocus then SetFocus;
    except
    end;
    FEditing := false;
    exit;
  end;
  //if not ByCancel then
  begin
    if (FInpEdit is TCustomEdit) then
    begin
      b:=not ByCancel and (FInpEdit as TCustomEdit).Modified;
      if b then s:=(FInpEdit as TCustomEdit).Text else s := '';
      if (FHeader.Visible) and (FEditingSect<>-1) then
        DoValidate(FEditingItem, FHeader.Sections[FEditingSect],s,b)
      else
        DoValidate(FEditingItem, nil,s,b);
      if FEditingItem <> nil then
      begin
        if (FEditingSect=FMainTreeCol) or (FEditingSect = -1) then
        begin
          if b then FEditingItem.Text:=s;
        end else
        if b then
        begin
          sn:=FEditingSect;
          if sn>FMainTreeCol then dec(sn);
          while FEditingItem.FColText.Count <= sn do FEditingItem.FColText.Add('');
          FEditingItem.FColText[sn] := s;
        end;
      end;
    end else
{$IFDEF USE_RX}
    if (FInpEdit is TDateEdit) then
    begin
      b:=not ByCancel;
      if b then DateToStr((FInpEdit as TDateEdit).Date) else s := '';
      DoValidate(FEditingItem,FHeader.Sections[FEditingSect],s,b);
      if FEditingItem <> nil then
      begin
        if (FEditingSect=FMainTreeCol) or (FEditingSect = -1) then
        begin
          if b then FEditingItem.Text:=s;
        end else
        if b then
        begin
          sn:=FEditingSect;
          if sn>FMainTreeCol then dec(sn);
          while FEditingItem.FColText.Count<= sn do FEditingItem.FColText.Add('');
          FEditingItem.FColText[sn]:=s;
        end;
      end;
    end else
{$ENDIF}
    if (FInpEdit is TComboBox) then
    begin
      b:= not ByCancel;
      if (FEditingSect=FMainTreeCol) or (FEditingSect = -1)
         then DoValidateCombo(FEditingItem, nil, FInpEdit as TComboBox, b)
         else DoValidateCombo(FEditingItem, FHeader.Sections[FEditingSect], FInpEdit as TComboBox, b);
    end;
  end;
  if FInpEdit <> nil then
  begin
    if FInpEdit is TEdit then (FInpEdit as TEdit).OnExit:=nil;
    if FInpEdit is TComboBox then (FInpEdit as TComboBox).OnExit:=nil;
    FInpEdit.Visible:=false;
  end;
  if (FEditingItem<>nil) and (not Destroying) then
  begin
    FEditingItem.RedrawItem(false);
    FUpdated := true;
    if not IsUpdating then Update;
  end;
  FEditingItem:=nil;
  try
    if (not Destroying) and CanFocus then SetFocus;
  except
  end;
  FEditing := false;
end;

procedure TCustomElTree.OnEditExit;
begin
  EndEdit(False);
end;

function TCustomElTree.GetShowHeader;
begin
  result:= (FHeader<>nil) and (FHeader.Visible=true);
end;

procedure TCustomElTree.SetShowHeader;
begin
  if (FHeader=nil) then exit;
  if not (csReading in ComponentState) and (Value) and (FHeader.Sections.Count=0) then exit;
  FHeader.Visible:=value;
  if value=false then FHRange:=-1;
  //if csReading in ComponentState then exit;
  IsUpdating:=true;
  FUpdated:=true;
  FVisUpdated:=true;
  FClearAll:=true;
  FClearVis:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetMainTreeCol;
begin
  if value=FMainTreeCol then exit;
  if value>=FHeader.Sections.Count then FMainTreeCol:=0 else FMainTreeCol:=value;
  IsUpdating:=true;
  FRangeUpdate := true;
  FUpdated:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.DoHeaderResize;
var i:Integer;
begin
  if FPainting then exit;
  IsUpdating:=true;
  FUpdated:=true;
  FRangeUpdate := true;
  FHeader.Left:=-FHPos;
  i:=FHPos+ClientWidth;
  FHeader.Width:=i+1;
  FClearVis:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.DoHeaderDraw;
begin
  if assigned(FOnColumnDraw) then FOnColumnDraw(Header,Section,Rect,Pressed);
end;

procedure TCustomElTree.DoEndDrag(Target: TObject; X, Y: Integer);
var TSI: TElTreeItem;
begin
  inherited DoEndDrag(Target, X, Y);
  FDragImages.Free;
  FDragImages:=nil;
  TSI:=FDropTrg;
  FDropTrg:=nil;
  if TSI<>nil then TSI.RedrawItem(true);
end;

procedure TCustomElTree.CMFontChanged(var Message: TMessage);
begin
  inherited;
  if Canvas <> nil then Canvas.Font.Assign(Font);
//  if FInpEdit <>nil then FInpEdit.Font.Assign(Font);
  if FHeader <>nil then FHeader.Font.Assign(Font);
  IsUpdating:=true;
  FRangeUpdate := true;
  FUpdated:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.CMColorChanged(var Message: TMessage);
begin
  inherited;
  if Canvas <> nil then Canvas.Brush.Color := Color;
end;

procedure TCustomElTree.SetHeaderHeight(value:integer);
begin
  IsUpdating:=true;
  FHeader.Height:=value;
  FClearAll:=true;
  FClearVis:=true;
  FUpdated:=true;
  IsUpdating:=false;
end;

function TCustomElTree.GetHeaderHeight;
begin
  Result:=FHeader.Height;
end;

procedure TCustomElTree.SetHeaderSections;
begin
  IsUpdating:=true;
  FHeader.Sections:=value;
  FUpdated:=true;
  FRangeUpdate := true;
  IsUpdating:=false;
end;

function TCustomElTree.GetHeaderSections;
begin
  Result:=FHeader.Sections;
end;

procedure TCustomElTree.SaveStringsToStream(Stream: TStream);

  procedure IntSaveString(Item:TElTreeItem; Index : integer; var ContinueIterate:boolean;
                                            IterateData:pointer; Tree:TCustomElTree);
  var   j: Integer;
        ANode: TElTreeItem;
        NodeStr: string;
        s      : string;
  begin
      ANode:=Item;
      NodeStr := '';
      j:=ANode.Level - 1;
      SetLength(NodeStr,j);
      FillMemory(PChar(NodeStr),j,32);
      s:=ANode.Text;
      NodeStr := NodeStr + s + #13#10;
      TStream(IterateData).Write(Pointer(NodeStr)^, Length(NodeStr));
  end;

begin
  FItems.Iterate(false,@IntSaveString,Stream);
end;

function TCustomElTree.GetItemAtY(y:integer):TElTreeItem;
var i:integer;
begin
  if FHeader.Visible then dec(y, GetHeaderHeight);
  i:=y div FLineHeight;
  if (i<0) or (i>=FVisible.Count) then result:=nil else result:=TElTreeItem(FVisible[i]);
end;

procedure TCustomElTree.FullCollapse;
var i:integer;
begin
  for i:=0 to FItems.FRoot.FChildren.Count-1 do
  TElTreeItem(FItems.FRoot.FChildren[i]).Collapse(true);
end;

procedure TCustomElTree.FullExpand;
var i:integer;
begin
  for i:=0 to FItems.FRoot.FChildren.Count-1 do
  TElTreeItem(FItems.FRoot.FChildren[i]).Expand(true);
end;

procedure TCustomElTree.SetTextColor(value:TColor);
begin
  if FTextColor=value then exit;
  FTextColor:=value;
  IsUpdating:=true;
  FUpdated:=true;
  FVisUpdated:=true;
  FClearVis:=true;
  FClearAll:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetBkColor(value:TColor);
begin
  if FBkColor=value then exit;
  FBkColor:=value;
  IsUpdating:=true;
  FUpdated:=true;
  FVisUpdated:=true;
  FClearVis:=true;
  FClearAll:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetItems(value:TElTreeItems);
begin
  FItems.Assign(value);
end;

Procedure TCustomElTree.SetHPosition(value:integer);
var Code: Word;
{$IFNDEF VER90}
    Form: TCustomForm;
{$ELSE}
    Form : TForm;
{$ENDIF}
    I: integer;
begin
  IsUpdating:=true;
  if csReading in ComponentState then FHPos := Value
  else
  begin
    if Value > FHRange-ClientWidth then Value := FHRange-ClientWidth+1;
    if Value < 0 then Value := 0;
    if Value <> FHPos then
    begin
      FHPos := Value;
      FHeader.Left:=-FHPos;
      i:=FHPos+ClientWidth;
      FHeader.Width:=i+1;
      FClearAll:=true;
      if csDesigning in ComponentState then
      begin
        Form := GetParentForm(self);
        if (Form <> nil) and (Form.Designer <> nil) then Form.Designer.Modified;
      end;
      FUpdated:=true;
      FRangeUpdate := true;
    end;
    Code:=SB_HORZ;
    if Windows.GetScrollPos(Handle, Code) <> FHPos then
    SetScrollPos(Handle, Code, FHPos, True);
  end;
  IsUpdating:=false;
end;

procedure TCustomElTree.WMHScroll;
begin
  if FInpEdit<>nil then EndEdit(false);
  with Message do
    case ScrollCode of
      SB_LINEUP: SetHPosition(FHPos-4);
      SB_LINEDOWN: SetHPosition(FHPos+4);
      SB_PAGEUP: SetHPosition(FHPos - ClientWidth div 2);
      SB_PAGEDOWN: SetHPosition(FHPos + ClientWidth div 2);
      SB_THUMBPOSITION: SetHPosition(Pos);
      SB_THUMBTRACK: if FScrollTracking then SetHPosition(Pos);
      SB_TOP: SetHPosition(0);
      SB_BOTTOM: SetHPosition(FHRange);
    end; // case
    TriggerScrollEvent(sbHorizontal, Message.ScrollCode);
end;

Procedure TCustomElTree.SetVPosition(value:integer);
var Code: Word;
{$IFNDEF VER90}
    Form: TCustomForm;
{$ELSE}
    Form : TForm;
{$ENDIF}

begin
  IsUpdating:=true;
  FTrackItem:=nil;
  if csReading in ComponentState then FTopIndex := Value
  else
  begin
    if Value >(TotalVisCount-GetVisCount+1) then Value := (TotalVisCount-GetVisCount+1);
    if Value < 0 then Value := 0;
    if Value <> FTopIndex then
    begin
      FUpdated:=true;
      FClearVis:=true;
      FClearAll:=true;
      DoSetTopIndex(Value);
    end;
    if csDesigning in ComponentState then
    begin
      Form := GetParentForm(self);
      if (Form <> nil) and (Form.Designer <> nil) then Form.Designer.Modified;
    end;
    Code:=SB_Vert;
//    UpdateScrollBars;
    SetScrollPos(Handle, Code, FTopIndex, True);
  end;
  IsUpdating:=false;
end;

procedure TCustomElTree.WMVScroll;
begin
  if FInpEdit<>nil then EndEdit(false);
  with Message do
    case ScrollCode of
      SB_LINEUP: SetVPosition(FTopIndex-1);
      SB_LINEDOWN: SetVPosition(FTopIndex+1);
      SB_PAGEUP: SetVPosition(FTopIndex-(GetVisCount-1));
      SB_PAGEDOWN: SetVPosition(FTopIndex+(GetVisCount-1));
      SB_THUMBPOSITION: SetVPosition(Pos);
      SB_TOP: SetVPosition(0);
      SB_BOTTOM: SetVPosition(TotalVisCount);
      SB_THUMBTRACK: if FScrollTracking then SetVPosition(Pos);
    end; // case
    TriggerScrollEvent(sbVertical, Message.ScrollCode);
end;

procedure TCustomElTree.OnHintTimer;
var Item:TElTreeItem;
begin
  if FHintWnd<>nil then
  begin
    DoHideLineHint;
    FHintTimer.Free;
    FHintTimer:=nil;
    exit;
  end;
  FHintTimer.Enabled:=false;
  Item:=GetItemAtY(FHintCoord.Y);
  if Item=nil then exit;
  if (not (csDesigning in ComponentState)) and
     ((FShowHintMode=shmAll) or
     ((FShowHintMode=shmLong) and
     (((Canvas<>nil) and (Canvas.TextWidth(Item.Text) > (Item.FTextRight-Item.FTextLeft)))
      or ((Item.FTextRight>ClientWidth+FHPos) or (Item.FTextLeft<FHPos))))) then
       DoShowLineHint(Item);
  if FHintHide then
  begin
    FHintTimer.Interval:=3000;
    FHintTimer.Enabled:=true;
  end;
end;

procedure TCustomElTree.WMLButtonUp(var Message: TWMLButtonUp);
var Item:TElTreeItem;
    ItemPart: TSTItemPart;
    HCol: Integer;

begin
  FPressed:=false;
  Item:=GetItemAt(Message.XPos, Message.YPos, ItemPart, HCol);
  if (ItemPart = ipMainText) and (not FIgnoreClick2) then
  begin
    if (Item = FClickItem) and FCanEdit and FClicked then
    begin
      //if (sqrt(sqr(FClickCoord.X-Message.XPos)+sqr(FClickCoord.Y-Message.YPos))<5) then
      if (FHeader.Visible) then
      begin
        if FHeader.Sections[FMainTreeCol].Editable then DoEditItem(Item, FMainTreeCol);
      end else DoEditItem(Item, -1);
      FClicked:=false;
      inherited;
      exit;
    end;
  end;
  if (FPressCoord.Y = Message.YPos) and (FPressCoord.X = Message.XPos)
  // if the mouse was moved when the button was pressed, we don't check for editing
     and (ItemPart<>ipInside) and (ItemPart <> ipButton) then
  begin
    if not FIgnoreClick2 then
    begin
      FClicked:=true;
      FClickCoord:=Point(Message.XPos, Message.YPos);
      FClickItem:=Item;
    end;
  end;
  if Item = nil then FClickItem := nil;
  FIgnoreClick2 := false;
  inherited;
end;

procedure TCustomElTree.WMLButtonDblClk(var Message: TWMLButtonDblClk);
var
  ItemPart: TSTItemPart;
  HCol: Integer;
  Item: TElTreeItem;

begin
  HCol:=0;
  Item:=GetItemAt(Message.XPos,Message.YPos,ItemPart,HCol);
  if ((ItemPart=ipMainText) or ((Item <> nil) and (FFullRowSelect) and (ItemPart <> ipCheckBox))) and Item.Enabled and FExpandOnDblClick then Item.Expanded := not (Item.Expanded);
  if (ItemPart = ipColumn) and (FCanEdit) and (FHeader.Sections[HCol].Editable) and Item.Enabled then DoEditItem(Item, HCol);
  FClicked := false;
  FClickItem:=nil;
  FIgnoreClick2 := true;
  FIgnoreClick:=true;
  if ItemPart <> ipButton then inherited;
  FIgnoreClick:=false;
end;

procedure TCustomElTree.WMLButtonDown(var Message: TWMLButtonDown);
begin
  FIgnoreClick2 := false;
  if (not (csDesigning in ComponentState)) and (not Focused) and CanFocus then SetFocus;
  FPressCoord := Point(Message.XPos,Message.YPos);
  FPressed:=true;
  inherited;
end;

procedure TCustomElTree.WMRButtonDown(var Message: TWMRButtonDown);
begin
  if not (csDesigning in ComponentState) and
    (CanFocus or (GetParentForm(Self) = nil)) then SetFocus;
  inherited;
end;

procedure TCustomElTree.WMMouseMove(var Message: TWMMouseMove);
var TSI: TElTreeItem;
    R : Trect;
{$IFNDEF VER90}
{$IFNDEF LITE}
    dataObj: IDataObject;
    dropSource: IDropSource;
    OKEffects : TDragTypes;
    Effect : TDragType;
    dwOKEffects,
    dwEffect : integer;
    HRes : HResult;
{$ENDIF}
{$ENDIF}
begin
  inherited;
  if not Dragging then
  begin
    if FPressed then
    begin
      if (FSelected<>nil) and (Focused) and (FDragAllowed) and FSelected.FEnabled
         and (sqrt(sqr(FPressCoord.X-Message.XPos)+sqr(FPressCoord.Y-Message.YPos))>=5) then
      begin
        if DragType <> dtOLE then
        begin
          FIgnoreClick2 := true;
          BeginDrag(true);
        end
{$IFDEF VER90}
;
{$ELSE}
        else
        begin
{$IFNDEF LITE}
          dataObj := nil;
          dropSource := nil;
          OKEffects := [];
          TriggerOleDragStartEvent(dataObj, dropSource, OKEffects);
          dwOkEffects := 0;
          if dtCopy in OKEffects then dwOkEffects := dwOkEffects or DROPEFFECT_COPY;
          if dtMove in OKEffects then dwOkEffects := dwOkEffects or DROPEFFECT_MOVE;
          if dtLink in OKEffects then dwOkEffects := dwOkEffects or DROPEFFECT_LINK;
          FPressed:=false;
          if Assigned(DataObj) and Assigned(dropSource) then
          begin
            dwEffect := DROPEFFECT_NONE;
            Effect := dtNone;
            HRes := DoDragDrop(dataObj, dropSource, dwOKEffects, dwEffect);
            if (dwEffect and DROPEFFECT_COPY) <> 0 then Effect := dtCopy;
            if (dwEffect and DROPEFFECT_LINK) <> 0 then Effect := dtLink;
            if (dwEffect and DROPEFFECT_Move) <> 0 then Effect := dtMove;
            TriggerOleDragFinishEvent(Effect, HRes);
          end;
{$ENDIF}          
        end;
{$ENDIF}
      end;
    end;
    if FShowHintMode<>shmNone then
    begin
      if (FHintCoord.X<>Message.XPos) or (FHintCoord.Y<>Message.YPos) then
      begin
        if (FHideHintOnMove) or (GetItemAtY(FHintCoord.Y)<>GetItemAtY(Message.YPos)) then
        begin
          DoHideLineHint;
          FHintCoord:=Point(Message.XPos,Message.YPos);
          FHintTimer:=TTimer.Create(self);
          FHintTimer.Enabled:=false;
          FHintTimer.OnTimer:=OnHintTimer;
          FHintTimer.Interval:=1000;
          FHintTimer.Enabled:=true;
        end;
      end;
    end;
    if FTracking and Focused then
    begin
      TSI:=FTrackItem;
      FTrackItem:=GetItemAtY(Message.YPos);
      if TSI=FTrackItem then exit;
      if TSI<>nil then
      begin
        if not FHeader.Visible then TSI.RedrawItem(false)
        else
        begin
          R := GetItemRect(FVisible.IndexOf(TSI));
          R.Left := FHeader.Sections[FMainTreeCol].Left - FHPos;
          R.Right := FHeader.Sections[FMainTreeCol].Right - FHPos;
          InvalidateRect(Handle, @R, false);
        end;
      end;
      if FTrackItem<>nil then
      begin
        if not FHeader.Visible then FTrackItem.RedrawItem(false)
        else
        begin
          R := GetItemRect(FVisible.IndexOf(FTrackItem));
          R.Left := FHeader.Sections[FMainTreeCol].Left - FHPos;
          R.Right := FHeader.Sections[FMainTreeCol].Right - FHPos;
          InvalidateRect(Handle, @R, false);
        end;
      end;
      TriggerHotTrackEvent(TSI, FTrackItem);
      Update;
    end; // if (FTracking and Focused)
  end; //if (not Dragging)
end;

procedure TCustomElTree.CMDrag(var Message: TCMDrag);
var TSI : TElTreeItem;
begin
  inherited;
  with Message, DragRec^ do
    case DragMessage of
      dmDragMove: with ScreenToClient(Pos) do DoDragOver(Source, X, Y, Message.Result<>0);
      dmDragLeave:
        begin
          TDragObject(Source).HideDragImage;
          TSI:=FDropTrg;
          FDropTrg := nil;
          if TSI<>nil then
          begin
            TSI.RedrawItem(false);
            Update;
          end;
          TDragObject(Source).ShowDragImage;
        end;
      dmDragDrop:
        begin
          TSI:=FDropTrg;
          FDropTrg := nil;
          if TSI<>nil then
          begin
            TSI.RedrawItem(false);
            Update;
          end;
        end;
      dmDragCancel:
        begin
          TSI:=FDropTrg;
          FDropTrg := nil;
          if TSI<>nil then
          begin
            TSI.RedrawItem(false);
            Update;
          end;
        end;
    end;
end;

procedure TCustomElTree.DoStartDrag(var DragObject: TDragObject);
begin
  FDropTrg := nil;
  if FSelected<>nil then FillDragImage;
  if (FDragImages<>nil) then
     if FHeader.Visible then
        FDragImages.SetDragImage(0, FPressCoord.X+FHPos-FHeader.Sections[FMainTreeCol].Left, FPressCoord.y - fddy)
     else
        FDragImages.SetDragImage(0, FPressCoord.X+FHPos, FPressCoord.y - fddy);
  inherited DoStartDrag(DragObject);
end;

function TCustomElTree.DragScroll;  { protected }
var NewY: integer;
begin
  result := false;
  NewY := 0;
  if FHeader.Visible then inc(NewY, FHeader.Height);
  if (Y < NewY+(FLineHeight div 2)) and (TopIndex>0) then
  begin
    TDragObject(Source).HideDragImage;
    result := true;
    SetVPosition(TopIndex-1);
  end else
  if (Y>ClientHeight-(FLineHeight div 2)) and (BottomIndex<TotalVisCount) then
  begin
    TDragObject(Source).HideDragImage;
    result := true;
    SetVPosition(TopIndex+1);
  end;
end;  { DragScroll }

procedure TCustomElTree.DoDragOver(Source: TDragObject; X, Y: Integer; CanDrop: Boolean);
var
  TSI: TElTreeItem;
  fh : boolean;
begin
  if (FHintTimer<>nil) or (FHintWnd <> nil) then DoHideLineHint;

  // if FHeader.Visible then inc(NewY, GetHeaderHeight);

  fh := DragScroll(Source, X, Y);

  FDropAcc:=CanDrop;
  TSI:=FDropTrg;

  FDropTrg:=GetItemAtY(Y);
  if Assigned(FDropTrg) and (not FDropTrg.Enabled) then FDropTrg := nil;
  if TSI=FDropTrg then
  begin
    if fh then TDragObject(Source).ShowDragImage;
    exit;
  end;
  if not fh then TDragObject(Source).HideDragImage;
  if TSI<>nil then TSI.RedrawItem(false);
  if FDropTrg<>nil then FDropTrg.RedrawItem(false);
  Update;
  TDragObject(Source).ShowDragImage;
end;

function TCustomElTree.CreateItems;
begin
  result:=TElTreeItems.Create(Self);
end;

procedure TCustomElTree.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
{$IFNDEF LITE}
    if AComponent = FStorage then FStorage := nil;
{$ENDIF}
    if AComponent = Images then Images := nil;
    if AComponent = Images2 then Images2 := nil;
  end;
end;

procedure TCustomElTree.SelectRangeEx(FromItem,ToItem: TElTreeItem; IncludeHidden : boolean);
type PGIRec = ^TGIRec;
     TGIRec = record
                DoAdd:boolean;
                SH : Boolean;
                FromItem,ToItem: TElTreeItem;
              end;
var GIRec : TGIRec;

  procedure IntSelRange(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                          IterateData:pointer; Tree:TCustomElTree);
  begin
    if ((PGIRec(IterateData)^.FromItem=Item)
    or (PGIRec(IterateData)^.ToItem=Item)) or (PGIRec(IterateData)^.FromItem=nil)
    or (PGIRec(IterateData)^.ToItem=nil) then
    begin
      if PGIRec(IterateData)^.DoAdd then
      begin
        ContinueIterate:=false;
        PGIRec(IterateData)^.DoAdd:=false;
      end
      else PGIRec(IterateData)^.DoAdd:=true;
    end;
    if (PGIRec(IterateData)^.DoAdd) and (not ((PGIRec(IterateData)^.FromItem=Item)
    or (PGIRec(IterateData)^.ToItem=Item))) then
    begin
      if (not PGIRec(IterateData)^.SH) or (not Item.Hidden) or (not FilteredVisibility) then
      begin
        Item.Selected := true;
        Item.UpdateItem;
      end;
    end;
  end;

begin
  if FMultiSelect = false then exit;
  if FromItem=ToItem then exit;
  IsUpdating:=true;
  GIRec.DoAdd:=false;
  GIRec.FromItem:=FromItem;
  GIRec.ToItem:=ToItem;
  GIRec.SH := IncludeHidden;
  FItems.Iterate(false, @intSelRange, @GIRec);
  IsUpdating:=false;
end;

procedure TCustomElTree.SelectRange(FromItem,ToItem: TElTreeItem);
type PGIRec = ^TGIRec;
     TGIRec = record
                DoAdd:boolean;
                FromItem,ToItem: TElTreeItem;
              end;
var GIRec : TGIRec;

  procedure IntSelRange(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                          IterateData:pointer; Tree:TCustomElTree);
  begin
    if ((PGIRec(IterateData)^.FromItem=Item)
    or (PGIRec(IterateData)^.ToItem=Item)) or (PGIRec(IterateData)^.FromItem=nil)
    or (PGIRec(IterateData)^.ToItem=nil) then
    begin
      if PGIRec(IterateData)^.DoAdd then
      begin
        ContinueIterate:=false;
        PGIRec(IterateData)^.DoAdd:=false;
      end
      else PGIRec(IterateData)^.DoAdd:=true;
    end;
    if (PGIRec(IterateData)^.DoAdd) and (not ((PGIRec(IterateData)^.FromItem=Item)
    or (PGIRec(IterateData)^.ToItem=Item))) then
    begin
      //Include(Item.FState, stsSelected);
      //inc(FSelCount);
      Item.Selected := true;
      Item.UpdateItem;
    end;
  end;

begin
  if FMultiSelect=false then exit;
  if (FromItem=ToItem) and (FromItem <> nil) then exit;
  IsUpdating:=true;
  GIRec.DoAdd:=false;
  GIRec.FromItem:=FromItem;
  GIRec.ToItem:=ToItem;
  FItems.Iterate(true, @intSelRange,@GIRec);
  IsUpdating:=false;
end;

procedure TCustomElTree.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
var
  ItemPart: TSTItemPart;
  HCol: Integer;
  Item: TElTreeItem;
  SelList : TElList;
  FFS : boolean;
begin
  if FIgnoreClick then exit;
  inherited;
  HCol:=0;
  Item:=GetItemAt(X,Y,ItemPart,HCol);
  if (ItemPart=ipButton) then
  begin
    if (not Item.Enabled) then exit;
    Item.Expanded := not(Item.Expanded)
  end
  else
  if (ItemPart = ipCheckBox) then
  begin
    if not Item.Enabled then exit;
    if Item.CheckBoxEnabled then
    begin
      if Item.CheckBoxType = ect3SCheckBox then
      begin
        if Item.CheckBoxState = cbChecked then Item.CheckBoxState := cbUnchecked else
        if Item.CheckBoxState = cbUnchecked then Item.CheckBoxState := cbGrayed else
        Item.CheckBoxState := cbChecked;
      end
      else
      if Item.CheckBoxType = ectCheckBox then Item.Checked := not Item.Checked
      else Item.Checked := true;
      DoItemChange(Item, icmCheckState);
    end;
  end else
  begin
    if (ItemPart = ipColumn) and (not HeaderSections[HCol].ClickSelect) then exit;
    if ssLeft in Shift then
    begin
      if FMultiSelect then IsUpdating:=true;
      if (Item<>nil) and ((ItemPart<>ipInside) or FFullRowSelect) then
      begin
        FFS := false;
        if (FSelMode=smUsual) and (FMultiSelect) then
        begin
          if (not (ssCtrl in shift)) and (not (ssShift in shift)) then
          begin
            if ItemFocused <> nil then FFS := ItemFocused.Selected;
            DeselectAll;
          end;
          if ssShift in shift then
          begin
            DeselectAll;
            if FSelected<>nil then FSelected.Selected:=true;
            SelectRange(Item,FSelected);
          end;
        end;
        if ItemFocused = Item then
        begin
          if ((not FFS) or (FAlwaysKeepSelection)) and (not Item.Selected) then DoSetSelected(Item)
        end else ItemFocused:=Item;
        if ((FMultiSelect) and (not (ssCtrl in shift)) and (not (ssShift in shift)))
           or (not FMultiSelect) then FSelected := Item;
      end else
// WARNING!!!
        if (Item = nil) then ItemFocused := nil;
// End of WARNING!!!
      if FMultiSelect then IsUpdating:=false;
    end
    else
    if ssRight in Shift then
    begin
      if FMultiSelect then
      begin
        IsUpdating:=true;
        SelList:=TElList.Create;
        AllSelected(SelList);
      end;
      if (not FMultiSelect) or (FMultiSelect and ((SelList.Count < 2) and (FSelMode = smUsual)) or (FSelMode = smSimple)) then
      begin
        if (Item<>nil) and ((ItemPart<>ipInside) or FFullRowSelect) then
        begin
          if (FSelMode=smUsual) and (FMultiSelect) then
          begin
            if (not (ssCtrl in shift)) and (not (ssShift in shift)) then
            begin
              //if ItemFocused <> nil then FFS := ItemFocused.Selected;
              DeselectAll;
            end;
            if ssShift in shift then
            begin
              DeselectAll;
              if FSelected<>nil then FSelected.Selected:=true;
              SelectRange(Item,FSelected);
            end;
          end;
          if (ItemFocused = Item) and (not Item.Selected) then DoSetSelected(Item) else ItemFocused:=Item;
          if ((FMultiSelect) and (not (ssCtrl in shift)) and (not (ssShift in shift)))
             or (not FMultiSelect) then FSelected:=Item;
        end else
// WARNING!!!
        if (Item = nil) then ItemFocused := nil;
// End of WARNING!!!
      end;
      if FMultiSelect then
      begin
        SelList.Free;
        IsUpdating:=false;
      end;
    end;
  end;
end;

procedure TCustomElTree.MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited;
end;

procedure TCustomElTree.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

procedure TCustomElTree.SetChStateImage(value:boolean);
begin
  if value=FChStateImage then exit;
  IsUpdating:=true;
  FChStateImage:=value;
  FRangeUpdate := true;
  FUpdated:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.DoHideLineHint;
begin
  if FHintTimer<>nil then
  begin
    FHintTimer.Free;
    FHintTimer:=nil;
  end;
  if FHintWnd<>nil then
  begin
    FHintWnd.ReleaseHandle;
    FHintWnd.Hide;
    FHintWnd.Free;
    FHintWnd:=nil;
  end;
end;

procedure TCustomElTree.DoShowLineHint;
var s:string;
    R,R1,R2:TRect;
    dx,dy:integer;
    b : boolean;

begin
  R1:=GetItemRect(FVisible.IndexOf(Item));
  if IsRectEmpty(R1) then exit;
  FHintWnd:=THintWindow.Create(self);
  s:=Item.Text;
  FHintWnd.Color:=FBkColor;
  R1.Left:=Item.FTextLeft-FHPos;
{$IFNDEF VER90}
  R:=FHintWnd.CalcHintRect(10000,S,nil);
{$ELSE}
  R.Left := 0;
  R.Right := FHintWnd.Canvas.TextWidth(S) + 6;
  R.Top := 0;
  R.Bottom := FHintWnd.Canvas.TextHeight(S) + 2;
{$ENDIF}

  dec(R.Bottom,2);
  dx:=R.Right-R.Left+1;
  dy:=R.Bottom-R.Top+1;
  R.Left:=R1.Left;
  R.Top:=R1.Top;
  R.Right:=R.Left+dx;
  R.Bottom:=R.Top+dy;
  R.TopLeft:=ClientToScreen(R.TopLeft);
  R.BottomRight:=ClientToScreen(R.BottomRight);
  FHintWnd.BoundsRect:=R;
  b:=true;
  DoShowHint(Item, S, FHintWnd, FHintCoord, b);
  if (b and (S<>Item.Text)) then
  begin
    R:=FHintWnd.BoundsRect;
{$IFNDEF VER90}
    R2:=FHintWnd.CalcHintRect(10000,S,nil);
{$ELSE}
    R2.Left := 0;
    R2.Right := FHintWnd.Canvas.TextWidth(S) + 6;
    R2.Top := 0;
    R2.Bottom := FHintWnd.Canvas.TextHeight(S) + 2;
{$ENDIF}
    dec(R2.Bottom,2);
    dx:=R2.Right-R2.Left+1;
    dy:=R2.Bottom-R2.Top+1;
    R.Right:=R.Left+dx;
    R.Bottom:=R.Top+dy;
  end;
  if b then FHintWnd.ActivateHint(R,S);
end;

procedure TCustomElTree.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS;
end;

procedure TCustomElTree.WMKeyDown(var Message: TWMKeyDown);
begin
  inherited;
end;

procedure TCustomElTree.KeyDown(var Key: Word; Shift: TShiftState);
type PGIRec = ^TGIRec;
     TGIRec = record
       i,oi,ni,PgCnt:integer;
       TSI,OTSI   : TElTreeItem;
       Sel        : boolean;
       NewFocused : TElTreeItem;
     end;
var GIRec : TGIRec;
    TSI,TSI1: TElTreeItem;
    OldSel,
    OldSel1 : boolean;

    procedure IntPrevVis(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                           IterateData:pointer; Tree:TCustomElTree);
    begin
      PGIRec(IterateData)^.OTSI:=PGIRec(IterateData)^.TSI;
      PGIRec(IterateData)^.TSI:=Item;
      PGIRec(IterateData)^.oi:=PGIRec(IterateData)^.ni;
      PGIRec(IterateData)^.ni:=Index;
      if Item=Tree.FFocused then
      begin
        if PGIRec(IterateData)^.OTSI<>nil then
        begin
          PGIRec(IterateData)^.Sel:=PGIRec(IterateData)^.OTSI.Selected;
          PGIRec(IterateData)^.NewFocused := PGIRec(IterateData)^.OTSI;
          PGIRec(IterateData)^.i:=PGIRec(IterateData)^.oi;
        end;
        ContinueIterate:=false;
      end;
    end;

    procedure IntNextVis(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                           IterateData:pointer; Tree:TCustomElTree);
    begin
      if PGIRec(IterateData)^.OTSI=Tree.FFocused then
      begin
        PGIRec(IterateData)^.Sel:=Item.Selected;
        PGIRec(IterateData)^.NewFocused := Item;
        PGIRec(IterateData)^.i:=Index;
        ContinueIterate:=false;
      end;
      if Item=Tree.FFocused then PGIRec(IterateData)^.OTSI:=Item;
    end;

    procedure IntPgVis(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                           IterateData:pointer; Tree:TCustomElTree);
    begin
      if Item=Tree.FFocused then PGIRec(IterateData)^.oi:=1;
      if PGIRec(IterateData)^.oi=1 then
      begin
        dec(PGIRec(IterateData)^.PgCnt);
        PGIRec(IterateData)^.TSI := Item;
        PGIRec(IterateData)^.NewFocused := Item;
        PGIRec(IterateData)^.Sel := Item.Selected;
        PGIRec(IterateData)^.i := Index;
      end;
      if PGIRec(IterateData)^.PgCnt<0 then
      begin
        ContinueIterate:=false;
      end;
    end;

    procedure IntVis0(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                        IterateData:pointer; Tree:TCustomElTree);
    begin
      PGIRec(IterateData)^.Sel:=Item.Selected;
      PGIRec(IterateData)^.NewFocused := Item;
      PGIRec(IterateData)^.i:=0;
      ContinueIterate:=false;
    end;


var NewFocused : TElTreeItem;

begin
  if not (csNoStdEvents in ControlStyle) then
  begin
    if FMultiSelect then IsUpdating:=true;
    TSI := nil;
    TSI1 := nil;
    NewFocused := nil;
    OldSel:=false;
    OldSel1:=false;
    if (ssCtrl in shift) and (ssShift in shift) then Shift:=Shift-[ssCtrl];
    case Key of    //
      VK_ADD: if (FFocused<>nil) and (FFocused.HasVisibleChildren or FFocused.ForceButtons) and
                (not (FFocused.Expanded)) and (FFocused.Enabled) then FFocused.Expanded:=true;
      VK_SUBTRACT: if (FFocused<>nil) and (FFocused.Enabled) and (FFocused.Expanded) then FFocused.Expanded:=false;
      VK_MULTIPLY: if (FFocused<>nil) and (FFocused.Enabled) and (FFocused.HasVisibleChildren or FFocused.ForceButtons) then FFocused.Expand(true);
      VK_SPACE: if (FFocused<>nil) and (FFocused.Enabled) and (FMultiSelect) then
        begin
          if (not (ssCtrl in shift)) and (not (ssShift in shift)) and (FSelMode=smUsual) then
          begin
            DeselectAll;
            FSelected:=FFocused;
          end;
          if FSelMode=smSimple then FSelected:=FFocused;
          FFocused.Selected :=not (FFocused.Selected);
        end;
      VK_HOME,
      VK_END,
      VK_PRIOR,
      VK_NEXT,
      VK_LEFT,
      VK_RIGHT,
      VK_UP,
      VK_DOWN:
         begin
           if RightAlignedTree then
           begin
             if Key = VK_LEFT then Key := VK_RIGHT else
             if Key = VK_RIGHT then Key := VK_LEFT;
           end;
           if Key=VK_LEFT then
             if (FFocused<>nil) and FFocused.Expanded and (FFocused.Enabled) then
             begin
               FFocused.Expanded:=false;
               if FMultiSelect then IsUpdating:=false;
               exit;
             end;
           if Key=VK_RIGHT then
             if (FFocused<>nil) and (not (FFocused.Expanded)) and (FFocused.Enabled) and (FFocused.HasVisibleChildren or FFocused.ForceButtons) then
             begin
               FFocused.Expanded:=true;
               if FMultiSelect then IsUpdating:=false;
               exit;
             end;
           GIRec.OTSI:=nil;
           GIRec.TSI:=nil;
           GIRec.i:=0;
           GIRec.NewFocused := nil;
           if FMultiSelect then
           begin
             if ((ssCtrl in Shift) and (FSelMode=smUsual)) or (FSelMode=smSimple) then
             begin
               TSI:=FFocused;
               if TSI<>nil then OldSel:=TSI.Selected else OldSel:=false;
             end;
             if (ssShift in Shift) then
             begin
               TSI1:=FSelected;
               if TSI1<>nil then OldSel1:=TSI1.Selected else OldSel1:=false;
             end;
             if (FSelMode=smUsual) and (not (ssCtrl in shift)) then DeselectAll;
           end;
           if Key=VK_UP then
           begin
             if FFocused=nil then
             begin
               if TotalVisCount>0 then
               begin
                 FItems.Iterate(true, @IntVis0, @GIRec);
                 NewFocused := GIRec.NewFocused;
               end else
               begin
                 if FMultiSelect then IsUpdating:=false;
                 exit;
               end;
             end
             else //FItems.Iterate(true,@IntPrevVis,@GIRec);
             begin
               FItems.IterateBack(true, @IntNextVis, @GIRec);
               if GIRec.NewFocused <> nil then
               NewFocused := GIRec.NewFocused else
               NewFocused := FFocused;
             end;
           end else
           if Key=VK_DOWN then
           begin
             if FFocused=nil then
             begin
               if TotalVisCount>0 then
               begin
                 FItems.Iterate(true, @IntVis0, @GIRec);
                 NewFocused := GIRec.NewFocused;
               end else
               begin
                 if FMultiSelect then IsUpdating:=false;
                 exit;
               end;
             end else
             begin
               FItems.Iterate(true, @IntNextVis,@GIRec);
               if GIRec.NewFocused <> nil then
               NewFocused := GIRec.NewFocused else
               NewFocused := FFocused;
             end;
           end else
           if Key=VK_Left then
           begin
            if (FFocused<>nil) and (FFocused.Parent<>nil) then
            if FFocused.Parent <> nil then
            NewFocused := FFocused.Parent
            else NewFocused := FFocused;
           end else
           if Key=VK_Right then
           begin
             if (FFocused<>nil) and (FFocused.GetFirstVisibleChild <> nil) then
             begin
               IsUpdating:=true;
               NewFocused := FFocused.GetFirstVisibleChild;
               if NewFocused = nil then NewFocused := FFocused;
               FUpdated:=true;
               FRangeUpdate := true;
               IsUpdating:=false;
             end;
           end else
           if Key=VK_NEXT then
           begin
             GIRec.PgCnt:=GetVisCount;
             GIRec.oi := 0;
             GIRec.NewFocused := nil;
             if FFocused<>nil then FItems.Iterate(true,@IntPgVis,@GIRec);
             if GIRec.NewFocused <> nil then
             NewFocused := GIRec.NewFocused else
             NewFocused := FFocused;
           end else
           if Key=VK_PRIOR then
           begin
             GIRec.PgCnt := GetVisCount;
             GIRec.oi:=0;
             GIRec.NewFocused := nil;
             if FFocused <> nil then FItems.IterateBack(true,@IntPgVis,@GIRec);
             if GIRec.NewFocused <> nil then
             NewFocused := GIRec.NewFocused else
             NewFocused := FFocused;
           end
           else
           if Key=VK_HOME then
           begin
             if FFocused<>nil then
             begin
               NewFocused := FItems.GetVisItem(0);
               GIRec.i:=0;
               if NewFocused = nil then
               begin
                 NewFocused := ItemFocused;
                 GIRec.i := FTopIndex;
               end;
             end;
           end else
           if Key=VK_END then
           begin
             if FFocused<>nil then
             begin
               NewFocused := FItems.GetVisItem(TotalVisCount-1);
               GIRec.i := TotalVisCount-1;
               if NewFocused = nil then
               begin
                 NewFocused := ItemFocused;
                 GIRec.i := FTopIndex;
               end;
             end;
           end;
           if FMultiSelect then
           begin
             if ((ssCtrl in Shift) and (FSelMode=smUsual)) or (FSelMode=smSimple) then
             begin
               if TSI<>nil then TSI.Selected := OldSel;
               //if FFocused<>TSI then FFocused.Selected:=GIRec.Sel;
             end;
             if ssShift in shift then
             begin
               SelectRangeEx(TSI1, NewFocused, false);
               if TSI1 <> nil then TSI1.Selected := OldSel1;
               //if FFocused <> nil then FFocused.Selected:=true;
             end;
           end;
           if NewFocused <> nil then
           begin
             if FMultiSelect then
             begin
               if ((ssCtrl in Shift) and (FSelMode=smUsual)) or (FSelMode=smSimple) then
               begin
                 if NewFocused <> TSI then NewFocused.Selected := not GIRec.Sel;
               end;
               if ssShift in shift then
               begin
                 if FFocused <> NewFocused then NewFocused.Selected := false;
               end;
             end;
             if FFocused <> NewFocused then ItemFocused := NewFocused else NewFocused.Selected := true;
           end;
           if (not ((ssCtrl in shift) or (ssShift in shift)))
              or not FMultiSelect then
              begin
                FSelected:=FFocused;
                if (FFocused<>nil) and (not ((FMultiSelect) and (FSelMode=smSimple))) and (FMultiSelect) then
                  FFocused.Selected := true;
              end;
           if (FFocused<>nil) and (FVisible.IndexOf(FFocused)=-1) then
           begin
             if Key=VK_UP then
             begin
               SetVPosition(GIRec.i);
             end else
             if Key=VK_DOWN then
             begin
               SetVPosition(GIRec.i-(GetVisCount-2));
             end else
             if Key=VK_Left then
             begin
               SetVPosition(FFocused.VisIndex);
             end else
             if Key=VK_Right then
             begin
               SetVPosition(GIRec.i-(GetVisCount-2));
             end else
             if Key=VK_PRIOR then
             begin
               SetVPosition(FTopIndex-(GetVisCount-2));
             end else
             if Key=VK_NEXT then
             begin
               SetVPosition(FTopIndex+(GetVisCount-2));
             end else
             if KEY=VK_Home then
             begin
               SetVPosition(0);
             end else
             if KEY=VK_End then
             begin
               SetVPosition(TotalVisCount-GetVisCount+2);
             end;
             FClearVis:=true;
             FClearAll:=true;
           end;
         end;
    end;    // case
    if FMultiSelect then IsUpdating:=false;
  end;
  inherited;
end;

procedure TCustomElTree.SetHideSelect(value:boolean);
begin
  if value=FHideSelect then exit;
  FHideSelect:=value;
  if (not Focused) and (HandleAllocated) then
  begin
    IsUpdating:=true;
    FRangeUpdate := true;
    FUpdated:=true;
    IsUpdating:=false;
  end;
end;

procedure TCustomElTree.DoEnter;
begin
  inherited;
  if FHideSelect then Repaint;
end;

procedure TCustomElTree.DoExit;
begin
  inherited;
  if FHideSelect then Repaint;
end;

procedure TCustomElTree.FillVisFwd;
type PGIRec = ^TGIRec;
     TGIRec = record
      ItemFound : boolean;
      VisToGet: integer;
      VisGot: integer;
      StartInd: integer;
     end;
var GIRec : TGIRec;

    procedure IntFillFwd(Item:TElTreeItem; Index: integer; var ContinueIterate:boolean;
                                           IterateData:pointer; Tree:TCustomElTree);
    begin
      if not PGIRec(IterateData)^.ItemFound then
         if Index=PGIRec(IterateData)^.StartInd then PGIRec(IterateData)^.ItemFound:=true;
      if (PGIRec(IterateData)^.ItemFound) and ((not Item.Hidden) or (not Tree.FilteredVisibility)) then
      begin
        Tree.FVisible.Add(Item);
        inc(PGIRec(IterateData)^.VisGot);
      end;
      if PGIRec(IterateData)^.VisToGet=PGIRec(IterateData)^.VisGot then ContinueIterate:=false;
    end;

begin
  FVisible.Clear;
  GIRec.VisToGet:=ToGet;
  GIRec.VisGot:=0;
  GIRec.ItemFound:=false;
  GIRec.StartInd:=StartIndex;
  FItems.Iterate(true, @IntFillFwd, @GIRec);
end;

procedure TCustomElTree.UpdateElTree;
var R:TRect;
begin
  if (not HandleAllocated) then
  begin
    FUpdated := true;
    exit;
  end;
  if FUpdating then FUpdated := true else
  begin
    if FRangeUpdate or (FClearAll) then
    begin
      if (FVLines and FHeader.Visible) or (FClearAll) then
      begin
        R:=Rect(0,0,ClientWidth,ClientHeight);
        if FHeader.Visible then
           inc(R.Top, GetHeaderHeight);
        FClearAll:=false;
      end
      else
      begin
        R:=Rect(0,0,ClientWidth,FVisible.Count*FLineHeight);
        if FHeader.Visible then
        begin
          inc(R.Top,GetHeaderHeight);
          inc(R.Bottom,GetHeaderHeight);
        end;
      end;
      InvalidateRect(Handle,@R,false);
      FRangeUpdate := false;
    end;
    Update;
  end;
end;

function TCustomElTree.GetVisCount;
var j:integer;
    CliHeight:integer;

begin
    if FHeader.Visible then CliHeight := ClientHeight - (GetHeaderHeight)
       else CliHeight := ClientHeight;
    j:=FLineHeight;
    result := CliHeight div j;
    if ((CliHeight mod j)>0) then inc(result);
end;

procedure TCustomElTree.OnEditKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if Shift=[] then
  begin
    if Key=VK_Return then EndEdit(false);
    if Key=VK_Escape then EndEdit(True);
  end;
end;

procedure TCustomElTree.Assign(Source:TPersistent);
var T:TCustomElTree;
{$IFNDEF VER90}
  Form:TCustomForm;
{$ELSE}
  Form:TForm;
{$ENDIF}

begin
  if Source is TCustomElTree then
  begin
    //inherited;
    IsUpdating:=true;
    T := TCustomElTree(Source);
    FItems.Assign(T.FItems);
    FLinesStyle := T.FLinesStyle;
    FLinesColor := T.LinesColor;
    FDeselectChildrenOnCollapse := T.FDeselectChildrenOnCollapse;
    FDrawFocusRect := T.FDrawFocusRect;
    FBarStyle := T.FBarStyle;
    FAlwaysKeepSelection := T.FAlwaysKeepSelection;
    FFullRowSelect := T.FFullRowSelect;
    FDragType := T.FDragType;
    FAutoLookup := T.FAutoLookup;
    FSelectColumn := T.FSelectColumn;
    FAutoExpand := AutoExpand;
    FPlusPicture.Assign(T.FPlusPicture);
    FMinusPicture.Assign(T.FMinusPicture);
    FCustomPlusMinus := T.FCustomPlusMinus;
    FShowCheckboxes := T.FShowCheckboxes;
    FHideHorzScrollBar := T.FHideHorzScrollBar;
    FHideVertScrollBar := T.FHideVertScrollBar;
    
    FBkColor:=T.FBkColor;
    FTextColor:=T.FTextColor;
    FShowButtons:=T.FShowButtons;
    FShowLines:=T.FShowLines;
    FShowImages:=T.FShowImages;
    FShowRoot:=T.FShowRoot;
    FShowHintMode:=T.FShowHintMode;
    ShowColumns:=T.ShowColumns;
    FBorderStyle:=T.FBorderStyle;
    FCanEdit:=T.FCanEdit;
    FHLines:=T.FHLines;
    FVLines:=T.FVLines;
    FScrollTracking:=T.FScrollTracking;
    FTracking:=T.FTracking;
    HeaderHotTrack:=T.HeaderHotTrack;
    FODFollowCol:=T.FODFollowCol;
    FODMask:=T.FODMask;
    FDragTrgDraw:=T.FDragTrgDraw;
    TopIndex:=T.TopIndex;
    FAutoLineHeight:=T.FAutoLineHeight;
    FLineHeight:=T.FLineHeight;
{$IFNDEF LITE}
    FStorage := T.FStorage;
    FStoragePath := T.FStoragePath;
{$ENDIF}
    FHRange:=-1;
    FUpdated:=true;
    FVisUpdated:=true;
    FClearVis:=true;
    FClearAll:=true;
    IsUpdating:=false;
    if csDesigning in ComponentState then
    begin
      Form := GetParentForm(self);
      if (Form <> nil) and (Form.Designer <> nil) then Form.Designer.Modified;
    end;
  end else inherited;
end;

procedure TCustomElTree.Sort;
begin
  FItems.FRoot.Sort(recursive);
end;

procedure TCustomElTree.CMMouseLeave;
var TSI:TElTreeItem;
begin
  if assigned(FInpEdit) and FInpEdit.Visible then exit;
  TSI:=FTrackItem;
  FTrackItem:=nil;
  if TSI<>nil then
  begin
    TSI.RedrawItem(true); Update;
  end;
  // Destroy Hint timer
  DoHideLineHint;
  inherited;
end;

procedure TCustomElTree.AllocateLineDC;
begin
  FLineDC := GetDCEx(Handle, 0, DCX_CACHE or DCX_CLIPSIBLINGS or DCX_LOCKWINDOWUPDATE);
end;

procedure TCustomElTree.ReleaseLineDC;
begin
  ReleaseDC(Handle, FLineDC);
end;

procedure TCustomElTree.DrawLine;
begin
  FHeaderLineVis:= not FHeaderLineVis;
  PatBlt(FLineDC, FHeaderLine, GetHeaderHeight, 1,
        (ClientHeight-GetHeaderHeight), PATINVERT);
end;

procedure TCustomElTree.OnHeaderSectionTrack;
begin
  case State of    //
   trsBegin:
     begin
       AllocateLineDC;
       FHeaderLine:=Section.Left+Width-FHPos;
       DrawLine;
     end;
   trsMove:
     begin
       DrawLine;
       FHeaderLine:=Section.Left+Width-FHPos;
       DrawLine;
     end;
   trsEnd:
     begin
       if FHeaderLineVis then DrawLine;
       ReleaseLineDC;
     end;
  end;    // case
end;

procedure TCustomElTree.WndProc(var Message: TMessage);
begin
  if ((FHintWnd<>nil) or (FHintTimer<>nil)) and (FHideHintOnMove) then
    if
      ((Message.Msg >= WM_KEYFIRST) and (Message.Msg <= WM_KEYLAST)) or
      ((Message.Msg = CM_ACTIVATE) or (Message.Msg = CM_DEACTIVATE)) or
      (Message.Msg = CM_APPKEYDOWN) or (Message.Msg = CM_APPSYSCOMMAND) or
      (Message.Msg = WM_COMMAND) or
      ((Message.Msg > WM_MOUSEMOVE) and (Message.Msg <= WM_MOUSELAST))
      or (Message.Msg = WM_NCMOUSEMOVE)
    then
      DoHideLineHint;
  if ((FHintWnd<>nil) or (FHintTimer<>nil)) and
      ((Message.Msg = CM_ACTIVATE) or (Message.Msg = CM_DEACTIVATE))
      or (Message.Msg = WM_NCMOUSEMOVE) then
         DoHideLineHint;
  inherited;
end;

procedure TCustomElTree.OnHeaderSectionDelete;
begin
  if FHeader.Sections.Count=0 then ShowColumns:=false;
end;

function TCustomElTree.DoGetPicture (Item:TElTreeItem):integer;
begin
  result:=-1;
  if Assigned(FOnItemPicDraw) then FOnItemPicDraw(self,Item,result);
end;

function TCustomElTree.DoGetPicture2(Item:TElTreeItem):integer;
begin
  result:=-1;
  if Assigned(FOnItemPicDraw2) then FOnItemPicDraw2(self,Item,result);
end;

function TCustomElTree.GetDraggableSections : Boolean ;
{ Returns the value of data member FDraggableSections. }
begin
  result := FHeader.AllowDrag;
end ;  { GetDraggableSections }

procedure TCustomElTree.SetDraggableSections( newValue : Boolean ) ;
begin
  if FHeader.AllowDrag <> newValue then
  begin
    FHeader.AllowDrag := newValue;
  end ;  { if }
end ;  { SetDraggableSections }

procedure TCustomElTree.OnHeaderSectionMove(Sender:TCustomElHeader; Section : TElHeaderSection; OldPos, NewPos : integer ) ;  { protected }
begin
  IsUpdating:=true;
  FClearAll:=true;
  FUpdated:=true;
  IsUpdating:=false;
  TriggerHeaderColumnMoveEvent(Section, OldPos, NewPos);
end ;  { OnHeaderSectionMove }

procedure TCustomElTree.TriggerHotTrackEvent(OldItem, NewItem : TElTreeItem);
begin
  if assigned(FOnHotTrack) then FOnHotTrack(Self, OldItem, NewItem );
end;  { TriggerHotTrackEvent }

procedure TCustomElTree.SetSortMode(newValue: TSortModes);
begin
  if FSortMode <> newValue then FSortMode := newValue;
end;  { SetSortMode }

procedure TCustomElTree.SetSortSection(newValue: Integer);
begin
  if FSortSection <> newValue then
  begin
    FSortSection := newValue;
  end;
end;  { SetSortSection }

function TCustomElTree.CompareItems(Item1, Item2 : TElTreeItem):integer;
  var FCol : integer;
      T1, T2 : TDateTime;
      N1, N2 : integer; // numeric data
      F1, F2 : extended;  // floating data
      S1, S2 : string;
      I      : integer;
      SM     : TElSSortMode;
      ST     : TSortTypes;
  begin
    if FSortSection = -1 then
    begin
      DoCompareItems(Item1, Item2, result);
      if FSortDir=sdDescend then result:=-result;
    end else
    begin
      if FSortSection = FMainTreeCol then
      begin
        S1:=Item1.Text;
        S2:=Item2.Text;
        if FHeader.Visible then
        begin
          SM := FHeader.Sections[FMainTreeCol].SortMode;
          ST := SectionTypeToSortType(FHeader.Sections[FMainTreeCol].FieldType);
        end else
        begin
          SM := hsmNone;
          ST := FSortType;
        end;
      end else
      begin
        FCol:=FSortSection;
        if FSortSection > FMainTreeCol then dec(FCol);
        SM := FHeader.Sections[FSortSection].SortMode;
        ST := SectionTypeToSortType(FHeader.Sections[FSortSection].FieldType);
        try
          if FCol > Item1.FColText.Count
             then S1 :=''
             else S1:=Item1.FColText[FCol];
        except
          on E:Exception do S1:='';
        end;
        try
          if Item2.FColText.Count<=FCol
             then S2 :=''
             else S2:=Item2.FColText[FCol];
        except
          on E:Exception do S2:='';
        end;
      end;
      case ST of
       stCustom:
          begin
            DoCompareItems(Item1, Item2, result);
            if FSortDir=sdDescend then result:=-result;
          end;
       stNumber: begin
                   val(S1,N1,I);
                   if I>0 then N1:=0;
                   val(S2,N2,I);
                   if I>0 then N2:=0;
                   if N1>N2 then result:=1 else if N1<N2 then result:=-1 else result:=0;
                 end;
       stFloating:  begin
                    F1:=0;
                    F2:=0;
                   try
                     F1:=StrToFloat(S1);
                   except
                     on E:Exception do ;
                   end;
                   try
                     F2:=StrToFloat(S2);
                   except
                     on E:Exception do ;
                   end;
                   if F1>F2 then result:=1 else if F1<F2 then result:=-1 else result:=0;
                 end;
       stDateTime: begin
                     T1:=Now;
                     T2:=T1;
                     try
                       T1:=StrToDateTime(S1);
                     except
                       on E:Exception do ;
                     end;
                     try
                       T2:=StrToDateTime(S2);
                     except
                       on E:Exception do ;
                     end;
                     if T1>T2 then result:=1 else if T1<T2 then result:=-1 else result:=0;
                   end;
       stDate: begin
                     T1:=Now;
                     T2:=T1;
                     try
                       T1:=StrToDate(S1);
                     except
                       on E:Exception do ;
                     end;
                     try
                       T2:=StrToDate(S2);
                     except
                       on E:Exception do ;
                     end;
                     if T1>T2 then result:=1 else if T1<T2 then result:=-1 else result:=0;
               end;
       stTime: begin
                     T1:=Now;
                     T2:=T1;
                     try
                       T1:=StrToTime(S1);
                     except
                       on E:Exception do ;
                     end;
                     try
                       T2:=StrToTime(S2);
                     except
                       on E:Exception do ;
                     end;
                     if T1>T2 then result:=1 else if T1<T2 then result:=-1 else result:=0;
               end;
       else result:=AnsiCompareText(S1,S2);
      end;
      if (SM=hsmDescend) or ((SM=hsmNone) and (FSortDir=sdDescend)) then result:=-result;
    end;
  end;

procedure TCustomElTree.Update;
begin
  inherited;
end;

function TCustomElTree.IsEditing : boolean;
begin
  result:= (FInpEdit<>nil) and FInpEdit.Visible;
end;
{$IFNDEF LITE}
procedure TCustomElTree.SetStorage(newValue : TElIniFile);
begin
  if newValue <> FStorage then
  begin
    FStorage := newValue;
    if FStorage <> nil then FStorage.FreeNotification(Self);
  end;
end;
{$ENDIF}
// Changed 10/26/98 by EM
function TCustomElTree.GetNodeAt(X, Y : integer) : TElTreeItem;  { public }
begin
  result := GetItemAtY(Y);
end;  { GetNodeAt }

procedure TCustomElTree.SetShowCheckboxes(newValue : Boolean);
begin
  if (FShowCheckboxes <> newValue) then
  begin
    IsUpdating := true;
    FShowCheckboxes := newValue;
    if not FHeader.Visible then FHRange := -1;
    FClearAll:=true;
    FUpdated:=true;
    IsUpdating := false;
  end;  {if}
end;  {SetShowCheckboxes}

procedure TCustomElTree.SetPlusPicture(newValue : TBitmap);
begin
  FPlusPicture.Assign(newValue);
end;  {SetPlusPicture}

procedure TCustomElTree.SetMinusPicture(newValue : TBitmap);
begin
  FMinusPicture.Assign(newValue);
end;  {SetMinusPicture}

procedure TCustomElTree.SetCustomPlusMinus(newValue : Boolean);
begin
  if (FCustomPlusMinus <> newValue) then
  begin
    IsUpdating := true;
    FCustomPlusMinus := newValue;
    if not FHeader.Visible then FHRange := -1;
    FClearAll:=true;
    FUpdated:=true;
    IsUpdating := false;
  end;  {if}
end;  {SetCustomPlusMinus}

procedure TCustomElTree.SetSelectColumn(newValue : Integer);
begin
  if (FSelectColumn <> newValue) then
  begin
    if (newValue > -2) and (newValue < FHeader.Sections.Count) then
    begin
      FSelectColumn := newValue;
      if not FRowSelect then
      begin
        if not FMultiSelect then
        begin
          if (FFocused <> nil) then FFocused.UpdateItem;
        end else
        begin
          IsUpdating := true;
          FClearAll:=true;
          FUpdated:=true;
          IsUpdating := false;
        end; // if/else
      end; //if
    end; // if
  end;
end;  {SetSelectColumn}

procedure TCustomElTree.TriggerHeaderLookupEvent;
begin
  if (assigned(FOnHeaderLookup)) then FOnHeaderLookup(Self, Section, Text);
end;  {TriggerHeaderLookupEvent}

procedure TCustomElTree.TriggerHeaderLookupDoneEvent;
begin
  if (assigned(FOnHeaderLookupDone)) then FOnHeaderLookupDone(Self, Section, Text, Accepted);
end;  {TriggerHeaderLookupDoneEvent}

procedure TCustomElTree.TriggerHeaderSectionExpandEvent(Section  : 
  TElHeaderSection);
begin
  if (assigned(FOnHeaderSectionExpand)) then FOnHeaderSectionExpand(Self, Section  );
end;

procedure TCustomElTree.TriggerHeaderSectionCollapseEvent(Section : TElHeaderSection);
begin
  if (assigned(FOnHeaderSectionCollapse)) then
    FOnHeaderSectionCollapse(Self, Section  );
end;

procedure TCustomElTree.Loaded;
begin
  inherited Loaded;
  IsUpdating := true;
  FHeader.Loaded;
  if (FSortMode=smAdd) or (FSortMode=smAddClick) then FItems.FRoot.Sort(true);
  FUpdated := true;
  FClearAll := true;
  IsUpdating := false;
end;

procedure TCustomElTree.SetDragType(newValue : TElDragType);
begin
  if (FDragType <> newValue) then
  begin

{$IFNDEF VER90}
    if newValue = dtDelphi then
    {$IFNDEF LITE}
      FDropTarget.Free
    {$ENDIF}
    else
    begin
      if FDragType = dtDelphi then // when switching from one to another OLE mode we don't create a drop target
      begin
      {$IFNDEF LITE}
        FDropTarget := TElDropTarget.Create(nil);
        FDropTarget.Target := Self;
        FDropTarget.OnTargetDrag := TargetDrag_FDropTargetHandler;
        FDropTarget.OnTargetDrop := TargetDrop_FDropTargetHandler;
      {$ENDIF}
      end;
    end;
{$ENDIF}
    FDragType := newValue;
  end;  {if}
end;  {SetDragType}

{$IFNDEF VER90}
{$IFNDEF LITE}
procedure TCustomElTree.TargetDrag_FDropTargetHandler(Sender : TObject; State : TDragState; Source : TOleDragObject; Shift : TShiftState; X : Integer; Y : Integer; var DragType : TDragType);
var Item : TElTreeItem;

begin
  TriggerOleTargetDragEvent(State, Source, Shift, X, Y, DragType);
  DoDragOver(Source, X, Y, DragType in [dtCopy, dtMove, dtLink]);
  if State = dsDragLeave then
  begin
    if Self.FDropTrg <> nil then
    begin
      Item := FDropTrg;
      FDropTrg := nil;
      Item.RedrawItem(true);
    end;
  end;
end;  {TargetDrag_FDropTargetHandler}

procedure TCustomElTree.TargetDrop_FDropTargetHandler(Sender : TObject; Source : TOleDragObject; Shift : TShiftState; X : Integer; Y : Integer; var DragType : TDragType);
var TSI: TElTreeItem;
begin
  TSI := FDropTrg;
  FDropTrg := nil;
  if TSI <> nil then TSI.RedrawItem(true);
  TriggerOleTargetDropEvent(Source, Shift, X, Y, DragType);
end;  {TargetDrop_FDropTargetHandler}

procedure TCustomElTree.TriggerOleTargetDragEvent;
begin
  if (assigned(FOnOleTargetDrag)) then
    FOnOleTargetDrag(Self, State, Source, Shift, X, Y, DragType);
end;  {TriggerOleTargetDragEvent}

procedure TCustomElTree.TriggerOleTargetDropEvent;
begin
  if (assigned(FOnOleTargetDrop)) then
    FOnOleTargetDrop(Self, Source, Shift, X, Y, DragType);
end;  {TriggerOleTargetDropEvent}

procedure TCustomElTree.TriggerOleDragStartEvent;
begin
  if (assigned(FOnOleDragStart)) then
      FOnOleDragStart(Self, dataObj, dropSource, dwOKEffects);
end;  {TriggerOleDragStartEvent}

procedure TCustomElTree.TriggerOleDragFinishEvent;
begin
  if (assigned(FOnOleDragFinish)) then FOnOleDragFinish(Self, dwEffect, Result);
end;  {TriggerOleDragFinishEvent}
{$ENDIF}
{$ENDIF}

procedure TCustomElTree.HeaderResizeTransfer(Sender: TObject);
begin
  if (assigned(FOnHeaderResize)) then
    FOnHeaderResize(Self);
end;

procedure TCustomElTree.HeaderResizeHandler(Sender: TObject);
begin
  DoHeaderResize(Sender);
  HeaderResizeTransfer(Self);
end;

function TCustomElTree.IsInView(Item : TElTreeItem): Boolean;
begin
  result := FVisible.IndexOf(Item) <> -1;
end;

function TCustomElTree.SectionTypeToSortType(SectionType : TFieldTypes):
  TSortTypes;
begin
  case SectionType of
    sftText: result := stText;
    sftNumber: result := stNumber;
    sftFloating: result := stFloating;
    sftDateTime: result := stDateTime;
    sftDate: result := stDate;
    sftTime: result := stTime;
    else result := stCustom;
  end;
end;

function TCustomElTree.GetStickyHeaderSections : Boolean;
begin
  result := FHeader.StickySections;
end;  {GetStickyHeaderSections}

procedure TCustomElTree.SetStickyHeaderSections(newValue : Boolean);
begin
  FHeader.StickySections := newValue;
end;  {SetStickyHeaderSections}

procedure TCustomElTree.SetBarStyle(newValue : Boolean);
begin
  if (FBarStyle <> newValue) then
  begin
    IsUpdating := true;
    FBarStyle := newValue;
    if FBarStyle then
    begin
      BkColor := clBtnFace;
      TextColor := clBtnText;
      LinesColor := clBtnShadow;
    end;
    if FAutoLineHeight then FLineHeight:=DefineLineHeight;
    if not FHeader.Visible then FHRange := -1;
    FClearAll:=true;
    FUpdated:=true;
    IsUpdating := false;
  end;  {if}
end;  {SetBarStyle}

procedure TCustomElTree.SetDrawFocusRect(newValue : Boolean);
begin
  if (FDrawFocusRect <> newValue) then
  begin
    FDrawFocusRect := newValue;
    if ItemFocused <> nil then ItemFocused.RedrawItem(true);
  end;  {if}
end;  {SetDrawFocusRect}

procedure TCustomElTree.TriggerMeasureItemPartEvent(Item : TElTreeItem; PartIndex : integer; var Size : TPoint);
begin
  if (assigned(FOnMeasureItemPart)) then
    FOnMeasureItemPart(Self, Item , PartIndex , Size );
end;  {TriggerMeasureItemPartEvent}

procedure TCustomElTree.SetLinesColor(newValue : TColor);
begin
  if (FLinesColor <> newValue) then
  begin
    IsUpdating := true;
    FLinesColor := newValue;
    if FShowLines then
    begin
      FClearAll:=true;
      FUpdated:=true;
    end;
    IsUpdating := false;
  end;  {if}
end;  {SetLinesColor}

procedure TCustomElTree.SetLinesStyle(newValue : TPenStyle);
begin
  if (FLinesStyle <> newValue) then
  begin
    IsUpdating := true;
    FLinesStyle := newValue;
    if FShowLines then
    begin
      FClearAll:=true;
      FUpdated:=true;
    end;
    IsUpdating := false;
  end;  {if}
end;  {SetLinesStyle}

{$IFNDEF LITE}
procedure TCustomElTree.SetImages2(newValue: TElImageList);
{$ELSE}
procedure TCustomElTree.SetImages2(newValue: TImageList);
{$ENDIF}
var i:integer;
begin
  if FImages2=newValue then exit;
  IsUpdating:=true;
  if FImages2 <> nil then FImages2.UnRegisterChanges(FImageChangeLink);
  FImages2:=newValue;
  if FImages2<>nil then FImages2.RegisterChanges(FImageChangeLink);
  i:=DefineLineHeight;
  if FAutoLineHeight and (i<>FLineHeight) then
  begin
    FLineHeight:=i;
    FClearVis:=true;
    FClearAll:=true;
  end;
  if not FHeader.Visible then
  begin
    FHRange := -1;
    DefineHRange;
    if RightAlignedTree then
    begin
      FRightAlignedTree := false;
      RightAlignedTree := true;
    end;
  end;
  FRangeUpdate := true;
  FUpdated:=true;
  IsUpdating:=false;
end;

procedure TCustomElTree.SetRightAlignedTree(newValue: Boolean);
begin
  if (FRightAlignedTree <> newValue) then
  begin
    FRightAlignedTree := newValue;
    FHeader.RightAlignedOrder := newValue;
    if not FHeader.Visible then
    begin
      if FHRange = -1 then DefineHRange;
      if newValue then SetHPosition(FHRange - FHPos) else SetHPosition(0);
    end;
    Repaint;
  end;  {if}
end;

procedure TCustomElTree.SetFlat(newValue : Boolean);
begin
  if (FFlat <> newValue) then
  begin
    FFlat := newValue;
    Repaint;
  end;  { if }
end;  { SetFlat }

procedure TCustomElTree.WMNCPaint(var Msg : TMessage);  { private }
var DC : HDC;
    R  : TRect;
    LBR : TLogBrush;
    ABR,
    OBR : HBRUSH;
    Rgn : HRGN;
begin
  inherited;
  if not flat then exit;
  R := BoundsRect;
  R.TopLeft := ScreenToClient(Parent.ClientToScreen(R.TopLeft));
  R.BottomRight := ScreenToClient(Parent.ClientToScreen(R.BottomRight));
  dc := GetWindowDC(Handle);
  Rgn := CreateRectRgn(R.Left, R.Top, R.Right, R.Bottom);
  GetWindowRgn(Handle, Rgn);
  if Focused then
  begin
    Lbr.lbStyle := BS_SOLID;
    Lbr.lbColor := GetSysColor(COLOR_ACTIVEBORDER);
    Lbr.lbHatch := 0;
    ABR := CreateBrushIndirect(Lbr);
    OBR := SelectObject(DC, ABR);
    FrameRect(DC, R, ABR);
    //FrameRgn(DC, Rgn, ABR, GetSystemMetrics(SM_CYBORDER), GetSystemMetrics(SM_CXBORDER));
    SelectObject(DC, OBR);
    DeleteObject(ABR);
  end else
  begin
    Lbr.lbStyle := BS_SOLID;
    Lbr.lbColor := GetSysColor(COLOR_INACTIVEBORDER);
    Lbr.lbHatch := 0;
    ABR := CreateBrushIndirect(Lbr);
    OBR := SelectObject(DC, ABR);
    FrameRect(DC, R, ABR);
    //FrameRgn(DC, Rgn, ABR, GetSystemMetrics(SM_CYBORDER), GetSystemMetrics(SM_CXBORDER));
    SelectObject(DC, OBR);
    DeleteObject(ABR);
  end;
  DeleteObject(Rgn);
  ReleaseDC(Handle, dc);
end;  { WMNCPaint }

procedure TCustomElTree.SetRightAlignedText(newValue: Boolean);
begin
  if (FRightAlignedText <> newValue) then
  begin
    FRightAlignedText := newValue;
    FHeader.RightAlignedText := newValue;
    RecreateWnd;
  end;  {if}
end;

procedure TCustomElTree.SetFilteredVisibility(newValue: Boolean);
begin
  if (FFilteredVisibility <> newValue) then
  begin
    IsUpdating := true;
    FFilteredVisibility := newValue;
    FVisUpdated := true;
    FClearVis:=true;
    FClearAll:=true;
    FRangeUpdate := true;
    FUpdated := true;
    IsUpdating := false;
  end;  {if}
end;

procedure TCustomElTree.TriggerApplyVisFilterEvent(Item : TElTreeItem; var 
  Hidden : boolean);
{CDK: Call as needed to trigger event.}
begin
  if (assigned(FOnApplyVisFilter)) then
    FOnApplyVisFilter(Self, Item , Hidden );
end;

{$IFDEF D_4_UP}
    function TCustomElTree.GetDragImages: TDragImageList;
{$ELSE}
    function TCustomElTree.GetDragImages: TCustomImageList;
{$ENDIF}
begin
  result:=nil;
  if (FDragImages<>nil) and (FDragImages.Count > 0) then Result := FDragImages;
end;

// ****************************************************************************
// TElTreeItems methods
// ****************************************************************************

constructor TElCellStyle.Create;
begin
  inherited Create;
  FOwner := Owner;
  FTextFlags := DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER or DT_LEFT or DT_END_ELLIPSIS;
  FCellType := sftText;
end;

procedure TElCellStyle.SetCellBkColor(newValue: TColor);
begin
  if (FCellBkColor <> newValue) then
  begin
    FCellBkColor := newValue;
    if Assigned(FOwner) and FOwner.FUseStyles then FOwner.UpdateItem;
  end;  {if}
end;

procedure TElCellStyle.SetTextBkColor(newValue: TColor);
begin
  if (FTextBkColor <> newValue) then
  begin
    FTextBkColor := newValue;
    if FStyle <> elhsText then exit;
    if Assigned(FOwner) and FOwner.FUseStyles then FOwner.UpdateItem;
  end;  {if}
end;

procedure TElCellStyle.SetTextColor(newValue: TColor);
begin
  if (FTextColor <> newValue) then
  begin
    FTextColor := newValue;
    if FStyle <> elhsText then exit;
    if Assigned(FOwner) and FOwner.FUseStyles then FOwner.UpdateItem;
  end;  {if}
end;

procedure TElCellStyle.SetTextFlags(newValue: DWORD);
begin
  if (FTextFlags <> newValue) then
  begin
    FTextFlags := newValue;
    if FStyle <> elhsText then exit;
    if Assigned(FOwner) and FOwner.FUseStyles then FOwner.UpdateItem;
  end;  {if}
end;

procedure TElCellStyle.SetPicture(newValue: TBitmap);
begin
  if (FPicture <> newValue) then
  begin
    FPicture := newValue;
    if FStyle <> elhsPictureOnly then exit;
    if Assigned(FOwner) and FOwner.FUseStyles then FOwner.UpdateItem;
  end;  {if}
end;

procedure TElCellStyle.SetCellType(newValue: TFieldTypes);
begin
  if (FCellType <> newValue) then
  begin
    FCellType := newValue;
  end;  {if}
end;

procedure TElCellStyle.SetStyle(newValue: TElSectionStyle);
begin
  if (FStyle <> newValue) then
  begin
    FStyle := newValue;
    if Assigned(FOwner) and FOwner.FUseStyles then FOwner.UpdateItem;
  end;  {if}
end;

procedure TElCellStyle.SetOwnerColors(newValue: Boolean);
begin
  if (FOwnerProps <> newValue) then
  begin
    FOwnerProps := newValue;
    if Assigned(FOwner) and FOwner.FUseStyles then FOwner.UpdateItem;
  end;  {if}
end;

procedure TElCellStyle.SetFontSize(newValue: integer);
begin
  if (FFontSize <> newValue) then
  begin
    FFontSize := newValue;
    if FStyle <> elhsText then exit;
    if Assigned(FOwner) and FOwner.FUseStyles then FOwner.UpdateItem;
  end;  {if}
end;

procedure TElCellStyle.SetFontStyles(newValue: TFontStyles);
begin
  if (FFontStyles <> newValue) then
  begin
    FFontStyles := newValue;
    if FStyle <> elhsText then exit;
    if Assigned(FOwner) and FOwner.FUseStyles then FOwner.UpdateItem;
  end;  {if}
end;

procedure TElCellStyle.SetFontName(newValue: TFontName);
begin
  if (FFontName <> newValue) then
  begin
    FFontName := newValue;
    if FStyle <> elhsText then exit;
    if Assigned(FOwner) and FOwner.FUseStyles then FOwner.UpdateItem;
  end;  {if}
end;

procedure TElCellStyle.Assign(Source: TElCellStyle);
begin
  if Source = nil then exit;
  FCellBkColor := Source.FCellBkColor;
  FTextBkColor := Source.FTextBkColor;
  FTextColor := Source.FTextColor;
  FTextFlags := Source.FTextFlags;
  FPicture := Source.FPicture;
  FCellType := Source.FCellType;
  FStyle := Source.FStyle;
  FOwnerProps := Source.FOwnerProps;
  FFontSize := Source.FFontSize;
  FFontStyles := Source.FFontStyles;
  FFontName := Source.FFontName;
end;  {Assign}

end.

